Package: gp3sequences
Title: Transparent Analysis of Ordered Categorical Sequences
Version: 0.3.0
Authors@R: 
    person(
      given = "Stefanos",
      family = "Balaskas",
      email = "s.balaskas@ac.upatras.gr",
      role = c("aut", "cre"),
      comment = c(ORCID = "0000-0003-2444-9796")
    )
Description: Provides transparent, reproducible, and auditable tools for validating,
    preparing, encoding, summarising, comparing, modelling, and diagnosing
    ordered categorical sequence data. Supports explicit preprocessing
    policies, contiguous motifs and bounded subsequences, consensus and group
    comparisons, edit and transition distances, clustering and stability
    diagnostics, transition networks, higher-order models, categorical,
    mixture, multichannel, and covariate hidden Markov models, longitudinal
    panel workflows, time-varying models, design-aware inference, analysis
    contracts and provenance audits, and guarded adapters to specialist
    sequence-analysis packages.
License: MIT + file LICENSE
URL: https://stefanosbalaskas.github.io/gp3sequences/,
        https://github.com/stefanosbalaskas/gp3sequences
BugReports: https://github.com/stefanosbalaskas/gp3sequences/issues
Suggests: arules, arulesSequences, cluster, GrpString, igraph, knitr,
        mgcv, rmarkdown, seqHMM, testthat (>= 3.0.0), TraMineR
Config/testthat/edition: 3
Config/Needs/website: pkgdown
Encoding: UTF-8
RoxygenNote: 8.0.0
Imports: graphics, methods, stats, utils
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-08-10 23:54:08 UTC; Stefanos-PC
Author: Stefanos Balaskas [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2444-9796>)
Maintainer: Stefanos Balaskas <s.balaskas@ac.upatras.gr>
Repository: CRAN
Date/Publication: 2026-08-23 10:40:25 UTC

# Generated by roxygen2: do not edit by hand

export(as_arules_sequences)
export(as_grpstring_data)
export(as_igraph_transition_network)
export(as_seqhmm_sequences)
export(as_traminer_sequences)
export(audit_sequence_analysis)
export(audit_sequence_data)
export(bootstrap_sequence_clusters)
export(bootstrap_sequence_group_difference)
export(bootstrap_transition_network)
export(cluster_sequences)
export(compare_sequence_analysis_results)
export(compare_sequence_groups)
export(compare_sequence_hmms)
export(compare_sequence_panel_changes)
export(compare_sequence_subsequences)
export(compute_sequence_distance)
export(create_consensus_sequence)
export(create_sequence_cluster_ensemble)
export(create_transition_network)
export(declare_sequence_comparison_design)
export(decode_covariate_sequence_states)
export(decode_multichannel_sequence_states)
export(decode_sequence_states)
export(detect_transition_communities)
export(encode_sequence_data)
export(extract_representative_sequences)
export(extract_sequence_ngrams)
export(extract_sequence_subsequences)
export(filter_sequence_motifs)
export(filter_sequence_subsequences)
export(fit_covariate_sequence_hmm)
export(fit_higher_order_transition_model)
export(fit_multichannel_sequence_hmm)
export(fit_sequence_hmm)
export(fit_sequence_hmm_mixture)
export(fit_time_varying_sequence_model)
export(format_consensus_sequence)
export(format_sequence_motif_positions)
export(format_sequence_motifs)
export(format_sequence_paths)
export(plot_consensus_sequence)
export(plot_multichannel_sequence_hmm)
export(plot_sequence_cluster_silhouette)
export(plot_sequence_distance_heatmap)
export(plot_sequence_entropy)
export(plot_sequence_group_comparison)
export(plot_sequence_group_inference)
export(plot_sequence_index)
export(plot_sequence_motif_positions)
export(plot_sequence_motifs)
export(plot_sequence_panel_changes)
export(plot_sequence_state_distribution)
export(plot_sequence_subsequences)
export(plot_time_varying_sequence_model)
export(plot_transition_network)
export(predict_covariate_transition_probabilities)
export(predict_next_state)
export(predict_time_varying_sequence_model)
export(prepare_gp3tools_sequences)
export(prepare_sequence_data)
export(prepare_sequence_panel)
export(sequence_capabilities)
export(summarise_consensus_agreement)
export(summarise_covariate_sequence_hmm)
export(summarise_multichannel_sequence_hmm)
export(summarise_sequence_cluster_stability)
export(summarise_sequence_distance)
export(summarise_sequence_group_inference)
export(summarise_sequence_hmm)
export(summarise_sequence_motif_positions)
export(summarise_sequence_motifs)
export(summarise_sequence_panel)
export(summarise_sequence_states)
export(summarise_sequence_subsequences)
export(summarise_sequence_transitions)
export(summarise_time_varying_sequence_model)
export(summarise_transition_centrality)
export(test_sequence_group_difference)
export(validate_sequence_clusters)
export(validate_sequence_data)

# gp3sequences <a href="https://stefanosbalaskas.github.io/gp3sequences/"><img src="man/figures/logo.png" align="right" height="139" alt="gp3sequences website" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/gp3sequences)](https://CRAN.R-project.org/package=gp3sequences)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/gp3sequences)](https://CRAN.R-project.org/package=gp3sequences)
[![R-CMD-check](https://github.com/stefanosbalaskas/gp3sequences/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/stefanosbalaskas/gp3sequences/actions/workflows/R-CMD-check.yaml)
[![pkgdown](https://github.com/stefanosbalaskas/gp3sequences/actions/workflows/pkgdown.yaml/badge.svg)](https://github.com/stefanosbalaskas/gp3sequences/actions/workflows/pkgdown.yaml)
[![CRAN
DOI](https://img.shields.io/badge/DOI-10.32614%2FCRAN.package.gp3sequences-blue.svg)](https://doi.org/10.32614/CRAN.package.gp3sequences)
[![License:
MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
<!-- badges: end -->

**Links:** [CRAN](https://CRAN.R-project.org/package=gp3sequences) ·
[Package website](https://stefanosbalaskas.github.io/gp3sequences/) ·
[Source code](https://github.com/stefanosbalaskas/gp3sequences) · [Issue
tracker](https://github.com/stefanosbalaskas/gp3sequences/issues) ·
[CRAN
checks](https://CRAN.R-project.org/web/checks/check_results_gp3sequences.html)

`gp3sequences` is an independent R package for transparent,
reproducible, and auditable analysis of ordered categorical sequences.
It is designed for ordinary long-format data frames and is not
restricted to Gazepoint exports, eye-tracking data, particular hardware,
or proprietary software.

## Installation

Install the current CRAN release:

``` r
install.packages("gp3sequences")
```

Install the current release version from GitHub:

``` r
pak::pak("stefanosbalaskas/gp3sequences")
```

The CRAN release is version **0.1.0**, published on **30 July 2026**.
The source repository currently contains release version **0.3.0**.

## Current release status

The current release version is **0.3.0**, building on the
**0.2.0** release. The package currently exposes **81 public functions**
and the website provides **15 synthetic, reproducible articles**.

The 0.3.0 release combines the audited sequence-data contract with
consensus and group comparisons, sequence distances, clustering and
stability workflows, transition networks, higher-order models,
categorical and mixture hidden Markov models, longitudinal and panel
workflows, bounded non-contiguous subsequences, time-varying models,
multichannel and covariate-dependent HMMs, design-aware inference,
extended visualisations, optional specialist-package adapters, and
analysis-contract/provenance auditing.

## Intended applications

Potential applications include:

- eye-tracking AOI scanpaths;
- website-navigation and customer-journey sequences;
- behavioural-state and task-action sequences;
- interface-interaction sequences;
- educational activity sequences;
- other repeated ordered categorical data.

## Neutral long-format input

The core data contract will require configurable columns corresponding
to `sequence_id`, `sequence_order`, and `state`. Optional participant,
trial, condition, group, timing, duration, weight, and metadata columns
may also be mapped explicitly.

A minimal input may look like this:

``` r
example_sequences <- data.frame(
  sequence_id = rep(c("s1", "s2"), each = 4L),
  sequence_order = rep(1:4, times = 2L),
  state = c(
    "home", "search", "product", "checkout",
    "home", "category", "product", "home"
  ),
  stringsAsFactors = FALSE
)

example_sequences
#>   sequence_id sequence_order    state
#> 1          s1              1     home
#> 2          s1              2   search
#> 3          s1              3  product
#> 4          s1              4 checkout
#> 5          s2              1     home
#> 6          s2              2 category
#> 7          s2              3  product
#> 8          s2              4     home
```

## Current data-contract API

The core sequence-data contract provides three related functions:

- `audit_sequence_data()` reports structured data-quality issues;
- `validate_sequence_data()` returns a compact validity contract;
- `prepare_sequence_data()` applies explicit preprocessing policies and
  creates deterministic canonical sequence data.

``` r
audit <- audit_sequence_data(
  example_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state"
)

validation <- validate_sequence_data(
  example_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state"
)

prepared <- prepare_sequence_data(
  example_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  repeated_state_policy = "preserve"
)
```

## Analysis contracts and provenance

The 0.3.0 release adds three public functions for machine-readable capability, contract, provenance, and reproducibility auditing:

- `sequence_capabilities()` inventories native, adapter, reference,
  handoff, and development capabilities without loading optional backend
  namespaces;
- `audit_sequence_analysis()` validates supported analysis objects
  against explicit structural contracts and recovers provenance where
  possible;
- `compare_sequence_analysis_results()` compares contracts, provenance,
  settings, and optionally result values across two analysis objects.

These functions complement the sequence-data contract and provide the
common foundation for the 0.3.0 hardening program.

## Encoding and structural summaries

Related functions support deterministic encoding and descriptive
sequence summaries without assigning substantive meaning to states:

- `encode_sequence_data()` creates a transparent state dictionary;
- `summarise_sequence_states()` reports state frequencies and durations;
- `summarise_sequence_transitions()` reports adjacent transition counts;
- `format_sequence_paths()` creates compact ordered path strings.

``` r
encoded <- encode_sequence_data(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state"
)

state_summary <- summarise_sequence_states(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state"
)

transition_summary <- summarise_sequence_transitions(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state"
)

paths <- format_sequence_paths(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state"
)
```

## Contiguous motif analysis

The motif workflow enumerates contiguous state windows only and retains
explicit structural definitions:

- `extract_sequence_ngrams()` extracts auditable motif occurrences;
- `summarise_sequence_motifs()` reports counts and sequence prevalence;
- `filter_sequence_motifs()` applies explicit transparent thresholds;
- `format_sequence_motifs()` creates stable report-ready tables.

``` r
motif_occurrences <- extract_sequence_ngrams(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  min_length = 2,
  max_length = 4,
  overlap = "allow"
)

motif_summary <- summarise_sequence_motifs(motif_occurrences)

motif_filter <- filter_sequence_motifs(
  motif_summary,
  min_sequences = 2,
  min_prevalence = 0.10,
  top_n = 20,
  ties = "include"
)

motif_table <- format_sequence_motifs(
  motif_filter,
  prevalence = "percent",
  digits = 1
)
```

### Motif positions and visualisation

Motif occurrences can be summarised and plotted without adding
analytical dependencies:

- `summarise_sequence_motif_positions()` reports structural locations;
- `format_sequence_motif_positions()` produces explicit report tables;
- `plot_sequence_motifs()` plots prevalence or occurrence metrics;
- `plot_sequence_motif_positions()` plots occurrence locations.

``` r
position_summary <- summarise_sequence_motif_positions(
  motif_occurrences,
  position = "centre",
  scale = "relative"
)

position_table <- format_sequence_motif_positions(
  position_summary,
  position_units = "percent",
  digits = 1
)

plot_sequence_motifs(
  motif_summary,
  metric = "sequence_prevalence",
  top_n = 10
)

plot_sequence_motif_positions(
  motif_occurrences,
  position = "centre",
  scale = "relative",
  top_n = 5
)
```

## Documentation

The package website now provides **15 synthetic, reproducible
articles**:

- getting started with the complete structural workflow;
- sequence-data validation and policy-driven preparation;
- choosing among motifs, consensus, distances, clustering, networks,
  higher-order models, HMMs, and adapters;
- a reproducible integrated synthetic case study;
- contiguous motif analysis and motif positions;
- consensus sequences and descriptive group comparisons;
- sequence distances, clustering, representatives, ensembles, and
  stability;
- transition networks and higher-order models;
- latent sequence models and optional ecosystem adapters;
- longitudinal and panel sequence workflows;
- non-contiguous subsequence mining;
- time-varying condition models;
- multichannel and covariate hidden Markov models;
- design-aware sequence inference and randomization;
- extended sequence visualisations.

The articles use ordinary data frames and synthetic data. They preserve
the package interpretation boundary and do not infer psychological or
causal attributes from structural sequence outputs.

## Current release extensions

The 0.3.0 release includes the following explicitly governed workflows:

- longitudinal and panel sequence preparation, summaries, change
  metrics, and plots;
- bounded non-contiguous subsequence mining with explicit gap, span, and
  combination limits;
- optional time-varying state and transition models using `mgcv`;
- multichannel and covariate-dependent categorical hidden Markov models;
- design-aware permutation and bootstrap group comparisons;
- sequence-index, state-distribution, entropy, distance, network, and
  silhouette plots.

These extensions preserve ordinary-data-frame inputs, explicit method
parameters, deterministic seeds, and structural interpretation
boundaries. Inferential helpers record the independent unit and study
design; observational contrasts remain associational, and randomized
contrasts require valid assignment and implementation.

## Interpretation boundary

Sequence outputs describe behavioural or structural patterns only. They
do not independently establish emotion, cognition, comprehension,
personality, intention, diagnosis, deception, or other psychological
attributes. Substantive interpretation requires an appropriate study
design and external evidence.

## Author

**Stefanos Balaskas**

[ORCID](https://orcid.org/0000-0003-2444-9796) · [Personal
website](https://sites.google.com/view/stefbalaskas/) ·
[LinkedIn](https://www.linkedin.com/in/stefanos-balaskas/)

## Licence

`gp3sequences` is released under the MIT License.

<!-- advanced-roadmap-start -->

## Advanced sequence-analysis methods

The 0.2.0 release established:

- aligned-position consensus sequences and agreement diagnostics;
- descriptive between-group comparisons;
- edit, LCS, optimal-matching, and transition-profile distances;
- clustering, validation, representatives, ensembles, and stability
  analysis;
- transition networks, centrality, communities, and higher-order models;
- categorical HMMs, mixture HMMs, and state decoding;
- optional adapters to specialist sequence-analysis packages.

The 0.3.0 release extends that foundation with longitudinal and panel workflows, bounded non-contiguous subsequences, time-varying models, multichannel and covariate HMMs, design-aware inference, extended visualisations, and analysis-contract/provenance auditing.

These outputs are structural and statistical. They do not independently
establish emotion, cognition, comprehension, personality, intention,
diagnosis, deception, causality, or other psychological attributes.
<!-- advanced-roadmap-end -->

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)

## ----data---------------------------------------------------------------------
library(gp3sequences)

sequence_data <- data.frame(
  sequence = c(
    rep("s1", 5L),
    rep("s2", 4L),
    rep("s3", 3L)
  ),
  order = c(1:5, 1:4, 1:3),
  state = c(
    "A", "B", "A", "B", "C",
    "A", "B", "C", "B",
    "B", "A", "B"
  ),
  group = c(
    rep("g1", 5L),
    rep("g1", 4L),
    rep("g2", 3L)
  ),
  stringsAsFactors = FALSE
)

sequence_data

## ----extract------------------------------------------------------------------
extracted <- extract_sequence_ngrams(
  sequence_data,
  sequence_id_col = "sequence",
  order_col = "order",
  state_col = "state",
  metadata_cols = "group",
  min_length = 2L,
  max_length = 3L,
  overlap = "allow"
)

head(extracted$occurrences)

## ----summarise-filter-format--------------------------------------------------
motif_summary <- summarise_sequence_motifs(extracted)

motif_filter <- filter_sequence_motifs(
  motif_summary,
  min_occurrences = 2L,
  min_sequences = 2L,
  min_prevalence = 0.50,
  motif_lengths = c(2L, 3L),
  top_n = 10L,
  rank_by = "sequence_prevalence",
  ties = "include"
)

motif_table <- format_sequence_motifs(
  motif_filter,
  prevalence = "percent",
  digits = 1L
)

motif_table$table

## ----position-summary---------------------------------------------------------
position_summary <- summarise_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  by = "group"
)

position_summary$summary

## ----position-format----------------------------------------------------------
position_table <- format_sequence_motif_positions(
  position_summary,
  digits = 1L,
  position_units = "percent",
  include_rank = TRUE
)

position_table$table

## ----motif-plot, fig.cap="Sequence prevalence for the most frequent contiguous motifs."----
plotted_motifs <- plot_sequence_motifs(
  motif_summary,
  metric = "sequence_prevalence",
  top_n = 8L,
  motif_lengths = c(2L, 3L),
  ties = "include",
  horizontal = TRUE
)

plotted_motifs[
  c("plot_rank", "motif", "motif_length", "plot_value")
]

## ----position-strip, fig.cap="Relative centre positions of selected motif occurrences."----
strip_data <- plot_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  top_n = 5L,
  display = "strip"
)

head(
  strip_data[
    c(
      "sequence_id",
      "motif",
      "position_value",
      "plot_rank"
    )
  ]
)

## ----position-distribution, fig.cap="Distributions of relative motif centre positions."----
plot_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  top_n = 5L,
  display = "distribution"
)

#' Declare a sequence group-comparison design
#'
#' Records the unit of assignment and the design under which a sequence-group
#' contrast will be evaluated. Causal interpretation is permitted only for an
#' explicitly randomized design and still depends on the validity of the study
#' implementation.
#'
#' @param group_col Group or treatment column.
#' @param unit_col Independent assignment or analysis-unit column.
#' @param design `"observational"`, `"randomized"`, or
#'   `"paired_randomized"`.
#' @param pair_col Pair or block column required for paired randomization.
#' @param cluster_col Optional higher-level assignment cluster.
#'
#' @return An object of class `gp3_sequence_comparison_design`.
#' @examples
#' declare_sequence_comparison_design("group", "participant_id",
#'                                    design = "randomized")
#' @export
declare_sequence_comparison_design <- function(
  group_col,
  unit_col,
  design = c("observational", "randomized", "paired_randomized"),
  pair_col = NULL,
  cluster_col = NULL
) {
  design <- match.arg(design)
  .sequence_adv_scalar_character(group_col, "group_col")
  .sequence_adv_scalar_character(unit_col, "unit_col")
  .sequence_adv_scalar_character(pair_col, "pair_col", allow_null = TRUE)
  .sequence_adv_scalar_character(cluster_col, "cluster_col", allow_null = TRUE)
  if (design == "paired_randomized" && is.null(pair_col)) {
    stop("`pair_col` is required for a paired randomized design.", call. = FALSE)
  }
  result <- list(
    group_col = group_col,
    unit_col = unit_col,
    design = design,
    pair_col = pair_col,
    cluster_col = cluster_col,
    interpretation = if (design == "observational") "associational" else "randomization-based",
    call = match.call()
  )
  class(result) <- c("gp3_sequence_comparison_design", "list")
  result
}

.sequence_ext_sequence_metric <- function(
  data,
  design,
  sequence_id_col,
  order_col,
  state_col,
  metric,
  target_state,
  target_subsequence,
  separator
) {
  metadata_cols <- unique(c(design$group_col, design$unit_col,
                            design$pair_col, design$cluster_col))
  metadata_cols <- metadata_cols[!is.na(metadata_cols)]
  x <- .sequence_adv_data(
    data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = metadata_cols,
    missing_state_policy = "error"
  )
  meta <- x$metadata
  names(meta)[names(meta) == sequence_id_col] <- "sequence_id"
  values <- vapply(x$sequence_ids, function(id) {
    sequence <- x$sequences[[id]]
    switch(
      metric,
      sequence_length = length(sequence),
      transition_count = max(length(sequence) - 1L, 0L),
      state_prevalence = mean(sequence == target_state),
      subsequence_presence = as.numeric(
        .sequence_ext_subsequence_present(
          sequence,
          strsplit(target_subsequence, separator, fixed = TRUE)[[1L]]
        )
      )
    )
  }, numeric(1))
  metric_data <- data.frame(
    sequence_id = x$sequence_ids,
    metric = as.numeric(values[x$sequence_ids]),
    stringsAsFactors = FALSE
  )
  merged <- merge(metric_data, meta, by = "sequence_id", sort = FALSE)
  unit_cols <- unique(c(design$unit_col, design$group_col,
                        design$pair_col, design$cluster_col))
  unit_cols <- unit_cols[!is.na(unit_cols)]
  key <- .sequence_adv_group_key(merged, unit_cols)
  pieces <- split(seq_len(nrow(merged)), key, drop = TRUE)
  unit_data <- do.call(rbind, lapply(pieces, function(rows) {
    prefix <- merged[rows[1L], unit_cols, drop = FALSE]
    cbind(prefix, data.frame(
      metric = mean(merged$metric[rows]),
      n_sequences = length(rows),
      stringsAsFactors = FALSE,
      check.names = FALSE
    ))
  }))
  row.names(unit_data) <- NULL
  list(sequence_data = merged, unit_data = unit_data, state_levels = x$state_levels)
}

.sequence_ext_difference <- function(metric, group, levels) {
  means <- tapply(metric, factor(group, levels = levels), mean)
  unname(means[2L] - means[1L])
}

.sequence_ext_permuted_groups <- function(unit_data, design, group_levels) {
  group <- as.character(unit_data[[design$group_col]])
  if (design$design == "paired_randomized") {
    pair <- as.character(unit_data[[design$pair_col]])
    output <- group
    for (value in unique(pair)) {
      rows <- which(pair == value)
      if (length(rows) != 2L || !setequal(group[rows], group_levels)) {
        stop("Each randomized pair must contain exactly one unit from each group.",
             call. = FALSE)
      }
      if (stats::runif(1) < 0.5) output[rows] <- rev(group[rows])
    }
    return(output)
  }
  assignment_col <- if (is.null(design$cluster_col)) design$unit_col else design$cluster_col
  assignment <- as.character(unit_data[[assignment_col]])
  cluster_group <- tapply(group, assignment, function(x) {
    value <- unique(x)
    if (length(value) != 1L) {
      stop("Assignment clusters must have one group label.", call. = FALSE)
    }
    value
  })
  permuted <- sample(cluster_group, length(cluster_group), replace = FALSE)
  names(permuted) <- names(cluster_group)
  unname(permuted[assignment])
}

#' Test a sequence group difference
#'
#' Aggregates sequence metrics to the declared independent unit and performs a
#' permutation or randomization test. For observational data the p-value tests
#' exchangeability-based association only; it is not a causal estimate.
#'
#' @param data Long-format sequence data.
#' @param design A comparison design.
#' @param metric `"sequence_length"`, `"transition_count"`,
#'   `"state_prevalence"`, or `"subsequence_presence"`.
#' @param target_state Required for state prevalence.
#' @param target_subsequence Required for subsequence presence, expressed using
#'   `separator`.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param separator Subsequence label separator.
#' @param n_permutations Number of permutations.
#' @param alternative Alternative hypothesis.
#' @param seed Reproducibility seed.
#'
#' @return An object of class `gp3_sequence_group_inference`.
#' @examples
#' data <- data.frame(
#'   participant_id = rep(paste0("p", 1:8), each = 4L),
#'   sequence_id = rep(paste0("s", 1:8), each = 4L),
#'   sequence_order = rep(1:4, times = 8L),
#'   state = c(rep(c("A", "B", "C", "D"), 4L),
#'             rep(c("A", "A", "C", "D"), 4L)),
#'   group = rep(rep(c("control", "treatment"), each = 4L), each = 4L)
#' )
#' design <- declare_sequence_comparison_design("group", "participant_id",
#'                                              design = "randomized")
#' test_sequence_group_difference(data, design, metric = "state_prevalence",
#'                                target_state = "A", n_permutations = 99L)
#' @export
test_sequence_group_difference <- function(
  data,
  design,
  metric = c("sequence_length", "transition_count", "state_prevalence",
             "subsequence_presence"),
  target_state = NULL,
  target_subsequence = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  separator = " > ",
  n_permutations = 999L,
  alternative = c("two.sided", "greater", "less"),
  seed = 1L
) {
  if (!inherits(design, "gp3_sequence_comparison_design")) {
    stop("`design` must be created by `declare_sequence_comparison_design()`.",
         call. = FALSE)
  }
  metric <- match.arg(metric)
  alternative <- match.arg(alternative)
  if (metric == "state_prevalence") {
    .sequence_adv_scalar_character(target_state, "target_state")
  }
  if (metric == "subsequence_presence") {
    .sequence_adv_scalar_character(target_subsequence, "target_subsequence")
  }
  .sequence_adv_scalar_character(separator, "separator")
  .sequence_adv_scalar_number(n_permutations, "n_permutations", lower = 1,
                              integer = TRUE)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  metric_value <- .sequence_ext_sequence_metric(
    data, design, sequence_id_col, order_col, state_col, metric,
    target_state, target_subsequence, separator
  )
  unit_data <- metric_value$unit_data
  groups <- as.character(unit_data[[design$group_col]])
  if (anyNA(groups) || any(!nzchar(trimws(groups)))) {
    stop("Group values must not be missing or blank.", call. = FALSE)
  }
  group_levels <- sort(unique(groups), method = "radix")
  if (length(group_levels) != 2L) {
    stop("The current inferential contrast requires exactly two groups.",
         call. = FALSE)
  }
  observed <- .sequence_ext_difference(unit_data$metric, groups, group_levels)
  permuted <- numeric(n_permutations)
  .sequence_adv_with_seed(seed, {
    for (i in seq_len(n_permutations)) {
      current_group <- .sequence_ext_permuted_groups(unit_data, design, group_levels)
      permuted[i] <- .sequence_ext_difference(unit_data$metric, current_group,
                                              group_levels)
    }
  })
  extreme <- switch(
    alternative,
    two.sided = abs(permuted) >= abs(observed),
    greater = permuted >= observed,
    less = permuted <= observed
  )
  p_value <- (1 + sum(extreme)) / (n_permutations + 1)
  means <- tapply(unit_data$metric, factor(groups, levels = group_levels), mean)
  result <- list(
    estimate = data.frame(
      group_1 = group_levels[1L],
      group_2 = group_levels[2L],
      mean_group_1 = unname(means[1L]),
      mean_group_2 = unname(means[2L]),
      difference_group_2_minus_group_1 = observed,
      p_value = p_value,
      alternative = alternative,
      n_permutations = as.integer(n_permutations),
      stringsAsFactors = FALSE
    ),
    permutation_distribution = permuted,
    unit_data = unit_data,
    sequence_data = metric_value$sequence_data,
    design = design,
    metric = metric,
    target_state = target_state,
    target_subsequence = target_subsequence,
    interpretation = if (design$design == "observational") {
      "Associational permutation contrast; causal interpretation is not supported."
    } else {
      "Randomization-based contrast, conditional on valid assignment and study implementation."
    },
    seed = as.integer(seed),
    call = match.call()
  )
  class(result) <- c("gp3_sequence_group_inference", "list")
  result
}

#' Bootstrap a sequence group difference
#'
#' Resamples declared independent units within groups and returns a percentile
#' interval for the mean difference. This interval does not by itself license
#' causal interpretation.
#'
#' @param inference A result from [test_sequence_group_difference()].
#' @param n_boot Number of bootstrap samples.
#' @param level Confidence level.
#' @param seed Reproducibility seed.
#'
#' @return The inference object with a `bootstrap` component.
#' @examples
#' # See `test_sequence_group_difference()`.
#' @export
bootstrap_sequence_group_difference <- function(
  inference,
  n_boot = 999L,
  level = 0.95,
  seed = 1L
) {
  if (!inherits(inference, "gp3_sequence_group_inference")) {
    stop("`inference` must be created by `test_sequence_group_difference()`.",
         call. = FALSE)
  }
  .sequence_adv_scalar_number(n_boot, "n_boot", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(level, "level", lower = 0.5, upper = 0.999999)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  unit_data <- inference$unit_data
  groups <- as.character(unit_data[[inference$design$group_col]])
  group_levels <- c(inference$estimate$group_1, inference$estimate$group_2)
  boot <- numeric(n_boot)
  .sequence_adv_with_seed(seed, {
    for (b in seq_len(n_boot)) {
      means <- numeric(2L)
      for (g in seq_along(group_levels)) {
        values <- unit_data$metric[groups == group_levels[g]]
        means[g] <- mean(sample(values, length(values), replace = TRUE))
      }
      boot[b] <- means[2L] - means[1L]
    }
  })
  alpha <- (1 - level) / 2
  interval <- stats::quantile(boot, probs = c(alpha, 1 - alpha), names = FALSE,
                              type = 6)
  inference$bootstrap <- list(
    estimates = boot,
    interval = data.frame(
      level = level,
      lower = interval[1L],
      upper = interval[2L],
      n_boot = as.integer(n_boot),
      seed = as.integer(seed),
      stringsAsFactors = FALSE
    )
  )
  inference
}

#' Summarise sequence group inference
#'
#' @param inference A sequence group-inference object.
#'
#' @return A list containing the estimate, optional interval, design, and
#'   interpretation statement.
#' @examples
#' # See `test_sequence_group_difference()`.
#' @export
summarise_sequence_group_inference <- function(inference) {
  if (!inherits(inference, "gp3_sequence_group_inference")) {
    stop("`inference` must be created by `test_sequence_group_difference()`.",
         call. = FALSE)
  }
  list(
    estimate = inference$estimate,
    bootstrap_interval = if (is.null(inference$bootstrap)) NULL else inference$bootstrap$interval,
    design = data.frame(
      design = inference$design$design,
      group_col = inference$design$group_col,
      unit_col = inference$design$unit_col,
      pair_col = if (is.null(inference$design$pair_col)) NA_character_ else inference$design$pair_col,
      cluster_col = if (is.null(inference$design$cluster_col)) NA_character_ else inference$design$cluster_col,
      stringsAsFactors = FALSE
    ),
    metric = inference$metric,
    interpretation = inference$interpretation
  )
}

#' Plot sequence group inference
#'
#' @param inference A sequence group-inference object.
#' @param type `"permutation"` or `"group_means"`.
#' @param ... Additional base-graphics arguments.
#'
#' @return The inference object, invisibly.
#' @examples
#' # See `test_sequence_group_difference()`.
#' @export
plot_sequence_group_inference <- function(
  inference,
  type = c("permutation", "group_means"),
  ...
) {
  if (!inherits(inference, "gp3_sequence_group_inference")) {
    stop("`inference` must be created by `test_sequence_group_difference()`.",
         call. = FALSE)
  }
  type <- match.arg(type)
  if (type == "permutation") {
    observed <- inference$estimate$difference_group_2_minus_group_1
    graphics::hist(inference$permutation_distribution,
                   xlab = "Permuted mean difference", main = "Permutation distribution", ...)
    graphics::abline(v = observed, lwd = 2, lty = 2)
  } else {
    means <- c(inference$estimate$mean_group_1,
               inference$estimate$mean_group_2)
    names(means) <- c(inference$estimate$group_1, inference$estimate$group_2)
    graphics::barplot(means, ylab = inference$metric, ...)
  }
  invisible(inference)
}

test_that("sequence inference records the design and resamples independent units", {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  data <- make_extension_sequence_data()
  design <- declare_sequence_comparison_design(
    group_col = "group",
    unit_col = "participant_id",
    design = "observational"
  )
  expect_s3_class(design, "gp3_sequence_comparison_design")
  inference <- test_sequence_group_difference(
    data,
    design,
    metric = "state_prevalence",
    target_state = "A",
    n_permutations = 49L,
    seed = 4L
  )
  expect_s3_class(inference, "gp3_sequence_group_inference")
  expect_equal(nrow(inference$unit_data), 8L)
  expect_true(inference$estimate$p_value >= 0 && inference$estimate$p_value <= 1)
  expect_match(inference$interpretation, "Associational")
  inference <- bootstrap_sequence_group_difference(
    inference, n_boot = 49L, seed = 5L
  )
  expect_equal(nrow(inference$bootstrap$interval), 1L)
  summary <- summarise_sequence_group_inference(inference)
  expect_true(all(c("estimate", "bootstrap_interval", "design", "interpretation") %in% names(summary)))
  expect_invisible(plot_sequence_group_inference(inference, type = "group_means"))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----data---------------------------------------------------------------------
paths <- list(
  s1 = c("A", "A", "B", "B", "C"),
  s2 = c("A", "B", "B", "C", "C"),
  s3 = c("C", "C", "B", "B", "A"),
  s4 = c("C", "B", "B", "A", "A"),
  s5 = c("A", "A", "B", "C", "C"),
  s6 = c("C", "C", "B", "A", "A")
)
data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(
    sequence_id = names(paths)[i],
    sequence_order = seq_along(paths[[i]]),
    state = paths[[i]],
    context = c("x", "x", "y", "y", "z"),
    condition = as.integer(i > 3L),
    stringsAsFactors = FALSE
  )
}))

## ----multichannel-------------------------------------------------------------
multi <- fit_multichannel_sequence_hmm(
  data,
  n_states = 2L,
  channel_cols = c("state", "context"),
  max_iter = 15L,
  seed = 2L
)
summarise_multichannel_sequence_hmm(multi)$fit
head(decode_multichannel_sequence_states(multi))

## ----emission-plot, fig.width=7, fig.height=4---------------------------------
plot_multichannel_sequence_hmm(multi, channel = "state")

## ----covariate----------------------------------------------------------------
covariate <- fit_covariate_sequence_hmm(
  data,
  n_states = 2L,
  initial_covariate_cols = "condition",
  transition_covariate_cols = "condition",
  max_iter = 10L,
  inner_maxit = 30L,
  seed = 3L
)
summarise_covariate_sequence_hmm(covariate)$fit
predict_covariate_transition_probabilities(
  covariate,
  data.frame(condition = c(0, 1))
)
head(decode_covariate_sequence_states(covariate))

#' Plot a sequence index heatmap
#'
#' @param data Long-format sequence data or a prepared result.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param sort_by `"input"`, `"length"`, or `"path"`.
#' @param state_levels Optional state ordering.
#' @param palette Base HCL palette.
#' @param show_sequence_labels Draw sequence identifiers.
#' @param ... Additional arguments passed to [graphics::image()].
#'
#' @return The plotted state-code matrix, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "C", "D"))
#' plot_sequence_index(data)
#' @export
plot_sequence_index <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  sort_by = c("input", "length", "path"),
  state_levels = NULL,
  palette = "Dark 3",
  show_sequence_labels = TRUE,
  ...
) {
  sort_by <- match.arg(sort_by)
  .sequence_adv_scalar_logical(show_sequence_labels, "show_sequence_labels")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  levels_value <- .sequence_adv_state_order(unlist(x$sequences, use.names = FALSE),
                                            state_levels)
  ids <- x$sequence_ids
  if (sort_by == "length") {
    ids <- ids[order(lengths(x$sequences)[ids], ids, method = "radix")]
  } else if (sort_by == "path") {
    paths <- vapply(x$sequences, paste, collapse = "\034", character(1))
    ids <- ids[order(paths[ids], ids, method = "radix")]
  }
  max_length <- max(lengths(x$sequences))
  matrix_value <- matrix(NA_integer_, nrow = length(ids), ncol = max_length,
                         dimnames = list(ids, seq_len(max_length)))
  for (i in seq_along(ids)) {
    sequence <- x$sequences[[ids[i]]]
    matrix_value[i, seq_along(sequence)] <- match(sequence, levels_value)
  }
  colors <- .sequence_ext_base_colors(length(levels_value), palette)
  graphics::image(
    x = seq_len(ncol(matrix_value)),
    y = seq_len(nrow(matrix_value)),
    z = t(matrix_value[nrow(matrix_value):1L, , drop = FALSE]),
    col = colors,
    xlab = "Sequence position",
    ylab = "Sequence",
    yaxt = "n",
    ...
  )
  if (show_sequence_labels) {
    graphics::axis(2, at = seq_len(length(ids)), labels = rev(ids), las = 2)
  }
  graphics::legend("topright", legend = levels_value, fill = colors, bty = "n")
  invisible(matrix_value)
}

#' Plot state distributions over aligned positions
#'
#' @param data Long-format sequence data or a prepared result.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param proportion Plot position-wise proportions rather than counts.
#' @param state_levels Optional state ordering.
#' @param palette Base HCL palette.
#' @param ... Additional arguments passed to [graphics::matplot()].
#'
#' @return A position-by-state matrix, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "C", "D"))
#' plot_sequence_state_distribution(data)
#' @export
plot_sequence_state_distribution <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  proportion = TRUE,
  state_levels = NULL,
  palette = "Dark 3",
  ...
) {
  .sequence_adv_scalar_logical(proportion, "proportion")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  levels_value <- .sequence_adv_state_order(unlist(x$sequences, use.names = FALSE),
                                            state_levels)
  positions <- sort(unique(x$data[[order_col]]))
  matrix_value <- matrix(0, nrow = length(positions), ncol = length(levels_value),
                         dimnames = list(as.character(positions), levels_value))
  for (i in seq_along(positions)) {
    current <- as.character(x$data[[state_col]][x$data[[order_col]] == positions[i]])
    matrix_value[i, ] <- table(factor(current, levels = levels_value))
    if (proportion && sum(matrix_value[i, ]) > 0) {
      matrix_value[i, ] <- matrix_value[i, ] / sum(matrix_value[i, ])
    }
  }
  colors <- .sequence_ext_base_colors(ncol(matrix_value), palette)
  graphics::matplot(positions, matrix_value, type = "l", lty = 1,
                    col = colors, xlab = "Sequence position",
                    ylab = if (proportion) "State proportion" else "State count", ...)
  graphics::legend("topright", legend = colnames(matrix_value), col = colors,
                   lty = 1, bty = "n")
  invisible(matrix_value)
}

#' Plot position-wise state entropy
#'
#' @param data Long-format sequence data or a prepared result.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param base Logarithm base.
#' @param normalise Divide entropy by the maximum entropy for the observed
#'   alphabet.
#' @param ... Additional arguments passed to [graphics::plot()].
#'
#' @return A data frame of position-wise entropy values, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "C", "D"))
#' plot_sequence_entropy(data)
#' @export
plot_sequence_entropy <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  base = 2,
  normalise = TRUE,
  ...
) {
  .sequence_adv_scalar_number(base, "base", lower = 1 + .Machine$double.eps)
  .sequence_adv_scalar_logical(normalise, "normalise")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  positions <- sort(unique(x$data[[order_col]]))
  entropy <- vapply(positions, function(position) {
    current <- as.character(x$data[[state_col]][x$data[[order_col]] == position])
    probabilities <- as.numeric(table(current)) / length(current)
    value <- .sequence_ext_entropy(probabilities, base = base)
    if (normalise && length(x$state_levels) > 1L) {
      value <- value / (log(length(x$state_levels)) / log(base))
    }
    value
  }, numeric(1))
  result <- data.frame(position = positions, entropy = entropy,
                       normalised = normalise, stringsAsFactors = FALSE)
  graphics::plot(result$position, result$entropy, type = "b",
                 xlab = "Sequence position",
                 ylab = if (normalise) "Normalised entropy" else "Entropy", ...)
  invisible(result)
}

#' Plot a sequence-distance heatmap
#'
#' @param distance A sequence distance object or square matrix.
#' @param order_by Optional clustering or named assignment vector used to order
#'   the matrix.
#' @param palette Base HCL palette.
#' @param show_labels Draw sequence identifiers.
#' @param ... Additional arguments passed to [graphics::image()].
#'
#' @return The ordered distance matrix, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(paste0("s", 1:4), each = 4L),
#'                    sequence_order = rep(1:4, 4L),
#'                    state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'                              "D", "C", "B", "A", "D", "C", "A", "A"))
#' plot_sequence_distance_heatmap(compute_sequence_distance(data))
#' @export
plot_sequence_distance_heatmap <- function(
  distance,
  order_by = NULL,
  palette = "Viridis",
  show_labels = TRUE,
  ...
) {
  .sequence_adv_scalar_logical(show_labels, "show_labels")
  matrix_value <- .sequence_adv_distance_matrix(distance)
  ids <- rownames(matrix_value)
  if (!is.null(order_by)) {
    assignments <- if (inherits(order_by, "gp3_sequence_clustering")) {
      order_by$assignments
    } else order_by
    if (is.null(names(assignments)) || !setequal(names(assignments), ids)) {
      stop("`order_by` assignments must be named for every sequence.", call. = FALSE)
    }
    ids <- ids[order(assignments[ids], ids, method = "radix")]
    matrix_value <- matrix_value[ids, ids, drop = FALSE]
  }
  colors <- .sequence_ext_base_colors(64L, palette)
  graphics::image(seq_len(nrow(matrix_value)), seq_len(ncol(matrix_value)),
                  t(matrix_value[nrow(matrix_value):1L, , drop = FALSE]),
                  col = colors, xlab = "Sequence", ylab = "Sequence",
                  xaxt = "n", yaxt = "n", ...)
  if (show_labels) {
    graphics::axis(1, at = seq_along(ids), labels = ids, las = 2)
    graphics::axis(2, at = seq_along(ids), labels = rev(ids), las = 2)
  }
  invisible(matrix_value)
}

#' Plot a first-order transition network
#'
#' @param network A first-order network from [create_transition_network()].
#' @param weight_col Edge-weight column.
#' @param minimum_weight Minimum plotted edge weight.
#' @param vertex_cex Vertex label size.
#' @param edge_scale Edge-width multiplier.
#' @param ... Additional arguments passed to [graphics::plot.window()].
#'
#' @return The plotted network rows, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "C", "D"))
#' plot_transition_network(create_transition_network(data, normalise = "from"))
#' @export
plot_transition_network <- function(
  network,
  weight_col = "weight",
  minimum_weight = 0,
  vertex_cex = 1,
  edge_scale = 5,
  ...
) {
  if (!inherits(network, "gp3_transition_network")) {
    stop("`network` must be created by `create_transition_network()`.", call. = FALSE)
  }
  settings <- attr(network, "settings")
  if (!identical(settings$order, 1L)) {
    stop("Only first-order networks can be plotted with this helper.", call. = FALSE)
  }
  if (length(unique(network$group_key)) > 1L) {
    stop("Filter a grouped network to one group before plotting.", call. = FALSE)
  }
  if (!(weight_col %in% names(network)) || !is.numeric(network[[weight_col]])) {
    stop("`weight_col` must name a numeric network column.", call. = FALSE)
  }
  .sequence_adv_scalar_number(minimum_weight, "minimum_weight", lower = 0)
  .sequence_adv_scalar_number(vertex_cex, "vertex_cex", lower = 0)
  .sequence_adv_scalar_number(edge_scale, "edge_scale", lower = 0)
  edges <- network[network[[weight_col]] >= minimum_weight, , drop = FALSE]
  states <- sort(unique(c(edges$from_state, edges$to_state)), method = "radix")
  if (length(states) == 0L) stop("No edges satisfy the plotting threshold.", call. = FALSE)
  angle <- seq(0, 2 * pi, length.out = length(states) + 1L)[seq_len(length(states))]
  coordinates <- cbind(x = cos(angle), y = sin(angle))
  rownames(coordinates) <- states
  graphics::plot.new()
  graphics::plot.window(xlim = c(-1.3, 1.3), ylim = c(-1.3, 1.3), asp = 1, ...)
  for (i in seq_len(nrow(edges))) {
    from <- coordinates[edges$from_state[i], ]
    to <- coordinates[edges$to_state[i], ]
    width <- max(0.5, edges[[weight_col]][i] * edge_scale)
    if (identical(edges$from_state[i], edges$to_state[i])) {
      graphics::symbols(from[1], from[2] + 0.12, circles = 0.12,
                        inches = FALSE, add = TRUE, lwd = width)
    } else {
      graphics::arrows(from[1], from[2], to[1], to[2], length = 0.08,
                       lwd = width)
    }
  }
  graphics::points(coordinates[, 1], coordinates[, 2], pch = 21, bg = "white",
                   cex = 2)
  graphics::text(coordinates[, 1], coordinates[, 2], labels = states,
                 cex = vertex_cex)
  invisible(edges)
}

#' Plot sequence-cluster silhouette values
#'
#' @param clustering A clustering result or named assignment vector.
#' @param distance Optional distance when `clustering` is an assignment vector.
#' @param ... Additional arguments passed to [graphics::barplot()].
#'
#' @return The ordered per-sequence silhouette table, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(paste0("s", 1:4), each = 4L),
#'                    sequence_order = rep(1:4, 4L),
#'                    state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'                              "D", "C", "B", "A", "D", "C", "A", "A"))
#' distance <- compute_sequence_distance(data)
#' plot_sequence_cluster_silhouette(cluster_sequences(distance, 2L))
#' @export
plot_sequence_cluster_silhouette <- function(clustering, distance = NULL, ...) {
  validation <- validate_sequence_clusters(clustering, distance)
  current <- validation$per_sequence
  current <- current[order(current$cluster, -current$silhouette,
                           current$sequence_id, method = "radix"), , drop = FALSE]
  colors <- .sequence_ext_base_colors(length(unique(current$cluster)))
  color_map <- stats::setNames(colors, sort(unique(current$cluster), method = "radix"))
  graphics::barplot(current$silhouette,
                    names.arg = current$sequence_id,
                    col = color_map[current$cluster],
                    las = 2, ylab = "Silhouette", ...)
  graphics::abline(h = 0, lty = 2)
  invisible(current)
}

test_that("extended visualisations accept core package objects", {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  data <- make_extension_sequence_data()
  distance <- compute_sequence_distance(data, method = "levenshtein")
  clustering <- cluster_sequences(distance, k = 2L, method = "hierarchical")
  network <- create_transition_network(data)

  expect_invisible(plot_sequence_index(data))
  expect_invisible(plot_sequence_state_distribution(data))
  expect_invisible(plot_sequence_entropy(data))
  expect_invisible(plot_sequence_distance_heatmap(distance))
  expect_invisible(plot_transition_network(network))
  expect_invisible(plot_sequence_cluster_silhouette(clustering, distance))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----problematic-data---------------------------------------------------------
problem_data <- data.frame(
  sequence_id = c("s2", "s1", "s1", "s1", "s1", "s2", "s2", "s2", "s2"),
  sequence_order = c(2, 2, 1, 2, 3, 1, 2, 3, 4),
  state = factor(
    c("search", "search", "home", "search", NA,
      "home", "home", "product", "other"),
    levels = c("home", "search", "product", "checkout", "other", "unused")
  ),
  duration = c(120, 100, 90, 110, 80, 100, 0, 150, 130),
  participant_id = c("p2", "p1", "p1", "p1", "p1", "p2", "p2", "p2", "p2"),
  group = c("interface_b", "interface_a", "interface_a", "interface_a",
            "interface_a", "interface_b", "interface_b", "interface_b",
            "interface_b"),
  stringsAsFactors = FALSE
)

expected_states <- c("home", "search", "product", "checkout")
problem_data

## ----audit--------------------------------------------------------------------
audit <- audit_sequence_data(
  problem_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states
)

audit
as.data.frame(table(audit$severity), stringsAsFactors = FALSE)
as.data.frame(table(audit$issue_code), stringsAsFactors = FALSE)

## ----validation---------------------------------------------------------------
validation <- validate_sequence_data(
  problem_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states
)

validation[c("valid", "status", "n_errors", "n_reviews", "n_info")]
validation$mapping

## ----prepare------------------------------------------------------------------
prepared <- prepare_sequence_data(
  problem_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states,
  missing_state_policy = "drop",
  duplicate_position_policy = "first",
  repeated_state_policy = "collapse",
  zero_duration_policy = "drop",
  unknown_state_policy = "drop",
  unused_state_levels = "drop"
)

prepared$status
prepared$decisions
prepared$data
prepared$audit

## ----revalidate---------------------------------------------------------------
revalidation <- validate_sequence_data(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states
)

revalidation[c("valid", "status", "n_errors", "n_reviews", "n_info")]
prepared$mapping
prepared$state_levels

.sequence_motif_position_empty_occurrences <- function(by) {
  result <- data.frame(
    sequence_id = character(),
    motif_id = character(),
    motif_key = character(),
    motif = character(),
    motif_length = integer(),
    start_index = integer(),
    end_index = integer(),
    n_states = integer(),
    position_basis = character(),
    position_scale = character(),
    absolute_position = double(),
    relative_position = double(),
    position_value = double(),
    stringsAsFactors = FALSE
  )

  for (column in by) {
    result[[column]] <- character()
  }

  result[
    c(
      "sequence_id",
      by,
      "motif_id",
      "motif_key",
      "motif",
      "motif_length",
      "start_index",
      "end_index",
      "n_states",
      "position_basis",
      "position_scale",
      "absolute_position",
      "relative_position",
      "position_value"
    )
  ]
}

.sequence_motif_position_empty_summary <- function(by) {
  result <- data.frame(
    motif_id = character(),
    motif_key = character(),
    motif = character(),
    motif_length = integer(),
    position_basis = character(),
    position_scale = character(),
    n_occurrences = integer(),
    n_sequences = integer(),
    min_position = double(),
    max_position = double(),
    mean_position = double(),
    median_position = double(),
    stringsAsFactors = FALSE
  )

  for (column in by) {
    result[[column]] <- character()
  }

  result[
    c(
      by,
      "motif_id",
      "motif_key",
      "motif",
      "motif_length",
      "position_basis",
      "position_scale",
      "n_occurrences",
      "n_sequences",
      "min_position",
      "max_position",
      "mean_position",
      "median_position"
    )
  ]
}

.sequence_motif_position_validate_by <- function(by, occurrences) {
  if (is.null(by)) {
    return(character())
  }

  if (
    !is.character(by) ||
      length(by) == 0L ||
      anyNA(by) ||
      any(!nzchar(by)) ||
      anyDuplicated(by)
  ) {
    stop(
      "`by` must be `NULL` or a unique character vector of column names.",
      call. = FALSE
    )
  }

  reserved <- c(
    "sequence_id",
    "motif_id",
    "motif_key",
    "motif",
    "motif_length",
    "start_index",
    "end_index",
    "start_order",
    "end_order",
    "start_original_row",
    "end_original_row",
    "occurrence_index"
  )

  reserved_by <- intersect(by, reserved)

  if (length(reserved_by) > 0L) {
    stop(
      "`by` must contain preserved metadata columns, not reserved motif columns: ",
      paste(reserved_by, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  missing_by <- setdiff(by, names(occurrences))

  if (length(missing_by) > 0L) {
    stop(
      "The following `by` columns were not preserved during extraction: ",
      paste(missing_by, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  by
}

.sequence_motif_group_factor <- function(data, columns) {
  if (length(columns) == 0L) {
    return(factor(rep("all", nrow(data))))
  }

  values <- lapply(
    data[columns],
    function(value) {
      addNA(factor(value, exclude = NULL), ifany = FALSE)
    }
  )

  do.call(
    interaction,
    c(values, list(drop = TRUE, lex.order = TRUE))
  )
}

.sequence_motif_position_sort <- function(data, by) {
  if (nrow(data) == 0L) {
    return(data)
  }

  order_arguments <- list()

  for (column in by) {
    value <- data[[column]]

    if (is.factor(value)) {
      value <- as.character(value)
    }

    order_arguments[[length(order_arguments) + 1L]] <- value
  }

  order_arguments[[length(order_arguments) + 1L]] <- data$mean_position
  order_arguments[[length(order_arguments) + 1L]] <- data$median_position
  order_arguments[[length(order_arguments) + 1L]] <- -data$n_occurrences
  order_arguments[[length(order_arguments) + 1L]] <- -data$n_sequences
  order_arguments[[length(order_arguments) + 1L]] <- data$motif_length
  order_arguments[[length(order_arguments) + 1L]] <- data$motif_key
  order_arguments$method <- "radix"

  data <- data[do.call(order, order_arguments), , drop = FALSE]
  row.names(data) <- NULL
  data
}

.sequence_motif_position_validate_summary <- function(x) {
  if (
    !is.list(x) ||
      is.null(x$summary) ||
      is.null(x$occurrences) ||
      is.null(x$settings)
  ) {
    stop(
      "`x` must be an object returned by `summarise_sequence_motif_positions()`.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_motif_position_values <- function(
  occurrences,
  sequences,
  position,
  scale
) {
  if (nrow(occurrences) == 0L) {
    result <- occurrences
    result$n_states <- integer()
    result$position_basis <- character()
    result$position_scale <- character()
    result$absolute_position <- double()
    result$relative_position <- double()
    result$position_value <- double()
    return(result)
  }

  sequence_match <- match(
    occurrences$sequence_id,
    sequences$sequence_id
  )

  if (anyNA(sequence_match)) {
    stop(
      "Some motif occurrences do not match the extraction sequence table.",
      call. = FALSE
    )
  }

  n_states <- as.integer(sequences$n_states[sequence_match])

  absolute_position <- switch(
    position,
    start = as.numeric(occurrences$start_index),
    centre = (
      as.numeric(occurrences$start_index) +
        as.numeric(occurrences$end_index)
    ) / 2,
    end = as.numeric(occurrences$end_index)
  )

  relative_position <- ifelse(
    n_states <= 1L,
    0,
    (absolute_position - 1) / (n_states - 1)
  )

  relative_position <- pmin(1, pmax(0, relative_position))

  result <- occurrences
  result$n_states <- n_states
  result$position_basis <- position
  result$position_scale <- scale
  result$absolute_position <- absolute_position
  result$relative_position <- relative_position
  result$position_value <- if (scale == "relative") {
    relative_position
  } else {
    absolute_position
  }

  result
}

.sequence_motif_plot_empty <- function(message, main, xlab = "") {
  graphics::plot.new()
  graphics::title(main = main, xlab = xlab)
  graphics::text(0.5, 0.5, labels = message)
  invisible(NULL)
}

.sequence_motif_plot_metric_label <- function(metric) {
  switch(
    metric,
    sequence_prevalence = "Sequence prevalence (proportion)",
    n_occurrences = "Occurrence count",
    n_sequences = "Sequences containing motif",
    occurrence_share = "Occurrence share (proportion)"
  )
}

.sequence_motif_plot_selectors <- function(motifs, dictionary) {
  if (is.null(motifs)) {
    return(NULL)
  }

  if (
    !is.character(motifs) ||
      length(motifs) == 0L ||
      anyNA(motifs) ||
      any(!nzchar(motifs))
  ) {
    stop(
      "`motifs` must be `NULL` or a non-empty character vector.",
      call. = FALSE
    )
  }

  selected_ids <- character()
  unknown <- character()

  for (selector in motifs) {
    matched <- dictionary$motif_id[
      dictionary$motif_id == selector |
        dictionary$motif_key == selector |
        dictionary$motif == selector
    ]

    matched <- unique(matched)

    if (length(matched) == 0L) {
      unknown <- c(unknown, selector)
    } else {
      selected_ids <- c(selected_ids, matched)
    }
  }

  if (length(unknown) > 0L) {
    stop(
      "The following requested motifs were not found: ",
      paste(unique(unknown), collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  unique(selected_ids)
}

# Shared internal plot-data contract. It is deliberately not exported.
.sequence_motif_plot_data <- function(
  x,
  type = c("motifs", "positions"),
  metric = "sequence_prevalence",
  motif_lengths = NULL,
  top_n = 20L,
  ties = c("include", "first"),
  motifs = NULL,
  position = "start",
  scale = "relative"
) {
  type <- .sequence_motif_match_arg(
    type,
    c("motifs", "positions"),
    "type"
  )

  .sequence_motif_assert_whole_number(
    top_n,
    "top_n",
    minimum = 1L
  )

  if (type == "motifs") {
    metric <- .sequence_motif_match_arg(
      metric,
      c(
        "sequence_prevalence",
        "n_occurrences",
        "n_sequences",
        "occurrence_share"
      ),
      "metric"
    )

    .sequence_motif_assert_motif_lengths(motif_lengths)

    ties <- .sequence_motif_match_arg(
      ties,
      c("include", "first"),
      "ties"
    )

    summary <- .sequence_motif_as_summary(x)
    data <- summary$overall

    if (!is.null(motif_lengths)) {
      data <- data[
        data$motif_length %in% sort(unique(as.integer(motif_lengths))),
        ,
        drop = FALSE
      ]
    }

    data <- .sequence_motif_sort_overall(data, primary = metric)

    if (nrow(data) > top_n) {
      if (ties == "include") {
        threshold <- data[[metric]][top_n]
        data <- data[data[[metric]] >= threshold, , drop = FALSE]
      } else {
        data <- data[seq_len(top_n), , drop = FALSE]
      }
    }

    row.names(data) <- NULL
    data$plot_rank <- seq_len(nrow(data))
    data$plot_value <- as.numeric(data[[metric]])
    data$plot_label <- as.character(data$motif)

    return(data)
  }

  position <- .sequence_motif_match_arg(
    position,
    c("start", "centre", "end"),
    "position"
  )

  scale <- .sequence_motif_match_arg(
    scale,
    c("absolute", "relative"),
    "scale"
  )

  if (
    is.list(x) &&
      !is.null(x$summary) &&
      !is.null(x$occurrences) &&
      !is.null(x$settings)
  ) {
    positional <- x
    occurrences <- positional$occurrences

    required <- c(
      "sequence_id",
      "motif_id",
      "motif_key",
      "motif",
      "motif_length",
      "start_index",
      "end_index",
      "n_states"
    )

    missing_required <- setdiff(required, names(occurrences))

    if (length(missing_required) > 0L) {
      stop(
        "The positional summary is missing required occurrence columns: ",
        paste(missing_required, collapse = ", "),
        ".",
        call. = FALSE
      )
    }

    absolute_position <- switch(
      position,
      start = as.numeric(occurrences$start_index),
      centre = (
        as.numeric(occurrences$start_index) +
          as.numeric(occurrences$end_index)
      ) / 2,
      end = as.numeric(occurrences$end_index)
    )

    relative_position <- ifelse(
      occurrences$n_states <= 1L,
      0,
      (absolute_position - 1) / (occurrences$n_states - 1)
    )

    relative_position <- pmin(1, pmax(0, relative_position))

    occurrences$position_basis <- position
    occurrences$position_scale <- scale
    occurrences$absolute_position <- absolute_position
    occurrences$relative_position <- relative_position
    occurrences$position_value <- if (scale == "relative") {
      relative_position
    } else {
      absolute_position
    }
  } else {
    .sequence_motif_validate_extraction(x)

    positional <- summarise_sequence_motif_positions(
      x,
      position = position,
      scale = scale
    )

    occurrences <- positional$occurrences
  }

  if (nrow(occurrences) == 0L) {
    occurrences$plot_rank <- integer()
    occurrences$plot_label <- character()
    return(occurrences)
  }

  dictionary <- unique(
    occurrences[
      c("motif_id", "motif_key", "motif", "motif_length")
    ]
  )

  selected_ids <- .sequence_motif_plot_selectors(
    motifs,
    dictionary
  )

  occurrence_groups <- split(
    seq_len(nrow(occurrences)),
    occurrences$motif_id
  )

  motif_rows <- lapply(
    occurrence_groups,
    function(indices) {
      first <- indices[1L]

      data.frame(
        motif_id = occurrences$motif_id[first],
        motif_key = occurrences$motif_key[first],
        motif = occurrences$motif[first],
        motif_length = as.integer(occurrences$motif_length[first]),
        n_occurrences = as.integer(length(indices)),
        n_sequences = as.integer(
          length(unique(occurrences$sequence_id[indices]))
        ),
        stringsAsFactors = FALSE
      )
    }
  )

  motif_table <- do.call(rbind, motif_rows)
  row.names(motif_table) <- NULL

  if (is.null(selected_ids)) {
    motif_table <- motif_table[
      order(
        -motif_table$n_occurrences,
        -motif_table$n_sequences,
        motif_table$motif_length,
        motif_table$motif_key,
        method = "radix"
      ),
      ,
      drop = FALSE
    ]

    if (nrow(motif_table) > top_n) {
      motif_table <- motif_table[seq_len(top_n), , drop = FALSE]
    }

    selected_ids <- motif_table$motif_id
  } else {
    motif_table <- motif_table[
      motif_table$motif_id %in% selected_ids,
      ,
      drop = FALSE
    ]

    motif_table <- motif_table[
      match(selected_ids, motif_table$motif_id),
      ,
      drop = FALSE
    ]
  }

  row.names(motif_table) <- NULL
  motif_table$plot_rank <- seq_len(nrow(motif_table))

  occurrences <- occurrences[
    occurrences$motif_id %in% selected_ids,
    ,
    drop = FALSE
  ]

  occurrences$plot_rank <- match(
    occurrences$motif_id,
    selected_ids
  )
  occurrences$plot_label <- motif_table$motif[
    match(occurrences$motif_id, motif_table$motif_id)
  ]

  occurrences <- occurrences[
    order(
      occurrences$plot_rank,
      occurrences$position_value,
      occurrences$sequence_id,
      occurrences$start_index,
      occurrences$end_index,
      method = "radix"
    ),
    ,
    drop = FALSE
  ]

  row.names(occurrences) <- NULL
  attr(occurrences, "motif_table") <- motif_table
  occurrences
}

#' Summarise Sequence Motif Positions
#'
#' Summarises where contiguous motif occurrences appear within sequences.
#'
#' @param x An object returned by `extract_sequence_ngrams()`.
#' @param position Position represented by each occurrence: motif `"start"`,
#'   `"centre"`, or `"end"`.
#' @param scale Position scale: one-based `"absolute"` sequence positions or
#'   `"relative"` positions from 0 to 1.
#' @param by Optional character vector naming preserved metadata columns used
#'   to produce separate summaries, such as `"group"` or `"condition"`.
#'
#' @return A named list containing:
#'
#' * `summary`: one row per motif and optional metadata group;
#' * `occurrences`: occurrence-level absolute, relative, and selected-scale
#'   positions;
#' * `sequences`, validation metadata, and extraction settings;
#' * `settings`: the resolved position basis, scale, and grouping columns.
#'
#' The summary reports motif identifiers and labels, motif length, occurrence
#' and sequence counts, and minimum, maximum, mean, and median positions.
#'
#' @details
#' Absolute positions use the one-based state index within each validated
#' sequence. Relative positions are calculated as
#' `(absolute_position - 1) / (n_states - 1)` and are constrained to the
#' interval from 0 to 1. A sequence containing one state is assigned relative
#' position 0.
#'
#' Grouping is descriptive. The function does not test differences or attach
#' behavioural, psychological, cognitive, or causal interpretations to motif
#' location.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "C", "A", "B", "C", "B"),
#'   group = c(rep("g1", 5L), rep("g2", 4L))
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state",
#'   metadata_cols = "group",
#'   min_length = 2,
#'   max_length = 3
#' )
#'
#' positions <- summarise_sequence_motif_positions(
#'   extracted,
#'   position = "centre",
#'   scale = "relative",
#'   by = "group"
#' )
#'
#' positions$summary
#'
#' @export
summarise_sequence_motif_positions <- function(
  x,
  position = c("start", "centre", "end"),
  scale = c("absolute", "relative"),
  by = NULL
) {
  .sequence_motif_validate_extraction(x)

  position <- .sequence_motif_match_arg(
    position,
    c("start", "centre", "end"),
    "position"
  )

  scale <- .sequence_motif_match_arg(
    scale,
    c("absolute", "relative"),
    "scale"
  )

  by <- .sequence_motif_position_validate_by(
    by,
    x$occurrences
  )

  occurrences <- .sequence_motif_position_values(
    x$occurrences,
    x$sequences,
    position,
    scale
  )

  occurrence_columns <- c(
    "sequence_id",
    by,
    "motif_id",
    "motif_key",
    "motif",
    "motif_length",
    "start_index",
    "end_index",
    "n_states",
    "position_basis",
    "position_scale",
    "absolute_position",
    "relative_position",
    "position_value"
  )

  if (nrow(occurrences) == 0L) {
    occurrence_positions <- .sequence_motif_position_empty_occurrences(by)
    summary <- .sequence_motif_position_empty_summary(by)
  } else {
    occurrence_positions <- occurrences[, occurrence_columns, drop = FALSE]

    group_columns <- c(by, "motif_id")
    groups <- split(
      seq_len(nrow(occurrence_positions)),
      .sequence_motif_group_factor(
        occurrence_positions,
        group_columns
      )
    )

    summary_rows <- lapply(
      groups,
      function(indices) {
        first <- indices[1L]
        group_values <- vector("list", length(by))
        names(group_values) <- by

        for (column in by) {
          group_values[[column]] <- occurrence_positions[[column]][first]
        }

        values <- occurrence_positions$position_value[indices]

        as.data.frame(
          c(
            group_values,
            list(
              motif_id = occurrence_positions$motif_id[first],
              motif_key = occurrence_positions$motif_key[first],
              motif = occurrence_positions$motif[first],
              motif_length = as.integer(
                occurrence_positions$motif_length[first]
              ),
              position_basis = position,
              position_scale = scale,
              n_occurrences = as.integer(length(indices)),
              n_sequences = as.integer(
                length(unique(occurrence_positions$sequence_id[indices]))
              ),
              min_position = min(values),
              max_position = max(values),
              mean_position = mean(values),
              median_position = stats::median(values)
            )
          ),
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      }
    )

    summary <- do.call(rbind, summary_rows)
    row.names(summary) <- NULL
    summary <- .sequence_motif_position_sort(summary, by)

    occurrence_order <- list()

    for (column in by) {
      value <- occurrence_positions[[column]]

      if (is.factor(value)) {
        value <- as.character(value)
      }

      occurrence_order[[length(occurrence_order) + 1L]] <- value
    }

    occurrence_order[[length(occurrence_order) + 1L]] <-
      occurrence_positions$motif_length
    occurrence_order[[length(occurrence_order) + 1L]] <-
      occurrence_positions$motif_key
    occurrence_order[[length(occurrence_order) + 1L]] <-
      occurrence_positions$position_value
    occurrence_order[[length(occurrence_order) + 1L]] <-
      occurrence_positions$sequence_id
    occurrence_order[[length(occurrence_order) + 1L]] <-
      occurrence_positions$start_index
    occurrence_order[[length(occurrence_order) + 1L]] <-
      occurrence_positions$end_index
    occurrence_order$method <- "radix"

    occurrence_positions <- occurrence_positions[
      do.call(order, occurrence_order),
      ,
      drop = FALSE
    ]

    row.names(occurrence_positions) <- NULL
  }

  list(
    summary = summary,
    occurrences = occurrence_positions,
    sequences = x$sequences,
    state_dictionary = x$state_dictionary,
    audit = x$audit,
    status = x$status,
    mapping = x$mapping,
    extraction_settings = x$settings,
    settings = list(
      position = position,
      scale = scale,
      by = by,
      absolute_unit = "one_based_state_index",
      relative_range = c(0, 1)
    ),
    n_occurrences = as.integer(nrow(occurrence_positions)),
    n_motifs = as.integer(length(unique(occurrence_positions$motif_id))),
    n_groups = if (nrow(summary) == 0L) {
      0L
    } else if (length(by) == 0L) {
      1L
    } else {
      as.integer(
        length(unique(.sequence_motif_group_factor(summary, by)))
      )
    }
  )
}

#' Format Sequence Motif Position Summaries
#'
#' Produces a deterministic report-ready table from positional motif summaries.
#'
#' @param x An object returned by `summarise_sequence_motif_positions()`.
#' @param digits Whole number from 0 to 15 controlling numeric rounding.
#' @param position_units Display units for relative positions: `"proportion"`
#'   or `"percent"`. Absolute positions remain one-based sequence indices.
#' @param include_rank Logical value indicating whether to include an
#'   earlier-to-later rank based on mean position. Ranking is performed within
#'   each supplied `by` group.
#'
#' @return A named list containing the formatted `table`, validation metadata,
#' and formatting settings.
#'
#' @details
#' Formatting copies and transforms the summary table only. The input object is
#' not modified. Rows are ordered deterministically by grouping values, mean
#' and median position, occurrence and sequence counts, motif length, and motif
#' key. Equal mean positions receive the same minimum rank.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "C", "A", "B", "C", "B")
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' positions <- summarise_sequence_motif_positions(
#'   extracted,
#'   position = "centre",
#'   scale = "relative"
#' )
#'
#' formatted <- format_sequence_motif_positions(
#'   positions,
#'   position_units = "percent",
#'   digits = 1
#' )
#'
#' formatted$table
#'
#' @export
format_sequence_motif_positions <- function(
  x,
  digits = 3L,
  position_units = c("proportion", "percent"),
  include_rank = TRUE
) {
  .sequence_motif_position_validate_summary(x)

  .sequence_motif_assert_whole_number(
    digits,
    "digits",
    minimum = 0L
  )

  if (digits > 15L) {
    stop("`digits` must not exceed 15.", call. = FALSE)
  }

  .sequence_assert_flag(include_rank, "include_rank")

  position_units <- .sequence_motif_match_arg(
    position_units,
    c("proportion", "percent"),
    "position_units"
  )

  by <- x$settings$by
  table <- x$summary
  table <- .sequence_motif_position_sort(table, by)

  if (include_rank) {
    rank_values <- integer(nrow(table))

    if (nrow(table) > 0L) {
      groups <- if (length(by) == 0L) {
        list(seq_len(nrow(table)))
      } else {
        split(
          seq_len(nrow(table)),
          .sequence_motif_group_factor(table, by)
        )
      }

      for (indices in groups) {
        rank_values[indices] <- as.integer(
          rank(
            table$mean_position[indices],
            ties.method = "min"
          )
        )
      }
    }

    table$rank <- rank_values
  }

  position_columns <- c(
    "min_position",
    "max_position",
    "mean_position",
    "median_position"
  )

  applied_units <- if (identical(x$settings$scale, "relative")) {
    position_units
  } else {
    "index"
  }

  if (
    nrow(table) > 0L &&
      identical(x$settings$scale, "relative") &&
      identical(position_units, "percent")
  ) {
    for (column in position_columns) {
      table[[column]] <- 100 * table[[column]]
    }
  }

  for (column in position_columns) {
    table[[column]] <- round(table[[column]], digits = digits)
  }

  table$position_unit <- applied_units

  front <- c(
    by,
    if (include_rank) "rank",
    "motif_id",
    "motif_key",
    "motif",
    "motif_length",
    "position_basis",
    "position_scale",
    "position_unit",
    "n_occurrences",
    "n_sequences"
  )

  table <- table[c(front, setdiff(names(table), front))]
  row.names(table) <- NULL

  list(
    table = table,
    audit = x$audit,
    status = x$status,
    mapping = x$mapping,
    source_settings = x$settings,
    settings = list(
      digits = as.integer(digits),
      requested_position_units = position_units,
      applied_position_units = applied_units,
      include_rank = include_rank,
      rank_metric = "mean_position",
      rank_direction = "earlier_to_later"
    )
  )
}

#' Plot Sequence Motif Summaries
#'
#' Draws a dependency-free base-R bar chart of structural motif measurements.
#'
#' @param x A motif extraction, summary, or filtered-motif object.
#' @param metric Structural metric to plot: `"sequence_prevalence"`,
#'   `"n_occurrences"`, `"n_sequences"`, or `"occurrence_share"`.
#' @param top_n Positive whole number giving the requested number of motifs.
#' @param motif_lengths Optional vector of positive whole-number motif lengths
#'   to retain.
#' @param ties Top-`n` boundary policy. `"include"` retains all motifs tied on
#'   the selected metric; `"first"` retains exactly `top_n` after deterministic
#'   secondary sorting.
#' @param horizontal Logical value indicating whether bars should be horizontal.
#'
#' @return Invisibly returns the exact motif table used for plotting, including
#' `plot_rank`, `plot_value`, `plot_label`, and `bar_midpoint`.
#'
#' @details
#' The plot is descriptive. It does not perform inferential testing or assign
#' substantive meaning to motif frequency or prevalence. Empty inputs produce
#' an informative blank plot and return an empty table.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "C", "A", "B", "C", "B")
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' plot_sequence_motifs(
#'   extracted,
#'   metric = "sequence_prevalence",
#'   top_n = 10
#' )
#'
#' @export
plot_sequence_motifs <- function(
  x,
  metric = c(
    "sequence_prevalence",
    "n_occurrences",
    "n_sequences",
    "occurrence_share"
  ),
  top_n = 20L,
  motif_lengths = NULL,
  ties = c("include", "first"),
  horizontal = TRUE
) {
  metric <- .sequence_motif_match_arg(
    metric,
    c(
      "sequence_prevalence",
      "n_occurrences",
      "n_sequences",
      "occurrence_share"
    ),
    "metric"
  )

  ties <- .sequence_motif_match_arg(
    ties,
    c("include", "first"),
    "ties"
  )

  .sequence_assert_flag(horizontal, "horizontal")

  plot_data <- .sequence_motif_plot_data(
    x = x,
    type = "motifs",
    metric = metric,
    motif_lengths = motif_lengths,
    top_n = top_n,
    ties = ties
  )

  axis_label <- .sequence_motif_plot_metric_label(metric)

  if (nrow(plot_data) == 0L) {
    .sequence_motif_plot_empty(
      "No motifs match the requested plotting settings.",
      "Sequence motifs",
      axis_label
    )
    plot_data$bar_midpoint <- double()
    return(invisible(plot_data))
  }

  old_parameters <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old_parameters), add = TRUE)

  if (horizontal) {
    label_width <- max(nchar(plot_data$plot_label), na.rm = TRUE)
    left_margin <- min(16, max(5.1, 3.5 + 0.13 * label_width))

    graphics::par(
      mar = c(5.1, left_margin, 4.1, 2.1)
    )

    reverse_index <- rev(seq_len(nrow(plot_data)))

    midpoint <- graphics::barplot(
      height = plot_data$plot_value[reverse_index],
      names.arg = plot_data$plot_label[reverse_index],
      horiz = TRUE,
      las = 1,
      xlab = axis_label,
      ylab = "",
      main = "Sequence motifs"
    )

    plot_data$bar_midpoint <- rev(as.numeric(midpoint))
  } else {
    bottom_margin <- min(
      16,
      max(6.1, 4.5 + 0.13 * max(nchar(plot_data$plot_label)))
    )

    graphics::par(
      mar = c(bottom_margin, 5.1, 4.1, 2.1)
    )

    midpoint <- graphics::barplot(
      height = plot_data$plot_value,
      names.arg = plot_data$plot_label,
      horiz = FALSE,
      las = 2,
      xlab = "",
      ylab = axis_label,
      main = "Sequence motifs"
    )

    plot_data$bar_midpoint <- as.numeric(midpoint)
  }

  invisible(plot_data)
}

#' Plot Sequence Motif Positions
#'
#' Shows occurrence locations for selected contiguous motifs.
#'
#' @param x An object returned by `extract_sequence_ngrams()` or
#'   `summarise_sequence_motif_positions()`.
#' @param motifs Optional character vector of motif identifiers, motif keys, or
#'   displayed motif labels. When omitted, motifs are selected deterministically
#'   by occurrence count, sequence count, motif length, and motif key.
#' @param position Occurrence position represented by motif `"start"`,
#'   `"centre"`, or `"end"`.
#' @param scale Position scale: one-based `"absolute"` sequence positions or
#'   `"relative"` positions from 0 to 1.
#' @param top_n Positive whole number giving the number of motifs selected when
#'   `motifs` is `NULL`.
#' @param display Base-R display type: occurrence `"strip"` or
#'   `"distribution"` boxplots.
#'
#' @return Invisibly returns the exact occurrence-level data used by the plot,
#' including deterministic motif ranks and plotting coordinates where relevant.
#'
#' @details
#' The strip display uses deterministic vertical stacking rather than random
#' jitter. The distribution display uses one horizontal boxplot per motif.
#' Empty inputs produce an informative blank plot and return an empty table.
#' The function provides structural location summaries only.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "C", "A", "B", "C", "B")
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' plot_sequence_motif_positions(
#'   extracted,
#'   position = "centre",
#'   scale = "relative",
#'   top_n = 5,
#'   display = "strip"
#' )
#'
#' @export
plot_sequence_motif_positions <- function(
  x,
  motifs = NULL,
  position = c("start", "centre", "end"),
  scale = c("absolute", "relative"),
  top_n = 10L,
  display = c("strip", "distribution")
) {
  position <- .sequence_motif_match_arg(
    position,
    c("start", "centre", "end"),
    "position"
  )

  scale <- .sequence_motif_match_arg(
    scale,
    c("absolute", "relative"),
    "scale"
  )

  display <- .sequence_motif_match_arg(
    display,
    c("strip", "distribution"),
    "display"
  )

  plot_data <- .sequence_motif_plot_data(
    x = x,
    type = "positions",
    motifs = motifs,
    position = position,
    scale = scale,
    top_n = top_n
  )

  position_label <- paste0(
    toupper(substr(position, 1L, 1L)),
    substring(position, 2L)
  )

  axis_label <- if (scale == "relative") {
    paste0(
      position_label,
      " position within sequence (0 to 1)"
    )
  } else {
    paste0(
      position_label,
      " position (one-based state index)"
    )
  }

  if (nrow(plot_data) == 0L) {
    .sequence_motif_plot_empty(
      "No motif occurrences match the requested plotting settings.",
      "Sequence motif positions",
      axis_label
    )
    plot_data$plot_y <- double()
    return(invisible(plot_data))
  }

  motif_table <- attr(plot_data, "motif_table")
  selected_ids <- motif_table$motif_id
  selected_labels <- motif_table$motif
  n_motifs <- nrow(motif_table)

  old_parameters <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old_parameters), add = TRUE)

  label_width <- max(nchar(selected_labels), na.rm = TRUE)
  left_margin <- min(16, max(5.1, 3.5 + 0.13 * label_width))

  graphics::par(
    mar = c(5.1, left_margin, 4.1, 2.1)
  )

  x_values <- plot_data$position_value

  if (scale == "relative") {
    x_limits <- c(0, 1)
  } else {
    value_range <- range(x_values, finite = TRUE)

    if (diff(value_range) == 0) {
      padding <- max(0.5, abs(value_range[1L]) * 0.04)
    } else {
      padding <- 0.04 * diff(value_range)
    }

    x_limits <- value_range + c(-padding, padding)
  }

  if (display == "strip") {
    base_y <- n_motifs - plot_data$plot_rank + 1L
    duplicate_group <- interaction(
      plot_data$motif_id,
      format(plot_data$position_value, digits = 15, trim = TRUE),
      drop = TRUE,
      lex.order = TRUE
    )

    stack_index <- stats::ave(
      seq_len(nrow(plot_data)),
      duplicate_group,
      FUN = seq_along
    )

    stack_size <- stats::ave(
      seq_len(nrow(plot_data)),
      duplicate_group,
      FUN = length
    )

    offset <- (stack_index - (stack_size + 1) / 2) * 0.06
    plot_data$plot_y <- base_y + offset

    graphics::plot(
      x = plot_data$position_value,
      y = plot_data$plot_y,
      xlim = x_limits,
      ylim = c(0.5, n_motifs + 0.5),
      xlab = axis_label,
      ylab = "",
      yaxt = "n",
      main = "Sequence motif positions",
      type = "n"
    )

    graphics::axis(
      side = 2,
      at = seq_len(n_motifs),
      labels = rev(selected_labels),
      las = 1
    )

    graphics::abline(
      h = seq_len(n_motifs),
      lty = 3
    )

    graphics::points(
      x = plot_data$position_value,
      y = plot_data$plot_y,
      pch = 16
    )
  } else {
    value_list <- lapply(
      rev(selected_ids),
      function(motif_id) {
        plot_data$position_value[plot_data$motif_id == motif_id]
      }
    )

    names(value_list) <- rev(selected_labels)

    graphics::boxplot(
      value_list,
      horizontal = TRUE,
      las = 1,
      xlim = x_limits,
      xlab = axis_label,
      ylab = "",
      main = "Sequence motif position distributions"
    )

    plot_data$plot_y <- n_motifs - plot_data$plot_rank + 1L
  }

  attr(plot_data, "motif_table") <- motif_table
  invisible(plot_data)
}

make_motif_visualisation_data <- function() {
  data.frame(
    id = c(
      rep("s1", 5L),
      rep("s2", 4L),
      rep("s3", 3L)
    ),
    position = c(1:5, 1:4, 1:3),
    state = c(
      "A", "B", "A", "B", "C",
      "A", "B", "C", "B",
      "B", "A", "B"
    ),
    group = c(
      rep("g1", 5L),
      rep("g1", 4L),
      rep("g2", 3L)
    ),
    stringsAsFactors = FALSE
  )
}

make_motif_visualisation_extraction <- function(
  min_length = 2L,
  max_length = 3L
) {
  extract_sequence_ngrams(
    make_motif_visualisation_data(),
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "group",
    min_length = min_length,
    max_length = max_length,
    overlap = "allow"
  )
}

capture_motif_plot <- function(expression) {
  path <- tempfile(fileext = ".pdf")
  grDevices::pdf(path)
  on.exit(grDevices::dev.off(), add = TRUE)

  value <- eval.parent(substitute(expression))

  list(
    value = value,
    path = path
  )
}

test_that("absolute motif positions are summarised exactly", {
  extracted <- make_motif_visualisation_extraction()

  result <- summarise_sequence_motif_positions(
    extracted,
    position = "start",
    scale = "absolute"
  )

  expect_identical(result$status, "pass")
  expect_identical(result$settings$position, "start")
  expect_identical(result$settings$scale, "absolute")
  expect_identical(result$settings$by, character())
  expect_equal(nrow(result$occurrences), nrow(extracted$occurrences))

  ab <- result$summary[
    result$summary$motif == "A > B",
    ,
    drop = FALSE
  ]

  expect_equal(nrow(ab), 1L)
  expect_identical(ab$n_occurrences, 4L)
  expect_identical(ab$n_sequences, 3L)
  expect_equal(ab$min_position, 1)
  expect_equal(ab$max_position, 3)
  expect_equal(ab$mean_position, 1.75)
  expect_equal(ab$median_position, 1.5)
  expect_true(
    all(result$occurrences$position_value ==
      result$occurrences$absolute_position)
  )
})

test_that("relative positions are bounded and use the requested basis", {
  extracted <- make_motif_visualisation_extraction()

  start <- summarise_sequence_motif_positions(
    extracted,
    position = "start",
    scale = "relative"
  )

  centre <- summarise_sequence_motif_positions(
    extracted,
    position = "centre",
    scale = "relative"
  )

  end <- summarise_sequence_motif_positions(
    extracted,
    position = "end",
    scale = "relative"
  )

  expect_true(all(start$occurrences$position_value >= 0))
  expect_true(all(start$occurrences$position_value <= 1))
  expect_true(all(centre$occurrences$position_value >= 0))
  expect_true(all(centre$occurrences$position_value <= 1))
  expect_true(all(end$occurrences$position_value >= 0))
  expect_true(all(end$occurrences$position_value <= 1))

  ab_start <- start$summary$mean_position[
    start$summary$motif == "A > B"
  ]

  ab_centre <- centre$summary$mean_position[
    centre$summary$motif == "A > B"
  ]

  ab_end <- end$summary$mean_position[
    end$summary$motif == "A > B"
  ]

  expect_equal(ab_start, 0.25)
  expect_equal(ab_centre, 5 / 12)
  expect_equal(ab_end, 7 / 12)
})

test_that("metadata grouping produces separate deterministic summaries", {
  extracted <- make_motif_visualisation_extraction()

  result <- summarise_sequence_motif_positions(
    extracted,
    position = "start",
    scale = "absolute",
    by = "group"
  )

  ab <- result$summary[
    result$summary$motif == "A > B",
    ,
    drop = FALSE
  ]

  expect_identical(ab$group, c("g1", "g2"))
  expect_identical(ab$n_occurrences, c(3L, 1L))
  expect_identical(ab$n_sequences, c(2L, 1L))
  expect_equal(ab$mean_position, c(5 / 3, 2))
  expect_identical(result$n_groups, 2L)
  expect_identical(result$settings$by, "group")
})

test_that("position summaries retain stable empty schemas", {
  extracted <- make_motif_visualisation_extraction(
    min_length = 8L,
    max_length = 8L
  )

  result <- summarise_sequence_motif_positions(
    extracted,
    position = "centre",
    scale = "relative",
    by = "group"
  )

  expected_summary <- c(
    "group",
    "motif_id",
    "motif_key",
    "motif",
    "motif_length",
    "position_basis",
    "position_scale",
    "n_occurrences",
    "n_sequences",
    "min_position",
    "max_position",
    "mean_position",
    "median_position"
  )

  expect_equal(nrow(result$summary), 0L)
  expect_equal(nrow(result$occurrences), 0L)
  expect_identical(names(result$summary), expected_summary)
  expect_identical(result$n_occurrences, 0L)
  expect_identical(result$n_motifs, 0L)
  expect_identical(result$n_groups, 0L)
})

test_that("position formatting changes display only", {
  positions <- summarise_sequence_motif_positions(
    make_motif_visualisation_extraction(),
    position = "start",
    scale = "relative",
    by = "group"
  )

  original <- positions

  formatted <- format_sequence_motif_positions(
    positions,
    digits = 1L,
    position_units = "percent",
    include_rank = TRUE
  )

  expect_identical(positions, original)
  expect_true("rank" %in% names(formatted$table))
  expect_true("position_unit" %in% names(formatted$table))
  expect_true(all(formatted$table$position_unit == "percent"))
  expect_identical(
    formatted$settings$applied_position_units,
    "percent"
  )

  ab_g1 <- formatted$table[
    formatted$table$group == "g1" &
      formatted$table$motif == "A > B",
    ,
    drop = FALSE
  ]

  expect_equal(ab_g1$mean_position, 16.7)
})

test_that("absolute positions remain indices during formatting", {
  positions <- summarise_sequence_motif_positions(
    make_motif_visualisation_extraction(),
    position = "centre",
    scale = "absolute"
  )

  formatted <- format_sequence_motif_positions(
    positions,
    digits = 2L,
    position_units = "percent",
    include_rank = FALSE
  )

  expect_false("rank" %in% names(formatted$table))
  expect_true(all(formatted$table$position_unit == "index"))
  expect_identical(
    formatted$settings$applied_position_units,
    "index"
  )
  expect_equal(
    formatted$table$mean_position,
    round(positions$summary$mean_position, 2L)
  )
})

test_that("motif plot data applies deterministic top-n ties", {
  summary <- summarise_sequence_motifs(
    make_motif_visualisation_extraction()
  )

  included <- .sequence_motif_plot_data(
    summary,
    type = "motifs",
    metric = "sequence_prevalence",
    top_n = 2L,
    ties = "include"
  )

  first <- .sequence_motif_plot_data(
    summary,
    type = "motifs",
    metric = "sequence_prevalence",
    top_n = 2L,
    ties = "first"
  )

  expect_true(nrow(included) > 2L)
  expect_equal(nrow(first), 2L)
  expect_identical(first$plot_rank, 1:2)
  expect_true(
    all(diff(first$sequence_prevalence) <= 0)
  )
})

test_that("motif bar plots return the exact plotted table", {
  summary <- summarise_sequence_motifs(
    make_motif_visualisation_extraction()
  )

  horizontal <- capture_motif_plot(
    plot_sequence_motifs(
      summary,
      metric = "n_occurrences",
      top_n = 3L,
      ties = "first",
      horizontal = TRUE
    )
  )

  vertical <- capture_motif_plot(
    plot_sequence_motifs(
      summary,
      metric = "occurrence_share",
      top_n = 3L,
      ties = "first",
      horizontal = FALSE
    )
  )

  expect_equal(nrow(horizontal$value), 3L)
  expect_equal(nrow(vertical$value), 3L)
  expect_true(all(is.finite(horizontal$value$bar_midpoint)))
  expect_true(all(is.finite(vertical$value$bar_midpoint)))
  expect_identical(horizontal$value$plot_rank, 1:3)
  expect_true(file.exists(horizontal$path))
  expect_true(file.exists(vertical$path))
})

test_that("motif plots handle empty filtered inputs", {
  empty <- filter_sequence_motifs(
    summarise_sequence_motifs(
      make_motif_visualisation_extraction()
    ),
    min_occurrences = 100L
  )

  plotted <- capture_motif_plot(
    plot_sequence_motifs(
      empty,
      top_n = 5L
    )
  )

  expect_equal(nrow(plotted$value), 0L)
  expect_true("bar_midpoint" %in% names(plotted$value))
  expect_true(file.exists(plotted$path))
})

test_that("strip position plots are deterministic and bounded", {
  extracted <- make_motif_visualisation_extraction()

  plotted <- capture_motif_plot(
    plot_sequence_motif_positions(
      extracted,
      position = "centre",
      scale = "relative",
      top_n = 2L,
      display = "strip"
    )
  )

  motif_table <- attr(plotted$value, "motif_table")

  expect_equal(nrow(motif_table), 2L)
  expect_true(all(plotted$value$position_value >= 0))
  expect_true(all(plotted$value$position_value <= 1))
  expect_true(all(is.finite(plotted$value$plot_y)))
  expect_identical(sort(unique(plotted$value$plot_rank)), 1:2)
  expect_true(file.exists(plotted$path))
})

test_that("position plots accept motif identifiers labels and summaries", {
  extracted <- make_motif_visualisation_extraction()
  positions <- summarise_sequence_motif_positions(
    extracted,
    position = "start",
    scale = "absolute"
  )

  motif_id <- positions$summary$motif_id[
    positions$summary$motif == "A > B"
  ][1L]

  by_label <- capture_motif_plot(
    plot_sequence_motif_positions(
      positions,
      motifs = "A > B",
      position = "end",
      scale = "absolute",
      display = "distribution"
    )
  )

  by_id <- capture_motif_plot(
    plot_sequence_motif_positions(
      extracted,
      motifs = motif_id,
      position = "start",
      scale = "relative",
      display = "strip"
    )
  )

  expect_identical(unique(by_label$value$motif), "A > B")
  expect_identical(unique(by_id$value$motif_id), motif_id)
  expect_true(all(by_label$value$position_basis == "end"))
  expect_true(all(by_id$value$position_scale == "relative"))
})

test_that("invalid positional and plotting settings are rejected", {
  extracted <- make_motif_visualisation_extraction()

  expect_error(
    summarise_sequence_motif_positions(
      extracted,
      position = "middle"
    ),
    "position"
  )

  expect_error(
    summarise_sequence_motif_positions(
      extracted,
      by = "missing_group"
    ),
    "not preserved"
  )

  positions <- summarise_sequence_motif_positions(extracted)

  expect_error(
    format_sequence_motif_positions(
      positions,
      digits = 16L
    ),
    "must not exceed"
  )

  expect_error(
    plot_sequence_motifs(
      extracted,
      top_n = 0L
    ),
    "top_n"
  )

  expect_error(
    plot_sequence_motif_positions(
      extracted,
      motifs = "not-a-motif"
    ),
    "not found"
  )

  expect_error(
    plot_sequence_motif_positions(
      extracted,
      display = "heatmap"
    ),
    "display"
  )
})

.sequence_assert_text_scalar <- function(value, argument) {
  if (
    !is.character(value) ||
    length(value) != 1L ||
    is.na(value) ||
    !nzchar(value)
  ) {
    stop(
      "`",
      argument,
      "` must be one non-missing, non-empty character value.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_assert_flag <- function(value, argument) {
  if (
    !is.logical(value) ||
    length(value) != 1L ||
    is.na(value)
  ) {
    stop(
      "`",
      argument,
      "` must be either `TRUE` or `FALSE`.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_assert_output_names <- function(
    metadata_cols,
    reserved,
    context
) {
  collisions <- intersect(metadata_cols, reserved)

  if (length(collisions) > 0L) {
    stop(
      "Metadata column names conflict with ",
      context,
      " output columns: ",
      paste(collisions, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_summary_input <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL,
    expected_states = NULL
) {
  validation <- validate_sequence_data(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    duration_col = duration_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  if (!validation$valid) {
    error_codes <- unique(
      validation$audit$issue_code[
        validation$audit$severity == "error"
      ]
    )

    stop(
      "Sequence data failed validation: ",
      paste(error_codes, collapse = ", "),
      ". Resolve the errors or use `prepare_sequence_data()` ",
      "with explicit policies before continuing.",
      call. = FALSE
    )
  }

  reserved <- c(
    "sequence_id",
    "sequence_order",
    "state",
    "original_row",
    "duration"
  )

  .sequence_assert_output_names(
    metadata_cols = metadata_cols,
    reserved = reserved,
    context = "canonical"
  )

  working <- data
  original_row <- seq_len(nrow(working))

  sequence_values <- working[[sequence_id_col]]

  sequence_sort <- if (
    is.numeric(sequence_values) ||
    is.logical(sequence_values)
  ) {
    sequence_values
  } else {
    as.character(sequence_values)
  }

  sort_index <- order(
    sequence_sort,
    working[[order_col]],
    original_row,
    method = "radix"
  )

  working <- working[sort_index, , drop = FALSE]
  original_row <- original_row[sort_index]
  row.names(working) <- NULL

  canonical <- data.frame(
    sequence_id = as.character(
      working[[sequence_id_col]]
    ),
    sequence_order = working[[order_col]],
    state = as.character(working[[state_col]]),
    original_row = as.integer(original_row),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  if (!is.null(duration_col)) {
    canonical$duration <- working[[duration_col]]
  }

  for (column in metadata_cols) {
    canonical[[column]] <- working[[column]]
  }

  list(
    data = canonical,
    audit = validation$audit,
    status = validation$status,
    mapping = validation$mapping,
    original_state_levels = if (
      is.factor(data[[state_col]])
    ) {
      levels(data[[state_col]])
    } else {
      NULL
    }
  )
}

.sequence_metadata_values <- function(
    data,
    rows,
    metadata_cols
) {
  values <- vector("list", length(metadata_cols))
  names(values) <- metadata_cols

  for (column in metadata_cols) {
    column_values <- data[[column]][rows]
    usable <- which(!is.na(column_values))

    selected <- if (length(usable) > 0L) {
      usable[1L]
    } else {
      1L
    }

    values[[column]] <- column_values[selected]
  }

  values
}

.sequence_sum_or_na <- function(value) {
  if (length(value) == 0L || all(is.na(value))) {
    return(NA_real_)
  }

  as.numeric(sum(value, na.rm = TRUE))
}

.sequence_mean_or_na <- function(value) {
  if (length(value) == 0L || all(is.na(value))) {
    return(NA_real_)
  }

  as.numeric(mean(value, na.rm = TRUE))
}

.sequence_safe_proportion <- function(
    numerator,
    denominator
) {
  if (
    length(denominator) != 1L ||
    is.na(denominator) ||
    denominator == 0
  ) {
    return(NA_real_)
  }

  as.numeric(numerator / denominator)
}

.sequence_bind_frames <- function(frames, empty) {
  if (length(frames) == 0L) {
    return(empty)
  }

  result <- do.call(rbind, frames)
  row.names(result) <- NULL
  result
}

.sequence_empty_state_by_sequence <- function(
    metadata_cols,
    include_duration
) {
  result <- data.frame(
    sequence_id = character(),
    state = character(),
    n_observations = integer(),
    observation_proportion = double(),
    stringsAsFactors = FALSE
  )

  for (column in metadata_cols) {
    result[[column]] <- character()
  }

  result <- result[
    c(
      "sequence_id",
      metadata_cols,
      "state",
      "n_observations",
      "observation_proportion"
    )
  ]

  if (include_duration) {
    result$duration_sum <- double()
    result$duration_proportion <- double()
    result$mean_duration <- double()
  }

  result
}

.sequence_empty_state_overall <- function(
    include_duration
) {
  result <- data.frame(
    state = character(),
    n_sequences = integer(),
    sequence_proportion = double(),
    n_observations = integer(),
    observation_proportion = double(),
    stringsAsFactors = FALSE
  )

  if (include_duration) {
    result$duration_sum <- double()
    result$duration_proportion <- double()
    result$mean_duration <- double()
  }

  result
}

.sequence_empty_transition_by_sequence <- function(
    metadata_cols
) {
  result <- data.frame(
    sequence_id = character(),
    from_state = character(),
    to_state = character(),
    n_transitions = integer(),
    sequence_transition_proportion = double(),
    origin_transition_proportion = double(),
    stringsAsFactors = FALSE
  )

  for (column in metadata_cols) {
    result[[column]] <- character()
  }

  result[
    c(
      "sequence_id",
      metadata_cols,
      "from_state",
      "to_state",
      "n_transitions",
      "sequence_transition_proportion",
      "origin_transition_proportion"
    )
  ]
}

.sequence_empty_transition_overall <- function() {
  data.frame(
    from_state = character(),
    to_state = character(),
    n_sequences = integer(),
    sequence_proportion = double(),
    n_transitions = integer(),
    transition_proportion = double(),
    origin_transition_proportion = double(),
    stringsAsFactors = FALSE
  )
}

#' Encode Ordered Sequence States
#'
#' Creates a deterministic dictionary and adds integer and labelled state
#' codes to long-format sequence data.
#'
#' @inheritParams audit_sequence_data
#' @param state_levels Optional atomic vector defining the complete state
#'   ordering. When omitted, factor levels are respected; otherwise observed
#'   state labels are sorted alphabetically.
#' @param prefix Character prefix used for labelled codes.
#' @param width Optional positive integer width for the numeric part of each
#'   labelled code. The default is determined from the dictionary size.
#'
#' @return A named list containing:
#'
#' * `data`: deterministically sorted canonical data with `state_index` and
#'   `state_code`;
#' * `dictionary`: state labels, integer indices, labelled codes, and an
#'   observed-state indicator;
#' * `audit`, `status`, and `mapping` from input validation;
#' * `settings`: the resolved code prefix and width.
#'
#' @details
#' The function does not reinterpret states. State codes are transparent
#' identifiers derived from an explicit or deterministic state ordering.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c("s1", "s1", "s2", "s2"),
#'   position = c(1, 2, 1, 2),
#'   state = c("home", "search", "home", "product")
#' )
#'
#' encoded <- encode_sequence_data(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' encoded$dictionary
#' encoded$data
#'
#' @export
encode_sequence_data <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL,
    expected_states = NULL,
    state_levels = NULL,
    prefix = "S",
    width = NULL
) {
  .sequence_assert_text_scalar(prefix, "prefix")

  input <- .sequence_summary_input(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    duration_col = duration_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  .sequence_assert_output_names(
    metadata_cols = metadata_cols,
    reserved = c("state_index", "state_code"),
    context = "encoded"
  )

  observed_states <- unique(input$data$state)

  if (is.null(state_levels)) {
    resolved_levels <- if (
      !is.null(input$original_state_levels)
    ) {
      as.character(input$original_state_levels)
    } else {
      sort(
        observed_states,
        method = "radix"
      )
    }
  } else {
    if (
      !is.atomic(state_levels) ||
      is.list(state_levels) ||
      anyNA(state_levels)
    ) {
      stop(
        "`state_levels` must be a non-missing atomic vector.",
        call. = FALSE
      )
    }

    resolved_levels <- as.character(state_levels)

    if (
      any(!nzchar(trimws(resolved_levels))) ||
      anyDuplicated(resolved_levels)
    ) {
      stop(
        "`state_levels` must contain unique, non-empty values.",
        call. = FALSE
      )
    }

    omitted_states <- setdiff(
      observed_states,
      resolved_levels
    )

    if (length(omitted_states) > 0L) {
      stop(
        "`state_levels` omits observed states: ",
        paste(omitted_states, collapse = ", "),
        ".",
        call. = FALSE
      )
    }
  }

  if (is.null(width)) {
    resolved_width <- max(
      1L,
      nchar(as.character(length(resolved_levels)))
    )
  } else {
    if (
      !is.numeric(width) ||
      length(width) != 1L ||
      is.na(width) ||
      !is.finite(width) ||
      width < 1 ||
      width != floor(width)
    ) {
      stop(
        "`width` must be one positive whole number.",
        call. = FALSE
      )
    }

    resolved_width <- as.integer(width)
  }

  state_index <- seq_along(resolved_levels)

  state_codes <- paste0(
    prefix,
    sprintf(
      paste0("%0", resolved_width, "d"),
      state_index
    )
  )

  dictionary <- data.frame(
    state = resolved_levels,
    state_index = as.integer(state_index),
    state_code = state_codes,
    observed = resolved_levels %in% observed_states,
    stringsAsFactors = FALSE
  )

  encoded <- input$data

  encoded$state_index <- dictionary$state_index[
    match(encoded$state, dictionary$state)
  ]

  encoded$state_code <- dictionary$state_code[
    match(encoded$state, dictionary$state)
  ]

  front <- c(
    "sequence_id",
    "sequence_order",
    "state",
    "state_index",
    "state_code",
    "original_row"
  )

  encoded <- encoded[
    c(
      front,
      setdiff(names(encoded), front)
    )
  ]

  list(
    data = encoded,
    dictionary = dictionary,
    audit = input$audit,
    status = input$status,
    mapping = input$mapping,
    settings = list(
      prefix = prefix,
      width = resolved_width
    )
  )
}

#' Summarise Sequence States
#'
#' Produces per-sequence and overall state-frequency summaries from validated
#' ordered sequence data.
#'
#' @inheritParams audit_sequence_data
#'
#' @return A named list containing:
#'
#' * `by_sequence`: state counts and proportions within each sequence;
#' * `overall`: state counts and proportions across all sequences;
#' * `audit`, `status`, and `mapping` from input validation.
#'
#' When `duration_col` is supplied, both tables also include duration sums,
#' duration proportions, and mean durations.
#'
#' @details
#' Observation proportions use state rows as the denominator. Sequence
#' proportions report the proportion of sequences in which each state occurs.
#' Missing durations are excluded from duration calculations; an all-missing
#' duration group returns `NA`.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c("s1", "s1", "s1", "s2", "s2"),
#'   position = c(1, 2, 3, 1, 2),
#'   state = c("A", "B", "A", "B", "C")
#' )
#'
#' summaries <- summarise_sequence_states(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' summaries$by_sequence
#' summaries$overall
#'
#' @export
summarise_sequence_states <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL,
    expected_states = NULL
) {
  .sequence_assert_output_names(
    metadata_cols = metadata_cols,
    reserved = c(
      "state",
      "n_observations",
      "observation_proportion",
      "duration_sum",
      "duration_proportion",
      "mean_duration"
    ),
    context = "state-summary"
  )

  input <- .sequence_summary_input(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    duration_col = duration_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  sequence_data <- input$data
  sequence_ids <- unique(sequence_data$sequence_id)
  include_duration <- "duration" %in% names(sequence_data)

  sequence_rows <- list()

  for (sequence_id in sequence_ids) {
    rows <- which(
      sequence_data$sequence_id == sequence_id
    )

    metadata <- .sequence_metadata_values(
      data = sequence_data,
      rows = rows,
      metadata_cols = metadata_cols
    )

    sequence_states <- sequence_data$state[rows]
    state_order <- unique(sequence_states)
    sequence_n <- length(rows)

    sequence_duration <- if (include_duration) {
      .sequence_sum_or_na(sequence_data$duration[rows])
    } else {
      NULL
    }

    for (state in state_order) {
      state_rows <- rows[
        sequence_data$state[rows] == state
      ]

      values <- c(
        list(sequence_id = sequence_id),
        metadata,
        list(
          state = state,
          n_observations = as.integer(
            length(state_rows)
          ),
          observation_proportion =
            length(state_rows) / sequence_n
        )
      )

      if (include_duration) {
        state_duration <- .sequence_sum_or_na(
          sequence_data$duration[state_rows]
        )

        values <- c(
          values,
          list(
            duration_sum = state_duration,
            duration_proportion =
              .sequence_safe_proportion(
                state_duration,
                sequence_duration
              ),
            mean_duration = .sequence_mean_or_na(
              sequence_data$duration[state_rows]
            )
          )
        )
      }

      sequence_rows[[length(sequence_rows) + 1L]] <-
        as.data.frame(
          values,
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
    }
  }

  by_sequence <- .sequence_bind_frames(
    frames = sequence_rows,
    empty = .sequence_empty_state_by_sequence(
      metadata_cols = metadata_cols,
      include_duration = include_duration
    )
  )

  overall_rows <- list()
  overall_states <- unique(sequence_data$state)
  total_observations <- nrow(sequence_data)
  total_sequences <- length(sequence_ids)

  total_duration <- if (include_duration) {
    .sequence_sum_or_na(sequence_data$duration)
  } else {
    NULL
  }

  for (state in overall_states) {
    state_rows <- which(sequence_data$state == state)
    state_sequences <- unique(
      sequence_data$sequence_id[state_rows]
    )

    values <- list(
      state = state,
      n_sequences = as.integer(
        length(state_sequences)
      ),
      sequence_proportion =
        length(state_sequences) / total_sequences,
      n_observations = as.integer(
        length(state_rows)
      ),
      observation_proportion =
        length(state_rows) / total_observations
    )

    if (include_duration) {
      state_duration <- .sequence_sum_or_na(
        sequence_data$duration[state_rows]
      )

      values <- c(
        values,
        list(
          duration_sum = state_duration,
          duration_proportion =
            .sequence_safe_proportion(
              state_duration,
              total_duration
            ),
          mean_duration = .sequence_mean_or_na(
            sequence_data$duration[state_rows]
          )
        )
      )
    }

    overall_rows[[length(overall_rows) + 1L]] <-
      as.data.frame(
        values,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
  }

  overall <- .sequence_bind_frames(
    frames = overall_rows,
    empty = .sequence_empty_state_overall(
      include_duration = include_duration
    )
  )

  list(
    by_sequence = by_sequence,
    overall = overall,
    audit = input$audit,
    status = input$status,
    mapping = input$mapping
  )
}

#' Summarise Adjacent Sequence Transitions
#'
#' Counts transitions between adjacent ordered states for each sequence and
#' across the complete data set.
#'
#' @inheritParams audit_sequence_data
#' @param include_self Logical value indicating whether transitions from a
#'   state to the same state should be included.
#'
#' @return A named list containing:
#'
#' * `by_sequence`: transition counts, proportions within each sequence, and
#'   conditional proportions within each origin state;
#' * `overall`: transition counts, sequence coverage, global proportions, and
#'   conditional origin-state proportions;
#' * `audit`, `status`, `mapping`, and the resolved `include_self` setting.
#'
#' @details
#' A transition is defined only between adjacent rows after deterministic
#' ordering by sequence identifier, sequence order, and original row.
#' Sequences with one state contribute no transitions.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c("s1", "s1", "s1", "s2", "s2"),
#'   position = c(1, 2, 3, 1, 2),
#'   state = c("A", "B", "C", "A", "C")
#' )
#'
#' transitions <- summarise_sequence_transitions(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' transitions$by_sequence
#' transitions$overall
#'
#' @export
summarise_sequence_transitions <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    metadata_cols = NULL,
    expected_states = NULL,
    include_self = TRUE
) {
  .sequence_assert_flag(include_self, "include_self")

  .sequence_assert_output_names(
    metadata_cols = metadata_cols,
    reserved = c(
      "from_state",
      "to_state",
      "n_transitions",
      "sequence_transition_proportion",
      "transition_proportion",
      "origin_transition_proportion",
      "n_sequences",
      "sequence_proportion"
    ),
    context = "transition-summary"
  )

  input <- .sequence_summary_input(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  sequence_data <- input$data
  sequence_ids <- unique(sequence_data$sequence_id)

  by_sequence_rows <- list()
  raw_transition_rows <- list()

  for (sequence_id in sequence_ids) {
    rows <- which(
      sequence_data$sequence_id == sequence_id
    )

    if (length(rows) < 2L) {
      next
    }

    states <- sequence_data$state[rows]

    pairs <- data.frame(
      from_state = states[-length(states)],
      to_state = states[-1L],
      stringsAsFactors = FALSE
    )

    if (!include_self) {
      pairs <- pairs[
        pairs$from_state != pairs$to_state,
        ,
        drop = FALSE
      ]
    }

    if (nrow(pairs) == 0L) {
      next
    }

    raw_transition_rows[[length(raw_transition_rows) + 1L]] <-
      data.frame(
        sequence_id = rep(sequence_id, nrow(pairs)),
        pairs,
        stringsAsFactors = FALSE
      )

    metadata <- .sequence_metadata_values(
      data = sequence_data,
      rows = rows,
      metadata_cols = metadata_cols
    )

    unique_pairs <- unique(pairs)
    sequence_total <- nrow(pairs)

    for (pair_index in seq_len(nrow(unique_pairs))) {
      from_state <- unique_pairs$from_state[pair_index]
      to_state <- unique_pairs$to_state[pair_index]

      pair_rows <- pairs$from_state == from_state &
        pairs$to_state == to_state

      pair_count <- sum(pair_rows)
      origin_count <- sum(
        pairs$from_state == from_state
      )

      values <- c(
        list(sequence_id = sequence_id),
        metadata,
        list(
          from_state = from_state,
          to_state = to_state,
          n_transitions = as.integer(pair_count),
          sequence_transition_proportion =
            pair_count / sequence_total,
          origin_transition_proportion =
            pair_count / origin_count
        )
      )

      by_sequence_rows[[length(by_sequence_rows) + 1L]] <-
        as.data.frame(
          values,
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
    }
  }

  by_sequence <- .sequence_bind_frames(
    frames = by_sequence_rows,
    empty = .sequence_empty_transition_by_sequence(
      metadata_cols = metadata_cols
    )
  )

  raw_transitions <- if (
    length(raw_transition_rows) == 0L
  ) {
    data.frame(
      sequence_id = character(),
      from_state = character(),
      to_state = character(),
      stringsAsFactors = FALSE
    )
  } else {
    result <- do.call(rbind, raw_transition_rows)
    row.names(result) <- NULL
    result
  }

  overall_rows <- list()

  if (nrow(raw_transitions) > 0L) {
    unique_pairs <- unique(
      raw_transitions[
        c("from_state", "to_state")
      ]
    )

    total_transitions <- nrow(raw_transitions)
    total_sequences <- length(sequence_ids)

    for (pair_index in seq_len(nrow(unique_pairs))) {
      from_state <- unique_pairs$from_state[pair_index]
      to_state <- unique_pairs$to_state[pair_index]

      pair_rows <-
        raw_transitions$from_state == from_state &
        raw_transitions$to_state == to_state

      pair_count <- sum(pair_rows)
      origin_count <- sum(
        raw_transitions$from_state == from_state
      )

      contributing_sequences <- unique(
        raw_transitions$sequence_id[pair_rows]
      )

      overall_rows[[length(overall_rows) + 1L]] <-
        data.frame(
          from_state = from_state,
          to_state = to_state,
          n_sequences = as.integer(
            length(contributing_sequences)
          ),
          sequence_proportion =
            length(contributing_sequences) /
            total_sequences,
          n_transitions = as.integer(pair_count),
          transition_proportion =
            pair_count / total_transitions,
          origin_transition_proportion =
            pair_count / origin_count,
          stringsAsFactors = FALSE
        )
    }
  }

  overall <- .sequence_bind_frames(
    frames = overall_rows,
    empty = .sequence_empty_transition_overall()
  )

  list(
    by_sequence = by_sequence,
    overall = overall,
    audit = input$audit,
    status = input$status,
    mapping = input$mapping,
    include_self = include_self
  )
}

#' Format Ordered Sequence Paths
#'
#' Creates a compact one-row-per-sequence representation of ordered state
#' paths.
#'
#' @inheritParams audit_sequence_data
#' @param separator Character value inserted between adjacent state labels.
#' @param collapse_repeats Logical value indicating whether consecutive
#'   repeated states should be collapsed for path display.
#'
#' @return A named list containing:
#'
#' * `paths`: one row per sequence with observation counts, formatted-state
#'   counts, unique-state counts, start and end states, and the path string;
#' * `audit`, `status`, and `mapping` from input validation;
#' * `settings`: the path separator and repeat-collapsing choice.
#'
#' @details
#' Repeat collapsing affects only the formatted representation. It does not
#' modify the supplied data or alter non-consecutive repeated states.
#' Input rows are ordered deterministically for formatting, but review-level
#' diagnostics such as `unordered_rows` remain reflected in the returned
#' `status` and `audit`.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c("s1", "s1", "s1", "s2", "s2"),
#'   position = c(1, 2, 3, 1, 2),
#'   state = c("A", "B", "C", "A", "C")
#' )
#'
#' paths <- format_sequence_paths(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' paths$paths
#'
#' @export
format_sequence_paths <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    metadata_cols = NULL,
    expected_states = NULL,
    separator = " > ",
    collapse_repeats = FALSE
) {
  .sequence_assert_text_scalar(separator, "separator")
  .sequence_assert_flag(
    collapse_repeats,
    "collapse_repeats"
  )

  .sequence_assert_output_names(
    metadata_cols = metadata_cols,
    reserved = c(
      "n_observations",
      "n_states",
      "n_unique_states",
      "start_state",
      "end_state",
      "path"
    ),
    context = "path"
  )

  input <- .sequence_summary_input(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  sequence_data <- input$data
  sequence_ids <- unique(sequence_data$sequence_id)
  path_rows <- list()

  for (sequence_id in sequence_ids) {
    rows <- which(
      sequence_data$sequence_id == sequence_id
    )

    observed_states <- sequence_data$state[rows]
    formatted_states <- observed_states

    if (
      collapse_repeats &&
      length(formatted_states) > 1L
    ) {
      keep <- c(
        TRUE,
        formatted_states[-1L] !=
          formatted_states[-length(formatted_states)]
      )

      formatted_states <- formatted_states[keep]
    }

    metadata <- .sequence_metadata_values(
      data = sequence_data,
      rows = rows,
      metadata_cols = metadata_cols
    )

    values <- c(
      list(sequence_id = sequence_id),
      metadata,
      list(
        n_observations = as.integer(
          length(observed_states)
        ),
        n_states = as.integer(
          length(formatted_states)
        ),
        n_unique_states = as.integer(
          length(unique(formatted_states))
        ),
        start_state = formatted_states[1L],
        end_state = formatted_states[
          length(formatted_states)
        ],
        path = paste(
          formatted_states,
          collapse = separator
        )
      )
    )

    path_rows[[length(path_rows) + 1L]] <-
      as.data.frame(
        values,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
  }

  paths <- do.call(rbind, path_rows)
  row.names(paths) <- NULL

  list(
    paths = paths,
    audit = input$audit,
    status = input$status,
    mapping = input$mapping,
    settings = list(
      separator = separator,
      collapse_repeats = collapse_repeats
    )
  )
}

make_encoding_summary_data <- function() {
  data.frame(
    id = c(
      rep("s1", 4L),
      rep("s2", 3L)
    ),
    position = c(1:4, 1:3),
    state = c(
      "A", "B", "A", "C",
      "B", "C", "A"
    ),
    duration_ms = c(
      1, 2, 3, 4,
      2, 2, 6
    ),
    group = c(
      rep("g1", 4L),
      rep("g2", 3L)
    ),
    stringsAsFactors = FALSE
  )
}

test_that("state encoding is deterministic across row order", {
  data <- make_encoding_summary_data()
  shuffled <- data[c(7, 2, 5, 1, 4, 3, 6), ]

  first <- encode_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  second <- encode_sequence_data(
    shuffled,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_identical(first$status, "pass")
  expect_identical(first$dictionary, second$dictionary)

  expect_identical(
    first$dictionary$state,
    c("A", "B", "C")
  )

  expect_identical(
    first$dictionary$state_index,
    1:3
  )

  expect_identical(
    first$dictionary$state_code,
    c("S1", "S2", "S3")
  )

  comparable <- c(
    "sequence_id",
    "sequence_order",
    "state",
    "state_index",
    "state_code"
  )

  expect_identical(
    first$data[comparable],
    second$data[comparable]
  )
})

test_that("custom encoding levels and labels are respected", {
  data <- make_encoding_summary_data()

  encoded <- encode_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    state_levels = c("C", "B", "A", "Z"),
    prefix = "Q",
    width = 2
  )

  expect_identical(
    encoded$dictionary$state,
    c("C", "B", "A", "Z")
  )

  expect_identical(
    encoded$dictionary$state_code,
    c("Q01", "Q02", "Q03", "Q04")
  )

  expect_identical(
    encoded$dictionary$observed,
    c(TRUE, TRUE, TRUE, FALSE)
  )

  expect_identical(encoded$settings$width, 2L)

  expect_error(
    encode_sequence_data(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state",
      state_levels = c("A", "B")
    ),
    "omits observed states"
  )
})

test_that("factor levels define the default encoding order", {
  data <- make_encoding_summary_data()

  data$state <- factor(
    data$state,
    levels = c("C", "A", "B", "Z")
  )

  encoded <- encode_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_identical(
    encoded$dictionary$state,
    c("C", "A", "B", "Z")
  )

  expect_false(
    encoded$dictionary$observed[
      encoded$dictionary$state == "Z"
    ]
  )
})

test_that("state summaries return exact counts and proportions", {
  data <- make_encoding_summary_data()

  result <- summarise_sequence_states(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms",
    metadata_cols = "group"
  )

  expect_identical(result$status, "pass")

  s1_a <- result$by_sequence[
    result$by_sequence$sequence_id == "s1" &
      result$by_sequence$state == "A",
    ,
    drop = FALSE
  ]

  expect_identical(s1_a$group, "g1")
  expect_identical(s1_a$n_observations, 2L)
  expect_equal(s1_a$observation_proportion, 0.5)
  expect_equal(s1_a$duration_sum, 4)
  expect_equal(s1_a$duration_proportion, 0.4)
  expect_equal(s1_a$mean_duration, 2)

  overall_a <- result$overall[
    result$overall$state == "A",
    ,
    drop = FALSE
  ]

  expect_identical(overall_a$n_sequences, 2L)
  expect_equal(overall_a$sequence_proportion, 1)
  expect_identical(overall_a$n_observations, 3L)
  expect_equal(
    overall_a$observation_proportion,
    3 / 7
  )
  expect_equal(overall_a$duration_sum, 10)
  expect_equal(overall_a$duration_proportion, 0.5)
  expect_equal(overall_a$mean_duration, 10 / 3)
})

test_that("state summaries are deterministic across input order", {
  data <- make_encoding_summary_data()
  shuffled <- data[c(6, 1, 7, 4, 2, 5, 3), ]

  first <- summarise_sequence_states(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  second <- summarise_sequence_states(
    shuffled,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_identical(
    first$by_sequence,
    second$by_sequence
  )

  expect_identical(
    first$overall,
    second$overall
  )
})

test_that("transition summaries return exact adjacent counts", {
  data <- make_encoding_summary_data()

  result <- summarise_sequence_transitions(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "group"
  )

  expect_identical(result$status, "pass")

  s1_a_b <- result$by_sequence[
    result$by_sequence$sequence_id == "s1" &
      result$by_sequence$from_state == "A" &
      result$by_sequence$to_state == "B",
    ,
    drop = FALSE
  ]

  expect_identical(s1_a_b$group, "g1")
  expect_identical(s1_a_b$n_transitions, 1L)
  expect_equal(
    s1_a_b$sequence_transition_proportion,
    1 / 3
  )
  expect_equal(
    s1_a_b$origin_transition_proportion,
    0.5
  )

  overall_b_a <- result$overall[
    result$overall$from_state == "B" &
      result$overall$to_state == "A",
    ,
    drop = FALSE
  ]

  expect_identical(overall_b_a$n_sequences, 1L)
  expect_equal(
    overall_b_a$sequence_proportion,
    0.5
  )
  expect_identical(
    overall_b_a$n_transitions,
    1L
  )
  expect_equal(
    overall_b_a$transition_proportion,
    0.2
  )
  expect_equal(
    overall_b_a$origin_transition_proportion,
    0.5
  )
})

test_that("self-transition filtering is explicit", {
  data <- data.frame(
    id = rep("s1", 3L),
    position = 1:3,
    state = c("A", "A", "B"),
    stringsAsFactors = FALSE
  )

  retained <- summarise_sequence_transitions(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    include_self = TRUE
  )

  removed <- summarise_sequence_transitions(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    include_self = FALSE
  )

  expect_identical(retained$status, "review")
  expect_true(
    any(
      retained$overall$from_state == "A" &
        retained$overall$to_state == "A"
    )
  )

  expect_false(
    any(
      removed$overall$from_state ==
        removed$overall$to_state
    )
  )

  expect_identical(
    removed$overall$n_transitions,
    1L
  )
})

test_that("single-state sequences produce stable empty transitions", {
  data <- data.frame(
    id = "s1",
    position = 1,
    state = "A",
    stringsAsFactors = FALSE
  )

  result <- summarise_sequence_transitions(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_identical(result$status, "review")
  expect_equal(nrow(result$by_sequence), 0L)
  expect_equal(nrow(result$overall), 0L)

  expect_identical(
    names(result$overall),
    c(
      "from_state",
      "to_state",
      "n_sequences",
      "sequence_proportion",
      "n_transitions",
      "transition_proportion",
      "origin_transition_proportion"
    )
  )
})

test_that("formatted paths are ordered and retain metadata", {
  data <- make_encoding_summary_data()
  shuffled <- data[c(7, 4, 2, 6, 1, 5, 3), ]

  result <- format_sequence_paths(
    shuffled,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "group"
  )

  # The rows are formatted deterministically, but the input audit
  # retains the review-level unordered-row diagnostic.
  expect_identical(result$status, "review")
  expect_true(
    any(
      result$audit$issue_code == "unordered_rows" &
        result$audit$severity == "review"
    )
  )

  expect_identical(
    result$paths$sequence_id,
    c("s1", "s2")
  )

  expect_identical(
    result$paths$group,
    c("g1", "g2")
  )

  expect_identical(
    result$paths$path,
    c(
      "A > B > A > C",
      "B > C > A"
    )
  )

  expect_identical(
    result$paths$n_observations,
    c(4L, 3L)
  )

  expect_identical(
    result$paths$n_states,
    c(4L, 3L)
  )

  ordered_result <- format_sequence_paths(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "group"
  )

  expect_identical(ordered_result$status, "pass")
  expect_equal(nrow(ordered_result$audit), 0L)
  expect_identical(
    ordered_result$paths,
    result$paths
  )
})


test_that("path formatting can collapse consecutive repeats", {
  data <- data.frame(
    id = rep("s1", 5L),
    position = 1:5,
    state = c("A", "A", "B", "B", "C"),
    stringsAsFactors = FALSE
  )

  result <- format_sequence_paths(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    separator = "/",
    collapse_repeats = TRUE
  )

  expect_identical(result$status, "review")
  expect_identical(result$paths$n_observations, 5L)
  expect_identical(result$paths$n_states, 3L)
  expect_identical(result$paths$n_unique_states, 3L)
  expect_identical(result$paths$start_state, "A")
  expect_identical(result$paths$end_state, "C")
  expect_identical(result$paths$path, "A/B/C")
})

test_that("summary functions reject unresolved validation errors", {
  data <- make_encoding_summary_data()
  data$position[2L] <- 1

  expect_error(
    encode_sequence_data(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state"
    ),
    "failed validation"
  )

  expect_error(
    summarise_sequence_states(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state"
    ),
    "failed validation"
  )

  expect_error(
    summarise_sequence_transitions(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state"
    ),
    "failed validation"
  )

  expect_error(
    format_sequence_paths(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state"
    ),
    "failed validation"
  )
})

.sequence_ext_multichannel_input <- function(
  data,
  sequence_id_col,
  order_col,
  channel_cols,
  symbol_levels = NULL
) {
  if (!is.character(channel_cols) || length(channel_cols) < 2L ||
      anyNA(channel_cols) || any(!nzchar(channel_cols)) || anyDuplicated(channel_cols)) {
    stop("`channel_cols` must contain at least two unique channel names.",
         call. = FALSE)
  }
  if (is.list(data) && !is.data.frame(data) && !is.null(data$data)) data <- data$data
  if (!is.data.frame(data)) stop("`data` must be a data frame.", call. = FALSE)
  missing <- setdiff(channel_cols, names(data))
  if (length(missing) > 0L) {
    stop("Missing channel columns: ", paste(missing, collapse = ", "), ".",
         call. = FALSE)
  }
  x <- .sequence_adv_data(
    data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = channel_cols[1L],
    missing_state_policy = "error"
  )
  working <- x$data
  for (channel in channel_cols) {
    values <- working[[channel]]
    text <- as.character(values)
    if (is.list(values) || anyNA(values) || any(!nzchar(trimws(text)))) {
      stop("Channel `", channel, "` contains missing, blank, or non-atomic values.",
           call. = FALSE)
    }
  }
  if (is.null(symbol_levels)) symbol_levels <- vector("list", length(channel_cols))
  if (!is.list(symbol_levels) || !(length(symbol_levels) %in% c(0L, length(channel_cols)))) {
    stop("`symbol_levels` must be `NULL` or one list element per channel.",
         call. = FALSE)
  }
  if (length(symbol_levels) == 0L) symbol_levels <- vector("list", length(channel_cols))
  names(symbol_levels) <- channel_cols
  symbols <- vector("list", length(channel_cols))
  names(symbols) <- channel_cols
  for (i in seq_along(channel_cols)) {
    channel <- channel_cols[i]
    observed <- unique(as.character(working[[channel]]))
    current <- symbol_levels[[i]]
    if (is.null(current)) {
      current <- if (is.factor(working[[channel]])) {
        levels(working[[channel]])[levels(working[[channel]]) %in% observed]
      } else sort(observed, method = "radix")
    }
    current <- as.character(current)
    if (length(current) < 1L || anyNA(current) || any(!nzchar(trimws(current))) ||
        anyDuplicated(current)) {
      stop("Invalid symbol levels for channel `", channel, "`.", call. = FALSE)
    }
    missing_symbols <- setdiff(observed, current)
    if (length(missing_symbols) > 0L) {
      stop("Symbol levels for channel `", channel,
           "` do not cover: ", paste(missing_symbols, collapse = ", "), ".",
           call. = FALSE)
    }
    symbols[[channel]] <- current
  }
  ids <- x$sequence_ids
  split_rows <- split(seq_len(nrow(working)),
                      factor(as.character(working[[sequence_id_col]]), levels = ids),
                      drop = TRUE)
  observations <- lapply(split_rows, function(rows) {
    out <- matrix(NA_integer_, nrow = length(rows), ncol = length(channel_cols),
                  dimnames = list(NULL, channel_cols))
    for (j in seq_along(channel_cols)) {
      out[, j] <- match(as.character(working[[channel_cols[j]]][rows]),
                        symbols[[channel_cols[j]]])
    }
    out
  })
  orders <- lapply(split_rows, function(rows) working[[order_col]][rows])
  names(observations) <- ids
  names(orders) <- ids
  list(
    data = working,
    sequence_ids = ids,
    observations = observations,
    orders = orders,
    channel_cols = channel_cols,
    symbols = symbols,
    columns = list(sequence_id = sequence_id_col, order = order_col)
  )
}

.sequence_ext_multichannel_emission <- function(observation, emission_probs) {
  n_time <- nrow(observation)
  n_states <- nrow(emission_probs[[1L]])
  likelihood <- matrix(1, nrow = n_time, ncol = n_states)
  for (channel in seq_along(emission_probs)) {
    current <- emission_probs[[channel]]
    likelihood <- likelihood * t(current[, observation[, channel], drop = FALSE])
  }
  likelihood
}

.sequence_ext_multichannel_initialise <- function(n_states, symbols, seed,
                                                   initial_probs = NULL,
                                                   transition_probs = NULL,
                                                   emission_probs = NULL) {
  generated <- .sequence_adv_with_seed(seed, {
    list(
      initial = if (is.null(initial_probs)) stats::rexp(n_states) + 0.1 else initial_probs,
      transition = if (is.null(transition_probs)) {
        matrix(stats::rexp(n_states * n_states) + 0.1,
               nrow = n_states, ncol = n_states)
      } else transition_probs,
      emissions = if (is.null(emission_probs)) {
        lapply(symbols, function(levels) {
          matrix(stats::rexp(n_states * length(levels)) + 0.1,
                 nrow = n_states, ncol = length(levels))
        })
      } else emission_probs
    )
  })
  if (!is.numeric(generated$initial) || length(generated$initial) != n_states ||
      anyNA(generated$initial) || any(!is.finite(generated$initial)) ||
      any(generated$initial < 0)) {
    stop("Invalid `initial_probs`.", call. = FALSE)
  }
  if (!is.matrix(generated$transition) ||
      !all(dim(generated$transition) == c(n_states, n_states)) ||
      anyNA(generated$transition) || any(!is.finite(generated$transition)) ||
      any(generated$transition < 0)) {
    stop("Invalid `transition_probs`.", call. = FALSE)
  }
  if (!is.list(generated$emissions) || length(generated$emissions) != length(symbols)) {
    stop("`emission_probs` must contain one matrix per channel.", call. = FALSE)
  }
  for (i in seq_along(symbols)) {
    current <- generated$emissions[[i]]
    if (!is.matrix(current) || !all(dim(current) == c(n_states, length(symbols[[i]]))) ||
        anyNA(current) || any(!is.finite(current)) || any(current < 0)) {
      stop("Invalid emission probabilities for channel `", names(symbols)[i], "`.",
           call. = FALSE)
    }
  }
  list(
    initial = .sequence_adv_vector_normalise(generated$initial),
    transition = .sequence_adv_row_normalise(generated$transition),
    emissions = lapply(generated$emissions, .sequence_adv_row_normalise)
  )
}

#' Fit a multichannel categorical hidden Markov model
#'
#' Fits a finite-state, time-homogeneous HMM to two or more categorical
#' channels under conditional independence of channels given the latent state.
#' Latent states are statistical model states only.
#'
#' @param data Long-format multichannel sequence data.
#' @param n_states Number of latent states.
#' @param channel_cols Names of categorical observation channels.
#' @param sequence_id_col,order_col Core sequence columns.
#' @param symbol_levels Optional named list of symbol orders by channel.
#' @param state_names Optional latent-state names.
#' @param initial_probs,transition_probs,emission_probs Optional starting values.
#' @param max_iter Maximum EM iterations.
#' @param tolerance Relative log-likelihood tolerance.
#' @param pseudocount Non-negative smoothing count.
#' @param seed Reproducibility seed.
#' @param keep_posteriors Retain final forward-backward results.
#'
#' @return An object of class `gp3_multichannel_sequence_hmm`.
#' @examples
#' multichannel <- data.frame(
#'   sequence_id = rep(paste0("s", 1:4), each = 5L),
#'   sequence_order = rep(1:5, times = 4L),
#'   action = c("A", "B", "C", "C", "D", "A", "B", "B", "C", "D",
#'              "D", "C", "B", "A", "A", "D", "C", "C", "B", "A"),
#'   context = rep(c("x", "x", "y", "y", "z"), times = 4L)
#' )
#' fit_multichannel_sequence_hmm(multichannel, 2L,
#'                               channel_cols = c("action", "context"),
#'                               max_iter = 5L, seed = 1L)
#' @export
fit_multichannel_sequence_hmm <- function(
  data,
  n_states,
  channel_cols,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  symbol_levels = NULL,
  state_names = NULL,
  initial_probs = NULL,
  transition_probs = NULL,
  emission_probs = NULL,
  max_iter = 200L,
  tolerance = 1e-6,
  pseudocount = 1e-6,
  seed = 1L,
  keep_posteriors = FALSE
) {
  .sequence_adv_scalar_number(n_states, "n_states", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(max_iter, "max_iter", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(tolerance, "tolerance", lower = 0)
  .sequence_adv_scalar_number(pseudocount, "pseudocount", lower = 0)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  .sequence_adv_scalar_logical(keep_posteriors, "keep_posteriors")
  input <- .sequence_ext_multichannel_input(
    data, sequence_id_col, order_col, channel_cols, symbol_levels
  )
  if (is.null(state_names)) state_names <- paste0("latent_", seq_len(n_states))
  if (!is.character(state_names) || length(state_names) != n_states ||
      anyNA(state_names) || any(!nzchar(trimws(state_names))) ||
      anyDuplicated(state_names)) {
    stop("`state_names` must uniquely name all latent states.", call. = FALSE)
  }
  parameters <- .sequence_ext_multichannel_initialise(
    n_states, input$symbols, seed, initial_probs, transition_probs,
    emission_probs
  )
  history <- numeric(max_iter)
  converged <- FALSE
  previous <- -Inf
  posteriors <- NULL
  for (iteration in seq_len(max_iter)) {
    initial_count <- rep(pseudocount, n_states)
    transition_count <- matrix(pseudocount, nrow = n_states, ncol = n_states)
    emission_count <- lapply(input$symbols, function(levels) {
      matrix(pseudocount, nrow = n_states, ncol = length(levels))
    })
    sequence_loglik <- numeric(length(input$observations))
    posterior_current <- vector("list", length(input$observations))
    for (s in seq_along(input$observations)) {
      obs <- input$observations[[s]]
      likelihood <- .sequence_ext_multichannel_emission(obs, parameters$emissions)
      fb <- .sequence_ext_forward_backward(
        likelihood, parameters$initial, parameters$transition
      )
      initial_count <- initial_count + fb$gamma[1L, ]
      if (nrow(obs) > 1L) {
        transition_count <- transition_count + apply(fb$xi, c(2L, 3L), sum)
      }
      for (channel in seq_along(input$symbols)) {
        for (t in seq_len(nrow(obs))) {
          emission_count[[channel]][, obs[t, channel]] <-
            emission_count[[channel]][, obs[t, channel]] + fb$gamma[t, ]
        }
      }
      sequence_loglik[s] <- fb$log_likelihood
      posterior_current[[s]] <- fb
    }
    current <- sum(sequence_loglik)
    history[iteration] <- current
    parameters$initial <- .sequence_adv_vector_normalise(initial_count)
    parameters$transition <- .sequence_adv_row_normalise(transition_count)
    parameters$emissions <- lapply(emission_count, .sequence_adv_row_normalise)
    if (iteration > 1L) {
      relative <- abs(current - previous) / max(1, abs(previous))
      if (relative <= tolerance) {
        converged <- TRUE
        history <- history[seq_len(iteration)]
        posteriors <- posterior_current
        break
      }
    }
    previous <- current
    posteriors <- posterior_current
    if (iteration == max_iter) history <- history[seq_len(iteration)]
  }
  final_loglik <- numeric(length(input$observations))
  final_posteriors <- vector("list", length(input$observations))
  for (s in seq_along(input$observations)) {
    likelihood <- .sequence_ext_multichannel_emission(
      input$observations[[s]], parameters$emissions
    )
    fb <- .sequence_ext_forward_backward(
      likelihood, parameters$initial, parameters$transition
    )
    final_loglik[s] <- fb$log_likelihood
    final_posteriors[[s]] <- fb
  }
  log_likelihood <- sum(final_loglik)
  history[length(history)] <- log_likelihood
  names(parameters$initial) <- state_names
  dimnames(parameters$transition) <- list(state_names, state_names)
  names(parameters$emissions) <- input$channel_cols
  for (channel in seq_along(parameters$emissions)) {
    dimnames(parameters$emissions[[channel]]) <-
      list(state_names, input$symbols[[channel]])
  }
  n_parameters <- (n_states - 1L) + n_states * (n_states - 1L) +
    sum(vapply(input$symbols, function(levels) {
      n_states * (length(levels) - 1L)
    }, numeric(1)))
  n_observations <- sum(vapply(input$observations, nrow, integer(1)))
  result <- list(
    initial_probs = parameters$initial,
    transition_probs = parameters$transition,
    emission_probs = parameters$emissions,
    state_names = state_names,
    channel_names = input$channel_cols,
    symbol_names = input$symbols,
    sequence_ids = input$sequence_ids,
    sequence_log_likelihoods = stats::setNames(final_loglik, input$sequence_ids),
    log_likelihood = log_likelihood,
    iterations = length(history),
    converged = converged,
    tolerance = tolerance,
    pseudocount = pseudocount,
    log_likelihood_history = history,
    n_parameters = as.integer(n_parameters),
    n_observations = as.integer(n_observations),
    aic = -2 * log_likelihood + 2 * n_parameters,
    bic = -2 * log_likelihood + log(max(1, n_observations)) * n_parameters,
    seed = as.integer(seed),
    training_observations = input$observations,
    training_orders = input$orders,
    posteriors = if (keep_posteriors) final_posteriors else NULL,
    columns = input$columns,
    call = match.call()
  )
  class(result) <- c("gp3_multichannel_sequence_hmm", "list")
  result
}

#' Decode latent states from a multichannel HMM
#'
#' @param model A fitted multichannel HMM.
#' @param data Optional new long-format multichannel data.
#' @param sequence_id_col,order_col Core sequence columns for new data.
#' @param channel_cols Channel columns for new data; defaults to training names.
#' @param method `"viterbi"` or `"posterior"`.
#'
#' @return A long data frame of decoded states and posterior probabilities.
#' @examples
#' # See `fit_multichannel_sequence_hmm()`.
#' @export
decode_multichannel_sequence_states <- function(
  model,
  data = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  channel_cols = NULL,
  method = c("viterbi", "posterior")
) {
  if (!inherits(model, "gp3_multichannel_sequence_hmm")) {
    stop("`model` must be created by `fit_multichannel_sequence_hmm()`.",
         call. = FALSE)
  }
  method <- match.arg(method)
  if (is.null(channel_cols)) channel_cols <- model$channel_names
  if (is.null(data)) {
    observations <- model$training_observations
    orders <- model$training_orders
    ids <- model$sequence_ids
  } else {
    input <- .sequence_ext_multichannel_input(
      data, sequence_id_col, order_col, channel_cols,
      symbol_levels = model$symbol_names
    )
    observations <- input$observations
    orders <- input$orders
    ids <- input$sequence_ids
  }
  rows <- list()
  h <- 0L
  for (i in seq_along(ids)) {
    likelihood <- .sequence_ext_multichannel_emission(
      observations[[i]], model$emission_probs
    )
    fb <- .sequence_ext_forward_backward(
      likelihood, model$initial_probs, model$transition_probs
    )
    if (method == "viterbi") {
      decoded <- .sequence_ext_viterbi(
        likelihood, model$initial_probs, model$transition_probs
      )$path
    } else {
      decoded <- max.col(fb$gamma, ties.method = "first")
    }
    posterior_probability <- fb$gamma[cbind(seq_along(decoded), decoded)]
    for (t in seq_along(decoded)) {
      h <- h + 1L
      row <- data.frame(
        sequence_id = ids[i],
        sequence_order = orders[[i]][t],
        latent_state = model$state_names[decoded[t]],
        posterior_probability = posterior_probability[t],
        decoding_method = method,
        stringsAsFactors = FALSE
      )
      for (channel in seq_along(model$channel_names)) {
        row[[model$channel_names[channel]]] <-
          model$symbol_names[[channel]][observations[[i]][t, channel]]
      }
      rows[[h]] <- row
    }
  }
  result <- do.call(rbind, rows)
  row.names(result) <- NULL
  result
}

#' Summarise a multichannel sequence HMM
#'
#' @param model A fitted multichannel HMM.
#'
#' @return A list of convergence, fit, transition, initial, and channel-specific
#'   emission summaries.
#' @examples
#' # See `fit_multichannel_sequence_hmm()`.
#' @export
summarise_multichannel_sequence_hmm <- function(model) {
  if (!inherits(model, "gp3_multichannel_sequence_hmm")) {
    stop("`model` must be created by `fit_multichannel_sequence_hmm()`.",
         call. = FALSE)
  }
  list(
    fit = data.frame(
      n_states = length(model$state_names),
      n_channels = length(model$channel_names),
      n_sequences = length(model$sequence_ids),
      n_observations = model$n_observations,
      log_likelihood = model$log_likelihood,
      aic = model$aic,
      bic = model$bic,
      iterations = model$iterations,
      converged = model$converged,
      stringsAsFactors = FALSE
    ),
    initial = data.frame(latent_state = names(model$initial_probs),
                         probability = as.numeric(model$initial_probs),
                         stringsAsFactors = FALSE),
    transition = as.data.frame(as.table(model$transition_probs),
                               stringsAsFactors = FALSE),
    emission = lapply(model$emission_probs, function(x) {
      as.data.frame(as.table(x), stringsAsFactors = FALSE)
    })
  )
}

#' Plot multichannel HMM emission profiles
#'
#' @param model A fitted multichannel HMM.
#' @param channel Channel name to plot.
#' @param ... Additional arguments passed to [graphics::barplot()].
#'
#' @return The selected emission matrix, invisibly.
#' @examples
#' # See `fit_multichannel_sequence_hmm()`.
#' @export
plot_multichannel_sequence_hmm <- function(model, channel = model$channel_names[1L], ...) {
  if (!inherits(model, "gp3_multichannel_sequence_hmm")) {
    stop("`model` must be created by `fit_multichannel_sequence_hmm()`.",
         call. = FALSE)
  }
  .sequence_adv_scalar_character(channel, "channel")
  if (!(channel %in% model$channel_names)) stop("Unknown channel.", call. = FALSE)
  matrix_value <- model$emission_probs[[channel]]
  graphics::barplot(
    t(matrix_value),
    beside = TRUE,
    legend.text = colnames(matrix_value),
    args.legend = list(x = "topright", bty = "n"),
    ylab = "Emission probability",
    xlab = "Latent state",
    names.arg = rownames(matrix_value),
    ...
  )
  invisible(matrix_value)
}

test_that("multichannel HMM fits, decodes, and summarises", {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  data <- make_extension_sequence_data()
  fit <- fit_multichannel_sequence_hmm(
    data,
    n_states = 2L,
    channel_cols = c("state", "channel_context"),
    max_iter = 4L,
    seed = 11L
  )
  expect_s3_class(fit, "gp3_multichannel_sequence_hmm")
  expect_equal(dim(fit$transition_probs), c(2L, 2L))
  expect_length(fit$emission_probs, 2L)
  expect_true(is.finite(fit$log_likelihood))
  decoded <- decode_multichannel_sequence_states(fit)
  expect_equal(nrow(decoded), nrow(data))
  expect_true(all(c("sequence_id", "sequence_order", "latent_state") %in% names(decoded)))
  summary <- summarise_multichannel_sequence_hmm(fit)
  expect_true(all(c("fit", "initial", "transition", "emission") %in% names(summary)))
  expect_invisible(plot_multichannel_sequence_hmm(fit, channel = "state"))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)
library(gp3sequences)

## ----data---------------------------------------------------------------------
paths <- list(
  s1 = c("home", "search", "product", "cart", "checkout"),
  s2 = c("home", "search", "product", "cart", "home"),
  s3 = c("home", "category", "product", "cart", "checkout"),
  s4 = c("home", "category", "product", "search", "checkout"),
  s5 = c("home", "category", "search", "product", "checkout"),
  s6 = c("home", "search", "category", "product", "home"),
  s7 = c("home", "category", "product", "cart", "home"),
  s8 = c("home", "search", "product", "checkout", "home")
)

raw_sequences <- do.call(
  rbind,
  lapply(seq_along(paths), function(i) {
    data.frame(
      sequence_id = names(paths)[i],
      sequence_order = seq_along(paths[[i]]),
      state = paths[[i]],
      duration = 80 + 10 * seq_along(paths[[i]]) + i,
      participant_id = sprintf("p%02d", i),
      group = if (i <= 4L) "interface_a" else "interface_b",
      stringsAsFactors = FALSE
    )
  })
)

raw_sequences

## ----contract-----------------------------------------------------------------
audit <- audit_sequence_data(
  raw_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

validation <- validate_sequence_data(
  raw_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

prepared <- prepare_sequence_data(
  raw_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  missing_state_policy = "error",
  duplicate_position_policy = "error",
  repeated_state_policy = "preserve",
  zero_duration_policy = "preserve",
  unknown_state_policy = "preserve",
  unused_state_levels = "preserve"
)

validation$status
prepared$status
prepared$mapping
prepared$decisions
head(prepared$data)

## ----summaries----------------------------------------------------------------
encoded <- encode_sequence_data(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

state_summary <- summarise_sequence_states(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

transition_summary <- summarise_sequence_transitions(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "group"),
  include_self = TRUE
)

paths_table <- format_sequence_paths(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "group")
)

encoded$dictionary
state_summary$overall
head(transition_summary$overall)
paths_table$paths

## ----motifs-------------------------------------------------------------------
motif_occurrences <- extract_sequence_ngrams(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = "group",
  min_length = 2L,
  max_length = 3L,
  overlap = "allow"
)

motif_summary <- summarise_sequence_motifs(motif_occurrences)

motif_filter <- filter_sequence_motifs(
  motif_summary,
  min_occurrences = 2L,
  min_sequences = 2L,
  min_prevalence = 0.20,
  motif_lengths = c(2L, 3L),
  top_n = 10L,
  rank_by = "sequence_prevalence",
  ties = "include"
)

motif_table <- format_sequence_motifs(
  motif_filter,
  prevalence = "percent",
  digits = 1L
)

motif_table$table

## ----distance-clustering------------------------------------------------------
lcs_distance <- compute_sequence_distance(
  prepared$data,
  method = "lcs",
  normalise = "max_length"
)

cluster_fit <- cluster_sequences(
  lcs_distance,
  k = 2L,
  method = "hierarchical",
  linkage = "average"
)

cluster_validation <- validate_sequence_clusters(cluster_fit)
representatives <- extract_representative_sequences(cluster_fit)

summarise_sequence_distance(lcs_distance)$overall
cluster_fit$assignments
cluster_validation$overall
representatives

## ----consensus-groups---------------------------------------------------------
consensus <- create_consensus_sequence(
  prepared$data,
  group_cols = "group",
  tie_method = "first",
  state_levels = encoded$dictionary$state
)

group_comparison <- compare_sequence_groups(
  prepared$data,
  group_col = "group"
)

summarise_consensus_agreement(consensus, by = "group")
format_consensus_sequence(consensus, include_agreement = TRUE)
head(group_comparison$state_contrasts)
head(group_comparison$transition_contrasts)
group_comparison$length_contrasts

## ----networks-----------------------------------------------------------------
network <- create_transition_network(
  prepared$data,
  normalise = "from",
  include_self = TRUE
)

network
summarise_transition_centrality(network)
detect_transition_communities(network)

#' Prepare longitudinal or panel sequence data
#'
#' Validates a collection of sequences observed repeatedly for the same panel
#' units and creates an auditable panel index. Each sequence must map to exactly
#' one panel unit and one occasion.
#'
#' @param data Long-format sequence data or a prepared gp3sequences result.
#' @param panel_id_col Column identifying the repeatedly observed unit.
#' @param occasion_col Column identifying the wave, visit, or occasion.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param metadata_cols Optional columns that must remain constant within each
#'   sequence.
#' @param require_unique_occasions Require at most one sequence per panel unit
#'   and occasion.
#'
#' @return An object of class `gp3_sequence_panel` containing canonical data,
#'   a panel index, state levels, and mapping information.
#' @examples
#' panel <- data.frame(
#'   participant_id = rep(c("p1", "p2"), each = 8L),
#'   occasion = rep(rep(c(1, 2), each = 4L), times = 2L),
#'   sequence_id = rep(c("p1_w1", "p1_w2", "p2_w1", "p2_w2"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "C", "C", "D",
#'             "D", "C", "B", "A", "D", "B", "B", "A")
#' )
#' prepare_sequence_panel(panel, "participant_id", "occasion")
#' @export
prepare_sequence_panel <- function(
  data,
  panel_id_col,
  occasion_col,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = NULL,
  require_unique_occasions = TRUE
) {
  .sequence_adv_scalar_character(panel_id_col, "panel_id_col")
  .sequence_adv_scalar_character(occasion_col, "occasion_col")
  .sequence_adv_scalar_logical(require_unique_occasions, "require_unique_occasions")
  metadata_cols <- unique(c(panel_id_col, occasion_col, metadata_cols))
  x <- .sequence_adv_data(
    data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = metadata_cols,
    missing_state_policy = "error"
  )
  metadata <- x$metadata
  panel_id <- metadata[[panel_id_col]]
  occasion <- metadata[[occasion_col]]
  panel_missing <- is.na(panel_id) | trimws(as.character(panel_id)) == ""
  occasion_missing <- is.na(occasion) | trimws(as.character(occasion)) == ""
  if (any(panel_missing)) {
    stop("Panel identifiers must not be missing or blank.", call. = FALSE)
  }
  if (any(occasion_missing)) {
    stop("Occasion values must not be missing or blank.", call. = FALSE)
  }
  panel_text <- as.character(panel_id)
  occasion_text <- as.character(occasion)
  duplicate_key <- paste(panel_text, occasion_text, sep = "\034")
  if (require_unique_occasions && anyDuplicated(duplicate_key)) {
    duplicated_rows <- which(duplicated(duplicate_key) | duplicated(duplicate_key, fromLast = TRUE))
    stop(
      "More than one sequence was found for panel/occasion combinations: ",
      paste(unique(duplicate_key[duplicated_rows]), collapse = ", "), ".",
      call. = FALSE
    )
  }
  occasion_rank <- integer(length(occasion))
  for (panel in unique(panel_text)) {
    rows <- which(panel_text == panel)
    current <- occasion[rows]
    if (is.numeric(current) || inherits(current, c("Date", "POSIXct", "POSIXlt"))) {
      ordering <- order(current, as.character(metadata[[sequence_id_col]][rows]), method = "radix")
    } else if (is.factor(current) && is.ordered(current)) {
      ordering <- order(as.integer(current), as.character(metadata[[sequence_id_col]][rows]), method = "radix")
    } else {
      ordering <- order(as.character(current), as.character(metadata[[sequence_id_col]][rows]), method = "radix")
    }
    occasion_rank[rows[ordering]] <- seq_along(rows)
  }
  sequence_lengths <- lengths(x$sequences)
  transition_counts <- pmax(sequence_lengths - 1L, 0L)
  index <- data.frame(
    sequence_id = x$sequence_ids,
    panel_id = panel_text,
    occasion = occasion_text,
    occasion_rank = as.integer(occasion_rank),
    sequence_length = as.integer(sequence_lengths[x$sequence_ids]),
    transition_count = as.integer(transition_counts[x$sequence_ids]),
    stringsAsFactors = FALSE
  )
  index <- index[order(index$panel_id, index$occasion_rank, index$sequence_id,
                       method = "radix"), , drop = FALSE]
  row.names(index) <- NULL
  result <- list(
    data = x$data,
    index = index,
    sequences = x$sequences,
    orders = x$orders,
    state_levels = x$state_levels,
    columns = list(
      panel_id = panel_id_col,
      occasion = occasion_col,
      sequence_id = sequence_id_col,
      order = order_col,
      state = state_col,
      metadata = metadata_cols
    ),
    settings = list(require_unique_occasions = require_unique_occasions),
    call = match.call()
  )
  class(result) <- c("gp3_sequence_panel", "list")
  result
}

#' Summarise a sequence panel
#'
#' @param panel A result from [prepare_sequence_panel()].
#'
#' @return A list with occasion-level sequence summaries and state prevalence.
#' @examples
#' panel_data <- data.frame(
#'   participant_id = rep(c("p1", "p2"), each = 6L),
#'   occasion = rep(rep(c(1, 2), each = 3L), times = 2L),
#'   sequence_id = rep(c("a", "b", "c", "d"), each = 3L),
#'   sequence_order = rep(1:3, times = 4L),
#'   state = c("A", "B", "C", "A", "C", "C", "C", "B", "A", "C", "B", "B")
#' )
#' summarise_sequence_panel(prepare_sequence_panel(panel_data, "participant_id", "occasion"))
#' @export
summarise_sequence_panel <- function(panel) {
  if (!inherits(panel, "gp3_sequence_panel")) {
    stop("`panel` must be created by `prepare_sequence_panel()`.", call. = FALSE)
  }
  index <- panel$index
  occasion_levels <- unique(index$occasion[order(index$occasion_rank, index$occasion,
                                                  method = "radix")])
  occasion_rows <- lapply(occasion_levels, function(occasion) {
    current <- index[index$occasion == occasion, , drop = FALSE]
    data.frame(
      occasion = occasion,
      n_panels = length(unique(current$panel_id)),
      n_sequences = nrow(current),
      mean_length = mean(current$sequence_length),
      median_length = stats::median(current$sequence_length),
      mean_transitions = mean(current$transition_count),
      stringsAsFactors = FALSE
    )
  })
  state_rows <- list()
  h <- 0L
  for (occasion in occasion_levels) {
    ids <- index$sequence_id[index$occasion == occasion]
    sequences <- panel$sequences[ids]
    for (state in panel$state_levels) {
      h <- h + 1L
      occurrence_count <- sum(vapply(sequences, function(x) sum(x == state), integer(1)))
      sequence_count <- sum(vapply(sequences, function(x) any(x == state), logical(1)))
      total_positions <- sum(lengths(sequences))
      state_rows[[h]] <- data.frame(
        occasion = occasion,
        state = state,
        occurrence_count = as.integer(occurrence_count),
        occurrence_share = if (total_positions > 0) occurrence_count / total_positions else NA_real_,
        sequence_count = as.integer(sequence_count),
        sequence_prevalence = if (length(sequences) > 0) sequence_count / length(sequences) else NA_real_,
        stringsAsFactors = FALSE
      )
    }
  }
  list(
    occasions = do.call(rbind, occasion_rows),
    states = do.call(rbind, state_rows),
    n_panels = length(unique(index$panel_id)),
    n_occasions = length(occasion_levels),
    state_levels = panel$state_levels
  )
}

#' Compare within-panel sequence changes
#'
#' Computes distance and simple structural changes between consecutive panel
#' occasions. The output is descriptive and does not establish a causal or
#' psychological change.
#'
#' @param panel A sequence panel.
#' @param method Distance method supported by [compute_sequence_distance()].
#' @param normalise Distance normalisation.
#' @param indel_cost,substitution_cost Costs for optimal matching.
#' @param substitution_matrix Optional named substitution matrix.
#' @param transition_smoothing Smoothing for transition-profile distance.
#'
#' @return A data frame of class `gp3_sequence_panel_changes`.
#' @examples
#' panel_data <- data.frame(
#'   participant_id = rep(c("p1", "p2"), each = 8L),
#'   occasion = rep(rep(c(1, 2), each = 4L), times = 2L),
#'   sequence_id = rep(c("a", "b", "c", "d"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "C", "C", "D",
#'             "D", "C", "B", "A", "D", "B", "B", "A")
#' )
#' compare_sequence_panel_changes(
#'   prepare_sequence_panel(panel_data, "participant_id", "occasion"),
#'   method = "lcs"
#' )
#' @export
compare_sequence_panel_changes <- function(
  panel,
  method = c("levenshtein", "lcs", "optimal_matching", "transition"),
  normalise = c("none", "max_length", "path_length"),
  indel_cost = 1,
  substitution_cost = 1,
  substitution_matrix = NULL,
  transition_smoothing = 0
) {
  if (!inherits(panel, "gp3_sequence_panel")) {
    stop("`panel` must be created by `prepare_sequence_panel()`.", call. = FALSE)
  }
  method <- match.arg(method)
  normalise <- match.arg(normalise)
  index <- panel$index
  pieces <- list()
  h <- 0L
  for (panel_id in unique(index$panel_id)) {
    current <- index[index$panel_id == panel_id, , drop = FALSE]
    current <- current[order(current$occasion_rank, current$sequence_id, method = "radix"), , drop = FALSE]
    if (nrow(current) < 2L) next
    for (i in seq_len(nrow(current) - 1L)) {
      ids <- current$sequence_id[c(i, i + 1L)]
      subset_data <- panel$data[as.character(panel$data[[panel$columns$sequence_id]]) %in% ids,
                                , drop = FALSE]
      distance <- compute_sequence_distance(
        subset_data,
        sequence_id_col = panel$columns$sequence_id,
        order_col = panel$columns$order,
        state_col = panel$columns$state,
        method = method,
        indel_cost = indel_cost,
        substitution_cost = substitution_cost,
        substitution_matrix = substitution_matrix,
        transition_smoothing = transition_smoothing,
        normalise = normalise
      )
      value <- as.numeric(distance)[1L]
      h <- h + 1L
      pieces[[h]] <- data.frame(
        panel_id = panel_id,
        from_sequence_id = ids[1L],
        to_sequence_id = ids[2L],
        from_occasion = current$occasion[i],
        to_occasion = current$occasion[i + 1L],
        from_rank = current$occasion_rank[i],
        to_rank = current$occasion_rank[i + 1L],
        distance = value,
        length_change = current$sequence_length[i + 1L] - current$sequence_length[i],
        transition_change = current$transition_count[i + 1L] - current$transition_count[i],
        stringsAsFactors = FALSE
      )
    }
  }
  if (length(pieces) == 0L) {
    result <- data.frame(
      panel_id = character(), from_sequence_id = character(), to_sequence_id = character(),
      from_occasion = character(), to_occasion = character(), from_rank = integer(),
      to_rank = integer(), distance = numeric(), length_change = numeric(),
      transition_change = numeric(), stringsAsFactors = FALSE
    )
  } else {
    result <- do.call(rbind, pieces)
    row.names(result) <- NULL
  }
  class(result) <- c("gp3_sequence_panel_changes", "data.frame")
  attr(result, "method") <- method
  attr(result, "normalise") <- normalise
  result
}

#' Plot longitudinal sequence changes
#'
#' @param changes A result from [compare_sequence_panel_changes()].
#' @param metric One of `"distance"`, `"length_change"`, or
#'   `"transition_change"`.
#' @param type Plot individual panel trajectories or occasion-transition means.
#' @param ... Additional graphical arguments passed to base plotting functions.
#'
#' @return The plotted data, invisibly.
#' @examples
#' panel_data <- data.frame(
#'   participant_id = rep(c("p1", "p2"), each = 6L),
#'   occasion = rep(rep(c(1, 2), each = 3L), times = 2L),
#'   sequence_id = rep(c("a", "b", "c", "d"), each = 3L),
#'   sequence_order = rep(1:3, times = 4L),
#'   state = c("A", "B", "C", "A", "C", "C", "C", "B", "A", "C", "B", "B")
#' )
#' changes <- compare_sequence_panel_changes(
#'   prepare_sequence_panel(panel_data, "participant_id", "occasion")
#' )
#' plot_sequence_panel_changes(changes)
#' @export
plot_sequence_panel_changes <- function(
  changes,
  metric = c("distance", "length_change", "transition_change"),
  type = c("individual", "summary"),
  ...
) {
  if (!inherits(changes, "gp3_sequence_panel_changes")) {
    stop("`changes` must be created by `compare_sequence_panel_changes()`.",
         call. = FALSE)
  }
  metric <- match.arg(metric)
  type <- match.arg(type)
  if (nrow(changes) == 0L) stop("There are no panel changes to plot.", call. = FALSE)
  transition_label <- paste(changes$from_occasion, changes$to_occasion, sep = " -> ")
  transition_levels <- unique(transition_label[order(changes$from_rank, changes$to_rank,
                                                      transition_label, method = "radix")])
  x <- match(transition_label, transition_levels)
  y <- changes[[metric]]
  if (type == "individual") {
    graphics::plot(x, y, type = "n", xaxt = "n", xlab = "Occasion transition",
                   ylab = metric, ...)
    graphics::axis(1, at = seq_along(transition_levels), labels = transition_levels)
    panels <- unique(changes$panel_id)
    colors <- .sequence_ext_base_colors(length(panels))
    for (i in seq_along(panels)) {
      rows <- which(changes$panel_id == panels[i])
      graphics::lines(x[rows], y[rows], type = "b", col = colors[i])
    }
  } else {
    means <- tapply(y, factor(transition_label, levels = transition_levels), mean)
    standard_errors <- tapply(y, factor(transition_label, levels = transition_levels),
                              function(z) if (length(z) > 1L) stats::sd(z) / sqrt(length(z)) else 0)
    graphics::plot(seq_along(means), means, type = "b", xaxt = "n",
                   ylim = range(c(means - standard_errors, means + standard_errors), finite = TRUE),
                   xlab = "Occasion transition", ylab = metric, ...)
    graphics::axis(1, at = seq_along(transition_levels), labels = transition_levels)
    error_index <- which(
      is.finite(standard_errors) &
        standard_errors > 0
    )
    if (length(error_index) > 0L) {
      graphics::arrows(
        error_index,
        means[error_index] - standard_errors[error_index],
        error_index,
        means[error_index] + standard_errors[error_index],
        angle = 90,
        code = 3,
        length = 0.05
      )
    }
  }
  invisible(changes)
}

test_that("sequence panels are prepared and summarised", {
  data <- make_extension_panel_data()
  panel <- prepare_sequence_panel(
    data,
    panel_id_col = "participant_id",
    occasion_col = "occasion"
  )
  expect_s3_class(panel, "gp3_sequence_panel")
  expect_equal(nrow(panel$index), 16L)
  expect_equal(length(unique(panel$index$panel_id)), 8L)
  expect_equal(sort(unique(panel$index$occasion_rank)), 1:2)

  summary <- summarise_sequence_panel(panel)
  expect_equal(summary$n_panels, 8L)
  expect_equal(summary$n_occasions, 2L)
  expect_true(all(c("occasions", "states") %in% names(summary)))
})

test_that("panel changes are deterministic", {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  panel <- prepare_sequence_panel(
    make_extension_panel_data(), "participant_id", "occasion"
  )
  changes <- compare_sequence_panel_changes(panel, method = "lcs")
  expect_s3_class(changes, "gp3_sequence_panel_changes")
  expect_equal(nrow(changes), 8L)
  expect_true(all(changes$distance >= 0))
  expect_silent(plot_sequence_panel_changes(changes, type = "summary"))

  zero_error_changes <- changes
  zero_error_changes$distance <- 1
  expect_silent(
    plot_sequence_panel_changes(
      zero_error_changes,
      metric = "distance",
      type = "summary"
    )
  )
})

test_that("duplicated panel occasions fail explicitly", {
  data <- make_extension_panel_data()
  data$occasion[data$sequence_id == "s01_w2"] <- 1L
  expect_error(
    prepare_sequence_panel(data, "participant_id", "occasion"),
    "More than one sequence"
  )
})

# Internal helpers for the post-0.2.0 advanced extension roadmap.

.sequence_ext_entropy <- function(probabilities, base = 2) {
  probabilities <- probabilities[is.finite(probabilities) & probabilities > 0]
  if (length(probabilities) == 0L) return(0)
  -sum(probabilities * (log(probabilities) / log(base)))
}

.sequence_ext_softmax <- function(eta) {
  eta <- as.numeric(eta)
  eta <- eta - max(eta)
  value <- exp(eta)
  value / sum(value)
}

.sequence_ext_numeric_matrix <- function(data, columns, center = NULL, scale = NULL) {
  columns <- .sequence_adv_match_cols(data, columns, "covariate columns")
  if (length(columns) == 0L) {
    x <- matrix(1, nrow = nrow(data), ncol = 1L,
                dimnames = list(NULL, "(Intercept)"))
    return(list(matrix = x, center = numeric(), scale = numeric(), columns = character()))
  }
  invalid <- vapply(data[columns], function(x) {
    !is.numeric(x) || anyNA(x) || any(!is.finite(x))
  }, logical(1))
  if (any(invalid)) {
    stop("Covariates must be finite, non-missing numeric columns: ",
         paste(columns[invalid], collapse = ", "), ".", call. = FALSE)
  }
  raw <- as.matrix(data[columns])
  storage.mode(raw) <- "double"
  if (is.null(center)) center <- colMeans(raw)
  if (is.null(scale)) {
    scale <- apply(raw, 2L, stats::sd)
    scale[!is.finite(scale) | scale <= 0] <- 1
  }
  if (!identical(names(center), columns)) names(center) <- columns
  if (!identical(names(scale), columns)) names(scale) <- columns
  standardized <- sweep(raw, 2L, center[columns], "-")
  standardized <- sweep(standardized, 2L, scale[columns], "/")
  x <- cbind("(Intercept)" = 1, standardized)
  list(matrix = x, center = center, scale = scale, columns = columns)
}

.sequence_ext_fit_softmax <- function(x, counts, start = NULL, ridge = 1e-6,
                                      maxit = 100L) {
  if (!is.matrix(x) || !is.numeric(x) || anyNA(x) || any(!is.finite(x))) {
    stop("`x` must be a finite numeric matrix.", call. = FALSE)
  }
  if (!is.matrix(counts) || !is.numeric(counts) || nrow(counts) != nrow(x) ||
      ncol(counts) < 2L || anyNA(counts) || any(!is.finite(counts)) ||
      any(counts < 0)) {
    stop("`counts` must be a non-negative finite matrix aligned with `x`.",
         call. = FALSE)
  }
  .sequence_adv_scalar_number(ridge, "ridge", lower = 0)
  .sequence_adv_scalar_number(maxit, "maxit", lower = 1, integer = TRUE)
  p <- ncol(x)
  k <- ncol(counts)
  if (is.null(start)) {
    start <- matrix(0, nrow = p, ncol = k)
  }
  if (!is.matrix(start) || !all(dim(start) == c(p, k))) {
    stop("`start` has incompatible dimensions.", call. = FALSE)
  }
  theta0 <- as.numeric(start[, seq_len(k - 1L), drop = FALSE])
  objective <- function(theta) {
    beta <- cbind(matrix(theta, nrow = p, ncol = k - 1L), 0)
    eta <- x %*% beta
    row_max <- apply(eta, 1L, max)
    exp_eta <- exp(eta - row_max)
    probabilities <- exp_eta / rowSums(exp_eta)
    probabilities <- pmax(probabilities, .Machine$double.xmin)
    value <- -sum(counts * log(probabilities)) + 0.5 * ridge * sum(theta^2)
    totals <- rowSums(counts)
    residual <- probabilities * totals - counts
    gradient <- t(x) %*% residual[, seq_len(k - 1L), drop = FALSE]
    gradient <- as.numeric(gradient) + ridge * theta
    attr(value, "gradient") <- gradient
    value
  }
  fit <- stats::optim(
    par = theta0,
    fn = function(theta) objective(theta),
    gr = function(theta) attr(objective(theta), "gradient"),
    method = "BFGS",
    control = list(maxit = as.integer(maxit), reltol = 1e-9)
  )
  beta <- cbind(matrix(fit$par, nrow = p, ncol = k - 1L), 0)
  rownames(beta) <- colnames(x)
  colnames(beta) <- colnames(counts)
  list(coefficients = beta, convergence = fit$convergence,
       value = fit$value, counts = fit$counts)
}

.sequence_ext_forward_backward <- function(emission_likelihood, initial,
                                           transition) {
  if (!is.matrix(emission_likelihood) || !is.numeric(emission_likelihood) ||
      nrow(emission_likelihood) < 1L || ncol(emission_likelihood) < 1L ||
      anyNA(emission_likelihood) || any(!is.finite(emission_likelihood)) ||
      any(emission_likelihood < 0)) {
    stop("Invalid emission-likelihood matrix.", call. = FALSE)
  }
  n_time <- nrow(emission_likelihood)
  n_states <- ncol(emission_likelihood)
  if (!is.numeric(initial) || length(initial) != n_states || any(initial < 0)) {
    stop("Invalid initial probabilities.", call. = FALSE)
  }
  transition_at <- function(t) {
    if (is.matrix(transition)) return(transition)
    transition[t, , , drop = TRUE]
  }
  if (is.matrix(transition)) {
    if (!all(dim(transition) == c(n_states, n_states))) {
      stop("Invalid transition matrix.", call. = FALSE)
    }
  } else if (!is.array(transition) || length(dim(transition)) != 3L ||
             !all(dim(transition) == c(max(n_time - 1L, 0L), n_states, n_states))) {
    stop("Invalid time-varying transition array.", call. = FALSE)
  }
  alpha <- matrix(0, nrow = n_time, ncol = n_states)
  scales <- numeric(n_time)
  alpha[1L, ] <- initial * emission_likelihood[1L, ]
  scales[1L] <- sum(alpha[1L, ])
  if (!is.finite(scales[1L]) || scales[1L] <= 0) scales[1L] <- .Machine$double.xmin
  alpha[1L, ] <- alpha[1L, ] / scales[1L]
  if (n_time > 1L) {
    for (t in 2:n_time) {
      current_transition <- transition_at(t - 1L)
      alpha[t, ] <- as.numeric(alpha[t - 1L, ] %*% current_transition) *
        emission_likelihood[t, ]
      scales[t] <- sum(alpha[t, ])
      if (!is.finite(scales[t]) || scales[t] <= 0) scales[t] <- .Machine$double.xmin
      alpha[t, ] <- alpha[t, ] / scales[t]
    }
  }
  beta <- matrix(1, nrow = n_time, ncol = n_states)
  if (n_time > 1L) {
    for (t in seq.int(n_time - 1L, 1L)) {
      current_transition <- transition_at(t)
      beta[t, ] <- current_transition %*%
        (emission_likelihood[t + 1L, ] * beta[t + 1L, ])
      beta[t, ] <- beta[t, ] / scales[t + 1L]
    }
  }
  gamma <- alpha * beta
  gamma_total <- rowSums(gamma)
  invalid <- !is.finite(gamma_total) | gamma_total <= 0
  if (any(invalid)) {
    gamma[invalid, ] <- 1 / n_states
    gamma_total[invalid] <- 1
  }
  gamma <- gamma / gamma_total
  xi <- array(0, dim = c(max(n_time - 1L, 0L), n_states, n_states))
  if (n_time > 1L) {
    for (t in seq_len(n_time - 1L)) {
      current_transition <- transition_at(t)
      current <- outer(alpha[t, ],
                       emission_likelihood[t + 1L, ] * beta[t + 1L, ]) *
        current_transition
      total <- sum(current)
      if (is.finite(total) && total > 0) current <- current / total
      xi[t, , ] <- current
    }
  }
  list(alpha = alpha, beta = beta, gamma = gamma, xi = xi,
       log_likelihood = sum(log(scales)))
}

.sequence_ext_viterbi <- function(emission_likelihood, initial, transition) {
  n_time <- nrow(emission_likelihood)
  n_states <- ncol(emission_likelihood)
  transition_at <- function(t) {
    if (is.matrix(transition)) transition else transition[t, , , drop = TRUE]
  }
  delta <- matrix(-Inf, nrow = n_time, ncol = n_states)
  psi <- matrix(0L, nrow = n_time, ncol = n_states)
  delta[1L, ] <- log(pmax(initial, .Machine$double.xmin)) +
    log(pmax(emission_likelihood[1L, ], .Machine$double.xmin))
  if (n_time > 1L) {
    for (t in 2:n_time) {
      current_transition <- transition_at(t - 1L)
      for (j in seq_len(n_states)) {
        candidates <- delta[t - 1L, ] +
          log(pmax(current_transition[, j], .Machine$double.xmin))
        psi[t, j] <- which.max(candidates)
        delta[t, j] <- max(candidates) +
          log(pmax(emission_likelihood[t, j], .Machine$double.xmin))
      }
    }
  }
  path <- integer(n_time)
  path[n_time] <- which.max(delta[n_time, ])
  if (n_time > 1L) {
    for (t in seq.int(n_time - 1L, 1L)) {
      path[t] <- psi[t + 1L, path[t + 1L]]
    }
  }
  list(path = path, log_probability = max(delta[n_time, ]))
}

.sequence_ext_subsequence_present <- function(sequence, pattern) {
  sequence <- as.character(sequence)
  pattern <- as.character(pattern)
  if (length(pattern) == 0L) return(TRUE)
  position <- 1L
  for (value in sequence) {
    if (identical(value, pattern[position])) position <- position + 1L
    if (position > length(pattern)) return(TRUE)
  }
  FALSE
}

.sequence_ext_base_colors <- function(n, palette = "Dark 3") {
  if (n < 1L) return(character())
  grDevices::hcl.colors(n, palette = palette)
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----data---------------------------------------------------------------------
paths <- replicate(20, sample(c("A", "B", "C"), 6L, replace = TRUE), simplify = FALSE)
data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(
    participant_id = paste0("p", i),
    sequence_id = paste0("s", i),
    sequence_order = seq_along(paths[[i]]),
    state = paths[[i]],
    group = if (i <= 10L) "control" else "treatment",
    stringsAsFactors = FALSE
  )
}))

## ----test---------------------------------------------------------------------
design <- declare_sequence_comparison_design(
  group_col = "group",
  unit_col = "participant_id",
  design = "randomized"
)

result <- test_sequence_group_difference(
  data,
  design,
  metric = "state_prevalence",
  target_state = "A",
  n_permutations = 999L,
  seed = 10L
)
result$estimate

## ----bootstrap----------------------------------------------------------------
result <- bootstrap_sequence_group_difference(
  result,
  n_boot = 999L,
  level = 0.95,
  seed = 11L
)
summarise_sequence_group_inference(result)

## ----plots, fig.width=7, fig.height=4-----------------------------------------
plot_sequence_group_inference(result, type = "permutation")
plot_sequence_group_inference(result, type = "group_means")

make_advanced_sequence_data <- function() {
  paths <- list(
    s01 = c("A", "B", "C", "D", "D"),
    s02 = c("A", "B", "C", "D", "C"),
    s03 = c("A", "B", "B", "C", "D"),
    s04 = c("A", "C", "C", "D", "D"),
    s05 = c("D", "C", "B", "A", "A"),
    s06 = c("D", "C", "B", "A", "B"),
    s07 = c("D", "C", "C", "B", "A"),
    s08 = c("D", "B", "B", "A", "A")
  )
  group <- rep(c("g1", "g2"), each = 4L)
  participant <- paste0("p", seq_along(paths))
  rows <- lapply(seq_along(paths), function(i) {
    data.frame(
      sequence_id = names(paths)[i],
      sequence_order = seq_along(paths[[i]]),
      state = paths[[i]],
      group = group[i],
      participant_id = participant[i],
      weight = 1,
      stringsAsFactors = FALSE
    )
  })
  do.call(rbind, rows)
}

.sequence_motif_match_arg <- function(value, choices, argument) {
  invalid_message <- paste0(
    "`",
    argument,
    "` must be one of: ",
    paste(
      paste0("`", choices, "`"),
      collapse = ", "
    ),
    "."
  )

  if (
    !is.character(value) ||
      length(value) < 1L ||
      anyNA(value)
  ) {
    stop(
      invalid_message,
      call. = FALSE
    )
  }

  tryCatch(
    match.arg(value, choices),
    error = function(error) {
      stop(
        invalid_message,
        call. = FALSE
      )
    }
  )
}

.sequence_motif_assert_whole_number <- function(
  value,
  argument,
  minimum = 0L,
  allow_null = FALSE
) {
  if (allow_null && is.null(value)) {
    return(invisible(NULL))
  }

  if (
    !is.numeric(value) ||
      length(value) != 1L ||
      is.na(value) ||
      !is.finite(value) ||
      value < minimum ||
      value != floor(value)
  ) {
    stop(
      "`",
      argument,
      "` must be one whole number greater than or equal to ",
      minimum,
      ".",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_motif_assert_proportion <- function(value, argument) {
  if (
    !is.numeric(value) ||
      length(value) != 1L ||
      is.na(value) ||
      !is.finite(value) ||
      value < 0 ||
      value > 1
  ) {
    stop(
      "`",
      argument,
      "` must be one finite number between 0 and 1.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_motif_assert_lengths <- function(min_length, max_length) {
  .sequence_motif_assert_whole_number(
    min_length,
    "min_length",
    minimum = 1L
  )

  .sequence_motif_assert_whole_number(
    max_length,
    "max_length",
    minimum = 1L
  )

  if (max_length < min_length) {
    stop(
      "`max_length` must be greater than or equal to `min_length`.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_motif_assert_motif_lengths <- function(value) {
  if (is.null(value)) {
    return(invisible(NULL))
  }

  if (
    !is.numeric(value) ||
      length(value) == 0L ||
      anyNA(value) ||
      any(!is.finite(value)) ||
      any(value < 1) ||
      any(value != floor(value))
  ) {
    stop(
      "`motif_lengths` must contain positive whole numbers or be `NULL`.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_motif_empty_occurrences <- function(metadata_cols) {
  result <- data.frame(
    sequence_id = character(),
    motif_id = character(),
    motif_key = character(),
    motif = character(),
    motif_length = integer(),
    start_index = integer(),
    end_index = integer(),
    start_order = double(),
    end_order = double(),
    start_original_row = integer(),
    end_original_row = integer(),
    occurrence_index = integer(),
    stringsAsFactors = FALSE
  )

  for (column in metadata_cols) {
    result[[column]] <- character()
  }

  result[
    c(
      "sequence_id",
      metadata_cols,
      "motif_id",
      "motif_key",
      "motif",
      "motif_length",
      "start_index",
      "end_index",
      "start_order",
      "end_order",
      "start_original_row",
      "end_original_row",
      "occurrence_index"
    )
  ]
}

.sequence_motif_empty_dictionary <- function() {
  data.frame(
    motif_id = character(),
    motif_key = character(),
    motif = character(),
    motif_length = integer(),
    stringsAsFactors = FALSE
  )
}

.sequence_motif_empty_sequences <- function(metadata_cols) {
  result <- data.frame(
    sequence_id = character(),
    n_states = integer(),
    n_candidate_occurrences = integer(),
    n_retained_occurrences = integer(),
    n_distinct_motifs = integer(),
    stringsAsFactors = FALSE
  )

  for (column in metadata_cols) {
    result[[column]] <- character()
  }

  result[
    c(
      "sequence_id",
      metadata_cols,
      "n_states",
      "n_candidate_occurrences",
      "n_retained_occurrences",
      "n_distinct_motifs"
    )
  ]
}

.sequence_motif_empty_by_sequence <- function(metadata_cols) {
  result <- data.frame(
    sequence_id = character(),
    motif_id = character(),
    motif_key = character(),
    motif = character(),
    motif_length = integer(),
    n_occurrences = integer(),
    first_start_index = integer(),
    last_start_index = integer(),
    stringsAsFactors = FALSE
  )

  for (column in metadata_cols) {
    result[[column]] <- character()
  }

  result[
    c(
      "sequence_id",
      metadata_cols,
      "motif_id",
      "motif_key",
      "motif",
      "motif_length",
      "n_occurrences",
      "first_start_index",
      "last_start_index"
    )
  ]
}

.sequence_motif_empty_overall <- function() {
  data.frame(
    motif_id = character(),
    motif_key = character(),
    motif = character(),
    motif_length = integer(),
    n_occurrences = integer(),
    n_sequences = integer(),
    sequence_prevalence = double(),
    occurrence_share = double(),
    mean_occurrences_per_sequence = double(),
    mean_occurrences_when_present = double(),
    stringsAsFactors = FALSE
  )
}

.sequence_motif_make_id <- function(motif_length, state_codes) {
  paste0(
    "L",
    motif_length,
    ":",
    paste(state_codes, collapse = "|")
  )
}

.sequence_motif_sort_overall <- function(data, primary) {
  if (nrow(data) == 0L) {
    return(data)
  }

  secondary <- c(
    "sequence_prevalence",
    "n_occurrences",
    "n_sequences"
  )

  secondary <- setdiff(secondary, primary)

  order_arguments <- list(-data[[primary]])

  for (column in secondary) {
    order_arguments[[length(order_arguments) + 1L]] <- -data[[column]]
  }

  order_arguments[[length(order_arguments) + 1L]] <- data$motif_length
  order_arguments[[length(order_arguments) + 1L]] <- data$motif_key
  order_arguments$method <- "radix"

  data <- data[do.call(order, order_arguments), , drop = FALSE]
  row.names(data) <- NULL
  data
}

.sequence_motif_select_non_overlapping <- function(occurrences) {
  if (nrow(occurrences) == 0L) {
    return(occurrences)
  }

  group <- interaction(
    occurrences$sequence_id,
    occurrences$motif_id,
    drop = TRUE,
    lex.order = TRUE
  )

  groups <- split(seq_len(nrow(occurrences)), group)
  keep <- logical(nrow(occurrences))

  for (indices in groups) {
    indices <- indices[
      order(
        occurrences$start_index[indices],
        occurrences$end_index[indices],
        method = "radix"
      )
    ]

    last_end <- -Inf

    for (index in indices) {
      if (occurrences$start_index[index] > last_end) {
        keep[index] <- TRUE
        last_end <- occurrences$end_index[index]
      }
    }
  }

  occurrences[keep, , drop = FALSE]
}

.sequence_motif_add_occurrence_index <- function(occurrences) {
  if (nrow(occurrences) == 0L) {
    occurrences$occurrence_index <- integer()
    return(occurrences)
  }

  group <- interaction(
    occurrences$sequence_id,
    occurrences$motif_id,
    drop = TRUE,
    lex.order = TRUE
  )

  groups <- split(seq_len(nrow(occurrences)), group)
  occurrence_index <- integer(nrow(occurrences))

  for (indices in groups) {
    indices <- indices[
      order(
        occurrences$start_index[indices],
        occurrences$end_index[indices],
        method = "radix"
      )
    ]

    occurrence_index[indices] <- seq_along(indices)
  }

  occurrences$occurrence_index <- as.integer(occurrence_index)
  occurrences
}

.sequence_motif_validate_extraction <- function(x) {
  if (
    !is.list(x) ||
      is.null(x$occurrences) ||
      is.null(x$motifs) ||
      is.null(x$sequences) ||
      is.null(x$settings)
  ) {
    stop(
      "`x` must be an object returned by `extract_sequence_ngrams()`.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_motif_as_summary <- function(x) {
  if (!is.list(x)) {
    stop(
      "`x` must be a motif extraction, summary, or filtered-motif object.",
      call. = FALSE
    )
  }

  if (!is.null(x$overall) && !is.null(x$by_sequence)) {
    return(x)
  }

  if (!is.null(x$motifs) && !is.null(x$by_sequence)) {
    result <- x
    result$overall <- x$motifs
    return(result)
  }

  if (!is.null(x$occurrences) && !is.null(x$sequences)) {
    return(summarise_sequence_motifs(x))
  }

  stop(
    "`x` must be a motif extraction, summary, or filtered-motif object.",
    call. = FALSE
  )
}

#' Extract Contiguous Sequence N-Grams
#'
#' Enumerates contiguous state motifs from validated long-format sequence data.
#'
#' @inheritParams audit_sequence_data
#' @param min_length Positive whole number giving the shortest motif length.
#' @param max_length Positive whole number giving the longest motif length.
#' @param overlap Character value specifying whether overlapping occurrences of
#'   the same motif within the same sequence are `"allow"`ed or
#'   `"disallow"`ed.
#' @param separator Character value used only to display state labels in the
#'   human-readable `motif` column.
#' @param state_levels Optional atomic vector defining the complete state
#'   ordering. When omitted, factor levels are respected; otherwise observed
#'   labels are sorted deterministically.
#'
#' @return A named list containing:
#'
#' * `occurrences`: one row per retained contiguous motif occurrence;
#' * `motifs`: the distinct motif dictionary;
#' * `sequences`: sequence-level counts of candidate and retained occurrences;
#' * `state_dictionary`: the deterministic state dictionary;
#' * `audit`, `status`, and `mapping` from input validation;
#' * `settings`: the resolved motif extraction settings.
#'
#' @details
#' Motifs are contiguous windows only. No subsequence gaps, edit distances,
#' statistical tests, or substantive interpretations are introduced.
#' Consecutive repeated states are used exactly as supplied; any repeat
#' collapsing should be performed explicitly with `prepare_sequence_data()`
#' before extraction.
#'
#' With `overlap = "disallow"`, overlap is resolved independently within each
#' sequence-motif pair by a deterministic left-to-right greedy rule. Different
#' motifs and different motif lengths do not compete for positions.
#'
#' The collision-resistant `motif_id` and `motif_key` columns are derived from
#' deterministic state codes. The display separator may therefore also occur
#' inside a state label without changing motif identity.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
#' )
#'
#' ngrams <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state",
#'   min_length = 2,
#'   max_length = 3,
#'   overlap = "allow"
#' )
#'
#' ngrams$occurrences
#' ngrams$motifs
#'
#' @export
extract_sequence_ngrams <- function(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL,
  min_length = 2L,
  max_length = 3L,
  overlap = c("allow", "disallow"),
  separator = " > ",
  state_levels = NULL
) {
  .sequence_motif_assert_lengths(min_length, max_length)
  overlap <- .sequence_motif_match_arg(
    overlap,
    c("allow", "disallow"),
    "overlap"
  )
  .sequence_assert_text_scalar(separator, "separator")

  .sequence_assert_output_names(
    metadata_cols = metadata_cols,
    reserved = c(
      "motif_id",
      "motif_key",
      "motif",
      "motif_length",
      "start_index",
      "end_index",
      "start_order",
      "end_order",
      "start_original_row",
      "end_original_row",
      "occurrence_index",
      "n_states",
      "n_candidate_occurrences",
      "n_retained_occurrences",
      "n_distinct_motifs"
    ),
    context = "motif"
  )

  encoded <- encode_sequence_data(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    duration_col = duration_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states,
    state_levels = state_levels,
    prefix = "S"
  )

  sequence_data <- encoded$data
  sequence_ids <- unique(sequence_data$sequence_id)
  occurrence_rows <- list()
  sequence_rows <- list()

  for (sequence_id in sequence_ids) {
    rows <- which(sequence_data$sequence_id == sequence_id)
    n_states <- length(rows)

    metadata <- .sequence_metadata_values(
      data = sequence_data,
      rows = rows,
      metadata_cols = metadata_cols
    )

    candidate_count <- 0L

    if (n_states >= min_length) {
      upper_length <- min(max_length, n_states)

      for (motif_length in seq.int(min_length, upper_length)) {
        n_windows <- n_states - motif_length + 1L
        candidate_count <- candidate_count + n_windows

        for (start_index in seq_len(n_windows)) {
          end_index <- start_index + motif_length - 1L
          window_rows <- rows[start_index:end_index]
          state_codes <- sequence_data$state_code[window_rows]
          state_labels <- sequence_data$state[window_rows]
          motif_key <- paste(state_codes, collapse = "|")

          values <- c(
            list(sequence_id = sequence_id),
            metadata,
            list(
              motif_id = .sequence_motif_make_id(
                motif_length,
                state_codes
              ),
              motif_key = motif_key,
              motif = paste(state_labels, collapse = separator),
              motif_length = as.integer(motif_length),
              start_index = as.integer(start_index),
              end_index = as.integer(end_index),
              start_order = sequence_data$sequence_order[
                window_rows[1L]
              ],
              end_order = sequence_data$sequence_order[
                window_rows[length(window_rows)]
              ],
              start_original_row = sequence_data$original_row[
                window_rows[1L]
              ],
              end_original_row = sequence_data$original_row[
                window_rows[length(window_rows)]
              ]
            )
          )

          occurrence_rows[[length(occurrence_rows) + 1L]] <-
            as.data.frame(
              values,
              stringsAsFactors = FALSE,
              check.names = FALSE
            )
        }
      }
    }

    sequence_rows[[length(sequence_rows) + 1L]] <-
      as.data.frame(
        c(
          list(sequence_id = sequence_id),
          metadata,
          list(
            n_states = as.integer(n_states),
            n_candidate_occurrences = as.integer(candidate_count),
            n_retained_occurrences = 0L,
            n_distinct_motifs = 0L
          )
        ),
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
  }

  occurrences <- .sequence_bind_frames(
    frames = occurrence_rows,
    empty = .sequence_motif_empty_occurrences(metadata_cols)
  )

  sequences <- .sequence_bind_frames(
    frames = sequence_rows,
    empty = .sequence_motif_empty_sequences(metadata_cols)
  )

  if (overlap == "disallow") {
    occurrences <- .sequence_motif_select_non_overlapping(occurrences)
  }

  if (nrow(occurrences) > 0L) {
    sequence_rank <- match(occurrences$sequence_id, sequence_ids)

    occurrences <- occurrences[
      order(
        sequence_rank,
        occurrences$start_index,
        occurrences$motif_length,
        occurrences$motif_key,
        method = "radix"
      ),
      ,
      drop = FALSE
    ]

    row.names(occurrences) <- NULL
    occurrences <- .sequence_motif_add_occurrence_index(occurrences)
  }

  if (nrow(sequences) > 0L) {
    for (index in seq_len(nrow(sequences))) {
      sequence_id <- sequences$sequence_id[index]
      occurrence_rows_for_sequence <- which(
        occurrences$sequence_id == sequence_id
      )

      sequences$n_retained_occurrences[index] <- as.integer(
        length(occurrence_rows_for_sequence)
      )

      sequences$n_distinct_motifs[index] <- as.integer(
        length(unique(occurrences$motif_id[occurrence_rows_for_sequence]))
      )
    }
  }

  motifs <- if (nrow(occurrences) == 0L) {
    .sequence_motif_empty_dictionary()
  } else {
    unique(
      occurrences[
        c("motif_id", "motif_key", "motif", "motif_length")
      ]
    )
  }

  if (nrow(motifs) > 0L) {
    motifs <- motifs[
      order(
        motifs$motif_length,
        motifs$motif_key,
        method = "radix"
      ),
      ,
      drop = FALSE
    ]
    row.names(motifs) <- NULL
  }

  list(
    occurrences = occurrences,
    motifs = motifs,
    sequences = sequences,
    state_dictionary = encoded$dictionary,
    audit = encoded$audit,
    status = encoded$status,
    mapping = encoded$mapping,
    settings = list(
      min_length = as.integer(min_length),
      max_length = as.integer(max_length),
      overlap = overlap,
      overlap_scope = "within_sequence_motif",
      overlap_rule = if (overlap == "disallow") {
        "left_to_right_greedy"
      } else {
        "all_contiguous_windows"
      },
      separator = separator,
      n_sequences = as.integer(length(sequence_ids))
    )
  )
}

#' Summarise Contiguous Sequence Motifs
#'
#' Aggregates extracted contiguous motif occurrences by sequence and overall.
#'
#' @param x An object returned by `extract_sequence_ngrams()`.
#'
#' @return A named list containing:
#'
#' * `by_sequence`: occurrence counts for each sequence-motif pair;
#' * `overall`: total occurrence counts, sequence counts, sequence prevalence,
#'   occurrence share, and mean occurrence rates;
#' * `sequences` and `state_dictionary` from extraction;
#' * `audit`, `status`, `mapping`, and extraction settings;
#' * scalar counts for sequences, occurrences, and distinct motifs.
#'
#' @details
#' Sequence prevalence uses every validated sequence in the extraction object as
#' its denominator, including sequences too short to contain a requested motif.
#' Results are sorted deterministically by sequence prevalence, occurrence
#' count, motif length, and motif key.
#'
#' The function reports structural recurrence only. It does not perform
#' significance testing or infer psychological, cognitive, emotional, or
#' diagnostic attributes.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state",
#'   min_length = 2,
#'   max_length = 3
#' )
#'
#' summaries <- summarise_sequence_motifs(extracted)
#' summaries$by_sequence
#' summaries$overall
#'
#' @export
summarise_sequence_motifs <- function(x) {
  .sequence_motif_validate_extraction(x)

  occurrences <- x$occurrences
  sequences <- x$sequences
  metadata_cols <- setdiff(
    names(sequences),
    c(
      "sequence_id",
      "n_states",
      "n_candidate_occurrences",
      "n_retained_occurrences",
      "n_distinct_motifs"
    )
  )

  total_sequences <- nrow(sequences)
  total_occurrences <- nrow(occurrences)

  if (total_occurrences == 0L) {
    by_sequence <- .sequence_motif_empty_by_sequence(metadata_cols)
    overall <- .sequence_motif_empty_overall()
  } else {
    pair_key <- interaction(
      occurrences$sequence_id,
      occurrences$motif_id,
      drop = TRUE,
      lex.order = TRUE
    )

    pair_groups <- split(seq_len(total_occurrences), pair_key)
    by_sequence_rows <- list()

    for (indices in pair_groups) {
      first <- indices[1L]
      sequence_id <- occurrences$sequence_id[first]
      metadata <- vector("list", length(metadata_cols))
      names(metadata) <- metadata_cols

      for (column in metadata_cols) {
        metadata[[column]] <- occurrences[[column]][first]
      }

      by_sequence_rows[[length(by_sequence_rows) + 1L]] <-
        as.data.frame(
          c(
            list(sequence_id = sequence_id),
            metadata,
            list(
              motif_id = occurrences$motif_id[first],
              motif_key = occurrences$motif_key[first],
              motif = occurrences$motif[first],
              motif_length = as.integer(
                occurrences$motif_length[first]
              ),
              n_occurrences = as.integer(length(indices)),
              first_start_index = as.integer(
                min(occurrences$start_index[indices])
              ),
              last_start_index = as.integer(
                max(occurrences$start_index[indices])
              )
            )
          ),
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
    }

    by_sequence <- .sequence_bind_frames(
      frames = by_sequence_rows,
      empty = .sequence_motif_empty_by_sequence(metadata_cols)
    )

    motif_groups <- split(
      seq_len(nrow(by_sequence)),
      by_sequence$motif_id
    )

    overall_rows <- list()

    for (indices in motif_groups) {
      first <- indices[1L]
      n_occurrences <- sum(by_sequence$n_occurrences[indices])
      n_sequences <- length(indices)

      overall_rows[[length(overall_rows) + 1L]] <-
        data.frame(
          motif_id = by_sequence$motif_id[first],
          motif_key = by_sequence$motif_key[first],
          motif = by_sequence$motif[first],
          motif_length = as.integer(by_sequence$motif_length[first]),
          n_occurrences = as.integer(n_occurrences),
          n_sequences = as.integer(n_sequences),
          sequence_prevalence = if (total_sequences > 0L) {
            n_sequences / total_sequences
          } else {
            NA_real_
          },
          occurrence_share = n_occurrences / total_occurrences,
          mean_occurrences_per_sequence = if (total_sequences > 0L) {
            n_occurrences / total_sequences
          } else {
            NA_real_
          },
          mean_occurrences_when_present = n_occurrences / n_sequences,
          stringsAsFactors = FALSE
        )
    }

    overall <- .sequence_bind_frames(
      frames = overall_rows,
      empty = .sequence_motif_empty_overall()
    )

    overall <- .sequence_motif_sort_overall(
      overall,
      primary = "sequence_prevalence"
    )

    motif_rank <- match(by_sequence$motif_id, overall$motif_id)
    sequence_rank <- match(
      by_sequence$sequence_id,
      sequences$sequence_id
    )

    by_sequence <- by_sequence[
      order(
        motif_rank,
        sequence_rank,
        method = "radix"
      ),
      ,
      drop = FALSE
    ]
    row.names(by_sequence) <- NULL
  }

  list(
    by_sequence = by_sequence,
    overall = overall,
    sequences = sequences,
    state_dictionary = x$state_dictionary,
    audit = x$audit,
    status = x$status,
    mapping = x$mapping,
    extraction_settings = x$settings,
    n_sequences = as.integer(total_sequences),
    n_occurrences = as.integer(total_occurrences),
    n_motifs = as.integer(nrow(overall))
  )
}

#' Filter Sequence Motif Summaries
#'
#' Applies transparent count, prevalence, length, and top-ranking filters to
#' sequence motif summaries.
#'
#' @param x A motif extraction, motif summary, or filtered-motif object.
#' @param min_occurrences Non-negative whole number giving the minimum total
#'   occurrence count.
#' @param min_sequences Non-negative whole number giving the minimum number of
#'   sequences containing the motif.
#' @param min_prevalence Minimum sequence prevalence between 0 and 1.
#' @param motif_lengths Optional vector of positive whole-number motif lengths
#'   to retain.
#' @param top_n Optional positive whole number giving the requested number of
#'   highest-ranked motifs.
#' @param rank_by Metric used for deterministic top-ranking: one of
#'   `"sequence_prevalence"`, `"n_occurrences"`, or `"n_sequences"`.
#' @param ties Character value controlling the top-`n` boundary. `"include"`
#'   retains every motif tied on `rank_by` with the final selected motif;
#'   `"first"` retains exactly `top_n` motifs after deterministic secondary
#'   sorting.
#'
#' @return A named list containing the filtered `motifs` table, the matching
#' sequence-level rows in `by_sequence`, sequence and state dictionaries,
#' validation metadata, filter settings, and counts before and after filtering.
#'
#' @details
#' Filtering is descriptive and deterministic. When `ties = "first"`, ties are
#' resolved by sequence prevalence, occurrence count, sequence count, shorter
#' motif length, and finally the collision-resistant motif key.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' filtered <- filter_sequence_motifs(
#'   extracted,
#'   min_sequences = 2,
#'   top_n = 5,
#'   ties = "include"
#' )
#'
#' filtered$motifs
#'
#' @export
filter_sequence_motifs <- function(
  x,
  min_occurrences = 1L,
  min_sequences = 1L,
  min_prevalence = 0,
  motif_lengths = NULL,
  top_n = NULL,
  rank_by = c(
    "sequence_prevalence",
    "n_occurrences",
    "n_sequences"
  ),
  ties = c("include", "first")
) {
  .sequence_motif_assert_whole_number(
    min_occurrences,
    "min_occurrences",
    minimum = 0L
  )
  .sequence_motif_assert_whole_number(
    min_sequences,
    "min_sequences",
    minimum = 0L
  )
  .sequence_motif_assert_proportion(
    min_prevalence,
    "min_prevalence"
  )
  .sequence_motif_assert_motif_lengths(motif_lengths)
  .sequence_motif_assert_whole_number(
    top_n,
    "top_n",
    minimum = 1L,
    allow_null = TRUE
  )

  rank_by <- .sequence_motif_match_arg(
    rank_by,
    c(
      "sequence_prevalence",
      "n_occurrences",
      "n_sequences"
    ),
    "rank_by"
  )

  ties <- .sequence_motif_match_arg(
    ties,
    c("include", "first"),
    "ties"
  )

  summary <- .sequence_motif_as_summary(x)
  available <- summary$overall
  selected <- available

  keep <- selected$n_occurrences >= min_occurrences &
    selected$n_sequences >= min_sequences &
    selected$sequence_prevalence >= min_prevalence

  if (!is.null(motif_lengths)) {
    keep <- keep & selected$motif_length %in% unique(
      as.integer(motif_lengths)
    )
  }

  selected <- selected[keep, , drop = FALSE]
  selected <- .sequence_motif_sort_overall(selected, primary = rank_by)

  if (!is.null(top_n) && nrow(selected) > top_n) {
    if (ties == "include") {
      threshold <- selected[[rank_by]][top_n]
      selected <- selected[
        selected[[rank_by]] >= threshold,
        ,
        drop = FALSE
      ]
    } else {
      selected <- selected[seq_len(top_n), , drop = FALSE]
    }
  }

  row.names(selected) <- NULL

  selected_ids <- selected$motif_id
  by_sequence <- summary$by_sequence[
    summary$by_sequence$motif_id %in% selected_ids,
    ,
    drop = FALSE
  ]

  if (nrow(by_sequence) > 0L) {
    motif_rank <- match(by_sequence$motif_id, selected_ids)
    sequence_rank <- match(
      by_sequence$sequence_id,
      summary$sequences$sequence_id
    )

    by_sequence <- by_sequence[
      order(
        motif_rank,
        sequence_rank,
        method = "radix"
      ),
      ,
      drop = FALSE
    ]
    row.names(by_sequence) <- NULL
  }

  list(
    motifs = selected,
    by_sequence = by_sequence,
    sequences = summary$sequences,
    state_dictionary = summary$state_dictionary,
    audit = summary$audit,
    status = summary$status,
    mapping = summary$mapping,
    extraction_settings = summary$extraction_settings,
    settings = list(
      min_occurrences = as.integer(min_occurrences),
      min_sequences = as.integer(min_sequences),
      min_prevalence = as.numeric(min_prevalence),
      motif_lengths = if (is.null(motif_lengths)) {
        NULL
      } else {
        sort(unique(as.integer(motif_lengths)))
      },
      top_n = if (is.null(top_n)) NULL else as.integer(top_n),
      rank_by = rank_by,
      ties = ties
    ),
    n_available = as.integer(nrow(available)),
    n_retained = as.integer(nrow(selected))
  )
}

#' Format Sequence Motif Summaries
#'
#' Produces a stable, report-ready table from motif extraction, summary, or
#' filtered-motif output.
#'
#' @param x A motif extraction, motif summary, or filtered-motif object.
#' @param digits Whole number from 0 to 15 controlling numeric rounding.
#' @param prevalence Character value specifying whether prevalence and
#'   occurrence share are shown as `"proportion"`s or `"percent"`ages.
#' @param include_rank Logical value indicating whether to add a rank column.
#' @param rank_by Metric used to order and rank motifs: one of
#'   `"sequence_prevalence"`, `"n_occurrences"`, or `"n_sequences"`.
#' @param ties Character value specifying `"min"` shared ranks or
#'   deterministic `"first"` ranks.
#' @param include_ids Logical value indicating whether `motif_id` and
#'   `motif_key` should be included in the formatted table.
#'
#' @return A named list containing `table`, validation metadata, and formatting
#' settings. The table contains only structural motif measurements.
#'
#' @details
#' Formatting changes display precision and units only. It does not change the
#' underlying motif counts or introduce substantive interpretation.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c(rep("s1", 5L), rep("s2", 4L)),
#'   position = c(1:5, 1:4),
#'   state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
#' )
#'
#' extracted <- extract_sequence_ngrams(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' formatted <- format_sequence_motifs(
#'   extracted,
#'   prevalence = "percent",
#'   digits = 1
#' )
#'
#' formatted$table
#'
#' @export
format_sequence_motifs <- function(
  x,
  digits = 3L,
  prevalence = c("proportion", "percent"),
  include_rank = TRUE,
  rank_by = c(
    "sequence_prevalence",
    "n_occurrences",
    "n_sequences"
  ),
  ties = c("min", "first"),
  include_ids = TRUE
) {
  .sequence_motif_assert_whole_number(
    digits,
    "digits",
    minimum = 0L
  )

  if (digits > 15L) {
    stop("`digits` must not exceed 15.", call. = FALSE)
  }

  .sequence_assert_flag(include_rank, "include_rank")
  .sequence_assert_flag(include_ids, "include_ids")

  prevalence <- .sequence_motif_match_arg(
    prevalence,
    c("proportion", "percent"),
    "prevalence"
  )

  rank_by <- .sequence_motif_match_arg(
    rank_by,
    c(
      "sequence_prevalence",
      "n_occurrences",
      "n_sequences"
    ),
    "rank_by"
  )

  ties <- .sequence_motif_match_arg(
    ties,
    c("min", "first"),
    "ties"
  )

  summary <- .sequence_motif_as_summary(x)
  table <- summary$overall
  table <- .sequence_motif_sort_overall(table, primary = rank_by)

  if (include_rank) {
    rank_values <- if (nrow(table) == 0L) {
      integer()
    } else if (ties == "first") {
      seq_len(nrow(table))
    } else {
      rank(-table[[rank_by]], ties.method = "min")
    }

    table$rank <- as.integer(rank_values)
  }

  if (nrow(table) > 0L) {
    if (prevalence == "percent") {
      table$sequence_prevalence <- 100 * table$sequence_prevalence
      table$occurrence_share <- 100 * table$occurrence_share
    }

    numeric_columns <- c(
      "sequence_prevalence",
      "occurrence_share",
      "mean_occurrences_per_sequence",
      "mean_occurrences_when_present"
    )

    for (column in numeric_columns) {
      table[[column]] <- round(table[[column]], digits = digits)
    }
  }

  if (prevalence == "percent") {
    names(table)[names(table) == "sequence_prevalence"] <-
      "sequence_prevalence_percent"
    names(table)[names(table) == "occurrence_share"] <-
      "occurrence_share_percent"
  }

  if (!include_ids) {
    table$motif_id <- NULL
    table$motif_key <- NULL
  }

  front <- c(
    if (include_rank) "rank",
    if (include_ids) c("motif_id", "motif_key"),
    "motif",
    "motif_length",
    "n_occurrences",
    "n_sequences"
  )

  table <- table[
    c(front, setdiff(names(table), front))
  ]
  row.names(table) <- NULL

  list(
    table = table,
    audit = summary$audit,
    status = summary$status,
    mapping = summary$mapping,
    settings = list(
      digits = as.integer(digits),
      prevalence = prevalence,
      include_rank = include_rank,
      rank_by = rank_by,
      ties = ties,
      include_ids = include_ids
    )
  )
}

make_motif_data <- function() {
  data.frame(
    id = c(rep("s1", 5L), rep("s2", 4L)),
    position = c(1:5, 1:4),
    state = c(
      "A", "B", "A", "B", "A",
      "A", "B", "A", "C"
    ),
    group = c(rep("g1", 5L), rep("g2", 4L)),
    stringsAsFactors = FALSE
  )
}

test_that("contiguous n-grams are enumerated with stable positions", {
  data <- make_motif_data()

  result <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "group",
    min_length = 2,
    max_length = 3,
    overlap = "allow"
  )

  expect_identical(result$status, "pass")
  expect_equal(nrow(result$occurrences), 12L)
  expect_identical(
    result$sequences$n_candidate_occurrences,
    c(7L, 5L)
  )
  expect_identical(
    result$sequences$n_retained_occurrences,
    c(7L, 5L)
  )
  expect_true(
    all(
      result$occurrences$end_index -
        result$occurrences$start_index + 1L ==
        result$occurrences$motif_length
    )
  )
  expect_identical(
    result$occurrences$group[
      result$occurrences$sequence_id == "s1"
    ],
    rep("g1", 7L)
  )

  s1_aba <- result$occurrences[
    result$occurrences$sequence_id == "s1" &
      result$occurrences$motif == "A > B > A",
    ,
    drop = FALSE
  ]

  expect_identical(s1_aba$start_index, c(1L, 3L))
  expect_identical(s1_aba$end_index, c(3L, 5L))
  expect_identical(s1_aba$occurrence_index, c(1L, 2L))
})

test_that("overlap policy is explicit and deterministic", {
  data <- make_motif_data()

  allowed <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 3,
    max_length = 3,
    overlap = "allow"
  )

  disallowed <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 3,
    max_length = 3,
    overlap = "disallow"
  )

  allowed_aba <- allowed$occurrences[
    allowed$occurrences$motif == "A > B > A",
    ,
    drop = FALSE
  ]

  disallowed_aba <- disallowed$occurrences[
    disallowed$occurrences$motif == "A > B > A",
    ,
    drop = FALSE
  ]

  expect_equal(nrow(allowed_aba), 3L)
  expect_equal(nrow(disallowed_aba), 2L)
  expect_identical(
    disallowed_aba$start_index,
    c(1L, 1L)
  )
  expect_identical(
    disallowed$settings$overlap_rule,
    "left_to_right_greedy"
  )

  repeated <- data.frame(
    id = rep("s1", 4L),
    position = 1:4,
    state = rep("A", 4L),
    stringsAsFactors = FALSE
  )

  repeated_result <- extract_sequence_ngrams(
    repeated,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 2,
    max_length = 2,
    overlap = "disallow"
  )

  expect_identical(repeated_result$status, "review")
  expect_identical(
    repeated_result$occurrences$start_index,
    c(1L, 3L)
  )
})

test_that("row order is corrected while review diagnostics are retained", {
  data <- make_motif_data()
  shuffled <- data[c(7, 2, 9, 1, 6, 4, 3, 8, 5), ]

  ordered <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 2,
    max_length = 3
  )

  reordered <- extract_sequence_ngrams(
    shuffled,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 2,
    max_length = 3
  )

  expect_identical(ordered$status, "pass")
  expect_identical(reordered$status, "review")
  expect_true(
    any(reordered$audit$issue_code == "unordered_rows")
  )

  comparison_columns <- c(
    "sequence_id",
    "motif_id",
    "motif",
    "motif_length",
    "start_index",
    "end_index",
    "start_order",
    "end_order",
    "occurrence_index"
  )

  expect_identical(
    ordered$occurrences[comparison_columns],
    reordered$occurrences[comparison_columns]
  )
})

test_that("motif identity remains stable when labels contain separators", {
  data <- data.frame(
    id = c("single", "pair", "pair"),
    position = c(1, 1, 2),
    state = c("A > B", "A", "B"),
    stringsAsFactors = FALSE
  )

  result <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 1,
    max_length = 2,
    separator = " > "
  )

  same_display <- result$occurrences[
    result$occurrences$motif == "A > B",
    c("motif_id", "motif_key", "motif_length"),
    drop = FALSE
  ]

  expect_equal(nrow(same_display), 2L)
  expect_equal(length(unique(same_display$motif_id)), 2L)
  expect_equal(length(unique(same_display$motif_key)), 2L)
  expect_identical(
    sort(same_display$motif_length),
    c(1L, 2L)
  )
})

test_that("motif summaries return exact counts and prevalence", {
  data <- make_motif_data()

  extracted <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 2,
    max_length = 3
  )

  summary <- summarise_sequence_motifs(extracted)

  aba <- summary$overall[
    summary$overall$motif == "A > B > A",
    ,
    drop = FALSE
  ]

  expect_identical(summary$n_sequences, 2L)
  expect_identical(summary$n_occurrences, 12L)
  expect_identical(aba$n_occurrences, 3L)
  expect_identical(aba$n_sequences, 2L)
  expect_equal(aba$sequence_prevalence, 1)
  expect_equal(aba$occurrence_share, 3 / 12)
  expect_equal(aba$mean_occurrences_per_sequence, 1.5)
  expect_equal(aba$mean_occurrences_when_present, 1.5)

  s1_aba <- summary$by_sequence[
    summary$by_sequence$sequence_id == "s1" &
      summary$by_sequence$motif == "A > B > A",
    ,
    drop = FALSE
  ]

  expect_identical(s1_aba$n_occurrences, 2L)
  expect_identical(s1_aba$first_start_index, 1L)
  expect_identical(s1_aba$last_start_index, 3L)
})

test_that("prevalence denominator includes sequences without eligible windows", {
  data <- rbind(
    make_motif_data(),
    data.frame(
      id = "s3",
      position = 1,
      state = "A",
      group = "g3",
      stringsAsFactors = FALSE
    )
  )

  extracted <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "group",
    min_length = 3,
    max_length = 3
  )

  summary <- summarise_sequence_motifs(extracted)
  aba <- summary$overall[
    summary$overall$motif == "A > B > A",
    ,
    drop = FALSE
  ]

  expect_identical(extracted$status, "review")
  expect_identical(summary$n_sequences, 3L)
  expect_equal(aba$sequence_prevalence, 2 / 3)
  expect_identical(
    extracted$sequences$n_candidate_occurrences,
    c(3L, 2L, 0L)
  )
})

test_that("filters apply count prevalence and length thresholds", {
  data <- make_motif_data()

  summary <- summarise_sequence_motifs(
    extract_sequence_ngrams(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state",
      min_length = 2,
      max_length = 3
    )
  )

  filtered <- filter_sequence_motifs(
    summary,
    min_occurrences = 3,
    min_sequences = 2,
    min_prevalence = 1,
    motif_lengths = 2
  )

  expect_identical(
    filtered$motifs$motif,
    c("A > B", "B > A")
  )
  expect_identical(filtered$n_available, 6L)
  expect_identical(filtered$n_retained, 2L)
  expect_true(
    all(
      filtered$by_sequence$motif_id %in%
        filtered$motifs$motif_id
    )
  )
})

test_that("top-n ties are included or resolved deterministically", {
  data <- make_motif_data()

  summary <- summarise_sequence_motifs(
    extract_sequence_ngrams(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state",
      min_length = 2,
      max_length = 3
    )
  )

  included <- filter_sequence_motifs(
    summary,
    top_n = 1,
    rank_by = "n_occurrences",
    ties = "include"
  )

  first <- filter_sequence_motifs(
    summary,
    top_n = 1,
    rank_by = "n_occurrences",
    ties = "first"
  )

  expect_identical(included$n_retained, 3L)
  expect_true(all(included$motifs$n_occurrences == 3L))
  expect_identical(first$n_retained, 1L)
  expect_identical(first$motifs$motif, "A > B")
})

test_that("formatted motif tables use explicit units and rank ties", {
  data <- make_motif_data()

  summary <- summarise_sequence_motifs(
    extract_sequence_ngrams(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state",
      min_length = 2,
      max_length = 3
    )
  )

  formatted <- format_sequence_motifs(
    summary,
    prevalence = "percent",
    digits = 1,
    rank_by = "n_occurrences",
    ties = "min"
  )

  expect_true("sequence_prevalence_percent" %in% names(formatted$table))
  expect_true("occurrence_share_percent" %in% names(formatted$table))
  expect_identical(
    formatted$table$rank[1:3],
    c(1L, 1L, 1L)
  )
  expect_equal(
    formatted$table$sequence_prevalence_percent[1:3],
    rep(100, 3L)
  )

  no_ids <- format_sequence_motifs(
    summary,
    include_ids = FALSE,
    include_rank = FALSE
  )

  expect_false("motif_id" %in% names(no_ids$table))
  expect_false("motif_key" %in% names(no_ids$table))
  expect_false("rank" %in% names(no_ids$table))
})

test_that("empty motif outputs retain stable schemas", {
  data <- make_motif_data()

  extracted <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 6,
    max_length = 6
  )

  summary <- summarise_sequence_motifs(extracted)
  filtered <- filter_sequence_motifs(summary, min_occurrences = 2)
  formatted <- format_sequence_motifs(filtered)

  expect_equal(nrow(extracted$occurrences), 0L)
  expect_equal(nrow(extracted$motifs), 0L)
  expect_equal(nrow(summary$by_sequence), 0L)
  expect_equal(nrow(summary$overall), 0L)
  expect_equal(nrow(filtered$motifs), 0L)
  expect_equal(nrow(formatted$table), 0L)
  expect_true("rank" %in% names(formatted$table))
  expect_true("sequence_prevalence" %in% names(formatted$table))
})

test_that("factor levels define deterministic motif codes", {
  data <- make_motif_data()
  data$state <- factor(
    data$state,
    levels = c("C", "A", "B", "Z")
  )

  result <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    min_length = 2,
    max_length = 2
  )

  expect_identical(
    result$state_dictionary$state,
    c("C", "A", "B", "Z")
  )
  expect_false(
    result$state_dictionary$observed[
      result$state_dictionary$state == "Z"
    ]
  )

  ab <- result$motifs[result$motifs$motif == "A > B", , drop = FALSE]
  expect_identical(ab$motif_key, "S2|S3")
})

test_that("invalid settings and unresolved input errors are rejected", {
  data <- make_motif_data()

  expect_error(
    extract_sequence_ngrams(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state",
      min_length = 3,
      max_length = 2
    ),
    "max_length"
  )

  expect_error(
    extract_sequence_ngrams(
      data,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state",
      overlap = "sometimes"
    ),
    "overlap"
  )

  duplicated <- data
  duplicated$position[2L] <- 1

  expect_error(
    extract_sequence_ngrams(
      duplicated,
      sequence_id_col = "id",
      order_col = "position",
      state_col = "state"
    ),
    "failed validation"
  )

  extracted <- extract_sequence_ngrams(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_error(
    filter_sequence_motifs(extracted, min_prevalence = 1.1),
    "between 0 and 1"
  )

  expect_error(
    filter_sequence_motifs(extracted, motif_lengths = c(2, 2.5)),
    "positive whole numbers"
  )

  expect_error(
    format_sequence_motifs(extracted, digits = 16),
    "must not exceed 15"
  )
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----data---------------------------------------------------------------------
data <- data.frame(
  sequence_id = rep(paste0("s", 1:8), each = 6L),
  sequence_order = rep(1:6, times = 8L),
  state = c(
    rep(c("A", "B", "B", "C", "D", "D"), 4L),
    rep(c("D", "C", "C", "B", "A", "A"), 4L)
  ),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(data, method = "levenshtein")
clustering <- cluster_sequences(distance, k = 2L, method = "hierarchical")
network <- create_transition_network(data)

## ----index, fig.width=7, fig.height=5-----------------------------------------
plot_sequence_index(data)

## ----distribution, fig.width=7, fig.height=4----------------------------------
plot_sequence_state_distribution(data)

## ----entropy, fig.width=7, fig.height=4---------------------------------------
plot_sequence_entropy(data)

## ----distance, fig.width=6, fig.height=5--------------------------------------
plot_sequence_distance_heatmap(distance)

## ----silhouette, fig.width=7, fig.height=4------------------------------------
plot_sequence_cluster_silhouette(clustering, distance)

## ----network, fig.width=6, fig.height=6---------------------------------------
plot_transition_network(network)

#' Fit a time-varying sequence condition model
#'
#' Fits a binary generalized additive model for either occupancy of a target
#' state or occurrence of a target transition over aligned sequence time. Group
#' specific smooths model time-varying condition differences, and a participant
#' random-effect smooth accounts for repeated observations. The optional `mgcv`
#' package is required.
#'
#' @param data Long-format sequence data or a prepared result.
#' @param group_col Group or condition column.
#' @param participant_id_col Participant or repeated-unit column.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param time_col Optional numeric time column; defaults to `order_col`.
#' @param outcome `"state"` or `"transition"`.
#' @param target_state Target state for occupancy models.
#' @param from_state,to_state Target transition for transition models.
#' @param k Basis dimension for time smooths.
#' @param method Smoothing-parameter estimation method passed to
#'   [mgcv::gam()].
#' @param include_random_effect Include a participant random-effect smooth.
#'
#' @return An object of class `gp3_sequence_time_model`.
#' @examples
#' set.seed(2026)
#'
#' n_participants <- 24L
#' n_positions <- 12L
#'
#' participant_index <- rep(
#'   seq_len(n_participants),
#'   each = n_positions
#' )
#'
#' sequence_order <- rep(
#'   seq_len(n_positions),
#'   times = n_participants
#' )
#'
#' group_by_participant <- rep(
#'   c("control", "treatment"),
#'   each = n_participants %/% 2L
#' )
#'
#' group <- rep(
#'   group_by_participant,
#'   each = n_positions
#' )
#'
#' probability <- stats::plogis(
#'   -0.8 +
#'     0.08 * sequence_order +
#'     0.4 * (group == "treatment")
#' )
#'
#' observed_b <- stats::rbinom(
#'   length(probability),
#'   size = 1L,
#'   prob = probability
#' )
#'
#' sequences <- data.frame(
#'   participant_id = paste0("p", participant_index),
#'   sequence_id = paste0("s", participant_index),
#'   sequence_order = sequence_order,
#'   state = ifelse(observed_b == 1L, "B", "A"),
#'   group = group
#' )
#'
#' if (requireNamespace("mgcv", quietly = TRUE)) {
#'   fit <- fit_time_varying_sequence_model(
#'     sequences,
#'     group_col = "group",
#'     participant_id_col = "participant_id",
#'     target_state = "B",
#'     k = 4L
#'   )
#' }
#' @export
fit_time_varying_sequence_model <- function(
  data,
  group_col,
  participant_id_col,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  time_col = NULL,
  outcome = c("state", "transition"),
  target_state = NULL,
  from_state = NULL,
  to_state = NULL,
  k = 5L,
  method = "REML",
  include_random_effect = TRUE
) {
  .sequence_adv_require("mgcv", "fit time-varying sequence models")
  outcome <- match.arg(outcome)
  .sequence_adv_scalar_character(group_col, "group_col")
  .sequence_adv_scalar_character(participant_id_col, "participant_id_col")
  if (is.null(time_col)) time_col <- order_col
  .sequence_adv_scalar_character(time_col, "time_col")
  .sequence_adv_scalar_number(k, "k", lower = 3, integer = TRUE)
  .sequence_adv_scalar_character(method, "method")
  .sequence_adv_scalar_logical(include_random_effect, "include_random_effect")
  if (outcome == "state") {
    .sequence_adv_scalar_character(target_state, "target_state")
  } else {
    .sequence_adv_scalar_character(from_state, "from_state")
    .sequence_adv_scalar_character(to_state, "to_state")
  }
  required_metadata <- unique(c(group_col, participant_id_col))
  x <- .sequence_adv_data(
    data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = required_metadata,
    missing_state_policy = "error"
  )
  working <- x$data
  if (!(time_col %in% names(working))) {
    stop("Missing time column `", time_col, "`.", call. = FALSE)
  }
  time_value <- working[[time_col]]
  if (!is.numeric(time_value) || anyNA(time_value) || any(!is.finite(time_value))) {
    stop("The time column must contain finite, non-missing numeric values.",
         call. = FALSE)
  }
  group_value <- as.character(working[[group_col]])
  participant_value <- as.character(working[[participant_id_col]])
  if (anyNA(group_value) || any(!nzchar(trimws(group_value)))) {
    stop("Group values must not be missing or blank.", call. = FALSE)
  }
  if (anyNA(participant_value) || any(!nzchar(trimws(participant_value)))) {
    stop("Participant identifiers must not be missing or blank.", call. = FALSE)
  }
  if (outcome == "state") {
    model_data <- data.frame(
      .outcome = as.integer(as.character(working[[state_col]]) == target_state),
      .time = as.numeric(time_value),
      .group = factor(group_value),
      .participant = factor(participant_value),
      stringsAsFactors = FALSE
    )
  } else {
    rows <- list()
    h <- 0L
    split_rows <- split(seq_len(nrow(working)),
                        factor(as.character(working[[sequence_id_col]]),
                               levels = x$sequence_ids), drop = TRUE)
    for (sequence_id in names(split_rows)) {
      current_rows <- split_rows[[sequence_id]]
      if (length(current_rows) < 2L) next
      current_state <- as.character(working[[state_col]][current_rows])
      for (j in seq_len(length(current_rows) - 1L)) {
        row <- current_rows[j]
        h <- h + 1L
        rows[[h]] <- data.frame(
          .outcome = as.integer(identical(current_state[j], from_state) &&
                                  identical(current_state[j + 1L], to_state)),
          .time = as.numeric(working[[time_col]][row]),
          .group = group_value[row],
          .participant = participant_value[row],
          stringsAsFactors = FALSE
        )
      }
    }
    if (length(rows) == 0L) stop("No transitions are available for modelling.", call. = FALSE)
    model_data <- do.call(rbind, rows)
    model_data$.group <- factor(model_data$.group)
    model_data$.participant <- factor(model_data$.participant)
  }
  if (nlevels(model_data$.group) < 2L) {
    stop("At least two groups are required.", call. = FALSE)
  }
  if (length(unique(model_data$.time)) < 4L) {
    stop("At least four distinct time values are required.", call. = FALSE)
  }
  if (length(unique(model_data$.outcome)) < 2L) {
    stop("The target outcome has no variation.", call. = FALSE)
  }
  k_used <- min(as.integer(k), length(unique(model_data$.time)) - 1L)
  formula_text <- paste0(
    ".outcome ~ .group + s(.time, by = .group, k = ",
    k_used, ")",
    if (include_random_effect) " + s(.participant, bs = 're')" else ""
  )
  model_formula <- stats::as.formula(formula_text)
  environment(model_formula) <- asNamespace("mgcv")
  fit <- mgcv::gam(
    formula = model_formula,
    family = stats::binomial(),
    data = model_data,
    method = method,
    drop.unused.levels = TRUE
  )
  result <- list(
    model = fit,
    model_data = model_data,
    outcome = outcome,
    target_state = target_state,
    from_state = from_state,
    to_state = to_state,
    group_levels = levels(model_data$.group),
    participant_levels = levels(model_data$.participant),
    time_range = range(model_data$.time),
    k = k_used,
    method = method,
    include_random_effect = include_random_effect,
    columns = list(group = group_col, participant = participant_id_col,
                   sequence_id = sequence_id_col, order = order_col,
                   state = state_col, time = time_col),
    call = match.call()
  )
  class(result) <- c("gp3_sequence_time_model", "list")
  result
}

#' Predict a time-varying sequence model
#'
#' @param model A fitted time-varying sequence model.
#' @param time Optional numeric prediction grid.
#' @param groups Optional subset of fitted groups.
#' @param level Confidence level for pointwise intervals.
#'
#' @return A data frame with fitted probabilities and pointwise intervals.
#' @examples
#' # See `fit_time_varying_sequence_model()`.
#' @export
predict_time_varying_sequence_model <- function(
  model,
  time = NULL,
  groups = NULL,
  level = 0.95
) {
  if (!inherits(model, "gp3_sequence_time_model")) {
    stop("`model` must be created by `fit_time_varying_sequence_model()`.",
         call. = FALSE)
  }
  .sequence_adv_scalar_number(level, "level", lower = 0.5, upper = 0.999999)
  if (is.null(time)) time <- seq(model$time_range[1L], model$time_range[2L], length.out = 100L)
  if (!is.numeric(time) || anyNA(time) || any(!is.finite(time))) {
    stop("`time` must contain finite numeric values.", call. = FALSE)
  }
  if (is.null(groups)) groups <- model$group_levels
  groups <- as.character(groups)
  if (any(!groups %in% model$group_levels)) {
    stop("Unknown groups requested for prediction.", call. = FALSE)
  }
  grid <- expand.grid(
    .time = sort(unique(as.numeric(time))),
    .group = groups,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )
  grid$.group <- factor(grid$.group, levels = model$group_levels)
  grid$.participant <- factor(model$participant_levels[1L],
                              levels = model$participant_levels)
  excluded <- if (model$include_random_effect) "s(.participant)" else NULL
  prediction <- stats::predict(
    model$model,
    newdata = grid,
    type = "link",
    se.fit = TRUE,
    exclude = excluded
  )
  z <- stats::qnorm(1 - (1 - level) / 2)
  grid$estimate <- stats::plogis(as.numeric(prediction$fit))
  grid$lower <- stats::plogis(as.numeric(prediction$fit) - z * as.numeric(prediction$se.fit))
  grid$upper <- stats::plogis(as.numeric(prediction$fit) + z * as.numeric(prediction$se.fit))
  names(grid)[names(grid) == ".time"] <- "time"
  names(grid)[names(grid) == ".group"] <- "group"
  grid$.participant <- NULL
  grid$group <- as.character(grid$group)
  grid$outcome <- model$outcome
  grid
}

#' Summarise a time-varying sequence model
#'
#' @param model A fitted time-varying sequence model.
#'
#' @return A list containing model metadata, parametric terms, smooth terms,
#'   deviance explained, and convergence information.
#' @examples
#' # See `fit_time_varying_sequence_model()`.
#' @export
summarise_time_varying_sequence_model <- function(model) {
  if (!inherits(model, "gp3_sequence_time_model")) {
    stop("`model` must be created by `fit_time_varying_sequence_model()`.",
         call. = FALSE)
  }
  summary_value <- summary(model$model)
  list(
    metadata = data.frame(
      outcome = model$outcome,
      target_state = if (is.null(model$target_state)) NA_character_ else model$target_state,
      from_state = if (is.null(model$from_state)) NA_character_ else model$from_state,
      to_state = if (is.null(model$to_state)) NA_character_ else model$to_state,
      n_observations = nrow(model$model_data),
      n_groups = length(model$group_levels),
      n_participants = length(model$participant_levels),
      k = model$k,
      deviance_explained = summary_value$dev.expl,
      adjusted_r_squared = summary_value$r.sq,
      stringsAsFactors = FALSE
    ),
    parametric_terms = as.data.frame(summary_value$p.table),
    smooth_terms = as.data.frame(summary_value$s.table),
    converged = isTRUE(model$model$converged),
    method = model$method
  )
}

#' Plot predicted time-varying sequence probabilities
#'
#' @param model A fitted time-varying sequence model.
#' @param time Optional prediction grid.
#' @param level Pointwise confidence level.
#' @param show_interval Draw pointwise confidence ribbons.
#' @param ... Additional arguments passed to [graphics::plot()].
#'
#' @return Prediction data, invisibly.
#' @examples
#' # See `fit_time_varying_sequence_model()`.
#' @export
plot_time_varying_sequence_model <- function(
  model,
  time = NULL,
  level = 0.95,
  show_interval = TRUE,
  ...
) {
  .sequence_adv_scalar_logical(show_interval, "show_interval")
  prediction <- predict_time_varying_sequence_model(model, time = time, level = level)
  groups <- unique(prediction$group)
  colors <- .sequence_ext_base_colors(length(groups))
  graphics::plot(
    range(prediction$time), range(c(prediction$lower, prediction$upper), finite = TRUE),
    type = "n", xlab = "Sequence time", ylab = "Estimated probability", ...
  )
  for (i in seq_along(groups)) {
    current <- prediction[prediction$group == groups[i], , drop = FALSE]
    current <- current[order(current$time), , drop = FALSE]
    if (show_interval) {
      polygon_x <- c(current$time, rev(current$time))
      polygon_y <- c(current$lower, rev(current$upper))
      graphics::polygon(polygon_x, polygon_y,
                        col = grDevices::adjustcolor(colors[i], alpha.f = 0.2),
                        border = NA)
    }
    graphics::lines(current$time, current$estimate, col = colors[i], lwd = 2)
  }
  graphics::legend("topright", legend = groups, col = colors, lwd = 2, bty = "n")
  invisible(prediction)
}

test_that("time-varying models are dependency guarded and auditable", {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  set.seed(101)
  participants <- paste0("p", seq_len(24L))
  data <- do.call(
    rbind,
    lapply(seq_along(participants), function(i) {
      time <- seq_len(12L)
      group <- if (i <= 12L) "g1" else "g2"
      linear_predictor <-
        -0.4 +
        0.06 * time +
        0.35 * (group == "g2") * sin(time / 3)

      data.frame(
        participant_id = participants[i],
        sequence_id = participants[i],
        sequence_order = time,
        state = ifelse(
          stats::runif(length(time)) <
            stats::plogis(linear_predictor),
          "A",
          "B"
        ),
        group = group,
        stringsAsFactors = FALSE
      )
    })
  )
  if (!requireNamespace("mgcv", quietly = TRUE)) {
    expect_error(
      fit_time_varying_sequence_model(
        data,
        group_col = "group",
        participant_id_col = "participant_id",
        target_state = "A",
        k = 3L
      ),
      "mgcv"
    )
    skip("mgcv is not installed")
  }

  fit <- fit_time_varying_sequence_model(
    data,
    group_col = "group",
    participant_id_col = "participant_id",
    target_state = "A",
    k = 3L,
    include_random_effect = FALSE
  )
  expect_s3_class(fit, "gp3_sequence_time_model")
  expect_equal(fit$outcome, "state")
  expect_equal(fit$group_levels, c("g1", "g2"))
  prediction <- predict_time_varying_sequence_model(fit)
  expect_true(all(c("time", "group", "estimate", "lower", "upper") %in% names(prediction)))
  expect_true(all(prediction$estimate >= 0 & prediction$estimate <= 1))
  summary <- summarise_time_varying_sequence_model(fit)
  expect_true(all(c("metadata", "smooth_terms", "converged") %in% names(summary)))
  expect_invisible(plot_time_varying_sequence_model(fit))
})

#' Create a transition network from ordered sequences
#'
#' @param data Long-format sequence data or a prepared result.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param group_cols Optional grouping columns constant within sequence.
#' @param order Markov order. `1` creates ordinary state-to-state edges;
#'   larger values create context-to-next-state edges.
#' @param include_self Include first-order self-transitions.
#' @param normalise Edge weight scale: counts, conditional probabilities from
#'   each context, or global shares.
#' @param smoothing Non-negative additive smoothing applied to observed edges.
#' @param context_separator Separator used for higher-order contexts.
#'
#' @return A data frame of class `gp3_transition_network` containing context,
#' next state, counts, weights, sequence prevalence, and group columns.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' create_transition_network(sequences, normalise = "from")
#'
#' @export
create_transition_network <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  group_cols = NULL,
  order = 1L,
  include_self = TRUE,
  normalise = c("count", "from", "global"),
  smoothing = 0,
  context_separator = " > "
) {
  normalise <- match.arg(normalise)
  .sequence_adv_scalar_number(order, "order", lower = 1, integer = TRUE)
  .sequence_adv_scalar_logical(include_self, "include_self")
  .sequence_adv_scalar_number(smoothing, "smoothing", lower = 0)
  .sequence_adv_scalar_character(context_separator, "context_separator")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          metadata_cols = group_cols, missing_state_policy = "error")
  if (any(grepl(context_separator, x$state_levels, fixed = TRUE))) {
    stop("`context_separator` must not occur inside an observed state label.",
         call. = FALSE)
  }
  group_cols <- .sequence_adv_match_cols(x$data, group_cols, "group_cols")
  sequence_meta <- if (length(group_cols) == 0L) {
    data.frame(sequence_id = x$sequence_ids, .group_key = "__all__",
               stringsAsFactors = FALSE)
  } else {
    meta <- x$metadata
    names(meta)[names(meta) == sequence_id_col] <- "sequence_id"
    meta$.group_key <- .sequence_adv_group_key(meta, group_cols)
    meta
  }
  event_rows <- list(); h <- 0L
  for (id in x$sequence_ids) {
    seq <- x$sequences[[id]]
    if (length(seq) <= order) next
    group_row <- sequence_meta[sequence_meta$sequence_id == id, , drop = FALSE]
    for (i in seq_len(length(seq) - order)) {
      context <- seq[i:(i + order - 1L)]
      next_state <- seq[i + order]
      if (order == 1L && !include_self && identical(context[1L], next_state)) next
      h <- h + 1L
      prefix <- if (length(group_cols) == 0L) {
        .sequence_adv_zero_column_frame(1L)
      } else {
        group_row[1L, group_cols, drop = FALSE]
      }
      event_rows[[h]] <- cbind(prefix, data.frame(
        group_key = group_row$.group_key[1L],
        sequence_id = id,
        context = paste(context, collapse = context_separator),
        from_state = if (order == 1L) context[1L] else NA_character_,
        to_state = next_state,
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }
  }
  if (length(event_rows) == 0L) {
    prefix <- if (length(group_cols) == 0L) {
      .sequence_adv_zero_column_frame(0L)
    } else {
      x$data[0L, group_cols, drop = FALSE]
    }
    result <- cbind(
      prefix,
      data.frame(group_key = character(), context = character(),
                 from_state = character(), to_state = character(),
                 count = numeric(), weight = numeric(),
                 sequence_count = integer(), sequence_prevalence = numeric(),
                 stringsAsFactors = FALSE, check.names = FALSE)
    )
    class(result) <- c("gp3_transition_network", "data.frame")
    attr(result, "group_cols") <- group_cols
    attr(result, "settings") <- list(order = as.integer(order), normalise = normalise,
                                     include_self = include_self, smoothing = smoothing,
                                     context_separator = context_separator)
    return(result)
  }
  events <- do.call(rbind, event_rows)
  grouping <- c(group_cols, "group_key", "context", "from_state", "to_state")
  key <- .sequence_adv_group_key(events, grouping)
  pieces <- split(seq_len(nrow(events)), key, drop = TRUE)
  edge_rows <- lapply(pieces, function(rows) {
    prefix <- events[rows[1L], grouping, drop = FALSE]
    group_key <- events$group_key[rows[1L]]
    group_sequences <- unique(sequence_meta$sequence_id[sequence_meta$.group_key == group_key])
    cbind(prefix, data.frame(
      count = length(rows) + smoothing,
      sequence_count = length(unique(events$sequence_id[rows])),
      sequence_prevalence = length(unique(events$sequence_id[rows])) / length(group_sequences),
      stringsAsFactors = FALSE,
      check.names = FALSE
    ))
  })
  result <- do.call(rbind, edge_rows)
  row.names(result) <- NULL
  if (normalise == "count") {
    result$weight <- result$count
  } else if (normalise == "from") {
    context_key <- .sequence_adv_group_key(result, c(group_cols, "group_key", "context"))
    totals <- stats::ave(result$count, context_key, FUN = sum)
    result$weight <- result$count / totals
  } else {
    group_key <- .sequence_adv_group_key(result, c(group_cols, "group_key"))
    totals <- stats::ave(result$count, group_key, FUN = sum)
    result$weight <- result$count / totals
  }
  result <- result[order(result$group_key, result$context, result$to_state, method = "radix"), , drop = FALSE]
  row.names(result) <- NULL
  class(result) <- c("gp3_transition_network", "data.frame")
  attr(result, "group_cols") <- group_cols
  attr(result, "settings") <- list(order = as.integer(order), normalise = normalise,
                                   include_self = include_self, smoothing = smoothing,
                                   context_separator = context_separator)
  result
}

.sequence_adv_graph_matrix <- function(network, use_weight = TRUE,
                                       symmetrise = FALSE) {
  if (!inherits(network, "gp3_transition_network")) {
    stop("`network` must be created by `create_transition_network()`.", call. = FALSE)
  }
  settings <- attr(network, "settings")
  if (!identical(settings$order, 1L)) {
    stop("Centrality and community helpers currently require a first-order network.",
         call. = FALSE)
  }
  group_keys <- unique(as.character(network$group_key))
  group_keys <- group_keys[!is.na(group_keys)]
  if (length(group_keys) > 1L) {
    stop("Filter a grouped transition network to one group before graph analysis.",
         call. = FALSE)
  }
  states <- sort(unique(c(network$from_state, network$to_state)), method = "radix")
  states <- states[!is.na(states)]
  adjacency <- matrix(0, length(states), length(states), dimnames = list(states, states))
  value <- if (use_weight) network$weight else rep(1, nrow(network))
  for (i in seq_len(nrow(network))) {
    adjacency[network$from_state[i], network$to_state[i]] <-
      adjacency[network$from_state[i], network$to_state[i]] + value[i]
  }
  if (symmetrise) {
    original_diagonal <- diag(adjacency)
    adjacency <- adjacency + t(adjacency)
    diag(adjacency) <- original_diagonal
  }
  adjacency
}

.sequence_adv_dijkstra <- function(cost, source) {
  n <- nrow(cost)
  distance <- rep(Inf, n)
  visited <- rep(FALSE, n)
  distance[source] <- 0
  for (step in seq_len(n)) {
    available <- which(!visited)
    if (length(available) == 0L) break
    current <- available[which.min(distance[available])]
    if (!is.finite(distance[current])) break
    visited[current] <- TRUE
    neighbours <- which(is.finite(cost[current, ]) & cost[current, ] > 0)
    for (j in neighbours) {
      candidate <- distance[current] + cost[current, j]
      if (candidate < distance[j]) distance[j] <- candidate
    }
  }
  distance
}

.sequence_adv_unweighted_betweenness <- function(adjacency) {
  n <- nrow(adjacency)
  score <- numeric(n)
  for (source in seq_len(n)) {
    stack <- integer()
    predecessors <- vector("list", n)
    sigma <- numeric(n); sigma[source] <- 1
    distance <- rep(-1L, n); distance[source] <- 0L
    queue <- source
    while (length(queue) > 0L) {
      v <- queue[1L]; queue <- queue[-1L]
      stack <- c(stack, v)
      neighbours <- which(adjacency[v, ] > 0)
      for (w in neighbours) {
        if (distance[w] < 0L) {
          queue <- c(queue, w)
          distance[w] <- distance[v] + 1L
        }
        if (distance[w] == distance[v] + 1L) {
          sigma[w] <- sigma[w] + sigma[v]
          predecessors[[w]] <- c(predecessors[[w]], v)
        }
      }
    }
    dependency <- numeric(n)
    while (length(stack) > 0L) {
      w <- stack[length(stack)]
      stack <- stack[-length(stack)]
      for (v in predecessors[[w]]) {
        if (sigma[w] > 0) dependency[v] <- dependency[v] +
          (sigma[v] / sigma[w]) * (1 + dependency[w])
      }
      if (w != source) score[w] <- score[w] + dependency[w]
    }
  }
  score
}

#' Summarise transition-network centrality
#'
#' @param network A first-order transition network.
#' @param directed Treat the network as directed.
#' @param pagerank_damping Damping factor for PageRank.
#' @param pagerank_tolerance Convergence tolerance.
#' @param pagerank_max_iter Maximum PageRank iterations.
#'
#' @return A data frame containing degree, strength, weighted closeness,
#' unweighted betweenness, and PageRank centrality.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' network <- create_transition_network(sequences)
#' summarise_transition_centrality(network)
#'
#' @export
summarise_transition_centrality <- function(network, directed = TRUE,
                                            pagerank_damping = 0.85,
                                            pagerank_tolerance = 1e-10,
                                            pagerank_max_iter = 1000L) {
  .sequence_adv_scalar_logical(directed, "directed")
  .sequence_adv_scalar_number(pagerank_damping, "pagerank_damping", lower = 0, upper = 1)
  .sequence_adv_scalar_number(pagerank_tolerance, "pagerank_tolerance", lower = 0)
  .sequence_adv_scalar_number(pagerank_max_iter, "pagerank_max_iter", lower = 1, integer = TRUE)
  adjacency <- .sequence_adv_graph_matrix(network, use_weight = TRUE,
                                          symmetrise = !directed)
  n <- nrow(adjacency)
  if (n == 0L) return(data.frame())
  binary <- adjacency > 0
  out_degree <- rowSums(binary)
  in_degree <- colSums(binary)
  out_strength <- rowSums(adjacency)
  in_strength <- colSums(adjacency)
  total_degree <- if (directed) out_degree + in_degree else out_degree
  total_strength <- if (directed) out_strength + in_strength else out_strength
  cost <- matrix(Inf, n, n, dimnames = dimnames(adjacency))
  cost[adjacency > 0] <- 1 / adjacency[adjacency > 0]
  diag(cost) <- 0
  distance_matrix <- do.call(rbind, lapply(seq_len(n), function(i) .sequence_adv_dijkstra(cost, i)))
  closeness <- vapply(seq_len(n), function(i) {
    reachable <- distance_matrix[i, is.finite(distance_matrix[i, ]) & distance_matrix[i, ] > 0]
    if (length(reachable) == 0L) 0 else length(reachable) / sum(reachable)
  }, numeric(1))
  betweenness <- .sequence_adv_unweighted_betweenness(binary)
  if (!directed) betweenness <- betweenness / 2
  transition <- adjacency
  dangling <- rowSums(transition) == 0
  if (any(!dangling)) transition[!dangling, ] <- transition[!dangling, , drop = FALSE] /
    rowSums(transition[!dangling, , drop = FALSE])
  if (any(dangling)) transition[dangling, ] <- 1 / n
  rank <- rep(1 / n, n)
  for (iter in seq_len(pagerank_max_iter)) {
    next_rank <- (1 - pagerank_damping) / n +
      pagerank_damping * as.numeric(t(transition) %*% rank)
    if (max(abs(next_rank - rank)) < pagerank_tolerance) {
      rank <- next_rank
      break
    }
    rank <- next_rank
  }
  rank <- .sequence_adv_vector_normalise(rank)
  data.frame(
    state = rownames(adjacency),
    out_degree = as.integer(out_degree),
    in_degree = as.integer(in_degree),
    total_degree = as.integer(total_degree),
    out_strength = as.numeric(out_strength),
    in_strength = as.numeric(in_strength),
    total_strength = as.numeric(total_strength),
    closeness = closeness,
    betweenness = betweenness,
    pagerank = rank,
    stringsAsFactors = FALSE
  )
}

#' Detect descriptive transition communities
#'
#' @param network A first-order transition network.
#' @param method `"label_propagation"` or `"components"`.
#' @param max_iter Maximum label-propagation iterations.
#' @param seed Reproducibility seed used only to rotate update order.
#'
#' @return A data frame with states and deterministic community labels.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' network <- create_transition_network(sequences)
#' detect_transition_communities(network)
#'
#' @export
detect_transition_communities <- function(network,
                                          method = c("label_propagation", "components"),
                                          max_iter = 100L, seed = 1L) {
  method <- match.arg(method)
  .sequence_adv_scalar_number(max_iter, "max_iter", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  adjacency <- .sequence_adv_graph_matrix(network, use_weight = TRUE, symmetrise = TRUE)
  n <- nrow(adjacency)
  if (n == 0L) return(data.frame(state = character(), community = integer(), stringsAsFactors = FALSE))
  labels <- seq_len(n)
  if (method == "components") {
    community <- rep(NA_integer_, n)
    current <- 0L
    for (i in seq_len(n)) {
      if (!is.na(community[i])) next
      current <- current + 1L
      queue <- i; community[i] <- current
      while (length(queue) > 0L) {
        v <- queue[1L]; queue <- queue[-1L]
        neighbours <- which(adjacency[v, ] > 0)
        unseen <- neighbours[is.na(community[neighbours])]
        if (length(unseen)) {
          community[unseen] <- current
          queue <- c(queue, unseen)
        }
      }
    }
    labels <- community
  } else {
    base_order <- seq_len(n)
    if (n > 1L) {
      shift <- as.integer(seed %% n)
      if (shift > 0L) {
        base_order <- c(base_order[(shift + 1L):n], base_order[seq_len(shift)])
      }
    }
    for (iter in seq_len(max_iter)) {
      changed <- FALSE
      for (v in base_order) {
        neighbours <- which(adjacency[v, ] > 0)
        if (!length(neighbours)) next
        totals <- tapply(adjacency[v, neighbours], labels[neighbours], sum)
        winners <- as.integer(names(totals)[totals == max(totals)])
        new_label <- min(winners)
        if (new_label != labels[v]) {
          labels[v] <- new_label
          changed <- TRUE
        }
      }
      if (!changed) break
    }
    unique_labels <- sort(unique(labels))
    labels <- match(labels, unique_labels)
  }
  data.frame(state = rownames(adjacency), community = as.integer(labels),
             stringsAsFactors = FALSE)
}

#' Fit a higher-order transition model
#'
#' @param data Long-format sequence data.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param order Context order.
#' @param smoothing Additive smoothing over observed next states.
#' @param backoff Retain lower-order context tables for prediction backoff.
#' @param context_separator Context separator.
#'
#' @return An object of class `gp3_higher_order_transition_model`.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' fit_higher_order_transition_model(sequences, order = 2L)
#'
#' @export
fit_higher_order_transition_model <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  order = 2L,
  smoothing = 0.5,
  backoff = TRUE,
  context_separator = " > "
) {
  .sequence_adv_scalar_number(order, "order", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(smoothing, "smoothing", lower = 0)
  .sequence_adv_scalar_logical(backoff, "backoff")
  .sequence_adv_scalar_character(context_separator, "context_separator")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  if (any(grepl(context_separator, x$state_levels, fixed = TRUE))) {
    stop("`context_separator` must not occur inside an observed state label.",
         call. = FALSE)
  }
  build_level <- function(level) {
    contexts <- list(); h <- 0L
    for (id in x$sequence_ids) {
      seq <- x$sequences[[id]]
      if (length(seq) <= level) next
      for (i in seq_len(length(seq) - level)) {
        h <- h + 1L
        contexts[[h]] <- data.frame(
          context = paste(seq[i:(i + level - 1L)], collapse = context_separator),
          next_state = seq[i + level], stringsAsFactors = FALSE
        )
      }
    }
    if (length(contexts) == 0L) return(data.frame())
    events <- do.call(rbind, contexts)
    keys <- unique(events$context)
    out <- list(); z <- 0L
    for (context in sort(keys, method = "radix")) {
      counts <- table(factor(events$next_state[events$context == context], levels = x$state_levels))
      probabilities <- (as.numeric(counts) + smoothing) /
        (sum(counts) + smoothing * length(counts))
      for (j in seq_along(x$state_levels)) {
        z <- z + 1L
        out[[z]] <- data.frame(order = level, context = context,
                               next_state = x$state_levels[j], count = as.integer(counts[j]),
                               probability = probabilities[j], stringsAsFactors = FALSE)
      }
    }
    do.call(rbind, out)
  }
  levels <- if (backoff) seq_len(order) else order
  tables <- lapply(levels, build_level)
  names(tables) <- paste0("order_", levels)
  result <- list(order = as.integer(order), tables = tables,
                 state_levels = x$state_levels, smoothing = smoothing,
                 backoff = backoff, context_separator = context_separator)
  class(result) <- c("gp3_higher_order_transition_model", "list")
  result
}

#' Predict the next state from a transition model
#'
#' @param model A higher-order transition model.
#' @param history Character vector of observed recent states.
#' @param top_n Optional number of states to retain. Returned probabilities
#'   remain on the full-model scale and are not renormalised after truncation.
#'
#' @return A probability table ordered from highest to lowest probability,
#' with the context order actually used.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' model <- fit_higher_order_transition_model(sequences, order = 2L)
#' predict_next_state(model, c("A", "B"))
#'
#' @export
predict_next_state <- function(model, history, top_n = NULL) {
  if (!inherits(model, "gp3_higher_order_transition_model")) {
    stop("`model` must be created by `fit_higher_order_transition_model()`.", call. = FALSE)
  }
  if (!is.character(history) || length(history) < 1L || anyNA(history) ||
      any(!nzchar(trimws(history)))) {
    stop("`history` must be a non-missing, non-blank character vector.",
         call. = FALSE)
  }
  if (!is.null(top_n)) .sequence_adv_scalar_number(top_n, "top_n", lower = 1, integer = TRUE)
  available_orders <- sort(as.integer(sub("order_", "", names(model$tables))), decreasing = TRUE)
  used <- NA_integer_; table_used <- NULL; context_used <- NULL
  for (level in available_orders) {
    if (length(history) < level) next
    context <- paste(utils::tail(history, level), collapse = model$context_separator)
    table <- model$tables[[paste0("order_", level)]]
    if (!is.data.frame(table) || nrow(table) == 0L || !("context" %in% names(table))) next
    current <- table[table$context == context, , drop = FALSE]
    if (nrow(current) > 0L) {
      used <- level; table_used <- current; context_used <- context; break
    }
  }
  if (is.null(table_used)) {
    table_used <- data.frame(
      order = 0L,
      context = "<unseen>",
      next_state = model$state_levels,
      count = 0L,
      probability = rep(1 / length(model$state_levels), length(model$state_levels)),
      stringsAsFactors = FALSE
    )
    used <- 0L
    context_used <- "<unseen>"
  }
  table_used <- table_used[order(-table_used$probability, table_used$next_state,
                                 method = "radix"), , drop = FALSE]
  if (!is.null(top_n)) table_used <- utils::head(table_used, top_n)
  table_used$used_order <- used
  table_used$used_context <- context_used
  row.names(table_used) <- NULL
  table_used
}

#' Bootstrap transition-network edge weights
#'
#' @param data Long-format sequence data.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param n_boot Number of bootstrap samples of whole sequences.
#' @param level Confidence level for percentile intervals.
#' @param seed Reproducibility seed.
#' @param include_self Include self-transitions.
#'
#' @return A data frame containing observed edge weights, bootstrap means,
#' standard deviations, and percentile intervals.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' bootstrap_transition_network(sequences, n_boot = 5L, seed = 1L)
#'
#' @export
bootstrap_transition_network <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  n_boot = 100L,
  level = 0.95,
  seed = 1L,
  include_self = TRUE
) {
  .sequence_adv_scalar_number(n_boot, "n_boot", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(level, "level", lower = 0.5, upper = 0.999)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  .sequence_adv_scalar_logical(include_self, "include_self")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  observed <- create_transition_network(x$data, sequence_id_col, order_col, state_col,
                                        include_self = include_self, normalise = "from")
  if (nrow(observed) == 0L) {
    stop("No first-order transitions are available to bootstrap.", call. = FALSE)
  }
  edge_key <- paste(observed$from_state, observed$to_state, sep = "\034")
  values <- matrix(0, nrow = nrow(observed), ncol = n_boot)
  .sequence_adv_with_seed(seed, {
    for (b in seq_len(n_boot)) {
      sampled <- sample(x$sequence_ids, length(x$sequence_ids), replace = TRUE)
      rows <- list(); h <- 0L
      for (i in seq_along(sampled)) {
        source_id <- sampled[i]
        current <- x$data[as.character(x$data[[sequence_id_col]]) == source_id, , drop = FALSE]
        current[[sequence_id_col]] <- paste0("boot", b, "_", i)
        h <- h + 1L
        rows[[h]] <- current
      }
      boot_data <- do.call(rbind, rows)
      boot <- create_transition_network(boot_data, sequence_id_col, order_col, state_col,
                                        include_self = include_self, normalise = "from")
      boot_key <- paste(boot$from_state, boot$to_state, sep = "\034")
      values[, b] <- boot$weight[match(edge_key, boot_key)]
      values[is.na(values[, b]), b] <- 0
    }
  })
  alpha <- (1 - level) / 2
  observed$bootstrap_mean <- rowMeans(values)
  observed$bootstrap_sd <- apply(values, 1L, stats::sd)
  observed$conf_low <- apply(values, 1L, stats::quantile, probs = alpha, names = FALSE)
  observed$conf_high <- apply(values, 1L, stats::quantile, probs = 1 - alpha, names = FALSE)
  observed$n_boot <- as.integer(n_boot)
  observed$confidence_level <- level
  observed
}

test_that("first-order transition networks contain auditable edge measures", {
  data <- make_advanced_sequence_data()
  network <- create_transition_network(data, normalise = "from")
  expect_s3_class(network, "gp3_transition_network")
  expect_true(all(c("from_state", "to_state", "count", "weight",
                    "sequence_count", "sequence_prevalence") %in% names(network)))
  sums <- tapply(network$weight, network$context, sum)
  expect_equal(as.numeric(sums), rep(1, length(sums)), tolerance = 1e-12)
  without_self <- create_transition_network(data, include_self = FALSE)
  expect_false(any(without_self$from_state == without_self$to_state))
})

test_that("higher-order networks and models use explicit contexts", {
  data <- make_advanced_sequence_data()
  network <- create_transition_network(data, order = 2L, normalise = "from")
  expect_true(all(is.na(network$from_state)))
  expect_true(all(grepl(" > ", network$context, fixed = TRUE)))
  model <- fit_higher_order_transition_model(data, order = 3L, smoothing = 0.5,
                                             backoff = TRUE)
  prediction <- predict_next_state(model, c("A", "B", "C"))
  expect_equal(sum(prediction$probability), 1, tolerance = 1e-12)
  expect_true(prediction$used_order[1L] <= 3L)
  unseen <- predict_next_state(model, c("Z"))
  expect_equal(unseen$used_order[1L], 0L)
})

test_that("centrality and communities are deterministic", {
  network <- create_transition_network(make_advanced_sequence_data(), normalise = "count")
  centrality <- summarise_transition_centrality(network)
  expect_true(all(c("state", "total_degree", "total_strength", "closeness",
                    "betweenness", "pagerank") %in% names(centrality)))
  expect_equal(sum(centrality$pagerank), 1, tolerance = 1e-8)
  first <- detect_transition_communities(network, seed = 9L)
  second <- detect_transition_communities(network, seed = 9L)
  expect_identical(first, second)
  components <- detect_transition_communities(network, method = "components")
  expect_equal(nrow(components), length(unique(c(network$from_state, network$to_state))))
})

test_that("network bootstrap is reproducible and bounded", {
  data <- make_advanced_sequence_data()
  first <- bootstrap_transition_network(data, n_boot = 10L, seed = 17L)
  second <- bootstrap_transition_network(data, n_boot = 10L, seed = 17L)
  expect_equal(first, second)
  expect_true(all(first$conf_low <= first$conf_high))
  expect_true(all(first$bootstrap_mean >= 0 & first$bootstrap_mean <= 1))
})

test_that("igraph conversion is optional", {
  network <- create_transition_network(make_advanced_sequence_data())
  if (requireNamespace("igraph", quietly = TRUE)) {
    graph <- as_igraph_transition_network(network)
    expect_true(inherits(graph, "igraph"))
  } else {
    expect_error(as_igraph_transition_network(network), "Optional package")
  }
})

test_that("graph summaries require one selected group", {
  grouped <- create_transition_network(
    make_advanced_sequence_data(),
    group_cols = "group"
  )
  expect_error(
    summarise_transition_centrality(grouped),
    "Filter a grouped transition network"
  )
  expect_error(
    detect_transition_communities(grouped),
    "Filter a grouped transition network"
  )
})

test_that("unseen higher-order contexts return a stable probability schema", {
  model <- fit_higher_order_transition_model(
    make_advanced_sequence_data(),
    order = 2L
  )
  unseen <- predict_next_state(model, "UNSEEN", top_n = 2L)
  expect_true(all(c("order", "context", "next_state", "count",
                    "probability", "used_order", "used_context") %in%
                  names(unseen)))
  expect_equal(unseen$used_order, rep(0L, nrow(unseen)))
  expect_equal(nrow(unseen), 2L)
})

test_that("network bootstrap restores the caller random-number state", {
  set.seed(901L)
  before <- .Random.seed
  bootstrap_transition_network(make_advanced_sequence_data(), n_boot = 3L,
                               seed = 8L)
  expect_identical(.Random.seed, before)
})

test_that("next-state history rejects blank states", {
  model <- fit_higher_order_transition_model(
    make_advanced_sequence_data(),
    order = 2L
  )
  expect_error(predict_next_state(model, c("A", "")), "non-blank")
})

test_that("network context labels require an unambiguous separator", {
  data <- make_advanced_sequence_data()
  data$state[data$state == "A"] <- "A > embedded"
  expect_error(
    create_transition_network(data),
    "must not occur inside"
  )
  expect_error(
    fit_higher_order_transition_model(data),
    "must not occur inside"
  )
})

.sequence_adv_wide <- function(data, sequence_id_col, order_col, state_col,
                               fill = NA_character_) {
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  positions <- sort(unique(unlist(x$orders, use.names = FALSE)))
  wide <- matrix(fill, nrow = length(x$sequence_ids), ncol = length(positions),
                 dimnames = list(x$sequence_ids, paste0("position_", positions)))
  for (id in x$sequence_ids) {
    wide[id, match(x$orders[[id]], positions)] <- x$sequences[[id]]
  }
  list(wide = as.data.frame(wide, stringsAsFactors = FALSE),
       positions = positions, state_levels = x$state_levels,
       sequence_ids = x$sequence_ids)
}

#' Convert sequence data to a TraMineR state-sequence object
#'
#' @param data Long-format sequence data.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param missing Missing-state code passed to `TraMineR::seqdef()`.
#' @param right Right-missing policy passed to `TraMineR::seqdef()`.
#' @param ... Additional arguments passed to `TraMineR::seqdef()`.
#'
#' @return A TraMineR `stslist` object with original sequence identifiers as
#' row names.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' if (requireNamespace("TraMineR", quietly = TRUE)) {
#'   as_traminer_sequences(sequences)
#' }
#'
#' @export
as_traminer_sequences <- function(data,
                                  sequence_id_col = "sequence_id",
                                  order_col = "sequence_order",
                                  state_col = "state",
                                  missing = NA,
                                  right = "DEL", ...) {
  .sequence_adv_require("TraMineR", "create a TraMineR state-sequence object")
  wide <- .sequence_adv_wide(data, sequence_id_col, order_col, state_col, fill = missing)
  result <- TraMineR::seqdef(wide$wide, states = wide$state_levels,
                             missing = missing, right = right, ...)
  rownames(result) <- wide$sequence_ids
  result
}

#' Convert sequence data to cSPADE transaction input
#'
#' @param data Long-format sequence data.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#'
#' @return An `arules` transactions object whose transaction information
#' contains positive integer `sequenceID` and `eventID` fields required by
#' `arulesSequences::cspade()`.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' if (requireNamespace("arules", quietly = TRUE) &&
#'     requireNamespace("arulesSequences", quietly = TRUE)) {
#'   as_arules_sequences(sequences)
#' }
#'
#' @export
as_arules_sequences <- function(data,
                                sequence_id_col = "sequence_id",
                                order_col = "sequence_order",
                                state_col = "state") {
  .sequence_adv_require("arules", "create arules transaction input")
  .sequence_adv_require("arulesSequences", "prepare cSPADE sequence metadata")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  itemsets <- lapply(as.character(x$data[[state_col]]), function(value) value)
  transactions <- methods::as(itemsets, "transactions")
  sequence_id <- match(as.character(x$data[[sequence_id_col]]), x$sequence_ids)
  event_id <- stats::ave(seq_len(nrow(x$data)), as.character(x$data[[sequence_id_col]]),
                  FUN = seq_along)
  information <- data.frame(
    sequenceID = as.integer(sequence_id),
    eventID = as.integer(event_id),
    stringsAsFactors = FALSE
  )
  transactions <- arules::`transactionInfo<-`(
    transactions,
    value = information
  )
  transactions
}

#' Create GrpString-compatible event and string inputs
#'
#' @param data Long-format sequence data.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param alphabet Optional single-character symbols. When omitted, printable
#' ASCII characters are assigned deterministically.
#'
#' @return A list of class `gp3_grpstring_input` containing a wide event data
#' frame, event-name vector, character vector, conversion key, and string vector.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' as_grpstring_data(sequences)
#'
#' @export
as_grpstring_data <- function(data,
                              sequence_id_col = "sequence_id",
                              order_col = "sequence_order",
                              state_col = "state",
                              alphabet = NULL) {
  wide <- .sequence_adv_wide(data, sequence_id_col, order_col, state_col)
  states <- wide$state_levels
  if (is.null(alphabet)) {
    alphabet <- c(LETTERS, letters, as.character(0:9), strsplit("!#$%&*+-./:;<=>?@^_~", "")[[1L]])
  }
  if (!is.character(alphabet) || anyNA(alphabet) || any(nchar(alphabet) != 1L) ||
      anyDuplicated(alphabet) || length(alphabet) < length(states)) {
    stop("`alphabet` must provide at least one unique character per observed state.",
         call. = FALSE)
  }
  characters <- alphabet[seq_along(states)]
  key <- data.frame(event_name = states, character = characters, stringsAsFactors = FALSE)
  strings <- apply(wide$wide, 1L, function(row) {
    row <- as.character(row)
    row <- row[!is.na(row) & nzchar(row)]
    paste(characters[match(row, states)], collapse = "")
  })
  result <- list(
    events = wide$wide,
    event_names = states,
    characters = characters,
    key = key,
    strings = stats::setNames(as.character(strings), wide$sequence_ids),
    sequence_ids = wide$sequence_ids
  )
  class(result) <- c("gp3_grpstring_input", "list")
  result
}

#' Convert sequence data to seqHMM observations
#'
#' @param data Long-format sequence data.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param ... Additional arguments passed to [as_traminer_sequences()].
#'
#' @return A TraMineR `stslist` suitable for `seqHMM::build_hmm()`.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' if (requireNamespace("TraMineR", quietly = TRUE) &&
#'     requireNamespace("seqHMM", quietly = TRUE)) {
#'   as_seqhmm_sequences(sequences)
#' }
#'
#' @export
as_seqhmm_sequences <- function(data,
                                sequence_id_col = "sequence_id",
                                order_col = "sequence_order",
                                state_col = "state", ...) {
  .sequence_adv_require("seqHMM", "prepare seqHMM observations")
  as_traminer_sequences(data, sequence_id_col, order_col, state_col, ...)
}

#' Convert a transition network to an igraph object
#'
#' @param network A first-order transition network.
#' @param directed Whether the resulting graph is directed.
#'
#' @return An `igraph` graph with edge attributes copied from the network.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' network <- create_transition_network(sequences)
#' if (requireNamespace("igraph", quietly = TRUE)) {
#'   as_igraph_transition_network(network)
#' }
#'
#' @export
as_igraph_transition_network <- function(network, directed = TRUE) {
  .sequence_adv_scalar_logical(directed, "directed")
  .sequence_adv_require("igraph", "create an igraph transition network")
  if (!inherits(network, "gp3_transition_network")) {
    stop("`network` must be created by `create_transition_network()`.", call. = FALSE)
  }
  settings <- attr(network, "settings")
  if (!identical(settings$order, 1L)) {
    stop("Only first-order networks can be converted directly to igraph.", call. = FALSE)
  }
  group_cols <- attr(network, "group_cols")
  if (length(group_cols) > 0L && nrow(unique(network[, group_cols, drop = FALSE])) > 1L) {
    stop("Select one network group before converting to igraph.", call. = FALSE)
  }
  edges <- network[, c("from_state", "to_state", "weight", "count",
                       "sequence_count", "sequence_prevalence"), drop = FALSE]
  igraph::graph_from_data_frame(edges, directed = directed)
}

#' Prepare common gp3tools-style sequence outputs
#'
#' This optional compatibility helper recognises ordinary data frames or lists
#' containing a data-frame component and maps common sequence-column names to
#' the neutral gp3sequences contract. It does not require a gp3tools class.
#'
#' @param data A data frame or list with a data-frame `data` component.
#' @param sequence_id_col,order_col,state_col Optional explicit mappings.
#' @param duration_col Optional duration mapping.
#' @param metadata_cols Optional constant-within-sequence metadata columns.
#' @param ... Additional arguments passed to [prepare_sequence_data()].
#'
#' @return A standard gp3sequences preparation result.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' names(sequences)[names(sequences) == "sequence_order"] <- "position"
#' names(sequences)[names(sequences) == "state"] <- "aoi_label"
#' prepare_gp3tools_sequences(sequences)
#'
#' @export
prepare_gp3tools_sequences <- function(data,
                                       sequence_id_col = NULL,
                                       order_col = NULL,
                                       state_col = NULL,
                                       duration_col = NULL,
                                       metadata_cols = NULL, ...) {
  if (is.list(data) && !is.data.frame(data) && !is.null(data$data)) data <- data$data
  if (!is.data.frame(data)) stop("A data frame or list with a data-frame `data` component is required.",
                                 call. = FALSE)
  choose_first <- function(explicit, candidates, role) {
    if (!is.null(explicit)) return(explicit)
    found <- candidates[candidates %in% names(data)]
    if (length(found) == 0L) stop("Could not infer the ", role, " column. Supply it explicitly.",
                                  call. = FALSE)
    if (length(found) > 1L) {
      stop("Multiple candidate ", role, " columns were found: ",
           paste(found, collapse = ", "), ". Supply the mapping explicitly.",
           call. = FALSE)
    }
    found[1L]
  }
  sequence_id_col <- choose_first(sequence_id_col,
                                  c("sequence_id", "scanpath_id", "trial_id", "participant_trial_id"),
                                  "sequence identifier")
  order_col <- choose_first(order_col,
                            c("sequence_order", "position", "event_order", "fixation_index", "row_order"),
                            "sequence order")
  state_col <- choose_first(state_col,
                            c("state", "aoi", "aoi_label", "event", "event_name"),
                            "state")
  if (is.null(duration_col)) {
    found_duration <- c("duration", "fixation_duration", "event_duration")
    found_duration <- found_duration[found_duration %in% names(data)]
    if (length(found_duration) > 1L) {
      stop("Multiple candidate duration columns were found: ",
           paste(found_duration, collapse = ", "),
           ". Supply `duration_col` explicitly.", call. = FALSE)
    }
    if (length(found_duration) == 1L) duration_col <- found_duration[1L]
  }
  prepare_sequence_data(data,
                        sequence_id_col = sequence_id_col,
                        order_col = order_col,
                        state_col = state_col,
                        duration_col = duration_col,
                        metadata_cols = metadata_cols, ...)
}

test_that("GrpString adapter creates deterministic strings and key", {
  data <- make_advanced_sequence_data()
  adapted <- as_grpstring_data(data)
  expect_s3_class(adapted, "gp3_grpstring_input")
  expect_equal(length(adapted$strings), 8L)
  expect_equal(nrow(adapted$key), 4L)
  expect_true(all(nchar(adapted$strings) == 5L))
  expect_identical(adapted, as_grpstring_data(data))
})

test_that("gp3tools compatibility helper maps common columns", {
  data <- make_advanced_sequence_data()
  names(data)[names(data) == "sequence_order"] <- "position"
  names(data)[names(data) == "state"] <- "aoi_label"
  prepared <- prepare_gp3tools_sequences(data, metadata_cols = c("group", "participant_id"))
  expect_true(prepared$status %in% c("pass", "review"))
  expect_true(all(c("sequence_id", "sequence_order", "state") %in% names(prepared$data)))
})

test_that("TraMineR and seqHMM adapters are guarded", {
  data <- make_advanced_sequence_data()
  if (requireNamespace("TraMineR", quietly = TRUE)) {
    object <- as_traminer_sequences(data)
    expect_true(inherits(object, "stslist"))
  } else {
    expect_error(as_traminer_sequences(data), "Optional package")
  }
  if (requireNamespace("TraMineR", quietly = TRUE) &&
      requireNamespace("seqHMM", quietly = TRUE)) {
    object <- as_seqhmm_sequences(data)
    expect_true(inherits(object, "stslist"))
  } else {
    expect_error(as_seqhmm_sequences(data), "Optional package")
  }
})

test_that("arules adapter is guarded or returns sequential metadata", {
  data <- make_advanced_sequence_data()
  if (requireNamespace("arules", quietly = TRUE) &&
      requireNamespace("arulesSequences", quietly = TRUE)) {
    transactions <- as_arules_sequences(data)
    info <- arules::transactionInfo(transactions)
    expect_true(all(c("sequenceID", "eventID") %in% names(info)))
    expect_equal(nrow(info), nrow(data))
    expect_identical(
      info$sequenceID,
      rep(seq_len(8L), each = 5L)
    )
    expect_identical(
      info$eventID,
      rep(seq_len(5L), times = 8L)
    )
    expect_true(all(info$sequenceID > 0L))
    expect_true(all(info$eventID > 0L))
  } else {
    expect_error(as_arules_sequences(data), "Optional package")
  }
})


test_that("igraph adapter refuses unresolved grouped networks", {
  data <- make_advanced_sequence_data()
  grouped <- create_transition_network(data, group_cols = "group")
  if (requireNamespace("igraph", quietly = TRUE)) {
    expect_error(
      as_igraph_transition_network(grouped),
      "Select one network group"
    )
  } else {
    expect_error(as_igraph_transition_network(grouped), "Optional package")
  }
})

test_that("gp3tools compatibility helper refuses ambiguous inferred mappings", {
  data <- make_advanced_sequence_data()
  data$trial_id <- data$sequence_id
  expect_error(
    prepare_gp3tools_sequences(data, order_col = "sequence_order", state_col = "state"),
    "Multiple candidate sequence identifier"
  )

  data <- make_advanced_sequence_data()
  data$fixation_duration <- 100
  data$event_duration <- 100
  expect_error(
    prepare_gp3tools_sequences(data),
    "Multiple candidate duration"
  )
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----guide--------------------------------------------------------------------
method_guide <- data.frame(
  question = c(
    "Which exact contiguous patterns recur?",
    "What state is most supported at each aligned position?",
    "How do observed state, transition, or length summaries differ by group?",
    "How dissimilar are complete sequences?",
    "Can a declared distance reveal reproducible descriptive partitions?",
    "Which transitions and contexts organise the observed paths?",
    "Does recent state history improve next-state description?",
    "Can a compact latent categorical model summarise serial dependence?",
    "Is interoperability with a specialist package required?"
  ),
  method = c(
    "Contiguous motifs",
    "Aligned-position consensus",
    "Descriptive group comparison",
    "Sequence distance",
    "Clustering and stability",
    "Transition network",
    "Higher-order transition model",
    "Categorical HMM or mixture HMM",
    "Optional adapter"
  ),
  primary_functions = c(
    "extract_sequence_ngrams(); summarise_sequence_motifs()",
    "create_consensus_sequence(); summarise_consensus_agreement()",
    "compare_sequence_groups()",
    "compute_sequence_distance(); summarise_sequence_distance()",
    "cluster_sequences(); validate_sequence_clusters(); bootstrap_sequence_clusters()",
    "create_transition_network(); summarise_transition_centrality()",
    "fit_higher_order_transition_model(); predict_next_state()",
    "fit_sequence_hmm(); fit_sequence_hmm_mixture(); decode_sequence_states()",
    "as_traminer_sequences(); as_arules_sequences(); as_igraph_transition_network()"
  ),
  stringsAsFactors = FALSE
)

method_guide

.sequence_contract_version <- function() {
  "0.3.0-contract-1"
}

.sequence_contract_issue_table <- function() {
  data.frame(
    code = character(),
    severity = character(),
    field = character(),
    message = character(),
    stringsAsFactors = FALSE
  )
}

.sequence_contract_issue <- function(code, severity, field = NA_character_, message) {
  data.frame(
    code = as.character(code),
    severity = as.character(severity),
    field = as.character(field),
    message = as.character(message),
    stringsAsFactors = FALSE
  )
}

.sequence_contract_bind_issues <- function(...) {
  pieces <- Filter(
    function(x) is.data.frame(x) && nrow(x) > 0L,
    list(...)
  )

  if (length(pieces) == 0L) {
    return(.sequence_contract_issue_table())
  }

  out <- do.call(rbind, pieces)
  row.names(out) <- NULL
  severity_rank <- match(out$severity, c("error", "review", "info"))
  out <- out[
    order(severity_rank, out$code, out$field, method = "radix"),
    ,
    drop = FALSE
  ]
  row.names(out) <- NULL
  out
}

.sequence_contract_status <- function(issues) {
  if (nrow(issues) == 0L) {
    return("pass")
  }
  if (any(issues$severity == "error")) {
    return("fail")
  }
  if (any(issues$severity == "review")) {
    return("review")
  }
  "pass"
}

.sequence_check_optional_backend <- function(package) {
  if (
    !is.character(package) ||
      length(package) != 1L ||
      is.na(package) ||
      !nzchar(package)
  ) {
    stop(
      "`package` must be one non-missing package name.",
      call. = FALSE
    )
  }

  requireNamespace(package, quietly = TRUE)
}

.sequence_installed_version <- function(package) {
  if (!.sequence_check_optional_backend(package)) {
    return(NA_character_)
  }

  as.character(utils::packageVersion(package))
}

.sequence_check_seed <- function(seed, allow_null = TRUE) {
  if (allow_null && is.null(seed)) {
    return(invisible(NULL))
  }

  if (
    !is.numeric(seed) ||
      length(seed) != 1L ||
      is.na(seed) ||
      !is.finite(seed) ||
      seed < 0 ||
      seed != floor(seed) ||
      seed > .Machine$integer.max
  ) {
    stop(
      "`seed` must be one non-negative finite integer or `NULL`.",
      call. = FALSE
    )
  }

  invisible(as.integer(seed))
}

.sequence_check_probability_simplex <- function(x, tolerance = 1e-8) {
  if (
    !is.numeric(x) ||
      length(x) == 0L ||
      anyNA(x) ||
      any(!is.finite(x))
  ) {
    return(FALSE)
  }

  if (any(x < -tolerance) || any(x > 1 + tolerance)) {
    return(FALSE)
  }

  abs(sum(x) - 1) <= tolerance
}

.sequence_validate_probability_matrix <- function(
  x,
  tolerance = 1e-8,
  field = "probability_matrix"
) {
  issues <- .sequence_contract_issue_table()

  if (
    !is.matrix(x) ||
      !is.numeric(x) ||
      nrow(x) == 0L ||
      ncol(x) == 0L
  ) {
    return(
      .sequence_contract_issue(
        "invalid_probability_matrix",
        "error",
        field,
        "The probability object must be a non-empty numeric matrix."
      )
    )
  }

  if (anyNA(x) || any(!is.finite(x))) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "non_finite_probability",
        "error",
        field,
        "Probability matrices must contain only finite, non-missing values."
      )
    )
  }

  if (
    any(x < -tolerance, na.rm = TRUE) ||
      any(x > 1 + tolerance, na.rm = TRUE)
  ) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "probability_out_of_bounds",
        "error",
        field,
        "Probability entries must lie in [0, 1] within tolerance."
      )
    )
  }

  row_sums <- rowSums(x)
  if (any(abs(row_sums - 1) > tolerance, na.rm = TRUE)) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "probability_rows_not_normalised",
        "error",
        field,
        "Every probability row must sum to one within tolerance."
      )
    )
  }

  issues
}

.sequence_validate_distance_matrix <- function(x, tolerance = 1e-8) {
  matrix_x <- if (inherits(x, "dist")) {
    as.matrix(x)
  } else {
    x
  }

  if (
    !is.matrix(matrix_x) ||
      !is.numeric(matrix_x) ||
      nrow(matrix_x) == 0L ||
      nrow(matrix_x) != ncol(matrix_x)
  ) {
    return(
      .sequence_contract_issue(
        "invalid_distance_matrix",
        "error",
        "distance",
        paste0(
          "A distance result must be a non-empty numeric square matrix ",
          "or `dist` object."
        )
      )
    )
  }

  issues <- .sequence_contract_issue_table()

  if (anyNA(matrix_x) || any(!is.finite(matrix_x))) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "non_finite_distance",
        "error",
        "distance",
        "Distance matrices must contain finite, non-missing values."
      )
    )
  }

  if (any(matrix_x < -tolerance, na.rm = TRUE)) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "negative_distance",
        "error",
        "distance",
        "Distances must be non-negative within tolerance."
      )
    )
  }

  if (any(abs(diag(matrix_x)) > tolerance, na.rm = TRUE)) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "nonzero_distance_diagonal",
        "error",
        "distance",
        "The distance diagonal must be zero within tolerance."
      )
    )
  }

  symmetry_error <- max(abs(matrix_x - t(matrix_x)), na.rm = TRUE)
  if (is.finite(symmetry_error) && symmetry_error > tolerance) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "asymmetric_distance",
        "error",
        "distance",
        "The distance matrix must be symmetric within tolerance."
      )
    )
  }

  rn <- rownames(matrix_x)
  cn <- colnames(matrix_x)

  if (
    is.null(rn) ||
      is.null(cn) ||
      anyNA(rn) ||
      anyNA(cn) ||
      any(!nzchar(rn)) ||
      any(!nzchar(cn)) ||
      anyDuplicated(rn) ||
      anyDuplicated(cn) ||
      !identical(rn, cn)
  ) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "distance_identifiers_invalid",
        "review",
        "distance",
        paste0(
          "Distance row/column identifiers should be unique, non-missing, ",
          "and identical."
        )
      )
    )
  }

  issues
}

.sequence_validate_transition_matrix <- function(x, tolerance = 1e-8) {
  if (inherits(x, "gp3_transition_network")) {
    settings <- attr(x, "settings")
    issues <- .sequence_contract_issue_table()
    required <- c("context", "to_state", "count", "weight")
    missing <- setdiff(required, names(x))

    if (length(missing) > 0L) {
      return(
        .sequence_contract_issue(
          "transition_columns_missing",
          "error",
          "network",
          paste(
            "Missing transition-network columns:",
            paste(missing, collapse = ", ")
          )
        )
      )
    }

    if (
      anyNA(x$count) ||
        any(!is.finite(x$count)) ||
        any(x$count < 0)
    ) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_contract_issue(
          "invalid_transition_counts",
          "error",
          "count",
          "Transition counts must be finite and non-negative."
        )
      )
    }

    if (
      anyNA(x$weight) ||
        any(!is.finite(x$weight)) ||
        any(x$weight < -tolerance)
    ) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_contract_issue(
          "invalid_transition_weights",
          "error",
          "weight",
          "Transition weights must be finite and non-negative."
        )
      )
    }

    if (
      is.list(settings) &&
        identical(settings$normalise, "from") &&
        nrow(x) > 0L
    ) {
      group_key <- if ("group_key" %in% names(x)) {
        x$group_key
      } else {
        rep("", nrow(x))
      }
      key <- paste(group_key, x$context, sep = "\034")
      sums <- tapply(x$weight, key, sum)

      if (any(abs(sums - 1) > tolerance)) {
        issues <- .sequence_contract_bind_issues(
          issues,
          .sequence_contract_issue(
            "transition_rows_not_normalised",
            "error",
            "weight",
            paste0(
              "Outgoing transition weights must sum to one for ",
              "`normalise = \"from\"`."
            )
          )
        )
      }
    }

    return(issues)
  }

  .sequence_validate_probability_matrix(
    x,
    tolerance = tolerance,
    field = "transition"
  )
}

.sequence_validate_clustering_object <- function(x) {
  if (
    !inherits(
      x,
      c(
        "gp3_sequence_clustering",
        "gp3_sequence_cluster_ensemble"
      )
    )
  ) {
    return(
      .sequence_contract_issue(
        "unsupported_clustering_object",
        "error",
        "class",
        "The object is not a supported gp3sequences clustering result."
      )
    )
  }

  issues <- .sequence_contract_issue_table()
  assignments <- x$assignments

  if (
    is.null(assignments) ||
      length(assignments) == 0L ||
      anyNA(assignments) ||
      is.null(names(assignments)) ||
      anyNA(names(assignments)) ||
      any(!nzchar(names(assignments))) ||
      anyDuplicated(names(assignments))
  ) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "invalid_cluster_assignments",
        "error",
        "assignments",
        paste0(
          "Cluster assignments must be non-missing and uniquely named ",
          "by sequence ID."
        )
      )
    )
  }

  if (
    !is.null(x$k) &&
      length(assignments) > 0L &&
      length(unique(assignments)) != x$k
  ) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "cluster_count_mismatch",
        "review",
        "k",
        paste0(
          "The stored `k` does not match the number of observed ",
          "assignment labels."
        )
      )
    )
  }

  issues
}

.sequence_validate_hmm_object <- function(x, tolerance = 1e-8) {
  hmm_classes <- c(
    "gp3_sequence_hmm",
    "gp3_sequence_hmm_mixture",
    "gp3_multichannel_sequence_hmm",
    "gp3_covariate_sequence_hmm"
  )

  if (!inherits(x, hmm_classes)) {
    return(
      .sequence_contract_issue(
        "unsupported_hmm_object",
        "error",
        "class",
        "The object is not a supported gp3sequences HMM result."
      )
    )
  }

  issues <- .sequence_contract_issue_table()

  if (inherits(x, "gp3_sequence_hmm")) {
    if (!.sequence_check_probability_simplex(x$initial_probs, tolerance)) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_contract_issue(
          "invalid_initial_probabilities",
          "error",
          "initial_probs",
          "Initial probabilities must form a probability simplex."
        )
      )
    }

    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_validate_probability_matrix(
        x$transition_probs,
        tolerance,
        "transition_probs"
      ),
      .sequence_validate_probability_matrix(
        x$emission_probs,
        tolerance,
        "emission_probs"
      )
    )

    if (!isTRUE(x$converged)) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_contract_issue(
          "hmm_not_converged",
          "review",
          "converged",
          "The HMM did not report convergence."
        )
      )
    }
  }

  if (inherits(x, "gp3_sequence_hmm_mixture")) {
    if (
      !is.null(x$component_probs) &&
        !.sequence_check_probability_simplex(
          x$component_probs,
          tolerance
        )
    ) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_contract_issue(
          "invalid_component_probabilities",
          "error",
          "component_probs",
          "Mixture probabilities must form a probability simplex."
        )
      )
    }

    if (!is.null(x$models)) {
      for (model in x$models) {
        issues <- .sequence_contract_bind_issues(
          issues,
          .sequence_validate_hmm_object(
            model,
            tolerance = tolerance
          )
        )
      }
    }

    if (!isTRUE(x$converged)) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_contract_issue(
          "mixture_not_converged",
          "review",
          "converged",
          "The HMM mixture did not report convergence."
        )
      )
    }
  }

  # Multichannel/covariate models have family-specific structures.
  # The foundation layer checks common matrix fields only when they exist.
  for (
    field in intersect(
      c(
        "transition_probs",
        "transition",
        "emission_probs",
        "emission"
      ),
      names(x)
    )
  ) {
    value <- x[[field]]

    if (is.matrix(value) && is.numeric(value)) {
      issues <- .sequence_contract_bind_issues(
        issues,
        .sequence_validate_probability_matrix(
          value,
          tolerance = tolerance,
          field = field
        )
      )
    }
  }

  issues
}

.sequence_object_family <- function(x) {
  if (inherits(x, "gp3_sequence_distance")) {
    return("distance")
  }
  if (inherits(x, "gp3_sequence_clustering")) {
    return("clustering")
  }
  if (inherits(x, "gp3_sequence_cluster_bootstrap")) {
    return("cluster_bootstrap")
  }
  if (inherits(x, "gp3_sequence_cluster_ensemble")) {
    return("cluster_ensemble")
  }
  if (inherits(x, "gp3_transition_network")) {
    return("transition_network")
  }
  if (inherits(x, "gp3_higher_order_transition_model")) {
    return("higher_order_transition")
  }
  if (inherits(x, "gp3_sequence_hmm")) {
    return("hmm")
  }
  if (inherits(x, "gp3_sequence_hmm_mixture")) {
    return("hmm_mixture")
  }
  if (inherits(x, "gp3_multichannel_sequence_hmm")) {
    return("multichannel_hmm")
  }
  if (inherits(x, "gp3_covariate_sequence_hmm")) {
    return("covariate_hmm")
  }

  if (
    is.list(x) &&
      all(
        c(
          "data",
          "audit",
          "decisions",
          "mapping",
          "status"
        ) %in% names(x)
      )
  ) {
    return("prepared_sequence_data")
  }

  if (
    is.list(x) &&
      all(
        c(
          "valid",
          "status",
          "audit",
          "mapping"
        ) %in% names(x)
      )
  ) {
    return("sequence_validation")
  }

  "generic"
}

.sequence_object_contract <- function(x) {
  family <- .sequence_object_family(x)

  required_fields <- switch(
    family,
    distance = c(
      "method",
      "sequence_ids",
      "state_levels",
      "settings"
    ),
    clustering = c(
      "assignments",
      "k",
      "method",
      "distance",
      "seed"
    ),
    cluster_bootstrap = c(
      "original",
      "pairwise_stability"
    ),
    cluster_ensemble = c(
      "assignments",
      "coassociation",
      "source_assignments"
    ),
    transition_network = c(
      "context",
      "to_state",
      "count",
      "weight"
    ),
    higher_order_transition = c(
      "order",
      "tables",
      "state_levels"
    ),
    hmm = c(
      "initial_probs",
      "transition_probs",
      "emission_probs",
      "state_names",
      "symbol_names",
      "log_likelihood",
      "converged",
      "seed"
    ),
    hmm_mixture = c(
      "models",
      "responsibilities",
      "log_likelihood",
      "converged",
      "seed"
    ),
    prepared_sequence_data = c(
      "data",
      "audit",
      "decisions",
      "mapping",
      "status",
      "state_levels"
    ),
    sequence_validation = c(
      "valid",
      "status",
      "audit",
      "mapping",
      "state_levels"
    ),
    character()
  )

  list(
    contract_version = .sequence_contract_version(),
    family = family,
    primary_class = class(x)[1L],
    required_fields = required_fields
  )
}

.sequence_validate_sequence_result <- function(x, tolerance = 1e-8) {
  family <- .sequence_object_family(x)

  if (identical(family, "distance")) {
    return(
      .sequence_validate_distance_matrix(
        x,
        tolerance = tolerance
      )
    )
  }

  if (family %in% c("clustering", "cluster_ensemble")) {
    return(.sequence_validate_clustering_object(x))
  }

  if (identical(family, "transition_network")) {
    return(
      .sequence_validate_transition_matrix(
        x,
        tolerance = tolerance
      )
    )
  }

  if (
    family %in% c(
      "hmm",
      "hmm_mixture",
      "multichannel_hmm",
      "covariate_hmm"
    )
  ) {
    return(
      .sequence_validate_hmm_object(
        x,
        tolerance = tolerance
      )
    )
  }

  if (identical(family, "prepared_sequence_data")) {
    if (!x$status %in% c("pass", "review", "fail")) {
      return(
        .sequence_contract_issue(
          "invalid_preparation_status",
          "error",
          "status",
          paste0(
            "Prepared sequence data must use status ",
            "pass, review, or fail."
          )
        )
      )
    }

    return(.sequence_contract_issue_table())
  }

  if (identical(family, "sequence_validation")) {
    if (!x$status %in% c("pass", "review", "fail")) {
      return(
        .sequence_contract_issue(
          "invalid_validation_status",
          "error",
          "status",
          paste0(
            "Validation results must use status ",
            "pass, review, or fail."
          )
        )
      )
    }

    return(.sequence_contract_issue_table())
  }

  .sequence_contract_issue(
    "generic_object_contract",
    "review",
    "class",
    paste0(
      "No specialised gp3sequences contract is currently ",
      "registered for this object."
    )
  )
}

.sequence_record_provenance <- function(x, provenance = list()) {
  if (!is.list(provenance)) {
    stop("`provenance` must be a named list.", call. = FALSE)
  }

  if (
    length(provenance) > 0L &&
      (
        is.null(names(provenance)) ||
          any(!nzchar(names(provenance)))
      )
  ) {
    stop(
      "Every provenance field must be named.",
      call. = FALSE
    )
  }

  base <- attr(
    x,
    "gp3sequences_provenance",
    exact = TRUE
  )

  if (is.null(base)) {
    base <- list()
  }

  if (!is.list(base)) {
    base <- list(previous = base)
  }

  if (is.null(provenance$contract_version)) {
    provenance$contract_version <- .sequence_contract_version()
  }

  attr(x, "gp3sequences_provenance") <- utils::modifyList(
    base,
    provenance
  )

  x
}

.sequence_restore_sequence_metadata <- function(x, template) {
  for (
    field in c(
      "method",
      "sequence_ids",
      "state_levels",
      "settings",
      "gp3sequences_provenance"
    )
  ) {
    value <- attr(
      template,
      field,
      exact = TRUE
    )

    if (!is.null(value)) {
      attr(x, field) <- value
    }
  }

  x
}

.sequence_align_state_levels <- function(x, y) {
  x <- as.character(x)
  y <- as.character(y)

  if (
    anyNA(x) ||
      anyNA(y) ||
      any(!nzchar(x)) ||
      any(!nzchar(y))
  ) {
    stop(
      "State levels must be non-missing, non-empty values.",
      call. = FALSE
    )
  }

  unique(c(x, y))
}

.sequence_align_sequence_ids <- function(x, y) {
  x <- as.character(x)
  y <- as.character(y)

  if (
    anyNA(x) ||
      anyNA(y) ||
      any(!nzchar(x)) ||
      any(!nzchar(y)) ||
      anyDuplicated(x) ||
      anyDuplicated(y)
  ) {
    stop(
      "Sequence IDs must be unique, non-missing, and non-empty.",
      call. = FALSE
    )
  }

  if (!setequal(x, y)) {
    stop(
      "Sequence-ID sets do not match.",
      call. = FALSE
    )
  }

  x
}

.sequence_align_partition_labels <- function(assignments) {
  if (
    is.null(names(assignments)) ||
      anyNA(assignments) ||
      anyNA(names(assignments)) ||
      any(!nzchar(names(assignments))) ||
      anyDuplicated(names(assignments))
  ) {
    stop(
      "Assignments must be non-missing and uniquely named.",
      call. = FALSE
    )
  }

  labels <- unique(as.character(assignments))

  order_key <- vapply(
    labels,
    function(label) {
      min(
        names(assignments)[
          as.character(assignments) == label
        ]
      )
    },
    character(1)
  )

  ordered <- labels[
    order(order_key, labels, method = "radix")
  ]

  mapping <- stats::setNames(
    seq_along(ordered),
    ordered
  )

  stats::setNames(
    unname(mapping[as.character(assignments)]),
    names(assignments)
  )
}

test_that("distance contract enforces core mathematical invariants", {
  x <- .sequence_test_case("minimal")

  distance <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )
  matrix_distance <- as.matrix(
    distance
  )

  expect_true(
    all(matrix_distance >= 0)
  )
  expect_equal(
    unname(diag(matrix_distance)),
    rep(
      0,
      nrow(matrix_distance)
    )
  )
  expect_equal(
    matrix_distance,
    t(matrix_distance)
  )
  expect_identical(
    rownames(matrix_distance),
    colnames(matrix_distance)
  )

  issues <- .sequence_validate_distance_matrix(
    distance
  )

  expect_equal(
    nrow(issues),
    0L
  )
})

test_that("probability simplex and matrix validators work", {
  expect_true(
    .sequence_check_probability_simplex(
      c(0.2, 0.8)
    )
  )
  expect_false(
    .sequence_check_probability_simplex(
      c(0.2, 0.9)
    )
  )

  valid <- matrix(
    c(
      0.8, 0.2,
      0.3, 0.7
    ),
    2L,
    byrow = TRUE
  )

  invalid <- valid
  invalid[1L, ] <- c(0.8, 0.3)

  expect_equal(
    nrow(
      .sequence_validate_probability_matrix(
        valid
      )
    ),
    0L
  )

  expect_true(
    any(
      .sequence_validate_probability_matrix(
        invalid
      )$code ==
        "probability_rows_not_normalised"
    )
  )
})

test_that("partition labels canonicalise without changing membership", {
  assignments <- c(
    s3 = "beta",
    s1 = "alpha",
    s2 = "alpha",
    s4 = "beta"
  )

  canonical <- .sequence_align_partition_labels(
    assignments
  )

  expect_identical(
    names(canonical),
    names(assignments)
  )
  expect_equal(
    unname(canonical["s1"]),
    unname(canonical["s2"])
  )
  expect_equal(
    unname(canonical["s3"]),
    unname(canonical["s4"])
  )
  expect_false(
    canonical["s1"] ==
      canonical["s3"]
  )
})

.sequence_assert_column_argument <- function(
    value,
    argument,
    allow_null = FALSE
) {
  if (allow_null && is.null(value)) {
    return(invisible(NULL))
  }

  if (
    !is.character(value) ||
    length(value) != 1L ||
    is.na(value) ||
    !nzchar(value)
  ) {
    stop(
      "`",
      argument,
      "` must be a single non-missing column name.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_assert_character_vector <- function(
    value,
    argument,
    allow_null = TRUE
) {
  if (allow_null && is.null(value)) {
    return(invisible(NULL))
  }

  if (
    !is.character(value) ||
    anyNA(value) ||
    any(!nzchar(value)) ||
    anyDuplicated(value)
  ) {
    stop(
      "`",
      argument,
      "` must contain unique, non-missing character values.",
      call. = FALSE
    )
  }

  invisible(NULL)
}

.sequence_missing_value <- function(value) {
  missing <- is.na(value)

  if (is.character(value) || is.factor(value)) {
    missing <- missing | trimws(as.character(value)) == ""
  }

  missing
}

.sequence_value_text <- function(value) {
  if (length(value) == 0L) {
    return(NA_character_)
  }

  value <- as.character(value)
  value[is.na(value)] <- "<NA>"

  paste(utils::head(value, 5L), collapse = " | ")
}

.sequence_empty_audit <- function() {
  data.frame(
    sequence_id = character(),
    row = integer(),
    column = character(),
    issue_code = character(),
    severity = character(),
    value = character(),
    message = character(),
    action = character(),
    stringsAsFactors = FALSE
  )
}

.sequence_issue_row <- function(
    sequence_id = NA_character_,
    row = NA_integer_,
    column = NA_character_,
    issue_code,
    severity,
    value = NA_character_,
    message,
    action
) {
  data.frame(
    sequence_id = as.character(sequence_id),
    row = as.integer(row),
    column = as.character(column),
    issue_code = as.character(issue_code),
    severity = as.character(severity),
    value = as.character(value),
    message = as.character(message),
    action = as.character(action),
    stringsAsFactors = FALSE
  )
}

.sequence_sort_audit <- function(audit) {
  if (nrow(audit) == 0L) {
    return(.sequence_empty_audit())
  }

  severity_rank <- match(
    audit$severity,
    c("error", "review", "info")
  )

  sequence_sort <- audit$sequence_id
  sequence_sort[is.na(sequence_sort)] <- "\U10FFFF"

  row_sort <- audit$row
  row_sort[is.na(row_sort)] <- .Machine$integer.max

  audit <- audit[
    order(
      severity_rank,
      sequence_sort,
      row_sort,
      audit$issue_code,
      method = "radix"
    ),
    ,
    drop = FALSE
  ]

  row.names(audit) <- NULL
  audit
}

.sequence_status <- function(audit) {
  if (any(audit$severity == "error")) {
    return("fail")
  }

  if (any(audit$severity == "review")) {
    return("review")
  }

  "pass"
}

.sequence_mapping <- function(
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL
) {
  roles <- c(
    "sequence_id",
    "sequence_order",
    "state"
  )

  columns <- c(
    sequence_id_col,
    order_col,
    state_col
  )

  if (!is.null(duration_col)) {
    roles <- c(roles, "duration")
    columns <- c(columns, duration_col)
  }

  if (length(metadata_cols) > 0L) {
    roles <- c(
      roles,
      paste0("metadata:", metadata_cols)
    )
    columns <- c(columns, metadata_cols)
  }

  data.frame(
    role = roles,
    source_column = columns,
    stringsAsFactors = FALSE
  )
}

.sequence_stage_audit <- function(audit, stage) {
  if (nrow(audit) == 0L) {
    return(
      data.frame(
        stage = character(),
        .sequence_empty_audit(),
        stringsAsFactors = FALSE
      )
    )
  }

  data.frame(
    stage = rep(stage, nrow(audit)),
    audit,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
}

.sequence_empty_decisions <- function() {
  data.frame(
    step = character(),
    policy = character(),
    affected_rows = integer(),
    details = character(),
    stringsAsFactors = FALSE
  )
}

.sequence_prepare_result <- function(
    data,
    audit,
    decisions,
    mapping,
    status,
    original_n_rows,
    prepared_n_rows,
    state_levels
) {
  list(
    data = data,
    audit = audit,
    decisions = decisions,
    mapping = mapping,
    status = status,
    original_n_rows = as.integer(original_n_rows),
    prepared_n_rows = as.integer(prepared_n_rows),
    state_levels = as.character(state_levels)
  )
}

#' Audit Long-Format Sequence Data
#'
#' Examines a long-format data frame against the neutral `gp3sequences`
#' sequence-data contract without modifying the input.
#'
#' @param data A data frame containing ordered state observations.
#' @param sequence_id_col Name of the sequence identifier column.
#' @param order_col Name of the numeric sequence-order column.
#' @param state_col Name of the categorical state column.
#' @param duration_col Optional name of a numeric duration column.
#' @param metadata_cols Optional character vector naming columns that should
#'   remain constant within each sequence.
#' @param expected_states Optional vector of known or permitted state values.
#'
#' @return A data frame with one row per detected issue and the stable columns
#'   `sequence_id`, `row`, `column`, `issue_code`, `severity`, `value`,
#'   `message`, and `action`. Severity values are `error`, `review`, and
#'   `info`.
#'
#' @details
#' The audit checks column mappings, empty inputs, missing identifiers,
#' missing or non-numeric order values, duplicated positions, integer order
#' gaps, unordered rows, missing states, consecutive repeated states,
#' single-row sequences, invalid durations, inconsistent metadata, unexpected
#' states, and unused factor levels.
#'
#' The function reports structural properties only. It does not infer
#' psychological, cognitive, emotional, or diagnostic states.
#'
#' @examples
#' sequences <- data.frame(
#'   id = rep(c("s1", "s2"), each = 3L),
#'   position = rep(1:3, times = 2L),
#'   state = c("home", "search", "product", "home", "category", "product")
#' )
#'
#' audit_sequence_data(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' @export
audit_sequence_data <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL,
    expected_states = NULL
) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  .sequence_assert_column_argument(
    sequence_id_col,
    "sequence_id_col"
  )
  .sequence_assert_column_argument(order_col, "order_col")
  .sequence_assert_column_argument(state_col, "state_col")
  .sequence_assert_column_argument(
    duration_col,
    "duration_col",
    allow_null = TRUE
  )
  .sequence_assert_character_vector(
    metadata_cols,
    "metadata_cols"
  )

  if (!is.null(expected_states) && !is.atomic(expected_states)) {
    stop(
      "`expected_states` must be an atomic vector or `NULL`.",
      call. = FALSE
    )
  }

  mapped_core <- c(
    sequence_id_col,
    order_col,
    state_col,
    duration_col
  )
  mapped_core <- mapped_core[!is.na(mapped_core)]

  if (anyDuplicated(mapped_core)) {
    stop(
      "Required column mappings must refer to distinct columns.",
      call. = FALSE
    )
  }

  if (
    length(metadata_cols) > 0L &&
    any(metadata_cols %in% mapped_core)
  ) {
    stop(
      "`metadata_cols` must not repeat required mapped columns.",
      call. = FALSE
    )
  }

  issues <- list()

  add_issue <- function(...) {
    issues[[length(issues) + 1L]] <<- .sequence_issue_row(...)
    invisible(NULL)
  }

  if (anyDuplicated(names(data))) {
    add_issue(
      issue_code = "duplicate_column_names",
      severity = "error",
      value = .sequence_value_text(
        names(data)[duplicated(names(data))]
      ),
      message = "The data frame contains duplicated column names.",
      action = "Rename duplicated columns before validation."
    )

    return(.sequence_sort_audit(do.call(rbind, issues)))
  }

  required_mapping <- c(
    sequence_id = sequence_id_col,
    sequence_order = order_col,
    state = state_col
  )

  if (!is.null(duration_col)) {
    required_mapping <- c(
      required_mapping,
      duration = duration_col
    )
  }

  missing_required <- required_mapping[
    !required_mapping %in% names(data)
  ]

  for (role in names(missing_required)) {
    add_issue(
      column = missing_required[[role]],
      issue_code = "missing_required_column",
      severity = "error",
      value = missing_required[[role]],
      message = paste0(
        "The mapped ",
        role,
        " column is absent from `data`."
      ),
      action = "Correct the column mapping or add the required column."
    )
  }

  missing_metadata <- setdiff(metadata_cols, names(data))

  for (column in missing_metadata) {
    add_issue(
      column = column,
      issue_code = "missing_metadata_column",
      severity = "error",
      value = column,
      message = "A requested metadata column is absent from `data`.",
      action = "Correct `metadata_cols` or add the metadata column."
    )
  }

  core_columns_present <- all(
    c(sequence_id_col, order_col, state_col) %in% names(data)
  )

  if (!core_columns_present) {
    return(.sequence_sort_audit(do.call(rbind, issues)))
  }

  if (nrow(data) == 0L) {
    add_issue(
      issue_code = "empty_data",
      severity = "error",
      message = "The input contains no sequence rows.",
      action = "Provide at least one ordered state observation."
    )

    return(.sequence_sort_audit(do.call(rbind, issues)))
  }

  sequence_values <- data[[sequence_id_col]]
  order_values <- data[[order_col]]
  state_values <- data[[state_col]]

  valid_sequence_type <- !is.list(sequence_values)
  valid_state_type <- !is.list(state_values)
  valid_order_type <- is.numeric(order_values)

  if (!valid_sequence_type) {
    add_issue(
      column = sequence_id_col,
      issue_code = "invalid_sequence_id_type",
      severity = "error",
      message = "Sequence identifiers must use an atomic vector.",
      action = "Convert sequence identifiers to character, factor, or numeric."
    )
  }

  if (!valid_state_type) {
    add_issue(
      column = state_col,
      issue_code = "invalid_state_type",
      severity = "error",
      message = "States must use an atomic vector.",
      action = "Convert states to character, factor, or numeric."
    )
  }

  if (!valid_order_type) {
    add_issue(
      column = order_col,
      issue_code = "invalid_order_type",
      severity = "error",
      message = "Sequence order must be numeric.",
      action = "Convert the order column to finite numeric positions."
    )
  }

  if (!valid_sequence_type || !valid_state_type) {
    return(.sequence_sort_audit(do.call(rbind, issues)))
  }

  sequence_missing <- .sequence_missing_value(sequence_values)
  state_missing <- .sequence_missing_value(state_values)

  sequence_text <- as.character(sequence_values)
  state_text <- as.character(state_values)

  for (row in which(sequence_missing)) {
    add_issue(
      row = row,
      column = sequence_id_col,
      issue_code = "missing_sequence_id",
      severity = "error",
      value = .sequence_value_text(sequence_values[row]),
      message = "The row has no usable sequence identifier.",
      action = "Supply a sequence identifier or remove the row explicitly."
    )
  }

  for (row in which(state_missing)) {
    add_issue(
      sequence_id = sequence_text[row],
      row = row,
      column = state_col,
      issue_code = "missing_state",
      severity = "error",
      value = .sequence_value_text(state_values[row]),
      message = "The row has no usable state value.",
      action = "Supply a state or apply an explicit missing-state policy."
    )
  }

  order_missing <- rep(TRUE, nrow(data))
  order_finite <- rep(FALSE, nrow(data))

  if (valid_order_type) {
    order_missing <- is.na(order_values)
    order_finite <- is.finite(order_values)

    for (row in which(order_missing)) {
      add_issue(
        sequence_id = sequence_text[row],
        row = row,
        column = order_col,
        issue_code = "missing_sequence_order",
        severity = "error",
        value = .sequence_value_text(order_values[row]),
        message = "The row has no sequence-order value.",
        action = "Supply the order explicitly; order is never inferred."
      )
    }

    for (row in which(!order_missing & !order_finite)) {
      add_issue(
        sequence_id = sequence_text[row],
        row = row,
        column = order_col,
        issue_code = "non_finite_order",
        severity = "error",
        value = .sequence_value_text(order_values[row]),
        message = "The sequence-order value is not finite.",
        action = "Replace infinite or non-finite order values."
      )
    }
  }

  valid_sequence_rows <- which(!sequence_missing)

  if (valid_order_type) {
    valid_order_rows <- which(
      !sequence_missing &
        !order_missing &
        order_finite
    )

    if (length(valid_order_rows) > 0L) {
      order_key <- format(
        order_values[valid_order_rows],
        digits = 17L,
        scientific = FALSE,
        trim = TRUE
      )

      keys <- paste(
        sequence_text[valid_order_rows],
        order_key,
        sep = "\034"
      )

      duplicate_local <- duplicated(keys) |
        duplicated(keys, fromLast = TRUE)

      duplicate_rows <- valid_order_rows[duplicate_local]

      for (row in duplicate_rows) {
        add_issue(
          sequence_id = sequence_text[row],
          row = row,
          column = order_col,
          issue_code = "duplicated_position",
          severity = "error",
          value = .sequence_value_text(order_values[row]),
          message = "The sequence contains a duplicated position.",
          action = paste0(
            "Correct the position or apply an explicit first/last policy."
          )
        )
      }

      groups <- split(
        valid_order_rows,
        sequence_text[valid_order_rows],
        drop = TRUE
      )

      for (sequence_id in names(groups)) {
        rows <- groups[[sequence_id]]
        current_order <- order_values[rows]

        if (is.unsorted(current_order, strictly = FALSE)) {
          add_issue(
            sequence_id = sequence_id,
            row = rows[1L],
            column = order_col,
            issue_code = "unordered_rows",
            severity = "review",
            value = .sequence_value_text(current_order),
            message = "Rows are not ordered within the sequence.",
            action = "Sort deterministically by sequence and order."
          )
        }

        unique_order <- sort(unique(current_order))

        integer_like <- all(
          abs(unique_order - round(unique_order)) <
            sqrt(.Machine$double.eps)
        )

        if (integer_like && length(unique_order) > 1L) {
          gap_index <- which(diff(unique_order) > 1)

          if (length(gap_index) > 0L) {
            gap_text <- paste(
              paste0(
                unique_order[gap_index],
                "->",
                unique_order[gap_index + 1L]
              ),
              collapse = ", "
            )

            add_issue(
              sequence_id = sequence_id,
              row = rows[1L],
              column = order_col,
              issue_code = "missing_positions",
              severity = "review",
              value = gap_text,
              message = "The sequence has gaps between integer positions.",
              action = "Confirm that omitted positions are intentional."
            )
          }
        }

        ordered_rows <- rows[
          order(order_values[rows], rows, method = "radix")
        ]

        ordered_states <- state_text[ordered_rows]
        ordered_missing <- state_missing[ordered_rows]

        repeated <- rep(FALSE, length(ordered_rows))

        if (length(ordered_rows) > 1L) {
          repeated[-1L] <- !ordered_missing[-1L] &
            !ordered_missing[-length(ordered_missing)] &
            ordered_states[-1L] ==
            ordered_states[-length(ordered_states)]
        }

        for (row in ordered_rows[repeated]) {
          add_issue(
            sequence_id = sequence_text[row],
            row = row,
            column = state_col,
            issue_code = "consecutive_repeated_state",
            severity = "review",
            value = .sequence_value_text(state_values[row]),
            message = "The state repeats consecutively within the sequence.",
            action = "Preserve or collapse repeats through an explicit policy."
          )
        }
      }
    }
  }

  if (length(valid_sequence_rows) > 0L) {
    sequence_groups <- split(
      valid_sequence_rows,
      sequence_text[valid_sequence_rows],
      drop = TRUE
    )

    for (sequence_id in names(sequence_groups)) {
      rows <- sequence_groups[[sequence_id]]
      usable_state_rows <- rows[!state_missing[rows]]

      if (length(usable_state_rows) == 1L) {
        add_issue(
          sequence_id = sequence_id,
          row = usable_state_rows,
          column = state_col,
          issue_code = "single_state_sequence",
          severity = "review",
          value = .sequence_value_text(
            state_values[usable_state_rows]
          ),
          message = "The sequence contains only one usable state row.",
          action = "Confirm that single-state sequences are admissible."
        )
      }
    }
  }

  if (!is.null(duration_col) && duration_col %in% names(data)) {
    duration_values <- data[[duration_col]]

    if (!is.numeric(duration_values)) {
      add_issue(
        column = duration_col,
        issue_code = "invalid_duration_type",
        severity = "error",
        message = "Duration must be numeric.",
        action = "Convert duration to a numeric scale."
      )
    } else {
      duration_missing <- is.na(duration_values)

      for (row in which(duration_missing)) {
        add_issue(
          sequence_id = sequence_text[row],
          row = row,
          column = duration_col,
          issue_code = "missing_duration",
          severity = "review",
          value = .sequence_value_text(duration_values[row]),
          message = "The row has no duration value.",
          action = "Confirm whether missing duration is admissible."
        )
      }

      for (row in which(
        !duration_missing & !is.finite(duration_values)
      )) {
        add_issue(
          sequence_id = sequence_text[row],
          row = row,
          column = duration_col,
          issue_code = "non_finite_duration",
          severity = "error",
          value = .sequence_value_text(duration_values[row]),
          message = "The duration value is not finite.",
          action = "Replace infinite or non-finite durations."
        )
      }

      for (row in which(
        !duration_missing &
        is.finite(duration_values) &
        duration_values < 0
      )) {
        add_issue(
          sequence_id = sequence_text[row],
          row = row,
          column = duration_col,
          issue_code = "negative_duration",
          severity = "error",
          value = .sequence_value_text(duration_values[row]),
          message = "The duration value is negative.",
          action = "Correct or explicitly exclude the row."
        )
      }

      for (row in which(
        !duration_missing &
        is.finite(duration_values) &
        duration_values == 0
      )) {
        add_issue(
          sequence_id = sequence_text[row],
          row = row,
          column = duration_col,
          issue_code = "zero_duration",
          severity = "review",
          value = "0",
          message = "The duration value is zero.",
          action = "Preserve, drop, or reject zero duration explicitly."
        )
      }
    }
  }

  available_metadata <- intersect(metadata_cols, names(data))

  for (column in available_metadata) {
    metadata_values <- data[[column]]

    if (is.list(metadata_values)) {
      add_issue(
        column = column,
        issue_code = "invalid_metadata_type",
        severity = "error",
        message = "Metadata columns must use atomic vectors.",
        action = "Convert list-columns before sequence preparation."
      )

      next
    }

    if (length(valid_sequence_rows) == 0L) {
      next
    }

    groups <- split(
      valid_sequence_rows,
      sequence_text[valid_sequence_rows],
      drop = TRUE
    )

    for (sequence_id in names(groups)) {
      rows <- groups[[sequence_id]]
      values <- metadata_values[rows]
      values <- unique(as.character(values[!is.na(values)]))

      if (length(values) > 1L) {
        add_issue(
          sequence_id = sequence_id,
          row = rows[1L],
          column = column,
          issue_code = "inconsistent_metadata",
          severity = "error",
          value = .sequence_value_text(values),
          message = paste0(
            "Metadata column `",
            column,
            "` varies within the sequence."
          ),
          action = "Correct the metadata or redefine the sequence identifier."
        )
      }
    }
  }

  if (!is.null(expected_states)) {
    expected_text <- unique(as.character(expected_states))

    unknown_rows <- which(
      !state_missing &
        !(state_text %in% expected_text)
    )

    for (row in unknown_rows) {
      add_issue(
        sequence_id = sequence_text[row],
        row = row,
        column = state_col,
        issue_code = "unknown_state",
        severity = "review",
        value = .sequence_value_text(state_values[row]),
        message = "The state is absent from `expected_states`.",
        action = "Preserve, drop, or reject unknown states explicitly."
      )
    }
  }

  if (is.factor(state_values)) {
    observed_levels <- unique(
      as.character(state_values[!state_missing])
    )

    unused_levels <- setdiff(
      levels(state_values),
      observed_levels
    )

    if (length(unused_levels) > 0L) {
      add_issue(
        column = state_col,
        issue_code = "unused_state_levels",
        severity = "info",
        value = .sequence_value_text(unused_levels),
        message = "The factor contains unused state levels.",
        action = "Preserve or drop unused levels explicitly."
      )
    }
  }

  if (length(issues) == 0L) {
    return(.sequence_empty_audit())
  }

  .sequence_sort_audit(do.call(rbind, issues))
}

#' Validate Long-Format Sequence Data
#'
#' Produces a compact validation result based on
#' `audit_sequence_data()` without modifying the input.
#'
#' @inheritParams audit_sequence_data
#'
#' @return A named list containing `valid`, `status`, issue counts, the
#'   complete `audit` table, the column `mapping`, row and sequence counts,
#'   and observed `state_levels`. A result is valid when no error-severity
#'   issue is present. Review-severity issues do not automatically invalidate
#'   the input.
#'
#' @examples
#' sequences <- data.frame(
#'   id = rep(c("s1", "s2"), each = 2L),
#'   position = rep(1:2, times = 2L),
#'   state = c("home", "search", "home", "product")
#' )
#'
#' validation <- validate_sequence_data(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' validation$status
#' validation$valid
#'
#' @export
validate_sequence_data <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL,
    expected_states = NULL
) {
  audit <- audit_sequence_data(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    duration_col = duration_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  issue_levels <- c("error", "review", "info")

  counts <- vapply(
    issue_levels,
    function(level) {
      as.integer(sum(audit$severity == level))
    },
    integer(1)
  )

  sequence_count <- 0L

  if (
    sequence_id_col %in% names(data) &&
    !is.list(data[[sequence_id_col]])
  ) {
    identifiers <- data[[sequence_id_col]]
    usable <- !.sequence_missing_value(identifiers)

    sequence_count <- length(
      unique(as.character(identifiers[usable]))
    )
  }

  state_levels <- character()

  if (
    state_col %in% names(data) &&
    !is.list(data[[state_col]])
  ) {
    states <- data[[state_col]]
    usable <- !.sequence_missing_value(states)

    state_levels <- if (is.factor(states)) {
      levels(states)
    } else {
      unique(as.character(states[usable]))
    }
  }

  status <- .sequence_status(audit)

  list(
    valid = identical(status, "pass") ||
      identical(status, "review"),
    status = status,
    n_errors = unname(counts[["error"]]),
    n_reviews = unname(counts[["review"]]),
    n_info = unname(counts[["info"]]),
    audit = audit,
    mapping = .sequence_mapping(
      sequence_id_col = sequence_id_col,
      order_col = order_col,
      state_col = state_col,
      duration_col = duration_col,
      metadata_cols = metadata_cols
    ),
    n_rows = as.integer(nrow(data)),
    n_sequences = as.integer(sequence_count),
    state_levels = as.character(state_levels)
  )
}

#' Prepare Long-Format Sequence Data
#'
#' Applies explicit preprocessing policies and returns a deterministic,
#' canonical long-format representation.
#'
#' @inheritParams audit_sequence_data
#' @param missing_state_policy Policy for missing states: `"error"` or
#'   `"drop"`.
#' @param duplicate_position_policy Policy for duplicated sequence positions:
#'   `"error"`, `"first"`, or `"last"`.
#' @param repeated_state_policy Policy for consecutive repeated states:
#'   `"preserve"` or `"collapse"`.
#' @param zero_duration_policy Policy for zero durations: `"preserve"`,
#'   `"drop"`, or `"error"`.
#' @param unknown_state_policy Policy for states absent from
#'   `expected_states`: `"preserve"`, `"drop"`, or `"error"`.
#' @param unused_state_levels Policy for unused factor levels: `"preserve"`
#'   or `"drop"`.
#'
#' @return A named list containing:
#'
#' * `data`: canonical prepared data, or `NULL` when unresolved errors remain;
#' * `audit`: input- and output-stage diagnostics;
#' * `decisions`: a machine-readable preprocessing decision log;
#' * `mapping`: source-to-contract column mappings;
#' * `status`: `pass`, `review`, or `fail`;
#' * row counts and final state levels.
#'
#' The canonical columns are `sequence_id`, `sequence_order`, `state`,
#' `original_row`, and optional `duration`. Unmapped columns are preserved.
#'
#' @details
#' Rows are sorted deterministically by sequence identifier, sequence order,
#' and original row number. When consecutive repeats are collapsed, the first
#' row supplies non-duration values and available durations are summed.
#'
#' Unresolved errors produce `status = "fail"` and `data = NULL`; diagnostics
#' and decision records remain available.
#'
#' @examples
#' sequences <- data.frame(
#'   id = c("s2", "s1", "s1", "s2"),
#'   position = c(2, 2, 1, 1),
#'   state = c("product", "search", "home", "home")
#' )
#'
#' prepared <- prepare_sequence_data(
#'   sequences,
#'   sequence_id_col = "id",
#'   order_col = "position",
#'   state_col = "state"
#' )
#'
#' prepared$status
#' prepared$data
#'
#' @export
prepare_sequence_data <- function(
    data,
    sequence_id_col,
    order_col,
    state_col,
    duration_col = NULL,
    metadata_cols = NULL,
    expected_states = NULL,
    missing_state_policy = c("error", "drop"),
    duplicate_position_policy = c("error", "first", "last"),
    repeated_state_policy = c("preserve", "collapse"),
    zero_duration_policy = c("preserve", "drop", "error"),
    unknown_state_policy = c("preserve", "drop", "error"),
    unused_state_levels = c("preserve", "drop")
) {
  missing_state_policy <- match.arg(missing_state_policy)
  duplicate_position_policy <- match.arg(
    duplicate_position_policy
  )
  repeated_state_policy <- match.arg(repeated_state_policy)
  zero_duration_policy <- match.arg(zero_duration_policy)
  unknown_state_policy <- match.arg(unknown_state_policy)
  unused_state_levels <- match.arg(unused_state_levels)

  initial_validation <- validate_sequence_data(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    duration_col = duration_col,
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  mapping <- initial_validation$mapping
  input_audit <- .sequence_stage_audit(
    initial_validation$audit,
    "input"
  )

  fatal_codes <- c(
    "duplicate_column_names",
    "empty_data",
    "missing_required_column",
    "missing_metadata_column",
    "invalid_sequence_id_type",
    "invalid_state_type",
    "invalid_order_type",
    "missing_sequence_id",
    "missing_sequence_order",
    "non_finite_order",
    "invalid_duration_type",
    "negative_duration",
    "non_finite_duration",
    "invalid_metadata_type",
    "inconsistent_metadata"
  )

  fatal_input <- initial_validation$audit$issue_code %in%
    fatal_codes

  if (any(fatal_input)) {
    return(
      .sequence_prepare_result(
        data = NULL,
        audit = input_audit,
        decisions = .sequence_empty_decisions(),
        mapping = mapping,
        status = "fail",
        original_n_rows = nrow(data),
        prepared_n_rows = 0L,
        state_levels = character()
      )
    )
  }

  working <- data
  working$.gp3_original_row <- seq_len(nrow(working))

  decision_rows <- list()

  add_decision <- function(
    step,
    policy,
    affected_rows,
    details
  ) {
    decision_rows[[length(decision_rows) + 1L]] <<- data.frame(
      step = step,
      policy = policy,
      affected_rows = as.integer(affected_rows),
      details = details,
      stringsAsFactors = FALSE
    )

    invisible(NULL)
  }

  state_missing <- .sequence_missing_value(
    working[[state_col]]
  )

  if (missing_state_policy == "drop") {
    affected <- sum(state_missing)
    working <- working[!state_missing, , drop = FALSE]

    add_decision(
      step = "missing_states",
      policy = "drop",
      affected_rows = affected,
      details = "Rows with missing or blank states were removed."
    )
  } else {
    add_decision(
      step = "missing_states",
      policy = "error",
      affected_rows = sum(state_missing),
      details = "Missing states were retained as unresolved errors."
    )
  }

  unknown_rows <- rep(FALSE, nrow(working))

  if (!is.null(expected_states) && nrow(working) > 0L) {
    expected_text <- unique(as.character(expected_states))
    state_text <- as.character(working[[state_col]])

    unknown_rows <- !.sequence_missing_value(
      working[[state_col]]
    ) & !(state_text %in% expected_text)
  }

  if (unknown_state_policy == "drop") {
    affected <- sum(unknown_rows)
    working <- working[!unknown_rows, , drop = FALSE]

    add_decision(
      step = "unknown_states",
      policy = "drop",
      affected_rows = affected,
      details = "States absent from `expected_states` were removed."
    )
  } else {
    add_decision(
      step = "unknown_states",
      policy = unknown_state_policy,
      affected_rows = sum(unknown_rows),
      details = paste0(
        "Unknown states were ",
        if (unknown_state_policy == "error") {
          "retained as unresolved errors."
        } else {
          "preserved for review."
        }
      )
    )
  }

  zero_duration <- rep(FALSE, nrow(working))

  if (!is.null(duration_col) && nrow(working) > 0L) {
    duration <- working[[duration_col]]
    zero_duration <- !is.na(duration) &
      is.finite(duration) &
      duration == 0
  }

  if (zero_duration_policy == "drop") {
    affected <- sum(zero_duration)
    working <- working[!zero_duration, , drop = FALSE]

    add_decision(
      step = "zero_durations",
      policy = "drop",
      affected_rows = affected,
      details = "Rows with zero duration were removed."
    )
  } else {
    add_decision(
      step = "zero_durations",
      policy = zero_duration_policy,
      affected_rows = sum(zero_duration),
      details = paste0(
        "Zero-duration rows were ",
        if (zero_duration_policy == "error") {
          "retained as unresolved errors."
        } else {
          "preserved."
        }
      )
    )
  }

  duplicate_rows <- rep(FALSE, nrow(working))

  if (nrow(working) > 0L) {
    sequence_text <- as.character(
      working[[sequence_id_col]]
    )

    order_key <- format(
      working[[order_col]],
      digits = 17L,
      scientific = FALSE,
      trim = TRUE
    )

    keys <- paste(sequence_text, order_key, sep = "\034")

    duplicate_rows <- if (
      duplicate_position_policy == "last"
    ) {
      duplicated(keys, fromLast = TRUE)
    } else {
      duplicated(keys)
    }
  }

  if (duplicate_position_policy %in% c("first", "last")) {
    affected <- sum(duplicate_rows)
    working <- working[!duplicate_rows, , drop = FALSE]

    add_decision(
      step = "duplicated_positions",
      policy = duplicate_position_policy,
      affected_rows = affected,
      details = paste0(
        "Duplicated positions were resolved by retaining the ",
        duplicate_position_policy,
        " occurrence."
      )
    )
  } else {
    duplicate_count <- if (nrow(working) == 0L) {
      0L
    } else {
      sequence_text <- as.character(
        working[[sequence_id_col]]
      )

      order_key <- format(
        working[[order_col]],
        digits = 17L,
        scientific = FALSE,
        trim = TRUE
      )

      keys <- paste(sequence_text, order_key, sep = "\034")

      sum(
        duplicated(keys) |
          duplicated(keys, fromLast = TRUE)
      )
    }

    add_decision(
      step = "duplicated_positions",
      policy = "error",
      affected_rows = duplicate_count,
      details = "Duplicated positions were retained as unresolved errors."
    )
  }

  if (nrow(working) > 0L) {
    sort_index <- order(
      working[[sequence_id_col]],
      working[[order_col]],
      working$.gp3_original_row,
      na.last = TRUE,
      method = "radix"
    )

    moved_rows <- sum(
      sort_index != seq_len(nrow(working))
    )

    working <- working[sort_index, , drop = FALSE]
    row.names(working) <- NULL
  } else {
    moved_rows <- 0L
  }

  add_decision(
    step = "row_order",
    policy = "deterministic_sort",
    affected_rows = moved_rows,
    details = paste0(
      "Rows were ordered by sequence identifier, sequence order, ",
      "and original row."
    )
  )

  collapsed_rows <- 0L

  if (
    repeated_state_policy == "collapse" &&
    nrow(working) > 1L
  ) {
    sequence_text <- as.character(
      working[[sequence_id_col]]
    )
    state_text <- as.character(working[[state_col]])

    same_previous <- rep(FALSE, nrow(working))

    same_previous[-1L] <-
      sequence_text[-1L] ==
      sequence_text[-nrow(working)] &
      !.sequence_missing_value(
        working[[state_col]][-1L]
      ) &
      !.sequence_missing_value(
        working[[state_col]][-nrow(working)]
      ) &
      state_text[-1L] ==
      state_text[-nrow(working)]

    group_id <- cumsum(!same_previous)
    keep_rows <- !duplicated(group_id)

    if (!is.null(duration_col)) {
      repeated_groups <- unique(group_id[duplicated(group_id)])

      for (group in repeated_groups) {
        rows <- which(group_id == group)
        duration_values <- working[[duration_col]][rows]

        combined_duration <- if (all(is.na(duration_values))) {
          NA_real_
        } else {
          sum(duration_values, na.rm = TRUE)
        }

        working[[duration_col]][rows[1L]] <- combined_duration
      }
    }

    collapsed_rows <- sum(!keep_rows)
    working <- working[keep_rows, , drop = FALSE]
    row.names(working) <- NULL
  }

  add_decision(
    step = "consecutive_repeats",
    policy = repeated_state_policy,
    affected_rows = collapsed_rows,
    details = if (repeated_state_policy == "collapse") {
      paste0(
        "The first row was retained for each repeated run; ",
        "available durations were summed."
      )
    } else {
      "Consecutive repeated states were preserved."
    }
  )

  if (
    unused_state_levels == "drop" &&
    is.factor(working[[state_col]])
  ) {
    before_levels <- levels(working[[state_col]])
    working[[state_col]] <- droplevels(
      working[[state_col]]
    )
    removed_levels <- length(
      setdiff(
        before_levels,
        levels(working[[state_col]])
      )
    )
  } else {
    removed_levels <- 0L
  }

  add_decision(
    step = "unused_state_levels",
    policy = unused_state_levels,
    affected_rows = removed_levels,
    details = if (unused_state_levels == "drop") {
      "Unused factor levels were removed."
    } else {
      "Unused factor levels were preserved."
    }
  )

  source_mapped <- c(
    sequence_id_col,
    order_col,
    state_col,
    duration_col
  )
  source_mapped <- source_mapped[!is.na(source_mapped)]

  extra_columns <- setdiff(
    names(working),
    c(source_mapped, ".gp3_original_row")
  )

  canonical_names <- c(
    "sequence_id",
    "sequence_order",
    "state",
    "original_row"
  )

  if (!is.null(duration_col)) {
    canonical_names <- c(canonical_names, "duration")
  }

  collisions <- intersect(extra_columns, canonical_names)

  if (length(collisions) > 0L) {
    stop(
      "Unmapped source columns use reserved canonical names: ",
      paste(collisions, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  canonical <- data.frame(
    sequence_id = working[[sequence_id_col]],
    sequence_order = working[[order_col]],
    state = working[[state_col]],
    original_row = as.integer(
      working$.gp3_original_row
    ),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  if (!is.null(duration_col)) {
    canonical$duration <- working[[duration_col]]
  }

  for (column in extra_columns) {
    canonical[[column]] <- working[[column]]
  }

  row.names(canonical) <- NULL

  add_decision(
    step = "column_mapping",
    policy = "canonicalise",
    affected_rows = nrow(canonical),
    details = paste0(
      "Mapped columns were standardised while unmapped columns ",
      "were preserved."
    )
  )

  output_validation <- validate_sequence_data(
    data = canonical,
    sequence_id_col = "sequence_id",
    order_col = "sequence_order",
    state_col = "state",
    duration_col = if (!is.null(duration_col)) {
      "duration"
    } else {
      NULL
    },
    metadata_cols = metadata_cols,
    expected_states = expected_states
  )

  final_audit <- output_validation$audit
  policy_issues <- list()

  if (
    zero_duration_policy == "error" &&
    "duration" %in% names(canonical)
  ) {
    rows <- which(
      !is.na(canonical$duration) &
        is.finite(canonical$duration) &
        canonical$duration == 0
    )

    for (row in rows) {
      policy_issues[[length(policy_issues) + 1L]] <-
        .sequence_issue_row(
          sequence_id = canonical$sequence_id[row],
          row = canonical$original_row[row],
          column = "duration",
          issue_code = "zero_duration_disallowed",
          severity = "error",
          value = "0",
          message = "Zero duration is disallowed by the selected policy.",
          action = "Use `preserve`, `drop`, or correct the duration."
        )
    }
  }

  if (
    unknown_state_policy == "error" &&
    !is.null(expected_states)
  ) {
    expected_text <- unique(as.character(expected_states))
    state_text <- as.character(canonical$state)

    rows <- which(
      !.sequence_missing_value(canonical$state) &
        !(state_text %in% expected_text)
    )

    for (row in rows) {
      policy_issues[[length(policy_issues) + 1L]] <-
        .sequence_issue_row(
          sequence_id = canonical$sequence_id[row],
          row = canonical$original_row[row],
          column = "state",
          issue_code = "unknown_state_disallowed",
          severity = "error",
          value = .sequence_value_text(canonical$state[row]),
          message = "The state is disallowed by the selected policy.",
          action = "Use `preserve`, `drop`, or revise expected states."
        )
    }
  }

  if (length(policy_issues) > 0L) {
    final_audit <- rbind(
      final_audit,
      do.call(rbind, policy_issues)
    )

    final_audit <- .sequence_sort_audit(final_audit)
  }

  output_audit <- .sequence_stage_audit(
    final_audit,
    "output"
  )

  combined_audit <- rbind(input_audit, output_audit)
  row.names(combined_audit) <- NULL

  decisions <- if (length(decision_rows) == 0L) {
    .sequence_empty_decisions()
  } else {
    do.call(rbind, decision_rows)
  }

  row.names(decisions) <- NULL

  final_status <- .sequence_status(final_audit)

  state_levels <- if (is.factor(canonical$state)) {
    levels(canonical$state)
  } else {
    unique(
      as.character(
        canonical$state[
          !.sequence_missing_value(canonical$state)
        ]
      )
    )
  }

  .sequence_prepare_result(
    data = if (final_status == "fail") {
      NULL
    } else {
      canonical
    },
    audit = combined_audit,
    decisions = decisions,
    mapping = mapping,
    status = final_status,
    original_n_rows = nrow(data),
    prepared_n_rows = nrow(canonical),
    state_levels = state_levels
  )
}

make_valid_sequence_data <- function() {
  data.frame(
    id = rep(c("s1", "s2"), each = 3L),
    position = rep(1:3, times = 2L),
    state = c(
      "home",
      "search",
      "product",
      "home",
      "category",
      "product"
    ),
    duration_ms = c(100, 200, 300, 120, 180, 260),
    participant = rep(c("p1", "p2"), each = 3L),
    stringsAsFactors = FALSE
  )
}

test_that("audit output has a stable empty contract", {
  data <- make_valid_sequence_data()

  audit <- audit_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms",
    metadata_cols = "participant"
  )

  expect_s3_class(audit, "data.frame")
  expect_identical(
    names(audit),
    c(
      "sequence_id",
      "row",
      "column",
      "issue_code",
      "severity",
      "value",
      "message",
      "action"
    )
  )
  expect_equal(nrow(audit), 0L)
})

test_that("validation reports normal and empty inputs", {
  data <- make_valid_sequence_data()

  result <- validate_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_true(result$valid)
  expect_identical(result$status, "pass")
  expect_identical(result$n_sequences, 2L)
  expect_identical(result$n_rows, 6L)

  empty <- data[0, , drop = FALSE]

  empty_result <- validate_sequence_data(
    empty,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_false(empty_result$valid)
  expect_identical(empty_result$status, "fail")
  expect_true(
    "empty_data" %in% empty_result$audit$issue_code
  )
})

test_that("missing columns and states are explicit errors", {
  data <- make_valid_sequence_data()

  missing_column <- validate_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "missing_position",
    state_col = "state"
  )

  expect_false(missing_column$valid)
  expect_true(
    "missing_required_column" %in%
      missing_column$audit$issue_code
  )

  data$state[2L] <- NA_character_

  missing_state <- validate_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_false(missing_state$valid)
  expect_true(
    "missing_state" %in% missing_state$audit$issue_code
  )
})

test_that("ordering, gaps, and duplicated positions are audited", {
  data <- make_valid_sequence_data()
  data <- data[c(2, 1, 3, 4, 5, 6), , drop = FALSE]

  audit <- audit_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_true("unordered_rows" %in% audit$issue_code)

  gap_data <- make_valid_sequence_data()
  gap_data$position[3L] <- 4

  gap_audit <- audit_sequence_data(
    gap_data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_true("missing_positions" %in% gap_audit$issue_code)

  duplicate_data <- rbind(
    make_valid_sequence_data(),
    make_valid_sequence_data()[2L, , drop = FALSE]
  )

  duplicate_audit <- audit_sequence_data(
    duplicate_data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_true(
    "duplicated_position" %in%
      duplicate_audit$issue_code
  )
})

test_that("duration and metadata errors are detected", {
  data <- make_valid_sequence_data()
  data$duration_ms[2L] <- -1

  duration_audit <- audit_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms"
  )

  expect_true(
    "negative_duration" %in%
      duration_audit$issue_code
  )

  data <- make_valid_sequence_data()
  data$participant[2L] <- "different"

  metadata_audit <- audit_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    metadata_cols = "participant"
  )

  expect_true(
    "inconsistent_metadata" %in%
      metadata_audit$issue_code
  )
})

test_that("preparation sorts deterministically and preserves identifiers", {
  data <- make_valid_sequence_data()[c(6, 2, 1, 5, 3, 4), ]

  prepared <- prepare_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms",
    metadata_cols = "participant"
  )

  expect_identical(prepared$status, "pass")
  expect_s3_class(prepared$data, "data.frame")

  expect_identical(
    names(prepared$data)[1:5],
    c(
      "sequence_id",
      "sequence_order",
      "state",
      "original_row",
      "duration"
    )
  )

  expect_identical(
    as.character(prepared$data$sequence_id),
    rep(c("s1", "s2"), each = 3L)
  )

  expect_identical(
    prepared$data$sequence_order,
    rep(1:3, times = 2L)
  )

  expect_identical(
    prepared$data$original_row,
    c(3L, 2L, 5L, 6L, 4L, 1L)
  )

  expect_true("participant" %in% names(prepared$data))
})

test_that("explicit policies drop missing and unknown states", {
  data <- make_valid_sequence_data()
  data$state[2L] <- NA_character_
  data$state[5L] <- "unknown"

  prepared <- prepare_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    expected_states = c(
      "home",
      "search",
      "category",
      "product"
    ),
    missing_state_policy = "drop",
    unknown_state_policy = "drop"
  )

  expect_identical(prepared$status, "review")
  expect_false(anyNA(prepared$data$state))
  expect_false("unknown" %in% prepared$data$state)
  expect_identical(prepared$prepared_n_rows, 4L)
})

test_that("duplicate and repeated-state policies are deterministic", {
  data <- data.frame(
    id = rep("s1", 5L),
    position = c(1, 2, 2, 3, 4),
    state = c("A", "B", "C", "C", "D"),
    duration_ms = c(1, 2, 3, 4, 5),
    stringsAsFactors = FALSE
  )

  prepared <- prepare_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms",
    duplicate_position_policy = "last",
    repeated_state_policy = "collapse"
  )

  # Position 3 is removed during the explicit repeated-state collapse.
  # The preserved source positions are therefore 1, 2, and 4, which
  # correctly produces a review-level missing-position diagnostic.
  expect_identical(prepared$status, "review")
  expect_s3_class(prepared$data, "data.frame")

  expect_identical(
    as.character(prepared$data$state),
    c("A", "C", "D")
  )
  expect_identical(
    prepared$data$sequence_order,
    c(1, 2, 4)
  )
  expect_identical(
    prepared$data$duration,
    c(1, 7, 5)
  )
  expect_identical(
    prepared$data$original_row,
    c(1L, 3L, 5L)
  )

  expect_true(
    any(
      prepared$audit$stage == "output" &
        prepared$audit$issue_code == "missing_positions" &
        prepared$audit$severity == "review"
    )
  )

  duplicate_decision <- prepared$decisions[
    prepared$decisions$step == "duplicated_positions",
    ,
    drop = FALSE
  ]

  repeat_decision <- prepared$decisions[
    prepared$decisions$step == "consecutive_repeats",
    ,
    drop = FALSE
  ]

  expect_identical(duplicate_decision$policy, "last")
  expect_identical(duplicate_decision$affected_rows, 1L)
  expect_identical(repeat_decision$policy, "collapse")
  expect_identical(repeat_decision$affected_rows, 1L)
})


test_that("unresolved policy errors suppress prepared data", {
  data <- make_valid_sequence_data()
  data$duration_ms[2L] <- 0

  zero_error <- prepare_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms",
    zero_duration_policy = "error"
  )

  expect_identical(zero_error$status, "fail")
  expect_null(zero_error$data)
  expect_true(
    "zero_duration_disallowed" %in%
      zero_error$audit$issue_code
  )

  data <- make_valid_sequence_data()
  data$duration_ms[2L] <- -1

  negative <- prepare_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state",
    duration_col = "duration_ms"
  )

  expect_identical(negative$status, "fail")
  expect_null(negative$data)
  expect_true(
    "negative_duration" %in% negative$audit$issue_code
  )
})

test_that("single-state and unused levels remain reviewable", {
  data <- data.frame(
    id = "s1",
    position = 1,
    state = factor("A", levels = c("A", "B"))
  )

  validation <- validate_sequence_data(
    data,
    sequence_id_col = "id",
    order_col = "position",
    state_col = "state"
  )

  expect_true(validation$valid)
  expect_identical(validation$status, "review")
  expect_true(
    "single_state_sequence" %in%
      validation$audit$issue_code
  )
  expect_true(
    "unused_state_levels" %in%
      validation$audit$issue_code
  )
})

test_that("torture corpus exposes expected validation failures", {
  empty <- .sequence_test_case(
    "empty"
  )
  duplicate <- .sequence_test_case(
    "duplicate_positions"
  )
  negative <- .sequence_test_case(
    "negative_duration"
  )

  empty_check <- validate_sequence_data(
    empty,
    "sequence_id",
    "sequence_order",
    "state",
    duration_col = "duration"
  )
  duplicate_check <- validate_sequence_data(
    duplicate,
    "sequence_id",
    "sequence_order",
    "state",
    duration_col = "duration"
  )
  negative_check <- validate_sequence_data(
    negative,
    "sequence_id",
    "sequence_order",
    "state",
    duration_col = "duration"
  )

  expect_identical(
    empty_check$status,
    "fail"
  )
  expect_identical(
    duplicate_check$status,
    "fail"
  )
  expect_identical(
    negative_check$status,
    "fail"
  )
})

test_that("torture corpus exposes review rather than silent destruction", {
  gaps <- .sequence_test_case(
    "order_gaps"
  )
  zeros <- .sequence_test_case(
    "zero_duration"
  )
  repeats <- .sequence_test_case(
    "high_repetition"
  )

  gap_check <- validate_sequence_data(
    gaps,
    "sequence_id",
    "sequence_order",
    "state",
    duration_col = "duration"
  )
  zero_check <- validate_sequence_data(
    zeros,
    "sequence_id",
    "sequence_order",
    "state",
    duration_col = "duration"
  )
  repeat_check <- validate_sequence_data(
    repeats,
    "sequence_id",
    "sequence_order",
    "state",
    duration_col = "duration"
  )

  expect_true(
    gap_check$status %in% c(
      "review",
      "fail"
    )
  )
  expect_true(
    zero_check$status %in% c(
      "review",
      "fail"
    )
  )
  expect_true(
    repeat_check$status %in% c(
      "review",
      "fail"
    )
  )
})

#' Extract bounded non-contiguous sequence subsequences
#'
#' Enumerates ordered non-contiguous subsequences under explicit length, gap,
#' span, and safety limits. The implementation is transparent and intended for
#' modest sequence collections; specialist mining packages remain preferable
#' for very large search spaces.
#'
#' @param data Long-format sequence data or a prepared result.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param metadata_cols Optional sequence-constant metadata retained as
#'   attributes.
#' @param min_length,max_length Minimum and maximum subsequence lengths.
#' @param max_gap Maximum number of skipped positions between adjacent selected
#'   states. Use `Inf` for no restriction.
#' @param max_span Maximum difference between the first and last selected
#'   sequence positions. Use `Inf` for no restriction.
#' @param repeated_state_policy Preserve or collapse consecutive repeated
#'   states before mining.
#' @param separator Separator used in stable motif labels.
#' @param max_combinations_per_sequence Safety limit for the number of index
#'   combinations considered for any one sequence.
#'
#' @return A data frame of class `gp3_sequence_subsequences`, with one row per
#'   qualifying occurrence.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2"), each = 5L),
#'   sequence_order = rep(1:5, times = 2L),
#'   state = c("A", "B", "C", "D", "E", "A", "C", "B", "D", "E")
#' )
#' extract_sequence_subsequences(sequences, min_length = 2L, max_length = 3L,
#'                               max_gap = 2L)
#' @export
extract_sequence_subsequences <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = NULL,
  min_length = 2L,
  max_length = 5L,
  max_gap = Inf,
  max_span = Inf,
  repeated_state_policy = c("preserve", "collapse"),
  separator = " > ",
  max_combinations_per_sequence = 100000L
) {
  repeated_state_policy <- match.arg(repeated_state_policy)
  .sequence_adv_scalar_number(min_length, "min_length", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(max_length, "max_length", lower = min_length, integer = TRUE)
  if (!is.numeric(max_gap) || length(max_gap) != 1L || is.na(max_gap) || max_gap < 0) {
    stop("`max_gap` must be one non-negative number or `Inf`.", call. = FALSE)
  }
  if (!is.numeric(max_span) || length(max_span) != 1L || is.na(max_span) || max_span < 0) {
    stop("`max_span` must be one non-negative number or `Inf`.", call. = FALSE)
  }
  .sequence_adv_scalar_character(separator, "separator")
  .sequence_adv_scalar_number(max_combinations_per_sequence,
                              "max_combinations_per_sequence", lower = 1, integer = TRUE)
  x <- .sequence_adv_data(
    data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = metadata_cols,
    missing_state_policy = "error"
  )
  if (any(grepl(separator, x$state_levels, fixed = TRUE))) {
    stop("`separator` must not occur inside an observed state label.", call. = FALSE)
  }
  occurrences <- list()
  h <- 0L
  for (sequence_id in x$sequence_ids) {
    states <- x$sequences[[sequence_id]]
    orders <- x$orders[[sequence_id]]
    if (repeated_state_policy == "collapse" && length(states) > 1L) {
      keep <- c(TRUE, states[-1L] != states[-length(states)])
      states <- states[keep]
      orders <- orders[keep]
    }
    n <- length(states)
    upper <- min(as.integer(max_length), n)
    if (upper < min_length) next
    expected <- 0
    for (k in seq.int(min_length, upper)) {
      log_count <- lchoose(n, k)
      if (is.finite(log_count)) expected <- expected + exp(log_count)
      if (!is.finite(expected) || expected > max_combinations_per_sequence) {
        stop(
          "The subsequence search exceeds `max_combinations_per_sequence` for sequence `",
          sequence_id, "`. Reduce `max_length` or tighten gap/span constraints.",
          call. = FALSE
        )
      }
    }
    for (k in seq.int(min_length, upper)) {
      combinations <- utils::combn(seq_len(n), k)
      if (is.null(dim(combinations))) combinations <- matrix(combinations, nrow = k)
      for (j in seq_len(ncol(combinations))) {
        indices <- combinations[, j]
        skipped <- if (length(indices) > 1L) diff(indices) - 1L else 0L
        span <- orders[indices[length(indices)]] - orders[indices[1L]]
        if (any(skipped > max_gap) || span > max_span) next
        h <- h + 1L
        occurrences[[h]] <- data.frame(
          sequence_id = sequence_id,
          subsequence = paste(states[indices], collapse = separator),
          subsequence_length = as.integer(k),
          start_order = orders[indices[1L]],
          end_order = orders[indices[length(indices)]],
          span = as.numeric(span),
          max_observed_gap = if (length(skipped)) max(skipped) else 0,
          selected_positions = paste(indices, collapse = ","),
          selected_orders = paste(orders[indices], collapse = ","),
          stringsAsFactors = FALSE
        )
      }
    }
  }
  if (length(occurrences) == 0L) {
    result <- data.frame(
      sequence_id = character(), subsequence = character(),
      subsequence_length = integer(), start_order = numeric(), end_order = numeric(),
      span = numeric(), max_observed_gap = numeric(), selected_positions = character(),
      selected_orders = character(), stringsAsFactors = FALSE
    )
  } else {
    result <- do.call(rbind, occurrences)
    result <- result[order(result$subsequence_length, result$subsequence,
                           result$sequence_id, result$start_order,
                           result$selected_positions, method = "radix"), , drop = FALSE]
    row.names(result) <- NULL
  }
  class(result) <- c("gp3_sequence_subsequences", "data.frame")
  attr(result, "sequence_ids") <- x$sequence_ids
  attr(result, "n_sequences") <- length(x$sequence_ids)
  attr(result, "state_levels") <- x$state_levels
  attr(result, "metadata") <- x$metadata
  attr(result, "metadata_cols") <- metadata_cols
  attr(result, "settings") <- list(
    min_length = as.integer(min_length), max_length = as.integer(max_length),
    max_gap = max_gap, max_span = max_span,
    repeated_state_policy = repeated_state_policy, separator = separator,
    max_combinations_per_sequence = as.integer(max_combinations_per_sequence)
  )
  result
}

#' Summarise non-contiguous subsequences
#'
#' @param occurrences A result from [extract_sequence_subsequences()].
#'
#' @return A data frame with occurrence counts, sequence prevalence, and span
#'   diagnostics.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "B", "D"))
#' summarise_sequence_subsequences(extract_sequence_subsequences(data))
#' @export
summarise_sequence_subsequences <- function(occurrences) {
  if (!inherits(occurrences, "gp3_sequence_subsequences")) {
    stop("`occurrences` must be created by `extract_sequence_subsequences()`.",
         call. = FALSE)
  }
  n_sequences <- attr(occurrences, "n_sequences")
  if (nrow(occurrences) == 0L) {
    return(data.frame(
      subsequence = character(), subsequence_length = integer(),
      occurrence_count = integer(), sequence_count = integer(),
      sequence_prevalence = numeric(), mean_span = numeric(), median_span = numeric(),
      mean_max_gap = numeric(), stringsAsFactors = FALSE
    ))
  }
  split_rows <- split(seq_len(nrow(occurrences)), occurrences$subsequence, drop = TRUE)
  rows <- lapply(split_rows, function(index) {
    current <- occurrences[index, , drop = FALSE]
    data.frame(
      subsequence = current$subsequence[1L],
      subsequence_length = current$subsequence_length[1L],
      occurrence_count = nrow(current),
      sequence_count = length(unique(current$sequence_id)),
      sequence_prevalence = length(unique(current$sequence_id)) / n_sequences,
      mean_span = mean(current$span),
      median_span = stats::median(current$span),
      mean_max_gap = mean(current$max_observed_gap),
      stringsAsFactors = FALSE
    )
  })
  result <- do.call(rbind, rows)
  result <- result[order(-result$sequence_prevalence, -result$occurrence_count,
                         result$subsequence_length, result$subsequence,
                         method = "radix"), , drop = FALSE]
  row.names(result) <- NULL
  result
}

#' Filter non-contiguous subsequence summaries
#'
#' @param summary A data frame from [summarise_sequence_subsequences()].
#' @param min_sequences Minimum sequence count.
#' @param min_prevalence Minimum sequence prevalence.
#' @param max_mean_gap Optional maximum mean gap.
#' @param top_n Optional number of rows retained after deterministic sorting.
#' @param ties Include or exclude ties at the `top_n` boundary.
#'
#' @return A filtered summary data frame.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "B", "D"))
#' x <- summarise_sequence_subsequences(extract_sequence_subsequences(data))
#' filter_sequence_subsequences(x, min_sequences = 2L)
#' @export
filter_sequence_subsequences <- function(
  summary,
  min_sequences = 1L,
  min_prevalence = 0,
  max_mean_gap = Inf,
  top_n = NULL,
  ties = c("include", "exclude")
) {
  ties <- match.arg(ties)
  required <- c("subsequence", "subsequence_length", "occurrence_count",
                "sequence_count", "sequence_prevalence", "mean_max_gap")
  if (!is.data.frame(summary) || !all(required %in% names(summary))) {
    stop("`summary` is not a subsequence summary table.", call. = FALSE)
  }
  .sequence_adv_scalar_number(min_sequences, "min_sequences", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(min_prevalence, "min_prevalence", lower = 0, upper = 1)
  if (!is.numeric(max_mean_gap) || length(max_mean_gap) != 1L || is.na(max_mean_gap) ||
      max_mean_gap < 0) {
    stop("`max_mean_gap` must be one non-negative number or `Inf`.", call. = FALSE)
  }
  if (!is.null(top_n)) .sequence_adv_scalar_number(top_n, "top_n", lower = 1, integer = TRUE)
  keep <- summary$sequence_count >= min_sequences &
    summary$sequence_prevalence >= min_prevalence &
    summary$mean_max_gap <= max_mean_gap
  result <- summary[keep, , drop = FALSE]
  result <- result[order(-result$sequence_prevalence, -result$occurrence_count,
                         result$subsequence_length, result$subsequence,
                         method = "radix"), , drop = FALSE]
  if (!is.null(top_n) && nrow(result) > top_n) {
    if (ties == "exclude") {
      result <- result[seq_len(top_n), , drop = FALSE]
    } else {
      boundary <- result$sequence_prevalence[top_n]
      result <- result[result$sequence_prevalence >= boundary, , drop = FALSE]
    }
  }
  row.names(result) <- NULL
  result
}

#' Compare subsequence prevalence between groups
#'
#' Performs transparent contingency-table tests on sequence-level presence.
#' Multiple testing is adjusted explicitly. Results are associational unless a
#' randomized design independently justifies causal interpretation.
#'
#' @param occurrences A non-contiguous subsequence occurrence table.
#' @param group_col Sequence-constant group column retained through
#'   `metadata_cols` during extraction.
#' @param test `"auto"`, `"chisq"`, or `"fisher"`.
#' @param p_adjust Multiple-testing adjustment method.
#' @param min_sequence_count Minimum total sequence count per subsequence.
#'
#' @return A data frame of prevalence contrasts and adjusted p-values.
#' @examples
#' data <- data.frame(
#'   sequence_id = rep(paste0("s", 1:4), each = 4L),
#'   sequence_order = rep(1:4, 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "D", "D",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L)
#' )
#' occurrences <- extract_sequence_subsequences(data, metadata_cols = "group")
#' compare_sequence_subsequences(occurrences, "group")
#' @export
compare_sequence_subsequences <- function(
  occurrences,
  group_col,
  test = c("auto", "chisq", "fisher"),
  p_adjust = "BH",
  min_sequence_count = 1L
) {
  if (!inherits(occurrences, "gp3_sequence_subsequences")) {
    stop("`occurrences` must be created by `extract_sequence_subsequences()`.",
         call. = FALSE)
  }
  test <- match.arg(test)
  .sequence_adv_scalar_character(group_col, "group_col")
  .sequence_adv_scalar_number(min_sequence_count, "min_sequence_count", lower = 1,
                              integer = TRUE)
  metadata <- attr(occurrences, "metadata")
  if (is.null(metadata) || !(group_col %in% names(metadata))) {
    stop("The requested group column was not retained as sequence metadata.",
         call. = FALSE)
  }
  id_col <- names(metadata)[1L]
  groups <- as.character(metadata[[group_col]])
  if (anyNA(groups) || any(!nzchar(trimws(groups)))) {
    stop("Group values must not be missing or blank.", call. = FALSE)
  }
  group_levels <- sort(unique(groups), method = "radix")
  if (length(group_levels) < 2L) stop("At least two groups are required.", call. = FALSE)
  sequence_ids <- as.character(metadata[[id_col]])
  motifs <- sort(unique(occurrences$subsequence), method = "radix")
  results <- list()
  h <- 0L
  for (motif in motifs) {
    present_ids <- unique(occurrences$sequence_id[occurrences$subsequence == motif])
    present <- sequence_ids %in% present_ids
    table_value <- table(
      factor(groups, levels = group_levels),
      factor(present, levels = c(FALSE, TRUE))
    )
    if (sum(table_value[, "TRUE"]) < min_sequence_count) next
    expected <- suppressWarnings(stats::chisq.test(table_value, correct = FALSE)$expected)
    chosen <- if (test == "auto") {
      if (any(expected < 5) && nrow(table_value) == 2L) "fisher" else "chisq"
    } else test
    if (chosen == "fisher" && nrow(table_value) != 2L) {
      stop("Fisher's exact test is currently limited to two groups.", call. = FALSE)
    }
    tested <- if (chosen == "fisher") {
      stats::fisher.test(table_value)
    } else {
      suppressWarnings(stats::chisq.test(table_value, correct = FALSE))
    }
    prevalence <- table_value[, "TRUE"] / rowSums(table_value)
    h <- h + 1L
    row <- data.frame(
      subsequence = motif,
      subsequence_length = occurrences$subsequence_length[match(motif, occurrences$subsequence)],
      test = chosen,
      statistic = if (is.null(tested$statistic)) NA_real_ else unname(tested$statistic),
      df = if (is.null(tested$parameter)) NA_real_ else unname(tested$parameter),
      p_value = tested$p.value,
      max_prevalence = max(prevalence),
      min_prevalence = min(prevalence),
      prevalence_range = max(prevalence) - min(prevalence),
      stringsAsFactors = FALSE
    )
    for (g in group_levels) {
      row[[paste0("prevalence_", make.names(g))]] <- prevalence[g]
      row[[paste0("n_", make.names(g))]] <- rowSums(table_value)[g]
    }
    if (length(group_levels) == 2L) {
      row$prevalence_difference <- prevalence[group_levels[2L]] - prevalence[group_levels[1L]]
    }
    results[[h]] <- row
  }
  if (length(results) == 0L) return(data.frame())
  result <- do.call(rbind, results)
  result$p_adjusted <- stats::p.adjust(result$p_value, method = p_adjust)
  result <- result[order(result$p_adjusted, -result$prevalence_range,
                         result$subsequence, method = "radix"), , drop = FALSE]
  row.names(result) <- NULL
  attr(result, "group_levels") <- group_levels
  attr(result, "p_adjust") <- p_adjust
  result
}

#' Plot non-contiguous subsequence summaries
#'
#' @param x A summary or group-comparison table.
#' @param metric Numeric column to plot.
#' @param top_n Maximum number of subsequences.
#' @param decreasing Sort metric in decreasing order.
#' @param ... Additional arguments passed to [graphics::barplot()].
#'
#' @return The plotted rows, invisibly.
#' @examples
#' data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
#'                    sequence_order = rep(1:4, 2L),
#'                    state = c("A", "B", "C", "D", "A", "C", "B", "D"))
#' summary <- summarise_sequence_subsequences(extract_sequence_subsequences(data))
#' plot_sequence_subsequences(summary, top_n = 5L)
#' @export
plot_sequence_subsequences <- function(
  x,
  metric = "sequence_prevalence",
  top_n = 10L,
  decreasing = TRUE,
  ...
) {
  if (!is.data.frame(x) || !("subsequence" %in% names(x)) ||
      !(metric %in% names(x)) || !is.numeric(x[[metric]])) {
    stop("`x` must contain `subsequence` and the requested numeric metric.",
         call. = FALSE)
  }
  .sequence_adv_scalar_number(top_n, "top_n", lower = 1, integer = TRUE)
  .sequence_adv_scalar_logical(decreasing, "decreasing")
  ordering <- order(x[[metric]], x$subsequence,
                    decreasing = decreasing, method = "radix")
  selected <- x[utils::head(ordering, top_n), , drop = FALSE]
  graphics::barplot(
    height = selected[[metric]],
    names.arg = selected$subsequence,
    las = 2,
    ylab = metric,
    ...
  )
  invisible(selected)
}

test_that("bounded non-contiguous subsequences are extracted", {
  data <- make_extension_sequence_data()
  occurrences <- extract_sequence_subsequences(
    data,
    metadata_cols = "group",
    min_length = 2L,
    max_length = 3L,
    max_gap = 2L,
    max_span = 4L
  )
  expect_s3_class(occurrences, "gp3_sequence_subsequences")
  expect_gt(nrow(occurrences), 0L)
  expect_true(all(occurrences$subsequence_length %in% 2:3))
  expect_true(all(occurrences$max_observed_gap <= 2))
  expect_true(all(occurrences$span <= 4))
})

test_that("subsequence summaries and filtering are stable", {
  grDevices::pdf(tempfile(fileext = ".pdf"))
  on.exit(grDevices::dev.off(), add = TRUE)
  occurrences <- extract_sequence_subsequences(
    make_extension_sequence_data(), metadata_cols = "group",
    min_length = 2L, max_length = 3L
  )
  summary <- summarise_sequence_subsequences(occurrences)
  expect_true(all(summary$sequence_prevalence >= 0 & summary$sequence_prevalence <= 1))
  filtered <- filter_sequence_subsequences(summary, min_sequences = 2L,
                                           top_n = 5L, ties = "exclude")
  expect_lte(nrow(filtered), 5L)
  expect_silent(plot_sequence_subsequences(summary, top_n = 3L))
})

test_that("group comparisons adjust multiple tests", {
  occurrences <- extract_sequence_subsequences(
    make_extension_sequence_data(), metadata_cols = "group",
    min_length = 2L, max_length = 2L
  )
  comparison <- compare_sequence_subsequences(occurrences, "group")
  expect_true(all(c("p_value", "p_adjusted") %in% names(comparison)))
  expect_true(all(comparison$p_adjusted >= comparison$p_value - 1e-12))
})

test_that("search-space safety limit is enforced", {
  data <- make_extension_sequence_data()
  expect_error(
    extract_sequence_subsequences(data, max_length = 5L,
                                  max_combinations_per_sequence = 2L),
    "exceeds"
  )
})

.sequence_adv_hmm_input <- function(data, sequence_id_col, order_col, state_col,
                                    symbol_levels = NULL) {
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  symbols <- if (is.null(symbol_levels)) x$state_levels else as.character(symbol_levels)
  if (!is.character(symbols) || length(symbols) < 1L || anyNA(symbols) ||
      any(!nzchar(trimws(symbols))) || anyDuplicated(symbols)) {
    stop("`symbol_levels` must contain unique, non-missing, non-blank symbols.",
         call. = FALSE)
  }
  missing <- setdiff(x$state_levels, symbols)
  if (length(missing) > 0L) {
    stop("`symbol_levels` does not cover observed symbols: ",
         paste(missing, collapse = ", "), ".", call. = FALSE)
  }
  encoded <- lapply(x$sequences, function(z) match(z, symbols))
  list(source = x, symbols = symbols, encoded = encoded)
}

.sequence_adv_hmm_initialise <- function(n_states, n_symbols, seed,
                                         initial_probs = NULL,
                                         transition_probs = NULL,
                                         emission_probs = NULL) {
  generated <- .sequence_adv_with_seed(seed, {
    list(
      initial = if (is.null(initial_probs)) stats::rexp(n_states) + 0.1 else initial_probs,
      transition = if (is.null(transition_probs)) {
        matrix(stats::rexp(n_states^2) + 0.1, n_states, n_states)
      } else {
        transition_probs
      },
      emission = if (is.null(emission_probs)) {
        matrix(stats::rexp(n_states * n_symbols) + 0.1, n_states, n_symbols)
      } else {
        emission_probs
      }
    )
  })
  initial_probs <- generated$initial
  transition_probs <- generated$transition
  emission_probs <- generated$emission
  if (!is.numeric(initial_probs) || length(initial_probs) != n_states ||
      any(!is.finite(initial_probs)) || any(initial_probs < 0)) {
    stop("Invalid `initial_probs`.", call. = FALSE)
  }
  if (!is.matrix(transition_probs) || !is.numeric(transition_probs) ||
      !all(dim(transition_probs) == c(n_states, n_states)) ||
      any(!is.finite(transition_probs)) || any(transition_probs < 0)) {
    stop("Invalid `transition_probs`.", call. = FALSE)
  }
  if (!is.matrix(emission_probs) || !is.numeric(emission_probs) ||
      !all(dim(emission_probs) == c(n_states, n_symbols)) ||
      any(!is.finite(emission_probs)) || any(emission_probs < 0)) {
    stop("Invalid `emission_probs`.", call. = FALSE)
  }
  list(
    initial = .sequence_adv_vector_normalise(initial_probs),
    transition = .sequence_adv_row_normalise(transition_probs),
    emission = .sequence_adv_row_normalise(emission_probs)
  )
}

.sequence_adv_hmm_sufficient <- function(encoded, parameters, weights = NULL,
                                         pseudocount = 0) {
  n_states <- length(parameters$initial)
  n_symbols <- ncol(parameters$emission)
  if (is.null(weights)) weights <- rep(1, length(encoded))
  if (!is.numeric(weights) || length(weights) != length(encoded) ||
      any(!is.finite(weights)) || any(weights < 0)) {
    stop("`weights` must contain one finite non-negative value per sequence.",
         call. = FALSE)
  }
  if (!is.numeric(pseudocount) || length(pseudocount) != 1L ||
      !is.finite(pseudocount) || pseudocount < 0) {
    stop("`pseudocount` must be one finite non-negative number.", call. = FALSE)
  }
  initial_count <- rep(pseudocount, n_states)
  transition_count <- matrix(pseudocount, n_states, n_states)
  emission_count <- matrix(pseudocount, n_states, n_symbols)
  log_likelihoods <- numeric(length(encoded))
  posterior <- vector("list", length(encoded))
  for (s in seq_along(encoded)) {
    fb <- .sequence_adv_forward_backward(encoded[[s]], parameters$initial,
                                         parameters$transition, parameters$emission)
    w <- weights[s]
    initial_count <- initial_count + w * fb$gamma[1L, ]
    if (length(encoded[[s]]) > 1L) transition_count <- transition_count +
      w * apply(fb$xi, c(2L, 3L), sum)
    for (t in seq_along(encoded[[s]])) {
      emission_count[, encoded[[s]][t]] <- emission_count[, encoded[[s]][t]] +
        w * fb$gamma[t, ]
    }
    log_likelihoods[s] <- fb$log_likelihood
    posterior[[s]] <- fb
  }
  list(initial = initial_count, transition = transition_count, emission = emission_count,
       log_likelihoods = log_likelihoods, posterior = posterior)
}

#' Fit a categorical hidden Markov model
#'
#' Fits a finite-state, time-homogeneous categorical HMM by Baum-Welch EM.
#' Latent states are statistical model states only; they are not psychological,
#' diagnostic, or causal constructs.
#'
#' @param data Long-format sequence data.
#' @param n_states Number of latent states.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param symbol_levels Optional observed-symbol ordering.
#' @param state_names Optional latent-state names.
#' @param initial_probs,transition_probs,emission_probs Optional starting values.
#' @param max_iter Maximum EM iterations.
#' @param tolerance Relative log-likelihood tolerance.
#' @param pseudocount Non-negative smoothing count.
#' @param seed Reproducibility seed.
#' @param keep_posteriors Retain final forward-backward results.
#'
#' @return An object of class `gp3_sequence_hmm` containing fitted parameters,
#' log likelihood, convergence diagnostics, symbol coding, and optional
#' posteriors.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' fit_sequence_hmm(sequences, n_states = 2L, max_iter = 5L, seed = 1L)
#'
#' @export
fit_sequence_hmm <- function(
  data,
  n_states,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  symbol_levels = NULL,
  state_names = NULL,
  initial_probs = NULL,
  transition_probs = NULL,
  emission_probs = NULL,
  max_iter = 200L,
  tolerance = 1e-6,
  pseudocount = 1e-6,
  seed = 1L,
  keep_posteriors = FALSE
) {
  .sequence_adv_scalar_number(n_states, "n_states", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(max_iter, "max_iter", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(tolerance, "tolerance", lower = 0)
  .sequence_adv_scalar_number(pseudocount, "pseudocount", lower = 0)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  .sequence_adv_scalar_logical(keep_posteriors, "keep_posteriors")
  input <- .sequence_adv_hmm_input(data, sequence_id_col, order_col, state_col,
                                   symbol_levels)
  n_symbols <- length(input$symbols)
  if (is.null(state_names)) state_names <- paste0("latent_", seq_len(n_states))
  if (!is.character(state_names) || length(state_names) != n_states || anyNA(state_names) ||
      any(!nzchar(trimws(state_names))) || anyDuplicated(state_names)) {
    stop("`state_names` must uniquely name all latent states.", call. = FALSE)
  }
  parameters <- .sequence_adv_hmm_initialise(n_states, n_symbols, seed,
                                             initial_probs, transition_probs,
                                             emission_probs)
  history <- numeric(max_iter)
  converged <- FALSE
  previous <- -Inf
  final_sufficient <- NULL
  for (iteration in seq_len(max_iter)) {
    sufficient <- .sequence_adv_hmm_sufficient(input$encoded, parameters,
                                                pseudocount = pseudocount)
    current <- sum(sufficient$log_likelihoods)
    history[iteration] <- current
    parameters$initial <- .sequence_adv_vector_normalise(sufficient$initial)
    parameters$transition <- .sequence_adv_row_normalise(sufficient$transition)
    parameters$emission <- .sequence_adv_row_normalise(sufficient$emission)
    final_sufficient <- sufficient
    if (iteration > 1L) {
      relative <- abs(current - previous) / max(1, abs(previous))
      if (relative <= tolerance) {
        converged <- TRUE
        history <- history[seq_len(iteration)]
        break
      }
    }
    previous <- current
    if (iteration == max_iter) history <- history[seq_len(iteration)]
  }
  final_sufficient <- .sequence_adv_hmm_sufficient(input$encoded, parameters,
                                                    pseudocount = 0)
  log_likelihood <- sum(final_sufficient$log_likelihoods)
  history[length(history)] <- log_likelihood
  dimnames(parameters$transition) <- list(state_names, state_names)
  dimnames(parameters$emission) <- list(state_names, input$symbols)
  names(parameters$initial) <- state_names
  n_parameters <- (n_states - 1L) + n_states * (n_states - 1L) +
    n_states * (n_symbols - 1L)
  n_observations <- sum(lengths(input$encoded))
  result <- list(
    initial_probs = parameters$initial,
    transition_probs = parameters$transition,
    emission_probs = parameters$emission,
    state_names = state_names,
    symbol_names = input$symbols,
    sequence_ids = input$source$sequence_ids,
    log_likelihood = log_likelihood,
    sequence_log_likelihoods = stats::setNames(final_sufficient$log_likelihoods,
                                        input$source$sequence_ids),
    iterations = length(history),
    converged = converged,
    tolerance = tolerance,
    pseudocount = pseudocount,
    log_likelihood_history = history,
    n_parameters = as.integer(n_parameters),
    n_observations = as.integer(n_observations),
    aic = -2 * log_likelihood + 2 * n_parameters,
    bic = -2 * log_likelihood + log(max(1, n_observations)) * n_parameters,
    seed = as.integer(seed),
    call = match.call(),
    posteriors = if (keep_posteriors) final_sufficient$posterior else NULL,
    training_sequences = input$source$sequences,
    training_orders = input$source$orders
  )
  class(result) <- c("gp3_sequence_hmm", "list")
  result
}

.sequence_adv_hmm_loglik <- function(encoded, parameters) {
  vapply(encoded, function(obs) {
    .sequence_adv_forward_backward(obs, parameters$initial,
                                   parameters$transition,
                                   parameters$emission)$log_likelihood
  }, numeric(1))
}

#' Fit a mixture of categorical hidden Markov models
#'
#' Fits sequence-level mixture components, each containing a categorical HMM.
#' Component membership is a statistical clustering device and should not be
#' interpreted as a substantive latent type without external validation.
#'
#' @param data Long-format sequence data.
#' @param n_components Number of mixture components.
#' @param n_states Number of hidden states per component; scalar or vector.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param symbol_levels Optional symbol ordering.
#' @param max_iter Maximum mixture-EM iterations.
#' @param inner_initial_iter Initial single-HMM iterations per component.
#' @param tolerance Relative log-likelihood tolerance.
#' @param pseudocount Smoothing count.
#' @param seed Reproducibility seed.
#'
#' @return An object of class `gp3_sequence_hmm_mixture`.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' fit_sequence_hmm_mixture(sequences, n_components = 2L, n_states = 2L,
#'                          max_iter = 5L, inner_initial_iter = 2L, seed = 1L)
#'
#' @export
fit_sequence_hmm_mixture <- function(
  data,
  n_components,
  n_states,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  symbol_levels = NULL,
  max_iter = 200L,
  inner_initial_iter = 20L,
  tolerance = 1e-6,
  pseudocount = 1e-6,
  seed = 1L
) {
  .sequence_adv_scalar_number(n_components, "n_components", lower = 2, integer = TRUE)
  if (length(n_states) == 1L) n_states <- rep(n_states, n_components)
  if (!is.numeric(n_states) || length(n_states) != n_components || anyNA(n_states) ||
      any(!is.finite(n_states)) || any(n_states < 1) ||
      any(n_states > .Machine$integer.max) || any(n_states != trunc(n_states))) {
    stop("`n_states` must be a positive integer scalar or one value per component.",
         call. = FALSE)
  }
  .sequence_adv_scalar_number(max_iter, "max_iter", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(inner_initial_iter, "inner_initial_iter", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(tolerance, "tolerance", lower = 0)
  .sequence_adv_scalar_number(pseudocount, "pseudocount", lower = 0)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  input <- .sequence_adv_hmm_input(data, sequence_id_col, order_col, state_col,
                                   symbol_levels)
  n_sequences <- length(input$encoded)
  if (n_components > n_sequences) stop("More components than sequences were requested.", call. = FALSE)
  initial_responsibilities <- .sequence_adv_with_seed(seed, {
    raw <- matrix(
      stats::rexp(n_sequences * n_components) + 0.1,
      nrow = n_sequences,
      ncol = n_components
    )
    raw / rowSums(raw)
  })
  parameters <- vector("list", n_components)
  for (component in seq_len(n_components)) {
    parameters[[component]] <- .sequence_adv_hmm_initialise(
      n_states[component], length(input$symbols), .sequence_adv_seed(seed, component)
    )
    for (initial_iteration in seq_len(inner_initial_iter)) {
      sufficient <- .sequence_adv_hmm_sufficient(
        input$encoded,
        parameters[[component]],
        weights = initial_responsibilities[, component],
        pseudocount = pseudocount
      )
      parameters[[component]]$initial <- .sequence_adv_vector_normalise(sufficient$initial)
      parameters[[component]]$transition <- .sequence_adv_row_normalise(sufficient$transition)
      parameters[[component]]$emission <- .sequence_adv_row_normalise(sufficient$emission)
    }
  }
  mixture_weights <- .sequence_adv_vector_normalise(colMeans(initial_responsibilities))
  history <- numeric(max_iter)
  previous <- -Inf
  converged <- FALSE
  responsibilities <- initial_responsibilities
  component_loglik <- matrix(NA_real_, n_sequences, n_components)
  for (iteration in seq_len(max_iter)) {
    for (component in seq_len(n_components)) {
      component_loglik[, component] <- .sequence_adv_hmm_loglik(input$encoded,
                                                                 parameters[[component]])
    }
    log_joint <- sweep(component_loglik, 2L,
                       log(pmax(mixture_weights, .Machine$double.xmin)), "+")
    row_max <- apply(log_joint, 1L, max)
    stabilised <- exp(log_joint - row_max)
    row_total <- rowSums(stabilised)
    if (any(!is.finite(row_total) | row_total <= 0)) {
      stop("Mixture responsibilities could not be normalised.", call. = FALSE)
    }
    responsibilities <- stabilised / row_total
    current <- sum(row_max + log(row_total))
    history[iteration] <- current
    mixture_weights <- colMeans(responsibilities)
    mixture_weights <- .sequence_adv_vector_normalise(mixture_weights, pseudocount)
    for (component in seq_len(n_components)) {
      sufficient <- .sequence_adv_hmm_sufficient(
        input$encoded, parameters[[component]],
        weights = responsibilities[, component], pseudocount = pseudocount
      )
      parameters[[component]]$initial <- .sequence_adv_vector_normalise(sufficient$initial)
      parameters[[component]]$transition <- .sequence_adv_row_normalise(sufficient$transition)
      parameters[[component]]$emission <- .sequence_adv_row_normalise(sufficient$emission)
    }
    if (iteration > 1L) {
      relative <- abs(current - previous) / max(1, abs(previous))
      if (relative <= tolerance) {
        converged <- TRUE
        history <- history[seq_len(iteration)]
        break
      }
    }
    previous <- current
    if (iteration == max_iter) history <- history[seq_len(iteration)]
  }
  # Recompute component likelihoods and memberships under the final parameter
  # values. This avoids returning responsibilities from the pre-update E-step.
  for (component in seq_len(n_components)) {
    component_loglik[, component] <- .sequence_adv_hmm_loglik(
      input$encoded, parameters[[component]]
    )
  }
  final_log_joint <- sweep(
    component_loglik, 2L,
    log(pmax(mixture_weights, .Machine$double.xmin)), "+"
  )
  final_row_max <- apply(final_log_joint, 1L, max)
  final_stabilised <- exp(final_log_joint - final_row_max)
  final_row_total <- rowSums(final_stabilised)
  if (any(!is.finite(final_row_total) | final_row_total <= 0)) {
    stop("Final mixture responsibilities could not be normalised.", call. = FALSE)
  }
  responsibilities <- final_stabilised / final_row_total
  final_log_likelihood <- sum(final_row_max + log(final_row_total))
  history[length(history)] <- final_log_likelihood

  component_models <- lapply(seq_len(n_components), function(component) {
    state_names <- paste0("component_", component, "_latent_",
                          seq_len(n_states[component]))
    p <- parameters[[component]]
    names(p$initial) <- state_names
    dimnames(p$transition) <- list(state_names, state_names)
    dimnames(p$emission) <- list(state_names, input$symbols)
    list(initial_probs = p$initial, transition_probs = p$transition,
         emission_probs = p$emission, state_names = state_names,
         symbol_names = input$symbols)
  })
  hard <- max.col(responsibilities, ties.method = "first")
  n_parameters <- (n_components - 1L) + sum(vapply(seq_len(n_components), function(component) {
    k <- n_states[component]
    (k - 1L) + k * (k - 1L) + k * (length(input$symbols) - 1L)
  }, numeric(1)))
  n_observations <- sum(lengths(input$encoded))
  result <- list(
    mixture_weights = stats::setNames(mixture_weights, paste0("component_", seq_len(n_components))),
    components = component_models,
    responsibilities = data.frame(
      sequence_id = input$source$sequence_ids,
      responsibilities,
      assigned_component = hard,
      stringsAsFactors = FALSE,
      check.names = FALSE
    ),
    log_likelihood = final_log_likelihood,
    log_likelihood_history = history,
    iterations = length(history),
    converged = converged,
    tolerance = tolerance,
    pseudocount = pseudocount,
    inner_initial_iter = as.integer(inner_initial_iter),
    n_components = as.integer(n_components),
    n_states = as.integer(n_states),
    n_parameters = as.integer(n_parameters),
    n_observations = as.integer(n_observations),
    aic = -2 * final_log_likelihood + 2 * n_parameters,
    bic = -2 * final_log_likelihood +
      log(max(1, n_observations)) * n_parameters,
    symbol_names = input$symbols,
    sequence_ids = input$source$sequence_ids,
    seed = as.integer(seed),
    training_sequences = input$source$sequences,
    training_orders = input$source$orders,
    call = match.call()
  )
  names(result$responsibilities)[seq_len(n_components) + 1L] <-
    paste0("component_", seq_len(n_components))
  class(result) <- c("gp3_sequence_hmm_mixture", "list")
  result
}

#' Decode hidden states from a fitted HMM
#'
#' @param model A fitted single HMM or HMM mixture.
#' @param data Optional new long-format data. Training sequences are used when
#'   omitted.
#' @param sequence_id_col,order_col,state_col Sequence columns for new data.
#' @param method `"viterbi"` or `"posterior"`.
#' @param component Mixture component to decode. When omitted for a mixture,
#' each sequence uses its highest-responsibility component.
#'
#' @return A long data frame containing decoded latent states and posterior
#' probabilities where available.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' model <- fit_sequence_hmm(sequences, 2L, max_iter = 5L)
#' decode_sequence_states(model)
#'
#' @export
decode_sequence_states <- function(model, data = NULL,
                                   sequence_id_col = "sequence_id",
                                   order_col = "sequence_order",
                                   state_col = "state",
                                   method = c("viterbi", "posterior"),
                                   component = NULL) {
  method <- match.arg(method)
  is_single <- inherits(model, "gp3_sequence_hmm")
  is_mixture <- inherits(model, "gp3_sequence_hmm_mixture")
  if (!is_single && !is_mixture) stop("Unsupported HMM model object.", call. = FALSE)
  symbols <- model$symbol_names
  if (is.null(data)) {
    sequences <- model$training_sequences
    orders <- model$training_orders
    ids <- names(sequences)
  } else {
    input <- .sequence_adv_hmm_input(data, sequence_id_col, order_col, state_col, symbols)
    sequences <- input$source$sequences
    orders <- input$source$orders
    ids <- input$source$sequence_ids
  }
  encoded <- lapply(sequences, function(z) match(z, symbols))
  if (is_mixture) {
    if (!is.null(component)) {
      .sequence_adv_scalar_number(component, "component", lower = 1,
                                  upper = model$n_components, integer = TRUE)
      components <- rep(as.integer(component), length(ids))
    } else if (is.null(data)) {
      components <- model$responsibilities$assigned_component[
        match(ids, model$responsibilities$sequence_id)]
    } else {
      loglik <- do.call(cbind, lapply(seq_len(model$n_components), function(k) {
        current <- model$components[[k]]
        p <- list(initial = current$initial_probs,
                  transition = current$transition_probs,
                  emission = current$emission_probs)
        .sequence_adv_hmm_loglik(encoded, p) +
          log(pmax(model$mixture_weights[k], .Machine$double.xmin))
      }))
      components <- max.col(loglik, ties.method = "first")
    }
  } else {
    components <- rep(1L, length(ids))
  }
  out <- list(); h <- 0L
  for (i in seq_along(ids)) {
    current <- if (is_single) {
      list(initial_probs = model$initial_probs,
           transition_probs = model$transition_probs,
           emission_probs = model$emission_probs,
           state_names = model$state_names)
    } else model$components[[components[i]]]
    p <- list(initial = current$initial_probs,
              transition = current$transition_probs,
              emission = current$emission_probs)
    fb <- .sequence_adv_forward_backward(encoded[[i]], p$initial, p$transition, p$emission)
    if (method == "viterbi") {
      decoded <- .sequence_adv_viterbi(encoded[[i]], p$initial, p$transition, p$emission)$path
      posterior_probability <- fb$gamma[cbind(seq_along(decoded), decoded)]
    } else {
      decoded <- max.col(fb$gamma, ties.method = "first")
      posterior_probability <- apply(fb$gamma, 1L, max)
    }
    for (t in seq_along(decoded)) {
      h <- h + 1L
      out[[h]] <- data.frame(
        sequence_id = ids[i],
        sequence_order = orders[[i]][t],
        observed_state = sequences[[i]][t],
        component = components[i],
        latent_state = current$state_names[decoded[t]],
        posterior_probability = posterior_probability[t],
        decoding_method = method,
        stringsAsFactors = FALSE
      )
    }
  }
  result <- do.call(rbind, out)
  row.names(result) <- NULL
  result
}

#' Summarise a fitted sequence HMM
#'
#' @param model A single or mixture HMM.
#'
#' @return A list of convergence, fit, initial, transition, emission, and
#' mixture summaries.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' model <- fit_sequence_hmm(sequences, 2L, max_iter = 5L)
#' summarise_sequence_hmm(model)
#'
#' @export
summarise_sequence_hmm <- function(model) {
  if (inherits(model, "gp3_sequence_hmm")) {
    initial <- data.frame(latent_state = names(model$initial_probs),
                          probability = as.numeric(model$initial_probs),
                          stringsAsFactors = FALSE)
    transition <- as.data.frame(as.table(model$transition_probs), stringsAsFactors = FALSE)
    names(transition) <- c("from_state", "to_state", "probability")
    emission <- as.data.frame(as.table(model$emission_probs), stringsAsFactors = FALSE)
    names(emission) <- c("latent_state", "observed_state", "probability")
    fit <- data.frame(log_likelihood = model$log_likelihood, aic = model$aic,
                      bic = model$bic, n_parameters = model$n_parameters,
                      n_observations = model$n_observations,
                      iterations = model$iterations, converged = model$converged,
                      stringsAsFactors = FALSE)
    return(list(fit = fit, initial = initial, transition = transition,
                emission = emission, mixture = NULL))
  }
  if (inherits(model, "gp3_sequence_hmm_mixture")) {
    component_summaries <- lapply(seq_along(model$components), function(k) {
      current <- model$components[[k]]
      transition <- as.data.frame(as.table(current$transition_probs), stringsAsFactors = FALSE)
      names(transition) <- c("from_state", "to_state", "probability")
      transition$component <- k
      emission <- as.data.frame(as.table(current$emission_probs), stringsAsFactors = FALSE)
      names(emission) <- c("latent_state", "observed_state", "probability")
      emission$component <- k
      initial <- data.frame(component = k, latent_state = names(current$initial_probs),
                            probability = as.numeric(current$initial_probs),
                            stringsAsFactors = FALSE)
      list(initial = initial, transition = transition, emission = emission)
    })
    fit <- data.frame(log_likelihood = model$log_likelihood, aic = model$aic,
                      bic = model$bic, n_parameters = model$n_parameters,
                      n_observations = model$n_observations,
                      iterations = model$iterations, converged = model$converged,
                      stringsAsFactors = FALSE)
    mixture <- data.frame(component = seq_len(model$n_components),
                          weight = as.numeric(model$mixture_weights),
                          n_states = model$n_states,
                          stringsAsFactors = FALSE)
    return(list(
      fit = fit,
      initial = do.call(rbind, lapply(component_summaries, `[[`, "initial")),
      transition = do.call(rbind, lapply(component_summaries, `[[`, "transition")),
      emission = do.call(rbind, lapply(component_summaries, `[[`, "emission")),
      mixture = mixture,
      responsibilities = model$responsibilities
    ))
  }
  stop("Unsupported HMM model object.", call. = FALSE)
}

#' Compare fitted sequence HMMs descriptively
#'
#' @param ... Fitted HMM or mixture-HMM objects.
#'
#' @return A data frame containing log likelihood, AIC, BIC, parameter count,
#' convergence status, and delta criteria. Criteria are descriptive and do not
#' automatically select a substantive model.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' model_1 <- fit_sequence_hmm(sequences, 1L, max_iter = 5L)
#' model_2 <- fit_sequence_hmm(sequences, 2L, max_iter = 5L)
#' compare_sequence_hmms(one = model_1, two = model_2)
#'
#' @export
compare_sequence_hmms <- function(...) {
  models <- list(...)
  if (length(models) < 2L) stop("Supply at least two fitted HMMs.", call. = FALSE)
  valid <- vapply(models, function(x) inherits(x, "gp3_sequence_hmm") ||
                    inherits(x, "gp3_sequence_hmm_mixture"), logical(1))
  if (!all(valid)) stop("All objects must be fitted sequence HMMs.", call. = FALSE)
  observation_counts <- vapply(models, `[[`, numeric(1), "n_observations")
  if (length(unique(observation_counts)) != 1L) {
    stop("HMM fit criteria are comparable only when models use the same number of observations.",
         call. = FALSE)
  }
  reference_ids <- models[[1L]]$sequence_ids
  reference_symbols <- models[[1L]]$symbol_names
  reference_sequences <- models[[1L]]$training_sequences
  reference_orders <- models[[1L]]$training_orders
  same_ids <- vapply(models, function(model) {
    identical(model$sequence_ids, reference_ids)
  }, logical(1))
  same_symbols <- vapply(models, function(model) {
    identical(model$symbol_names, reference_symbols)
  }, logical(1))
  same_sequences <- vapply(models, function(model) {
    identical(model$training_sequences, reference_sequences) &&
      identical(model$training_orders, reference_orders)
  }, logical(1))
  if (!all(same_ids) || !all(same_symbols) || !all(same_sequences)) {
    stop("HMM fit criteria are comparable only for identical training sequences, orders, and symbol coding.",
         call. = FALSE)
  }
  labels <- names(models)
  if (is.null(labels) || anyNA(labels) || any(!nzchar(labels)) ||
      anyDuplicated(labels)) {
    labels <- paste0("model_", seq_along(models))
  }
  out <- do.call(rbind, lapply(seq_along(models), function(i) {
    model <- models[[i]]
    data.frame(model = labels[i], class = class(model)[1L],
               log_likelihood = model$log_likelihood, aic = model$aic,
               bic = model$bic, n_parameters = model$n_parameters,
               n_observations = model$n_observations,
               converged = model$converged, stringsAsFactors = FALSE)
  }))
  out$delta_aic <- out$aic - min(out$aic)
  out$delta_bic <- out$bic - min(out$bic)
  out[order(out$bic, out$aic, out$model, method = "radix"), , drop = FALSE]
}

test_that("single HMM fitting is deterministic and normalised", {
  data <- make_advanced_sequence_data()
  first <- fit_sequence_hmm(data, n_states = 2L, max_iter = 40L,
                            tolerance = 1e-7, seed = 11L)
  second <- fit_sequence_hmm(data, n_states = 2L, max_iter = 40L,
                             tolerance = 1e-7, seed = 11L)
  expect_s3_class(first, "gp3_sequence_hmm")
  expect_equal(first$initial_probs, second$initial_probs)
  expect_equal(first$transition_probs, second$transition_probs)
  expect_equal(first$emission_probs, second$emission_probs)
  expect_equal(sum(first$initial_probs), 1, tolerance = 1e-10)
  expect_equal(unname(rowSums(first$transition_probs)), rep(1, 2L), tolerance = 1e-10)
  expect_equal(unname(rowSums(first$emission_probs)), rep(1, 2L), tolerance = 1e-10)
  expect_true(is.finite(first$log_likelihood))
})

test_that("HMM decoding returns one latent state per observation", {
  data <- make_advanced_sequence_data()
  model <- fit_sequence_hmm(data, n_states = 2L, max_iter = 30L, seed = 5L)
  viterbi <- decode_sequence_states(model, method = "viterbi")
  posterior <- decode_sequence_states(model, method = "posterior")
  expect_equal(nrow(viterbi), nrow(data))
  expect_equal(nrow(posterior), nrow(data))
  expect_true(all(viterbi$posterior_probability >= 0 &
                    viterbi$posterior_probability <= 1))
  expect_true(all(posterior$posterior_probability >= 0 &
                    posterior$posterior_probability <= 1))
})

test_that("HMM summaries and model comparisons are structured", {
  data <- make_advanced_sequence_data()
  model_2 <- fit_sequence_hmm(data, 2L, max_iter = 25L, seed = 3L)
  model_3 <- fit_sequence_hmm(data, 3L, max_iter = 25L, seed = 4L)
  summary <- summarise_sequence_hmm(model_2)
  expect_true(all(c("fit", "initial", "transition", "emission") %in% names(summary)))
  comparison <- compare_sequence_hmms(two = model_2, three = model_3)
  expect_equal(nrow(comparison), 2L)
  expect_true(all(c("delta_aic", "delta_bic", "converged") %in% names(comparison)))
})

test_that("mixture HMM responsibilities are normalised and deterministic", {
  data <- make_advanced_sequence_data()
  first <- fit_sequence_hmm_mixture(data, n_components = 2L, n_states = 2L,
                                    max_iter = 30L, inner_initial_iter = 5L,
                                    seed = 21L)
  second <- fit_sequence_hmm_mixture(data, n_components = 2L, n_states = 2L,
                                     max_iter = 30L, inner_initial_iter = 5L,
                                     seed = 21L)
  expect_s3_class(first, "gp3_sequence_hmm_mixture")
  probs <- as.matrix(first$responsibilities[, c("component_1", "component_2")])
  expect_equal(rowSums(probs), rep(1, nrow(probs)), tolerance = 1e-10)
  expect_equal(first$mixture_weights, second$mixture_weights)
  expect_equal(first$responsibilities, second$responsibilities)
  decoded <- decode_sequence_states(first)
  expect_equal(nrow(decoded), nrow(data))
  expect_true(all(decoded$component %in% 1:2))
})

test_that("HMM inputs reject unsupported symbols and invalid probabilities", {
  data <- make_advanced_sequence_data()
  expect_error(fit_sequence_hmm(data, 2L, symbol_levels = c("A", "B")),
               "does not cover")
  expect_error(fit_sequence_hmm(data, 2L,
                                initial_probs = c(-1, 2)), "Invalid")
})

test_that("HMM initialisation rejects non-finite probabilities", {
  data <- make_advanced_sequence_data()
  expect_error(
    fit_sequence_hmm(data, 2L, initial_probs = c(Inf, 1)),
    "Invalid `initial_probs`"
  )
  expect_error(
    fit_sequence_hmm(
      data,
      2L,
      transition_probs = matrix(c(1, 0, NA, 1), 2L, 2L)
    ),
    "Invalid `transition_probs`"
  )
})

test_that("HMM fitting restores the caller random-number state", {
  set.seed(321L)
  before <- .Random.seed
  fit_sequence_hmm(make_advanced_sequence_data(), 2L,
                   max_iter = 3L, seed = 2L)
  expect_identical(.Random.seed, before)
})

test_that("HMM fit criteria require a common observation basis", {
  data <- make_advanced_sequence_data()
  full <- fit_sequence_hmm(data, 2L, max_iter = 3L, seed = 1L)
  reduced <- fit_sequence_hmm(
    data[data$sequence_id != "s08", , drop = FALSE],
    2L,
    max_iter = 3L,
    seed = 1L
  )
  expect_error(
    compare_sequence_hmms(full = full, reduced = reduced),
    "same number of observations"
  )
})

test_that("HMM symbol levels are unique and exclude unused factor levels", {
  data <- make_advanced_sequence_data()
  expect_error(
    fit_sequence_hmm(data, 2L, symbol_levels = c("A", "A", "B", "C", "D"),
                     max_iter = 2L),
    "unique"
  )
  data$state <- factor(data$state, levels = c("A", "B", "C", "D", "UNUSED"))
  model <- fit_sequence_hmm(data, 2L, max_iter = 2L, seed = 5L)
  expect_false("UNUSED" %in% model$symbol_names)
})

test_that("HMM comparison requires identical sequence identifiers", {
  data <- make_advanced_sequence_data()
  first <- fit_sequence_hmm(data, 2L, max_iter = 2L, seed = 1L)
  renamed <- data
  renamed$sequence_id <- paste0("renamed_", renamed$sequence_id)
  second <- fit_sequence_hmm(renamed, 2L, max_iter = 2L, seed = 1L)
  expect_error(
    compare_sequence_hmms(first = first, second = second),
    "identical training sequences"
  )
})

testthat::test_that("mixture seed offsets remain valid at the integer boundary", {
  data <- make_advanced_sequence_data()

  testthat::expect_error(
    fit_sequence_hmm(data, 2L, max_iter = 1L, seed = -1L),
    "seed"
  )
  testthat::expect_silent(
    fit_sequence_hmm_mixture(
      data,
      n_components = 2L,
      n_states = 1L,
      max_iter = 1L,
      inner_initial_iter = 1L,
      seed = .Machine$integer.max
    )
  )
})

test_that("mixture hidden-state counts reject values outside R integer range", {
  data <- make_advanced_sequence_data()
  expect_error(
    fit_sequence_hmm_mixture(
      data,
      n_components = 2L,
      n_states = c(1, .Machine$integer.max + 1),
      max_iter = 1L,
      inner_initial_iter = 1L
    ),
    "positive integer"
  )
})

#' Report gp3sequences capabilities and optional integrations
#'
#' Returns a deterministic, machine-readable inventory of native analytical
#' capabilities, optional adapters, reference implementations, and specialist
#' handoffs relevant to the current gp3sequences development series.
#'
#' @param include_optional Logical; include optional/reference capabilities.
#' @param check_versions Logical; report installed optional-package versions.
#'
#' @return A data frame with capability family, capability, implementation role,
#' backend, availability, and version information. The function never installs,
#' attaches, or loads optional packages.
#'
#' @examples
#' capabilities <- sequence_capabilities()
#' capabilities[c("family", "capability", "role", "available")]
#'
#' @export
sequence_capabilities <- function(
  include_optional = TRUE,
  check_versions = TRUE
) {
  if (
    !is.logical(include_optional) ||
      length(include_optional) != 1L ||
      is.na(include_optional)
  ) {
    stop(
      "`include_optional` must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  if (
    !is.logical(check_versions) ||
      length(check_versions) != 1L ||
      is.na(check_versions)
  ) {
    stop(
      "`check_versions` must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  rows <- data.frame(
    family = c(
      "Data contract",
      "Distances",
      "Distances",
      "Clustering",
      "Patterns",
      "Patterns",
      "HMMs",
      "HMMs",
      "Networks",
      "Networks",
      "Networks",
      "Inference",
      "Missingness",
      "Model-based clustering",
      "Graphics",
      "Graphics",
      "Property testing",
      "Performance"
    ),
    capability = c(
      "Validation and preparation",
      "Native sequence distances",
      "Reference distance validation",
      "Native clustering and stability",
      "Frequent pattern reference validation",
      "String/AOI pattern interoperability",
      "Native categorical HMMs",
      "HMM reference validation",
      "Native transition networks",
      "Graph interoperability",
      "Markov-chain interoperability",
      "Permutation/distance reference validation",
      "Sequence-imputation handoff",
      "Specialist model-based clustering handoff",
      "Sequence-plot handoff",
      "Seriation/ordering handoff",
      "Property-based testing",
      "Benchmarking"
    ),
    role = c(
      "native",
      "native",
      "reference",
      "native",
      "reference",
      "adapter",
      "native",
      "reference",
      "native",
      "adapter",
      "planned_adapter",
      "reference",
      "handoff",
      "handoff",
      "handoff",
      "handoff",
      "development",
      "development"
    ),
    backend = c(
      NA,
      NA,
      "TraMineR|stringdist",
      "cluster|WeightedCluster|clusterCrit|clue",
      "TraMineR|arulesSequences",
      "GrpString",
      NA,
      "seqHMM",
      NA,
      "igraph",
      "markovchain",
      "TraMineRextras|vegan|energy|coin",
      "seqimpute",
      "MEDseq",
      "ggseqplot",
      "seriation",
      "quickcheck|hedgehog",
      "bench|microbenchmark"
    ),
    reference_only = c(
      FALSE, FALSE, TRUE, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE,
      FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE
    ),
    stringsAsFactors = FALSE
  )

  package_groups <- lapply(
    rows$backend,
    function(value) {
      if (is.na(value)) {
        return(character())
      }

      strsplit(
        value,
        "\\|"
      )[[1L]]
    }
  )

  installed <- utils::installed.packages(
    fields = "Version"
  )

  installed_names <- rownames(installed)

  rows$available <- vapply(
    package_groups,
    function(packages) {
      if (length(packages) == 0L) {
        return(TRUE)
      }

      all(
        packages %in% installed_names
      )
    },
    logical(1)
  )

  rows$installed_version <- vapply(
    package_groups,
    function(packages) {
      if (
        !check_versions ||
          length(packages) == 0L
      ) {
        return(NA_character_)
      }

      versions <- vapply(
        packages,
        function(package) {
          if (package %in% installed_names) {
            paste0(
              package,
              " ",
              installed[
                package,
                "Version"
              ]
            )
          } else {
            paste0(
              package,
              " <not installed>"
            )
          }
        },
        character(1)
      )

      paste(
        versions,
        collapse = "; "
      )
    },
    character(1)
  )

  rows$native <- rows$role == "native"

  rows$backend_required <- rows$role %in% c(
    "adapter",
    "planned_adapter"
  )

  rows$minimum_tested_version <- NA_character_

  rows$notes <- ifelse(
    rows$role == "native",
    "Available without the optional backend.",
    paste0(
      "Optional integration, reference validation, development QA, ",
      "or documented handoff."
    )
  )

  if (!include_optional) {
    rows <- rows[
      rows$role == "native",
      ,
      drop = FALSE
    ]
  }

  rows <- rows[
    order(
      rows$family,
      rows$capability,
      rows$role,
      method = "radix"
    ),
    c(
      "family",
      "capability",
      "role",
      "native",
      "backend",
      "backend_required",
      "available",
      "installed_version",
      "minimum_tested_version",
      "reference_only",
      "notes"
    ),
    drop = FALSE
  ]

  row.names(rows) <- NULL

  rows
}

test_that("sequence_capabilities is deterministic and dependency-safe", {
  first <- sequence_capabilities()
  second <- sequence_capabilities()

  expect_s3_class(
    first,
    "data.frame"
  )

  expect_identical(
    first,
    second
  )

  expect_true(
    all(
      c(
        "family",
        "capability",
        "role",
        "native",
        "backend",
        "backend_required",
        "available",
        "installed_version",
        "minimum_tested_version",
        "reference_only",
        "notes"
      ) %in% names(first)
    )
  )

  expect_true(
    any(first$role == "native")
  )

  expect_true(
    any(first$role == "reference")
  )

  expect_true(
    all(first$available %in% c(TRUE, FALSE))
  )
})

test_that("sequence_capabilities can report native capabilities only", {
  native <- sequence_capabilities(
    include_optional = FALSE
  )

  expect_true(
    nrow(native) > 0L
  )

  expect_true(
    all(native$role == "native")
  )

  expect_true(
    all(native$native)
  )
})

test_that("sequence_capabilities does not load optional backend namespaces", {
  optional_backends <- c(
    "TraMineR",
    "stringdist",
    "cluster",
    "WeightedCluster",
    "clusterCrit",
    "clue",
    "arulesSequences",
    "GrpString",
    "seqHMM",
    "igraph",
    "markovchain",
    "TraMineRextras",
    "vegan",
    "energy",
    "coin",
    "seqimpute",
    "MEDseq",
    "ggseqplot",
    "seriation",
    "quickcheck",
    "hedgehog",
    "bench",
    "microbenchmark"
  )

  before <- loadedNamespaces()

  result <- sequence_capabilities()

  after <- loadedNamespaces()

  newly_loaded_optional <- setdiff(
    intersect(
      after,
      optional_backends
    ),
    intersect(
      before,
      optional_backends
    )
  )

  expect_s3_class(
    result,
    "data.frame"
  )

  expect_identical(
    newly_loaded_optional,
    character()
  )
})

.sequence_ext_covariate_hmm_input <- function(
  data,
  sequence_id_col,
  order_col,
  state_col,
  initial_covariate_cols,
  transition_covariate_cols,
  symbol_levels = NULL,
  initial_center = NULL,
  initial_scale = NULL,
  transition_center = NULL,
  transition_scale = NULL
) {
  all_covariates <- unique(c(initial_covariate_cols, transition_covariate_cols))
  if (is.list(data) && !is.data.frame(data) && !is.null(data$data)) data <- data$data
  if (!is.data.frame(data)) stop("`data` must be a data frame.", call. = FALSE)
  missing <- setdiff(all_covariates, names(data))
  if (length(missing) > 0L) {
    stop("Missing covariate columns: ", paste(missing, collapse = ", "), ".",
         call. = FALSE)
  }
  x <- .sequence_adv_data(
    data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    missing_state_policy = "error"
  )
  working <- x$data
  observed <- x$state_levels
  if (is.null(symbol_levels)) symbol_levels <- observed
  symbol_levels <- as.character(symbol_levels)
  if (length(symbol_levels) < 1L || anyNA(symbol_levels) ||
      any(!nzchar(trimws(symbol_levels))) || anyDuplicated(symbol_levels)) {
    stop("`symbol_levels` must contain unique non-missing symbols.", call. = FALSE)
  }
  missing_symbols <- setdiff(observed, symbol_levels)
  if (length(missing_symbols) > 0L) {
    stop("`symbol_levels` does not cover observed symbols: ",
         paste(missing_symbols, collapse = ", "), ".", call. = FALSE)
  }
  split_rows <- split(seq_len(nrow(working)),
                      factor(as.character(working[[sequence_id_col]]),
                             levels = x$sequence_ids), drop = TRUE)
  for (column in initial_covariate_cols) {
    inconsistent <- vapply(split_rows, function(rows) {
      length(unique(working[[column]][rows])) > 1L
    }, logical(1))
    if (any(inconsistent)) {
      stop("Initial covariate `", column,
           "` must remain constant within each sequence.", call. = FALSE)
    }
  }
  initial_data <- working[vapply(split_rows, function(rows) rows[1L], integer(1)),
                          , drop = FALSE]
  initial_design <- .sequence_ext_numeric_matrix(
    initial_data, initial_covariate_cols,
    center = initial_center, scale = initial_scale
  )
  transition_source_rows <- unlist(lapply(split_rows, function(rows) {
    if (length(rows) < 2L) integer() else rows[-length(rows)]
  }), use.names = FALSE)
  if (length(transition_source_rows) == 0L) {
    stop("At least one transition is required.", call. = FALSE)
  }
  transition_design_all <- .sequence_ext_numeric_matrix(
    working[transition_source_rows, , drop = FALSE],
    transition_covariate_cols,
    center = transition_center, scale = transition_scale
  )
  transition_design <- vector("list", length(split_rows))
  observations <- vector("list", length(split_rows))
  orders <- vector("list", length(split_rows))
  cursor <- 1L
  for (i in seq_along(split_rows)) {
    rows <- split_rows[[i]]
    observations[[i]] <- match(as.character(working[[state_col]][rows]), symbol_levels)
    orders[[i]] <- working[[order_col]][rows]
    n_transition <- max(length(rows) - 1L, 0L)
    if (n_transition > 0L) {
      transition_design[[i]] <- transition_design_all$matrix[
        seq.int(cursor, cursor + n_transition - 1L), , drop = FALSE
      ]
      cursor <- cursor + n_transition
    } else {
      transition_design[[i]] <- matrix(
        numeric(), nrow = 0L, ncol = ncol(transition_design_all$matrix),
        dimnames = list(NULL, colnames(transition_design_all$matrix))
      )
    }
  }
  names(observations) <- x$sequence_ids
  names(orders) <- x$sequence_ids
  names(transition_design) <- x$sequence_ids
  list(
    data = working,
    sequence_ids = x$sequence_ids,
    observations = observations,
    orders = orders,
    symbol_levels = symbol_levels,
    initial_design = initial_design$matrix,
    transition_design = transition_design,
    initial_center = initial_design$center,
    initial_scale = initial_design$scale,
    transition_center = transition_design_all$center,
    transition_scale = transition_design_all$scale,
    initial_covariate_cols = initial_covariate_cols,
    transition_covariate_cols = transition_covariate_cols,
    columns = list(sequence_id = sequence_id_col, order = order_col,
                   state = state_col)
  )
}

.sequence_ext_covariate_parameters <- function(input, n_states, seed,
                                               emission_probs = NULL) {
  initial_p <- ncol(input$initial_design)
  transition_p <- ncol(input$transition_design[[1L]])
  generated <- .sequence_adv_with_seed(seed, {
    list(
      initial_coef = matrix(stats::rnorm(initial_p * n_states, sd = 0.02),
                            nrow = initial_p, ncol = n_states),
      transition_coef = lapply(seq_len(n_states), function(i) {
        matrix(stats::rnorm(transition_p * n_states, sd = 0.02),
               nrow = transition_p, ncol = n_states)
      }),
      emission = if (is.null(emission_probs)) {
        matrix(stats::rexp(n_states * length(input$symbol_levels)) + 0.1,
               nrow = n_states, ncol = length(input$symbol_levels))
      } else emission_probs
    )
  })
  generated$initial_coef[, n_states] <- 0
  generated$transition_coef <- lapply(generated$transition_coef, function(x) {
    x[, n_states] <- 0
    x
  })
  if (!is.matrix(generated$emission) ||
      !all(dim(generated$emission) == c(n_states, length(input$symbol_levels))) ||
      anyNA(generated$emission) || any(!is.finite(generated$emission)) ||
      any(generated$emission < 0)) {
    stop("Invalid `emission_probs`.", call. = FALSE)
  }
  list(
    initial_coef = generated$initial_coef,
    transition_coef = generated$transition_coef,
    emission = .sequence_adv_row_normalise(generated$emission)
  )
}

.sequence_ext_covariate_probabilities <- function(initial_x, transition_x,
                                                  parameters) {
  initial <- .sequence_ext_softmax(initial_x %*% parameters$initial_coef)
  n_time <- nrow(transition_x) + 1L
  n_states <- ncol(parameters$initial_coef)
  transition <- array(0, dim = c(max(n_time - 1L, 0L), n_states, n_states))
  if (n_time > 1L) {
    for (t in seq_len(n_time - 1L)) {
      for (origin in seq_len(n_states)) {
        transition[t, origin, ] <- .sequence_ext_softmax(
          transition_x[t, ] %*% parameters$transition_coef[[origin]]
        )
      }
    }
  }
  list(initial = initial, transition = transition)
}

#' Fit a covariate-dependent categorical hidden Markov model
#'
#' Fits a categorical HMM whose initial-state and transition probabilities may
#' depend on explicitly declared numeric covariates. Multinomial-logit
#' coefficients are estimated inside the EM algorithm with a small ridge
#' penalty. Emission probabilities remain time-homogeneous.
#'
#' @param data Long-format sequence data.
#' @param n_states Number of latent states.
#' @param initial_covariate_cols Numeric sequence-constant covariates for
#'   initial-state probabilities.
#' @param transition_covariate_cols Numeric row-level covariates for transition
#'   probabilities.
#' @param sequence_id_col,order_col,state_col Core sequence columns.
#' @param symbol_levels Optional observed-symbol order.
#' @param state_names Optional latent-state names.
#' @param emission_probs Optional starting emission matrix.
#' @param max_iter Maximum EM iterations.
#' @param inner_maxit Maximum BFGS iterations in each multinomial M-step.
#' @param tolerance Relative log-likelihood tolerance.
#' @param pseudocount Emission smoothing count.
#' @param ridge Non-negative coefficient penalty.
#' @param seed Reproducibility seed.
#' @param keep_posteriors Retain final posteriors.
#'
#' @return An object of class `gp3_covariate_sequence_hmm`.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(paste0("s", 1:8), each = 5L),
#'   sequence_order = rep(1:5, times = 8L),
#'   state = rep(c("A", "B", "C", "B", "A"), times = 8L),
#'   condition = rep(rep(c(0, 1), each = 4L), each = 5L),
#'   time_scaled = rep(seq(-1, 1, length.out = 5L), times = 8L)
#' )
#' fit_covariate_sequence_hmm(
#'   sequences, 2L,
#'   initial_covariate_cols = "condition",
#'   transition_covariate_cols = c("condition", "time_scaled"),
#'   max_iter = 3L, inner_maxit = 10L, seed = 1L
#' )
#' @export
fit_covariate_sequence_hmm <- function(
  data,
  n_states,
  initial_covariate_cols = NULL,
  transition_covariate_cols = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  symbol_levels = NULL,
  state_names = NULL,
  emission_probs = NULL,
  max_iter = 100L,
  inner_maxit = 100L,
  tolerance = 1e-6,
  pseudocount = 1e-6,
  ridge = 1e-6,
  seed = 1L,
  keep_posteriors = FALSE
) {
  .sequence_adv_scalar_number(n_states, "n_states", lower = 2, integer = TRUE)
  .sequence_adv_scalar_number(max_iter, "max_iter", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(inner_maxit, "inner_maxit", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(tolerance, "tolerance", lower = 0)
  .sequence_adv_scalar_number(pseudocount, "pseudocount", lower = 0)
  .sequence_adv_scalar_number(ridge, "ridge", lower = 0)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  .sequence_adv_scalar_logical(keep_posteriors, "keep_posteriors")
  initial_covariate_cols <- if (is.null(initial_covariate_cols)) character() else initial_covariate_cols
  transition_covariate_cols <- if (is.null(transition_covariate_cols)) character() else transition_covariate_cols
  if (!is.character(initial_covariate_cols) || anyNA(initial_covariate_cols) ||
      any(!nzchar(initial_covariate_cols)) || anyDuplicated(initial_covariate_cols)) {
    stop("Invalid `initial_covariate_cols`.", call. = FALSE)
  }
  if (!is.character(transition_covariate_cols) || anyNA(transition_covariate_cols) ||
      any(!nzchar(transition_covariate_cols)) || anyDuplicated(transition_covariate_cols)) {
    stop("Invalid `transition_covariate_cols`.", call. = FALSE)
  }
  input <- .sequence_ext_covariate_hmm_input(
    data, sequence_id_col, order_col, state_col,
    initial_covariate_cols, transition_covariate_cols, symbol_levels
  )
  if (is.null(state_names)) state_names <- paste0("latent_", seq_len(n_states))
  if (!is.character(state_names) || length(state_names) != n_states ||
      anyNA(state_names) || any(!nzchar(trimws(state_names))) ||
      anyDuplicated(state_names)) {
    stop("`state_names` must uniquely name all latent states.", call. = FALSE)
  }
  parameters <- .sequence_ext_covariate_parameters(input, n_states, seed,
                                                   emission_probs)
  colnames(parameters$initial_coef) <- state_names
  rownames(parameters$initial_coef) <- colnames(input$initial_design)
  for (origin in seq_len(n_states)) {
    colnames(parameters$transition_coef[[origin]]) <- state_names
    rownames(parameters$transition_coef[[origin]]) <-
      colnames(input$transition_design[[1L]])
  }
  history <- numeric(max_iter)
  converged <- FALSE
  previous <- -Inf
  optimizer_convergence <- integer()
  final_posteriors <- NULL
  for (iteration in seq_len(max_iter)) {
    n_sequences <- length(input$observations)
    initial_counts <- matrix(0, nrow = n_sequences, ncol = n_states,
                             dimnames = list(input$sequence_ids, state_names))
    transition_x <- do.call(rbind, input$transition_design)
    transition_counts <- lapply(seq_len(n_states), function(origin) {
      matrix(0, nrow = nrow(transition_x), ncol = n_states,
             dimnames = list(NULL, state_names))
    })
    emission_count <- matrix(pseudocount, nrow = n_states,
                             ncol = length(input$symbol_levels))
    sequence_loglik <- numeric(n_sequences)
    posterior_current <- vector("list", n_sequences)
    cursor <- 1L
    for (s in seq_len(n_sequences)) {
      obs <- input$observations[[s]]
      probabilities <- .sequence_ext_covariate_probabilities(
        input$initial_design[s, ], input$transition_design[[s]], parameters
      )
      emission_likelihood <- t(parameters$emission[, obs, drop = FALSE])
      fb <- .sequence_ext_forward_backward(
        emission_likelihood, probabilities$initial, probabilities$transition
      )
      initial_counts[s, ] <- fb$gamma[1L, ]
      n_transition <- max(length(obs) - 1L, 0L)
      if (n_transition > 0L) {
        rows <- seq.int(cursor, cursor + n_transition - 1L)
        for (origin in seq_len(n_states)) {
          current_xi <- fb$xi[, origin, , drop = FALSE]
          dim(current_xi) <- c(n_transition, n_states)
          transition_counts[[origin]][rows, ] <- current_xi
        }
        cursor <- cursor + n_transition
      }
      for (t in seq_along(obs)) {
        emission_count[, obs[t]] <- emission_count[, obs[t]] + fb$gamma[t, ]
      }
      sequence_loglik[s] <- fb$log_likelihood
      posterior_current[[s]] <- fb
    }
    current <- sum(sequence_loglik)
    history[iteration] <- current
    initial_fit <- .sequence_ext_fit_softmax(
      input$initial_design, initial_counts,
      start = parameters$initial_coef, ridge = ridge, maxit = inner_maxit
    )
    parameters$initial_coef <- initial_fit$coefficients
    transition_fits <- vector("list", n_states)
    for (origin in seq_len(n_states)) {
      transition_fits[[origin]] <- .sequence_ext_fit_softmax(
        transition_x, transition_counts[[origin]],
        start = parameters$transition_coef[[origin]],
        ridge = ridge, maxit = inner_maxit
      )
      parameters$transition_coef[[origin]] <- transition_fits[[origin]]$coefficients
    }
    optimizer_convergence <- c(
      initial_fit$convergence,
      vapply(transition_fits, function(x) x$convergence, integer(1))
    )
    parameters$emission <- .sequence_adv_row_normalise(emission_count)
    final_posteriors <- posterior_current
    if (iteration > 1L) {
      relative <- abs(current - previous) / max(1, abs(previous))
      if (relative <= tolerance) {
        converged <- TRUE
        history <- history[seq_len(iteration)]
        break
      }
    }
    previous <- current
    if (iteration == max_iter) history <- history[seq_len(iteration)]
  }
  final_loglik <- numeric(length(input$observations))
  final_posteriors <- vector("list", length(input$observations))
  for (s in seq_along(input$observations)) {
    probabilities <- .sequence_ext_covariate_probabilities(
      input$initial_design[s, ], input$transition_design[[s]], parameters
    )
    emission_likelihood <- t(parameters$emission[, input$observations[[s]], drop = FALSE])
    fb <- .sequence_ext_forward_backward(
      emission_likelihood, probabilities$initial, probabilities$transition
    )
    final_loglik[s] <- fb$log_likelihood
    final_posteriors[[s]] <- fb
  }
  log_likelihood <- sum(final_loglik)
  history[length(history)] <- log_likelihood
  dimnames(parameters$emission) <- list(state_names, input$symbol_levels)
  n_parameters <- ncol(input$initial_design) * (n_states - 1L) +
    n_states * ncol(input$transition_design[[1L]]) * (n_states - 1L) +
    n_states * (length(input$symbol_levels) - 1L)
  n_observations <- sum(lengths(input$observations))
  result <- list(
    initial_coefficients = parameters$initial_coef,
    transition_coefficients = stats::setNames(parameters$transition_coef,
                                               state_names),
    emission_probs = parameters$emission,
    state_names = state_names,
    symbol_names = input$symbol_levels,
    sequence_ids = input$sequence_ids,
    sequence_log_likelihoods = stats::setNames(final_loglik, input$sequence_ids),
    log_likelihood = log_likelihood,
    iterations = length(history),
    converged = converged,
    optimizer_convergence = optimizer_convergence,
    tolerance = tolerance,
    pseudocount = pseudocount,
    ridge = ridge,
    log_likelihood_history = history,
    n_parameters = as.integer(n_parameters),
    n_observations = as.integer(n_observations),
    aic = -2 * log_likelihood + 2 * n_parameters,
    bic = -2 * log_likelihood + log(max(1, n_observations)) * n_parameters,
    seed = as.integer(seed),
    initial_covariate_cols = input$initial_covariate_cols,
    transition_covariate_cols = input$transition_covariate_cols,
    initial_center = input$initial_center,
    initial_scale = input$initial_scale,
    transition_center = input$transition_center,
    transition_scale = input$transition_scale,
    training_observations = input$observations,
    training_orders = input$orders,
    training_initial_design = input$initial_design,
    training_transition_design = input$transition_design,
    training_data = input$data,
    columns = input$columns,
    posteriors = if (keep_posteriors) final_posteriors else NULL,
    call = match.call()
  )
  class(result) <- c("gp3_covariate_sequence_hmm", "list")
  result
}

#' Predict covariate-dependent transition probabilities
#'
#' @param model A fitted covariate HMM.
#' @param newdata Data frame containing transition covariates.
#'
#' @return A long data frame with one row per input row, origin state, and
#'   destination state.
#' @examples
#' # See `fit_covariate_sequence_hmm()`.
#' @export
predict_covariate_transition_probabilities <- function(model, newdata) {
  if (!inherits(model, "gp3_covariate_sequence_hmm")) {
    stop("`model` must be created by `fit_covariate_sequence_hmm()`.",
         call. = FALSE)
  }
  if (!is.data.frame(newdata) || nrow(newdata) < 1L) {
    stop("`newdata` must be a non-empty data frame.", call. = FALSE)
  }
  design <- .sequence_ext_numeric_matrix(
    newdata,
    model$transition_covariate_cols,
    center = model$transition_center,
    scale = model$transition_scale
  )$matrix
  rows <- list()
  h <- 0L
  for (r in seq_len(nrow(design))) {
    for (origin in seq_along(model$state_names)) {
      probabilities <- .sequence_ext_softmax(
        design[r, ] %*% model$transition_coefficients[[origin]]
      )
      for (destination in seq_along(model$state_names)) {
        h <- h + 1L
        rows[[h]] <- data.frame(
          row = r,
          from_state = model$state_names[origin],
          to_state = model$state_names[destination],
          probability = probabilities[destination],
          stringsAsFactors = FALSE
        )
      }
    }
  }
  do.call(rbind, rows)
}

#' Decode states from a covariate-dependent HMM
#'
#' @param model A fitted covariate HMM.
#' @param data Optional new data. Training data are used when omitted.
#' @param sequence_id_col,order_col,state_col Core columns for new data.
#' @param method `"viterbi"` or `"posterior"`.
#'
#' @return A long decoded-state data frame.
#' @examples
#' # See `fit_covariate_sequence_hmm()`.
#' @export
decode_covariate_sequence_states <- function(
  model,
  data = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  method = c("viterbi", "posterior")
) {
  if (!inherits(model, "gp3_covariate_sequence_hmm")) {
    stop("`model` must be created by `fit_covariate_sequence_hmm()`.",
         call. = FALSE)
  }
  method <- match.arg(method)
  if (is.null(data)) {
    input <- list(
      sequence_ids = model$sequence_ids,
      observations = model$training_observations,
      orders = model$training_orders,
      initial_design = model$training_initial_design,
      transition_design = model$training_transition_design
    )
  } else {
    input <- .sequence_ext_covariate_hmm_input(
      data, sequence_id_col, order_col, state_col,
      model$initial_covariate_cols, model$transition_covariate_cols,
      symbol_levels = model$symbol_names,
      initial_center = model$initial_center,
      initial_scale = model$initial_scale,
      transition_center = model$transition_center,
      transition_scale = model$transition_scale
    )
  }
  rows <- list()
  h <- 0L
  parameters <- list(
    initial_coef = model$initial_coefficients,
    transition_coef = model$transition_coefficients,
    emission = model$emission_probs
  )
  for (s in seq_along(input$sequence_ids)) {
    probabilities <- .sequence_ext_covariate_probabilities(
      input$initial_design[s, ], input$transition_design[[s]], parameters
    )
    emission_likelihood <- t(model$emission_probs[, input$observations[[s]], drop = FALSE])
    fb <- .sequence_ext_forward_backward(
      emission_likelihood, probabilities$initial, probabilities$transition
    )
    decoded <- if (method == "viterbi") {
      .sequence_ext_viterbi(
        emission_likelihood, probabilities$initial, probabilities$transition
      )$path
    } else max.col(fb$gamma, ties.method = "first")
    posterior_probability <- fb$gamma[cbind(seq_along(decoded), decoded)]
    for (t in seq_along(decoded)) {
      h <- h + 1L
      rows[[h]] <- data.frame(
        sequence_id = input$sequence_ids[s],
        sequence_order = input$orders[[s]][t],
        observed_state = model$symbol_names[input$observations[[s]][t]],
        latent_state = model$state_names[decoded[t]],
        posterior_probability = posterior_probability[t],
        decoding_method = method,
        stringsAsFactors = FALSE
      )
    }
  }
  result <- do.call(rbind, rows)
  row.names(result) <- NULL
  result
}

#' Summarise a covariate-dependent HMM
#'
#' @param model A fitted covariate HMM.
#'
#' @return A list of fit, coefficient, and emission summaries.
#' @examples
#' # See `fit_covariate_sequence_hmm()`.
#' @export
summarise_covariate_sequence_hmm <- function(model) {
  if (!inherits(model, "gp3_covariate_sequence_hmm")) {
    stop("`model` must be created by `fit_covariate_sequence_hmm()`.",
         call. = FALSE)
  }
  transition <- do.call(rbind, lapply(seq_along(model$transition_coefficients),
                                     function(origin) {
    current <- model$transition_coefficients[[origin]]
    data.frame(
      from_state = model$state_names[origin],
      covariate = rep(rownames(current), times = ncol(current)),
      to_state = rep(colnames(current), each = nrow(current)),
      coefficient = as.numeric(current),
      stringsAsFactors = FALSE
    )
  }))
  initial <- data.frame(
    covariate = rep(rownames(model$initial_coefficients),
                    times = ncol(model$initial_coefficients)),
    latent_state = rep(colnames(model$initial_coefficients),
                       each = nrow(model$initial_coefficients)),
    coefficient = as.numeric(model$initial_coefficients),
    stringsAsFactors = FALSE
  )
  list(
    fit = data.frame(
      n_states = length(model$state_names),
      n_sequences = length(model$sequence_ids),
      n_observations = model$n_observations,
      log_likelihood = model$log_likelihood,
      aic = model$aic,
      bic = model$bic,
      iterations = model$iterations,
      converged = model$converged,
      optimizers_converged = all(model$optimizer_convergence == 0L),
      ridge = model$ridge,
      stringsAsFactors = FALSE
    ),
    initial_coefficients = initial,
    transition_coefficients = transition,
    emission = as.data.frame(as.table(model$emission_probs),
                              stringsAsFactors = FALSE)
  )
}

test_that("covariate HMM fits with explicit design matrices", {
  data <- make_extension_sequence_data()
  fit <- fit_covariate_sequence_hmm(
    data,
    n_states = 2L,
    initial_covariate_cols = "condition_numeric",
    transition_covariate_cols = "condition_numeric",
    max_iter = 3L,
    inner_maxit = 10L,
    seed = 12L
  )
  expect_s3_class(fit, "gp3_covariate_sequence_hmm")
  expect_equal(dim(fit$emission_probs), c(2L, 4L))
  expect_true(is.finite(fit$log_likelihood))
  newdata <- data.frame(condition_numeric = c(0, 1))
  prediction <- predict_covariate_transition_probabilities(fit, newdata)
  expect_true(all(c("row", "from_state", "to_state", "probability") %in% names(prediction)))
  expect_equal(nrow(prediction), 8L)
  decoded <- decode_covariate_sequence_states(fit)
  expect_equal(nrow(decoded), nrow(data))
  summary <- summarise_covariate_sequence_hmm(fit)
  expect_true(all(c("fit", "initial_coefficients", "transition_coefficients", "emission") %in% names(summary)))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----data---------------------------------------------------------------------
sequences <- data.frame(
  sequence_id = rep(paste0("s", 1:6), each = 5L),
  sequence_order = rep(1:5, times = 6L),
  state = c(
    "A", "B", "C", "D", "E",
    "A", "C", "B", "D", "E",
    "A", "B", "D", "C", "E",
    "E", "D", "C", "B", "A",
    "E", "C", "D", "B", "A",
    "E", "D", "B", "C", "A"
  ),
  group = rep(rep(c("forward", "reverse"), each = 3L), each = 5L),
  stringsAsFactors = FALSE
)

## ----extract------------------------------------------------------------------
occurrences <- extract_sequence_subsequences(
  sequences,
  metadata_cols = "group",
  min_length = 2L,
  max_length = 3L,
  max_gap = 2L,
  max_span = 4L,
  repeated_state_policy = "preserve"
)
head(occurrences)
attributes(occurrences)[c("n_sequences", "settings")]

## ----summary------------------------------------------------------------------
subsequence_summary <- summarise_sequence_subsequences(occurrences)
head(subsequence_summary, 10L)

frequent <- filter_sequence_subsequences(
  subsequence_summary,
  min_sequences = 2L,
  min_prevalence = 0.25,
  top_n = 12L
)
frequent

## ----compare------------------------------------------------------------------
comparison <- compare_sequence_subsequences(
  occurrences,
  group_col = "group",
  p_adjust = "holm"
)
head(comparison, 10L)

## ----plot, fig.width=7, fig.height=5------------------------------------------
plot_sequence_subsequences(frequent, metric = "sequence_prevalence")

make_extension_sequence_data <- function() {
  paths <- list(
    s01 = c("A", "B", "C", "D", "D"),
    s02 = c("A", "B", "C", "D", "C"),
    s03 = c("A", "B", "B", "C", "D"),
    s04 = c("A", "C", "C", "D", "D"),
    s05 = c("D", "C", "B", "A", "A"),
    s06 = c("D", "C", "B", "A", "B"),
    s07 = c("D", "C", "C", "B", "A"),
    s08 = c("D", "B", "B", "A", "A")
  )
  rows <- lapply(seq_along(paths), function(i) {
    data.frame(
      participant_id = paste0("p", i),
      sequence_id = names(paths)[i],
      sequence_order = seq_along(paths[[i]]),
      state = paths[[i]],
      group = rep(c("g1", "g2"), each = 4L)[i],
      condition_numeric = rep(c(0, 1), each = 4L)[i],
      time_scaled = seq(-1, 1, length.out = length(paths[[i]])),
      channel_context = c("x", "x", "y", "y", "z"),
      stringsAsFactors = FALSE
    )
  })
  do.call(rbind, rows)
}

make_extension_panel_data <- function() {
  base <- make_extension_sequence_data()
  first <- base
  first$sequence_id <- paste0(first$sequence_id, "_w1")
  first$occasion <- 1L
  second <- base
  second$sequence_id <- paste0(second$sequence_id, "_w2")
  second$occasion <- 2L
  second$state[second$sequence_order == 2L] <- ifelse(
    second$group[second$sequence_order == 2L] == "g1", "C", "B"
  )
  rbind(first, second)
}

.sequence_test_case <- function(
  case = c(
    "minimal",
    "empty",
    "single_row",
    "single_sequence",
    "single_state",
    "equal_sequences",
    "variable_length",
    "unicode_states",
    "whitespace_states",
    "high_repetition",
    "missing_internal",
    "duplicate_positions",
    "unordered_rows",
    "order_gaps",
    "noninteger_order",
    "negative_duration",
    "zero_duration",
    "inconsistent_metadata"
  ),
  seed = 2026L
) {
  case <- match.arg(case)

  .sequence_check_seed(
    seed,
    allow_null = FALSE
  )

  base <- data.frame(
    sequence_id = rep(
      c("s1", "s2", "s3", "s4"),
      each = 4L
    ),
    sequence_order = rep(
      1:4,
      times = 4L
    ),
    state = c(
      "A", "B", "C", "D",
      "A", "B", "C", "C",
      "D", "C", "B", "A",
      "D", "C", "A", "A"
    ),
    duration = rep(
      c(1, 2, 1, 3),
      times = 4L
    ),
    group = rep(
      c("g1", "g2"),
      each = 8L
    ),
    stringsAsFactors = FALSE
  )

  if (case == "minimal") {
    return(base)
  }

  if (case == "empty") {
    return(
      base[0L, , drop = FALSE]
    )
  }

  if (case == "single_row") {
    return(
      base[1L, , drop = FALSE]
    )
  }

  if (case == "single_sequence") {
    return(
      base[
        base$sequence_id == "s1",
        ,
        drop = FALSE
      ]
    )
  }

  if (case == "single_state") {
    out <- base
    out$state <- "A"
    return(out)
  }

  if (case == "equal_sequences") {
    out <- base
    out$state <- rep(
      c("A", "B", "C", "D"),
      times = 4L
    )
    return(out)
  }

  if (case == "variable_length") {
    return(
      base[
        -c(8L, 15L, 16L),
        ,
        drop = FALSE
      ]
    )
  }

  if (case == "unicode_states") {
    out <- base
    out$state[
      out$state == "A"
    ] <- "\u0391"
    out$state[
      out$state == "B"
    ] <- "\u03b2"
    return(out)
  }

  if (case == "whitespace_states") {
    out <- base
    out$state[2L] <- " "
    return(out)
  }

  if (case == "high_repetition") {
    out <- base
    out$state <- rep(
      c("A", "A", "A", "B"),
      times = 4L
    )
    return(out)
  }

  if (case == "missing_internal") {
    out <- base
    out$state[2L] <- NA_character_
    return(out)
  }

  if (case == "duplicate_positions") {
    out <- base
    out$sequence_order[2L] <-
      out$sequence_order[1L]
    return(out)
  }

  if (case == "unordered_rows") {
    out <- base
    index <- c(
      2L,
      1L,
      seq.int(
        3L,
        nrow(base)
      )
    )
    return(
      out[
        index,
        ,
        drop = FALSE
      ]
    )
  }

  if (case == "order_gaps") {
    out <- base
    out$sequence_order[
      out$sequence_id == "s1"
    ] <- c(1, 2, 4, 5)
    return(out)
  }

  if (case == "noninteger_order") {
    out <- base
    out$sequence_order[2L] <- 1.5
    return(out)
  }

  if (case == "negative_duration") {
    out <- base
    out$duration[2L] <- -1
    return(out)
  }

  if (case == "zero_duration") {
    out <- base
    out$duration[2L] <- 0
    return(out)
  }

  if (case == "inconsistent_metadata") {
    out <- base
    out$group[2L] <- "other"
    return(out)
  }

  stop(
    "Unhandled sequence test case.",
    call. = FALSE
  )
}

test_that("distance is invariant to input row order with explicit order", {
  x <- .sequence_test_case("minimal")

  set.seed(2026)
  shuffled <- x[
    sample(
      seq_len(
        nrow(x)
      )
    ),
    ,
    drop = FALSE
  ]

  original <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )
  permuted <- compute_sequence_distance(
    shuffled,
    method = "levenshtein"
  )

  expect_equal(
    as.matrix(original),
    as.matrix(permuted)
  )
  expect_setequal(
    attr(
      original,
      "state_levels"
    ),
    attr(
      permuted,
      "state_levels"
    )
  )
})

test_that("irrelevant metadata does not change sequence distance", {
  x <- .sequence_test_case("minimal")
  y <- x

  y$irrelevant_metadata <- paste0(
    "row",
    seq_len(
      nrow(y)
    )
  )

  dx <- compute_sequence_distance(
    x,
    method = "lcs"
  )
  dy <- compute_sequence_distance(
    y,
    method = "lcs"
  )

  expect_equal(
    as.matrix(dx),
    as.matrix(dy)
  )
})

test_that("state relabelling preserves Levenshtein geometry", {
  x <- .sequence_test_case("minimal")
  y <- x

  mapping <- c(
    A = "north",
    B = "south",
    C = "east",
    D = "west"
  )

  y$state <- unname(
    mapping[y$state]
  )

  dx <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )
  dy <- compute_sequence_distance(
    y,
    method = "levenshtein"
  )

  expect_equal(
    unname(
      as.matrix(dx)
    ),
    unname(
      as.matrix(dy)
    )
  )
})

# Internal helpers for the advanced gp3sequences roadmap.

.sequence_adv_scalar_character <- function(x, argument, allow_null = FALSE) {
  if (allow_null && is.null(x)) {
    return(invisible(NULL))
  }
  if (!is.character(x) || length(x) != 1L || is.na(x) || !nzchar(x)) {
    stop("`", argument, "` must be a single non-missing character value.", call. = FALSE)
  }
  invisible(NULL)
}

.sequence_adv_scalar_number <- function(x, argument, lower = -Inf, upper = Inf,
                                        integer = FALSE) {
  valid_scalar <- is.numeric(x) && length(x) == 1L && !is.na(x) && is.finite(x)
  invalid_integer <- if (integer && valid_scalar) {
    x != trunc(x) || x < -.Machine$integer.max || x > .Machine$integer.max
  } else FALSE
  if (!valid_scalar || x < lower || x > upper || invalid_integer) {
    stop("`", argument, "` has an invalid numeric value.", call. = FALSE)
  }
  invisible(NULL)
}

.sequence_adv_scalar_logical <- function(x, argument) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    stop("`", argument, "` must be a single non-missing logical value.",
         call. = FALSE)
  }
  invisible(NULL)
}

.sequence_adv_match_cols <- function(data, columns, argument, allow_null = TRUE) {
  if (allow_null && is.null(columns)) {
    return(character())
  }
  if (!is.character(columns) || anyNA(columns) || any(!nzchar(columns)) ||
      anyDuplicated(columns)) {
    stop("`", argument, "` must contain unique, non-missing column names.", call. = FALSE)
  }
  missing <- setdiff(columns, names(data))
  if (length(missing) > 0L) {
    stop("Missing columns in `", argument, "`: ", paste(missing, collapse = ", "), ".",
         call. = FALSE)
  }
  columns
}

.sequence_adv_data <- function(data, sequence_id_col = "sequence_id",
                               order_col = "sequence_order", state_col = "state",
                               metadata_cols = NULL,
                               missing_state_policy = c("error", "drop", "state"),
                               missing_state_label = "<MISSING>") {
  missing_state_policy <- match.arg(missing_state_policy)
  if (missing_state_policy == "state") {
    .sequence_adv_scalar_character(missing_state_label, "missing_state_label")
  }
  if (is.list(data) && !is.data.frame(data) && !is.null(data$data)) {
    data <- data$data
  }
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame or an object with a data-frame `data` component.",
         call. = FALSE)
  }
  .sequence_adv_scalar_character(sequence_id_col, "sequence_id_col")
  .sequence_adv_scalar_character(order_col, "order_col")
  .sequence_adv_scalar_character(state_col, "state_col")
  metadata_cols <- .sequence_adv_match_cols(data, metadata_cols, "metadata_cols")
  required <- c(sequence_id_col, order_col, state_col)
  overlap <- intersect(metadata_cols, required)
  if (length(overlap) > 0L) {
    stop("`metadata_cols` must not repeat core sequence columns: ",
         paste(overlap, collapse = ", "), ".", call. = FALSE)
  }
  missing <- setdiff(required, names(data))
  if (length(missing) > 0L) {
    stop("Missing required sequence columns: ", paste(missing, collapse = ", "), ".",
         call. = FALSE)
  }
  if (nrow(data) == 0L) {
    stop("`data` must contain at least one sequence row.", call. = FALSE)
  }
  if (anyDuplicated(names(data))) {
    stop("`data` contains duplicated column names.", call. = FALSE)
  }
  id <- data[[sequence_id_col]]
  ord <- data[[order_col]]
  state <- data[[state_col]]
  if (is.list(id) || is.list(state)) {
    stop("Sequence identifiers and states must be atomic vectors.", call. = FALSE)
  }
  if (!is.numeric(ord) || anyNA(ord) || any(!is.finite(ord))) {
    stop("The sequence order column must contain finite, non-missing numeric values.",
         call. = FALSE)
  }
  id_text <- as.character(id)
  state_text <- as.character(state)
  missing_id <- is.na(id) | trimws(id_text) == ""
  if (any(missing_id)) {
    stop("Sequence identifiers must not be missing or blank.", call. = FALSE)
  }
  missing_state <- is.na(state) | trimws(state_text) == ""
  if (missing_state_policy == "error" && any(missing_state)) {
    stop("Missing states were found. Select an explicit non-error policy to continue.",
         call. = FALSE)
  }
  working <- data
  working$.gp3_adv_original_row <- seq_len(nrow(working))
  if (missing_state_policy == "drop") {
    working <- working[!missing_state, , drop = FALSE]
  } else if (missing_state_policy == "state") {
    replacement <- as.character(working[[state_col]])
    replacement[missing_state] <- missing_state_label
    working[[state_col]] <- replacement
  }
  if (nrow(working) == 0L) {
    stop("No sequence rows remain after applying the missing-state policy.", call. = FALSE)
  }
  keys <- paste(as.character(working[[sequence_id_col]]),
                format(working[[order_col]], digits = 17L, scientific = FALSE),
                sep = "\034")
  if (anyDuplicated(keys)) {
    stop("Duplicated sequence positions were found. Prepare the data before advanced analysis.",
         call. = FALSE)
  }
  index <- order(as.character(working[[sequence_id_col]]), working[[order_col]],
                 working$.gp3_adv_original_row, method = "radix")
  working <- working[index, , drop = FALSE]
  row.names(working) <- NULL
  seq_ids <- unique(as.character(working[[sequence_id_col]]))
  split_rows <- split(seq_len(nrow(working)),
                      factor(as.character(working[[sequence_id_col]]), levels = seq_ids),
                      drop = TRUE)
  sequences <- lapply(split_rows, function(rows) as.character(working[[state_col]][rows]))
  orders <- lapply(split_rows, function(rows) working[[order_col]][rows])
  names(sequences) <- seq_ids
  names(orders) <- seq_ids
  observed_states <- unique(as.character(working[[state_col]]))
  state_levels <- if (is.factor(working[[state_col]])) {
    levels(working[[state_col]])[levels(working[[state_col]]) %in% observed_states]
  } else {
    sort(observed_states, method = "radix")
  }
  metadata <- NULL
  if (length(metadata_cols) > 0L) {
    metadata <- do.call(rbind, lapply(seq_along(split_rows), function(i) {
      rows <- split_rows[[i]]
      values <- working[rows, metadata_cols, drop = FALSE]
      inconsistent <- vapply(values, function(x) {
        encoded <- as.character(x)
        encoded[is.na(x)] <- "<NA>"
        length(unique(encoded)) > 1L
      }, logical(1))
      if (any(inconsistent)) {
        stop("Metadata vary within sequence `", seq_ids[i], "`: ",
             paste(names(inconsistent)[inconsistent], collapse = ", "), ".", call. = FALSE)
      }
      out <- values[1L, , drop = FALSE]
      out[[sequence_id_col]] <- seq_ids[i]
      out[c(sequence_id_col, metadata_cols)]
    }))
    row.names(metadata) <- NULL
  }
  list(
    data = working,
    sequences = sequences,
    orders = orders,
    sequence_ids = seq_ids,
    state_levels = state_levels,
    metadata = metadata,
    columns = list(sequence_id = sequence_id_col, order = order_col, state = state_col,
                   metadata = metadata_cols)
  )
}

.sequence_adv_group_key <- function(data, group_cols) {
  if (length(group_cols) == 0L) {
    return(rep("__all__", nrow(data)))
  }
  values <- lapply(data[group_cols], function(x) {
    y <- as.character(x)
    y[is.na(y)] <- "<NA>"
    y
  })
  do.call(paste, c(values, sep = "\035"))
}

.sequence_adv_state_order <- function(values, state_levels = NULL) {
  observed <- unique(as.character(values[!is.na(values)]))
  if (is.null(state_levels)) {
    return(sort(observed, method = "radix"))
  }
  if (is.list(state_levels)) {
    stop("`state_levels` must be an atomic vector.", call. = FALSE)
  }
  state_levels <- as.character(state_levels)
  if (anyNA(state_levels) || any(!nzchar(trimws(state_levels))) ||
      anyDuplicated(state_levels)) {
    stop("`state_levels` must contain unique, non-missing, non-blank values.",
         call. = FALSE)
  }
  c(state_levels[state_levels %in% observed],
    sort(setdiff(observed, state_levels), method = "radix"))
}

.sequence_adv_tie <- function(states, weights, state_levels, tie_method) {
  totals <- tapply(weights, states, sum)
  totals <- totals[!is.na(names(totals))]
  max_total <- max(totals)
  tied <- names(totals)[abs(totals - max_total) <= sqrt(.Machine$double.eps)]
  ordered <- .sequence_adv_state_order(tied, state_levels)
  selected <- switch(
    tie_method,
    first = ordered[1L],
    last = ordered[length(ordered)],
    missing = if (length(ordered) == 1L) ordered else NA_character_,
    all = paste(ordered, collapse = " | ")
  )
  list(selected = selected, tied = ordered, total = max_total,
       agreement = max_total / sum(weights))
}

.sequence_adv_pair_grid <- function(ids) {
  if (length(ids) < 2L) {
    return(matrix(integer(), nrow = 0L, ncol = 2L))
  }
  t(utils::combn(seq_along(ids), 2L))
}

.sequence_adv_edit_distance <- function(a, b, indel_cost = 1,
                                        substitution_cost = 1,
                                        substitution_matrix = NULL) {
  n <- length(a)
  m <- length(b)
  d <- matrix(0, nrow = n + 1L, ncol = m + 1L)
  d[, 1L] <- (0:n) * indel_cost
  d[1L, ] <- (0:m) * indel_cost
  if (n == 0L || m == 0L) {
    return(d[n + 1L, m + 1L])
  }
  cost_fun <- function(x, y) {
    if (identical(x, y)) return(0)
    if (is.null(substitution_matrix)) return(substitution_cost)
    if (!(x %in% rownames(substitution_matrix)) || !(y %in% colnames(substitution_matrix))) {
      stop("The substitution matrix does not cover all observed states.", call. = FALSE)
    }
    substitution_matrix[x, y]
  }
  for (i in seq_len(n)) {
    for (j in seq_len(m)) {
      d[i + 1L, j + 1L] <- min(
        d[i, j + 1L] + indel_cost,
        d[i + 1L, j] + indel_cost,
        d[i, j] + cost_fun(a[i], b[j])
      )
    }
  }
  d[n + 1L, m + 1L]
}

.sequence_adv_lcs_length <- function(a, b) {
  n <- length(a)
  m <- length(b)
  if (n == 0L || m == 0L) return(0L)
  previous <- integer(m + 1L)
  for (i in seq_len(n)) {
    current <- integer(m + 1L)
    for (j in seq_len(m)) {
      current[j + 1L] <- if (identical(a[i], b[j])) {
        previous[j] + 1L
      } else {
        max(previous[j + 1L], current[j])
      }
    }
    previous <- current
  }
  previous[m + 1L]
}

.sequence_adv_transition_profile <- function(sequence, states, smoothing = 0) {
  p <- length(states)
  out <- matrix(smoothing, nrow = p, ncol = p, dimnames = list(states, states))
  if (length(sequence) > 1L) {
    for (i in seq_len(length(sequence) - 1L)) {
      out[sequence[i], sequence[i + 1L]] <- out[sequence[i], sequence[i + 1L]] + 1
    }
  }
  row_totals <- rowSums(out)
  nonzero <- row_totals > 0
  out[nonzero, ] <- out[nonzero, , drop = FALSE] / row_totals[nonzero]
  as.numeric(out)
}

.sequence_adv_distance_matrix <- function(x) {
  if (inherits(x, "dist")) {
    matrix_x <- as.matrix(x)
  } else if (is.matrix(x) && nrow(x) == ncol(x)) {
    matrix_x <- x
  } else {
    stop("A `dist` object or square distance matrix is required.", call. = FALSE)
  }
  if (!is.numeric(matrix_x) || anyNA(matrix_x) || any(!is.finite(matrix_x)) ||
      any(matrix_x < 0)) {
    stop("Sequence distances must be finite, non-missing, and non-negative.",
         call. = FALSE)
  }
  if (nrow(matrix_x) == 0L) {
    stop("The distance object must contain at least one sequence.", call. = FALSE)
  }
  row_ids <- rownames(matrix_x)
  col_ids <- colnames(matrix_x)
  if (is.null(row_ids) && is.null(col_ids)) {
    row_ids <- col_ids <- as.character(seq_len(nrow(matrix_x)))
  } else if (is.null(row_ids) || is.null(col_ids) || !identical(row_ids, col_ids) ||
             anyNA(row_ids) || any(!nzchar(row_ids)) || anyDuplicated(row_ids)) {
    stop("Distance rows and columns must have identical, unique sequence identifiers.",
         call. = FALSE)
  }
  rownames(matrix_x) <- row_ids
  colnames(matrix_x) <- col_ids
  tolerance <- sqrt(.Machine$double.eps)
  if (max(abs(matrix_x - t(matrix_x))) > tolerance) {
    stop("The distance matrix must be symmetric.", call. = FALSE)
  }
  if (max(abs(diag(matrix_x))) > tolerance) {
    stop("The distance-matrix diagonal must be zero.", call. = FALSE)
  }
  matrix_x
}

.sequence_adv_seed <- function(seed, offset = 0L) {
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  .sequence_adv_scalar_number(offset, "offset", lower = 0, integer = TRUE)
  modulus <- as.double(.Machine$integer.max)
  as.integer((as.double(seed) + as.double(offset)) %% modulus)
}

.sequence_adv_with_seed <- function(seed, code) {
  seed <- .sequence_adv_seed(seed)
  had_seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  if (had_seed) {
    old_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  }
  on.exit({
    if (had_seed) {
      assign(".Random.seed", old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }, add = TRUE)
  set.seed(as.integer(seed))
  force(code)
}

.sequence_adv_silhouette <- function(assignments, distance_matrix) {
  ids <- names(assignments)
  if (is.null(ids)) ids <- rownames(distance_matrix)
  assignments <- assignments[ids]
  out <- numeric(length(ids))
  names(out) <- ids
  clusters <- sort(unique(assignments))
  for (i in seq_along(ids)) {
    own <- assignments[i]
    same <- which(assignments == own)
    same <- setdiff(same, i)
    if (length(same) == 0L) {
      out[i] <- 0
      next
    }
    a <- mean(distance_matrix[i, same])
    other_means <- vapply(setdiff(clusters, own), function(k) {
      mean(distance_matrix[i, assignments == k])
    }, numeric(1))
    b <- if (length(other_means) == 0L) 0 else min(other_means)
    out[i] <- if (max(a, b) == 0) 0 else (b - a) / max(a, b)
  }
  out
}

.sequence_adv_row_normalise <- function(x, pseudocount = 0) {
  x <- x + pseudocount
  totals <- rowSums(x)
  zero <- totals <= 0 | !is.finite(totals)
  if (any(zero)) x[zero, ] <- 1
  x / rowSums(x)
}

.sequence_adv_vector_normalise <- function(x, pseudocount = 0) {
  x <- x + pseudocount
  total <- sum(x)
  if (!is.finite(total) || total <= 0) x[] <- 1
  x / sum(x)
}

.sequence_adv_forward_backward <- function(obs, initial, transition, emission) {
  n_states <- length(initial)
  n_time <- length(obs)
  alpha <- matrix(0, nrow = n_time, ncol = n_states)
  scales <- numeric(n_time)
  alpha[1L, ] <- initial * emission[, obs[1L]]
  scales[1L] <- sum(alpha[1L, ])
  if (!is.finite(scales[1L]) || scales[1L] <= 0) scales[1L] <- .Machine$double.xmin
  alpha[1L, ] <- alpha[1L, ] / scales[1L]
  if (n_time > 1L) {
    for (t in 2:n_time) {
      alpha[t, ] <- as.numeric(alpha[t - 1L, ] %*% transition) * emission[, obs[t]]
      scales[t] <- sum(alpha[t, ])
      if (!is.finite(scales[t]) || scales[t] <= 0) scales[t] <- .Machine$double.xmin
      alpha[t, ] <- alpha[t, ] / scales[t]
    }
  }
  beta <- matrix(1, nrow = n_time, ncol = n_states)
  if (n_time > 1L) {
    for (t in seq.int(n_time - 1L, 1L)) {
      beta[t, ] <- transition %*% (emission[, obs[t + 1L]] * beta[t + 1L, ])
      beta[t, ] <- beta[t, ] / scales[t + 1L]
    }
  }
  gamma <- alpha * beta
  gamma_totals <- rowSums(gamma)
  invalid_gamma <- !is.finite(gamma_totals) | gamma_totals <= 0
  if (any(invalid_gamma)) {
    gamma[invalid_gamma, ] <- 1 / n_states
    gamma_totals[invalid_gamma] <- 1
  }
  gamma <- gamma / gamma_totals
  xi <- array(0, dim = c(max(n_time - 1L, 0L), n_states, n_states))
  if (n_time > 1L) {
    for (t in seq_len(n_time - 1L)) {
      current <- outer(alpha[t, ], emission[, obs[t + 1L]] * beta[t + 1L, ]) * transition
      total <- sum(current)
      if (total > 0) current <- current / total
      xi[t, , ] <- current
    }
  }
  list(alpha = alpha, beta = beta, gamma = gamma, xi = xi,
       log_likelihood = sum(log(scales)))
}

.sequence_adv_viterbi <- function(obs, initial, transition, emission) {
  n_states <- length(initial)
  n_time <- length(obs)
  log_initial <- log(pmax(initial, .Machine$double.xmin))
  log_transition <- log(pmax(transition, .Machine$double.xmin))
  log_emission <- log(pmax(emission, .Machine$double.xmin))
  delta <- matrix(-Inf, nrow = n_time, ncol = n_states)
  psi <- matrix(0L, nrow = n_time, ncol = n_states)
  delta[1L, ] <- log_initial + log_emission[, obs[1L]]
  if (n_time > 1L) {
    for (t in 2:n_time) {
      for (j in seq_len(n_states)) {
        candidates <- delta[t - 1L, ] + log_transition[, j]
        psi[t, j] <- which.max(candidates)
        delta[t, j] <- max(candidates) + log_emission[j, obs[t]]
      }
    }
  }
  path <- integer(n_time)
  path[n_time] <- which.max(delta[n_time, ])
  if (n_time > 1L) {
    for (t in seq.int(n_time - 1L, 1L)) path[t] <- psi[t + 1L, path[t + 1L]]
  }
  list(path = path, log_probability = max(delta[n_time, ]))
}

.sequence_adv_require <- function(package, purpose) {
  if (!requireNamespace(package, quietly = TRUE)) {
    stop("Optional package `", package, "` is required to ", purpose, ".",
         call. = FALSE)
  }
  invisible(TRUE)
}

.sequence_adv_zero_column_frame <- function(n = 0L) {
  data.frame(
    .gp3_adv_row = seq_len(n),
    stringsAsFactors = FALSE
  )[, FALSE, drop = FALSE]
}

#' Create an aligned-position consensus sequence
#'
#' Computes a descriptive consensus state at each observed sequence position.
#' The function does not treat the consensus as a behavioural norm and does not
#' infer psychological or causal meaning.
#'
#' @param data A long-format data frame, a prepared gp3sequences result, or a
#'   data frame containing canonical columns.
#' @param sequence_id_col,order_col,state_col Column names defining sequences.
#' @param group_cols Optional columns defining independent consensus groups.
#' @param weight_col Optional non-negative row-weight column.
#' @param missing_state_policy One of `"exclude"`, `"state"`, or `"error"`.
#' @param missing_state_label Label used when missing states are retained.
#' @param tie_method Deterministic tie policy: `"first"`, `"last"`,
#'   `"missing"`, or `"all"`.
#' @param state_levels Optional preferred state ordering used for ties.
#' @param min_support Minimum number of contributing sequences at a position.
#'
#' @return A data frame of class `gp3_consensus_sequence` containing group
#'   columns, sequence position, consensus state, support counts and weights,
#'   agreement proportion, tie count, tied states, and total group sequences.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' create_consensus_sequence(sequences, group_cols = "group")
#'
#' @export
create_consensus_sequence <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  group_cols = NULL,
  weight_col = NULL,
  missing_state_policy = c("exclude", "state", "error"),
  missing_state_label = "<MISSING>",
  tie_method = c("first", "last", "missing", "all"),
  state_levels = NULL,
  min_support = 1L
) {
  missing_state_policy <- match.arg(missing_state_policy)
  tie_method <- match.arg(tie_method)
  .sequence_adv_scalar_number(min_support, "min_support", lower = 1, integer = TRUE)
  missing_input_policy <- switch(missing_state_policy,
                                 exclude = "drop", state = "state", error = "error")
  x <- .sequence_adv_data(
    data = data,
    sequence_id_col = sequence_id_col,
    order_col = order_col,
    state_col = state_col,
    metadata_cols = group_cols,
    missing_state_policy = missing_input_policy,
    missing_state_label = missing_state_label
  )
  group_cols <- .sequence_adv_match_cols(x$data, group_cols, "group_cols")
  if (!is.null(weight_col)) {
    .sequence_adv_scalar_character(weight_col, "weight_col")
    if (!(weight_col %in% names(x$data))) stop("`weight_col` is absent from `data`.", call. = FALSE)
    weights <- x$data[[weight_col]]
    if (!is.numeric(weights) || anyNA(weights) || any(!is.finite(weights)) || any(weights < 0)) {
      stop("`weight_col` must contain finite, non-negative numeric values.", call. = FALSE)
    }
  } else {
    weights <- rep(1, nrow(x$data))
  }
  working <- x$data
  working$.gp3_adv_weight <- weights
  working$.gp3_adv_group <- .sequence_adv_group_key(working, group_cols)
  state_order <- .sequence_adv_state_order(working[[state_col]], state_levels)
  groups <- split(seq_len(nrow(working)), working$.gp3_adv_group, drop = TRUE)
  output <- list()
  output_index <- 0L
  for (group_key in names(groups)) {
    rows_group <- groups[[group_key]]
    group_data <- working[rows_group, , drop = FALSE]
    positions <- sort(unique(group_data[[order_col]]))
    n_group_sequences <- length(unique(as.character(group_data[[sequence_id_col]])))
    for (position in positions) {
      rows <- rows_group[working[[order_col]][rows_group] == position]
      states <- as.character(working[[state_col]][rows])
      current_weights <- working$.gp3_adv_weight[rows]
      contributing <- current_weights > 0
      states <- states[contributing]
      current_weights <- current_weights[contributing]
      support_n <- length(unique(as.character(working[[sequence_id_col]][rows][contributing])))
      if (support_n < min_support || length(states) == 0L) next
      tie <- .sequence_adv_tie(states, current_weights, state_order, tie_method)
      output_index <- output_index + 1L
      group_values <- if (length(group_cols) == 0L) {
        .sequence_adv_zero_column_frame(1L)
      } else {
        group_data[1L, group_cols, drop = FALSE]
      }
      output[[output_index]] <- cbind(
        group_values,
        data.frame(
          sequence_order = position,
          consensus_state = tie$selected,
          support_n = as.integer(support_n),
          support_weight = sum(current_weights),
          agreement = tie$agreement,
          tie_n = as.integer(length(tie$tied)),
          tied_states = paste(tie$tied, collapse = " | "),
          n_sequences = as.integer(n_group_sequences),
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      )
    }
  }
  if (length(output) == 0L) {
    prefix <- if (length(group_cols) == 0L) {
      .sequence_adv_zero_column_frame(0L)
    } else {
      x$data[0L, group_cols, drop = FALSE]
    }
    result <- cbind(
      prefix,
      data.frame(
        sequence_order = numeric(), consensus_state = character(), support_n = integer(),
        support_weight = numeric(), agreement = numeric(), tie_n = integer(),
        tied_states = character(), n_sequences = integer(), stringsAsFactors = FALSE,
        check.names = FALSE
      )
    )
  } else {
    result <- do.call(rbind, output)
    row.names(result) <- NULL
  }
  class(result) <- c("gp3_consensus_sequence", "data.frame")
  attr(result, "group_cols") <- group_cols
  attr(result, "state_levels") <- state_order
  attr(result, "settings") <- list(
    missing_state_policy = missing_state_policy,
    missing_state_label = missing_state_label,
    tie_method = tie_method,
    min_support = as.integer(min_support),
    weight_col = weight_col
  )
  result
}

#' Summarise consensus agreement
#'
#' @param consensus A result from [create_consensus_sequence()].
#' @param by Summary level: `"overall"`, `"group"`, or `"position"`.
#' @param threshold Agreement threshold used to count low-agreement positions.
#'
#' @return A data frame with position counts, mean, median, minimum and maximum
#'   agreement, weighted agreement, tie counts, and low-agreement counts.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' consensus <- create_consensus_sequence(sequences)
#' summarise_consensus_agreement(consensus)
#'
#' @export
summarise_consensus_agreement <- function(consensus,
                                          by = c("overall", "group", "position"),
                                          threshold = 0.5) {
  by <- match.arg(by)
  .sequence_adv_scalar_number(threshold, "threshold", lower = 0, upper = 1)
  if (!inherits(consensus, "gp3_consensus_sequence") || !is.data.frame(consensus)) {
    stop("`consensus` must be created by `create_consensus_sequence()`.", call. = FALSE)
  }
  group_cols <- attr(consensus, "group_cols")
  if (by == "group" && length(group_cols) == 0L) {
    stop("`by = \"group\"` requires a consensus created with `group_cols`.",
         call. = FALSE)
  }
  split_cols <- switch(by, overall = character(), group = group_cols,
                       position = c(group_cols, "sequence_order"))
  if (nrow(consensus) == 0L) {
    prefix <- if (length(split_cols) == 0L) {
      .sequence_adv_zero_column_frame(0L)
    } else {
      consensus[0L, split_cols, drop = FALSE]
    }
    return(cbind(
      prefix,
      data.frame(
        n_positions = integer(), mean_agreement = numeric(), median_agreement = numeric(),
        min_agreement = numeric(), max_agreement = numeric(), weighted_agreement = numeric(),
        n_ties = integer(), n_below_threshold = integer(), threshold = numeric(),
        stringsAsFactors = FALSE, check.names = FALSE
      )
    ))
  }
  key <- .sequence_adv_group_key(consensus, split_cols)
  pieces <- split(seq_len(nrow(consensus)), key, drop = TRUE)
  result <- lapply(pieces, function(rows) {
    prefix <- if (length(split_cols) == 0L) {
      .sequence_adv_zero_column_frame(1L)
    } else {
      consensus[rows[1L], split_cols, drop = FALSE]
    }
    weighted <- if (sum(consensus$support_weight[rows]) > 0) {
      stats::weighted.mean(consensus$agreement[rows], consensus$support_weight[rows])
    } else NA_real_
    cbind(prefix, data.frame(
      n_positions = as.integer(length(rows)),
      mean_agreement = mean(consensus$agreement[rows]),
      median_agreement = stats::median(consensus$agreement[rows]),
      min_agreement = min(consensus$agreement[rows]),
      max_agreement = max(consensus$agreement[rows]),
      weighted_agreement = weighted,
      n_ties = as.integer(sum(consensus$tie_n[rows] > 1L)),
      n_below_threshold = as.integer(sum(consensus$agreement[rows] < threshold)),
      threshold = threshold,
      stringsAsFactors = FALSE,
      check.names = FALSE
    ))
  })
  out <- do.call(rbind, result)
  row.names(out) <- NULL
  out
}

#' Format consensus sequences as paths
#'
#' @param consensus A consensus result.
#' @param separator State separator.
#' @param include_order Include position labels.
#' @param include_agreement Include rounded agreement values.
#' @param digits Number of decimal places for agreement values.
#'
#' @return One row per consensus group with a formatted path and position count.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' consensus <- create_consensus_sequence(sequences)
#' format_consensus_sequence(consensus)
#'
#' @export
format_consensus_sequence <- function(consensus, separator = " -> ",
                                      include_order = FALSE,
                                      include_agreement = FALSE,
                                      digits = 3L) {
  if (!inherits(consensus, "gp3_consensus_sequence")) {
    stop("`consensus` must be created by `create_consensus_sequence()`.", call. = FALSE)
  }
  .sequence_adv_scalar_character(separator, "separator")
  .sequence_adv_scalar_logical(include_order, "include_order")
  .sequence_adv_scalar_logical(include_agreement, "include_agreement")
  .sequence_adv_scalar_number(digits, "digits", lower = 0, integer = TRUE)
  group_cols <- attr(consensus, "group_cols")
  if (nrow(consensus) == 0L) {
    prefix <- if (length(group_cols) == 0L) {
      .sequence_adv_zero_column_frame(0L)
    } else {
      consensus[0L, group_cols, drop = FALSE]
    }
    return(cbind(
      prefix,
      data.frame(path = character(), n_positions = integer(),
                 stringsAsFactors = FALSE, check.names = FALSE)
    ))
  }
  key <- .sequence_adv_group_key(consensus, group_cols)
  pieces <- split(seq_len(nrow(consensus)), key, drop = TRUE)
  result <- lapply(pieces, function(rows) {
    rows <- rows[order(consensus$sequence_order[rows], method = "radix")]
    labels <- consensus$consensus_state[rows]
    labels[is.na(labels)] <- "<TIE>"
    if (include_order) labels <- paste0(consensus$sequence_order[rows], ":", labels)
    if (include_agreement) labels <- paste0(labels, " [", format(round(consensus$agreement[rows], digits),
                                                                 nsmall = digits), "]")
    prefix <- if (length(group_cols) == 0L) {
      .sequence_adv_zero_column_frame(1L)
    } else {
      consensus[rows[1L], group_cols, drop = FALSE]
    }
    cbind(prefix, data.frame(path = paste(labels, collapse = separator),
                             n_positions = as.integer(length(rows)),
                             stringsAsFactors = FALSE, check.names = FALSE))
  })
  out <- do.call(rbind, result)
  row.names(out) <- NULL
  out
}

#' Plot a consensus sequence
#'
#' @param consensus A consensus result.
#' @param type `"agreement"` or `"states"`.
#' @param group Optional group value, encoded group key, or named list of
#'   values when grouped consensus was created. It is required when more than one
#'   consensus group is present.
#' @param main,xlab,ylab Plot labels.
#' @param ... Additional arguments passed to base graphics.
#'
#' @return The plotted data, invisibly.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' consensus <- create_consensus_sequence(sequences)
#' plot_consensus_sequence(consensus)
#'
#' @export
plot_consensus_sequence <- function(consensus, type = c("agreement", "states"),
                                    group = NULL, main = NULL,
                                    xlab = "Sequence position", ylab = NULL, ...) {
  type <- match.arg(type)
  if (!inherits(consensus, "gp3_consensus_sequence")) {
    stop("`consensus` must be created by `create_consensus_sequence()`.", call. = FALSE)
  }
  data <- consensus
  group_cols <- attr(consensus, "group_cols")
  keys <- .sequence_adv_group_key(data, group_cols)
  available_keys <- unique(keys)
  if (length(group_cols) > 0L && length(available_keys) > 1L && is.null(group)) {
    stop("Select one consensus group before plotting grouped results.", call. = FALSE)
  }
  if (!is.null(group)) {
    if (is.list(group)) {
      if (is.null(names(group)) || !setequal(names(group), group_cols) ||
          any(lengths(group) != 1L)) {
        stop("A list-valued `group` must provide one named value per group column.",
             call. = FALSE)
      }
      selected <- rep(TRUE, nrow(data))
      for (column in group_cols) {
        observed <- as.character(data[[column]])
        target <- as.character(group[[column]])
        selected <- selected & if (is.na(group[[column]])) is.na(data[[column]]) else
          !is.na(observed) & observed == target
      }
      data <- data[selected, , drop = FALSE]
    } else {
      .sequence_adv_scalar_character(as.character(group), "group")
      data <- data[keys == as.character(group), , drop = FALSE]
    }
  }
  if (nrow(data) == 0L) stop("No consensus positions are available to plot.", call. = FALSE)
  data <- data[order(data$sequence_order, method = "radix"), , drop = FALSE]
  if (type == "agreement") {
    if (is.null(main)) main <- "Consensus agreement by position"
    if (is.null(ylab)) ylab <- "Agreement proportion"
    graphics::plot(data$sequence_order, data$agreement, type = "b", ylim = c(0, 1),
                   xlab = xlab, ylab = ylab, main = main, ...)
    graphics::abline(h = c(0.5, 1), lty = c(3, 2))
  } else {
    if (is.null(main)) main <- "Consensus states by position"
    if (is.null(ylab)) ylab <- "Consensus state"
    display_state <- as.character(data$consensus_state)
    display_state[is.na(display_state)] <- "<TIE>"
    levels <- .sequence_adv_state_order(display_state, attr(consensus, "state_levels"))
    y <- match(display_state, levels)
    graphics::plot(data$sequence_order, y, type = "b", yaxt = "n", xlab = xlab,
                   ylab = ylab, main = main, ...)
    graphics::axis(2, at = seq_along(levels), labels = levels, las = 1)
  }
  invisible(data)
}

#' Compare sequence groups descriptively
#'
#' Produces descriptive state, transition and sequence-length comparisons.
#' No hypothesis tests or causal interpretations are produced.
#'
#' @param data Long-format sequence data.
#' @param group_col Grouping column that is constant within sequence.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param reference Optional reference group. When supplied, each other
#'   group is reported as `group_1` and the reference as `group_2`; differences
#'   and ratios are therefore other-minus-reference and other/reference. When
#'   omitted, all pairwise group contrasts are returned.
#' @param metrics Any of `"state"`, `"transition"`, and `"length"`.
#' @param include_self Include self-transitions.
#' @param transition_separator Separator used in reported transition labels. It
#'   must not occur inside an observed state label.
#' @param zero_policy Ratio policy when the reference value is zero.
#'
#' @return A list of class `gp3_sequence_group_comparison` containing group
#' summaries and pairwise descriptive contrasts.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' compare_sequence_groups(sequences, group_col = "group")
#'
#' @export
compare_sequence_groups <- function(
  data,
  group_col,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  reference = NULL,
  metrics = c("state", "transition", "length"),
  include_self = TRUE,
  transition_separator = " -> ",
  zero_policy = c("missing", "infinite")
) {
  zero_policy <- match.arg(zero_policy)
  metrics <- unique(match.arg(metrics, c("state", "transition", "length"), several.ok = TRUE))
  if (length(metrics) == 0L) stop("Select at least one comparison metric.", call. = FALSE)
  .sequence_adv_scalar_logical(include_self, "include_self")
  .sequence_adv_scalar_character(transition_separator, "transition_separator")
  .sequence_adv_scalar_character(group_col, "group_col")
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          metadata_cols = group_col, missing_state_policy = "error")
  if (any(grepl(transition_separator, x$state_levels, fixed = TRUE))) {
    stop("`transition_separator` must not occur inside an observed state label.",
         call. = FALSE)
  }
  group_map <- x$metadata
  names(group_map)[names(group_map) == sequence_id_col] <- ".sequence_id"
  names(group_map)[names(group_map) == group_col] <- ".group"
  group_text <- as.character(group_map$.group)
  if (any(is.na(group_map$.group) | trimws(group_text) == "")) {
    stop("The grouping column must be non-missing and non-blank for every sequence.",
         call. = FALSE)
  }
  if (!is.null(reference)) {
    .sequence_adv_scalar_character(as.character(reference), "reference")
  }
  groups <- unique(group_text)
  groups <- sort(groups, method = "radix")
  if (length(groups) < 2L) stop("At least two sequence groups are required.", call. = FALSE)
  if (!is.null(reference) && !(as.character(reference) %in% groups)) {
    stop("`reference` is not an observed group.", call. = FALSE)
  }
  sequence_group <- stats::setNames(as.character(group_map$.group), as.character(group_map$.sequence_id))
  group_n <- table(factor(sequence_group, levels = groups))
  state_rows <- list()
  transition_rows <- list()
  length_rows <- list()
  for (g in groups) {
    ids <- names(sequence_group)[sequence_group == g]
    seqs <- x$sequences[ids]
    states_all <- unlist(seqs, use.names = FALSE)
    state_count <- table(factor(states_all, levels = x$state_levels))
    state_sequence_count <- vapply(x$state_levels, function(s) sum(vapply(seqs, function(z) s %in% z, logical(1))), integer(1))
    state_rows[[g]] <- data.frame(
      group = g, state = x$state_levels, event_count = as.integer(state_count),
      event_share = as.numeric(state_count) / max(1, sum(state_count)),
      sequence_count = as.integer(state_sequence_count),
      sequence_prevalence = as.numeric(state_sequence_count) / length(seqs),
      stringsAsFactors = FALSE
    )
    transitions <- unlist(lapply(seqs, function(z) {
      if (length(z) < 2L) return(character())
      paste(z[-length(z)], z[-1L], sep = transition_separator)
    }), use.names = FALSE)
    if (!include_self && length(transitions) > 0L) {
      parts <- strsplit(transitions, transition_separator, fixed = TRUE)
      transitions <- transitions[vapply(parts, function(p) p[1L] != p[2L], logical(1))]
    }
    transition_levels <- sort(unique(transitions), method = "radix")
    if (length(transition_levels) > 0L) {
      transition_count <- table(factor(transitions, levels = transition_levels))
      transition_sequence_count <- vapply(transition_levels, function(tr) {
        sum(vapply(seqs, function(z) {
          if (length(z) < 2L) return(FALSE)
          tr %in% paste(z[-length(z)], z[-1L], sep = transition_separator)
        }, logical(1)))
      }, integer(1))
      transition_rows[[g]] <- data.frame(
        group = g, transition = transition_levels,
        occurrence_count = as.integer(transition_count),
        occurrence_share = as.numeric(transition_count) / max(1, sum(transition_count)),
        sequence_count = as.integer(transition_sequence_count),
        sequence_prevalence = as.numeric(transition_sequence_count) / length(seqs),
        stringsAsFactors = FALSE
      )
    }
    lengths <- lengths(seqs)
    length_rows[[g]] <- data.frame(
      group = g, n_sequences = length(lengths), mean_length = mean(lengths),
      median_length = stats::median(lengths), min_length = min(lengths),
      max_length = max(lengths), sd_length = if (length(lengths) > 1L) stats::sd(lengths) else 0,
      stringsAsFactors = FALSE
    )
  }
  state_summary <- do.call(rbind, state_rows)
  transition_summary <- if (length(transition_rows) > 0L) do.call(rbind, transition_rows) else
    data.frame(group = character(), transition = character(), occurrence_count = integer(),
               occurrence_share = numeric(), sequence_count = integer(),
               sequence_prevalence = numeric(), stringsAsFactors = FALSE)
  length_summary <- do.call(rbind, length_rows)
  pair_matrix <- if (is.null(reference)) {
    .sequence_adv_pair_grid(groups)
  } else {
    cbind(setdiff(seq_along(groups), match(as.character(reference), groups)),
          match(as.character(reference), groups))
  }
  make_contrasts <- function(summary, key_col, measures) {
    if (nrow(pair_matrix) == 0L || nrow(summary) == 0L) return(data.frame())
    keys <- unique(as.character(summary[[key_col]]))
    out <- list(); h <- 0L
    for (r in seq_len(nrow(pair_matrix))) {
      g1 <- groups[pair_matrix[r, 1L]]; g2 <- groups[pair_matrix[r, 2L]]
      for (key in keys) {
        a <- summary[summary$group == g1 & summary[[key_col]] == key, , drop = FALSE]
        b <- summary[summary$group == g2 & summary[[key_col]] == key, , drop = FALSE]
        if (nrow(a) == 0L) a <- summary[0L, , drop = FALSE]
        if (nrow(b) == 0L) b <- summary[0L, , drop = FALSE]
        h <- h + 1L
        row <- data.frame(group_1 = g1, group_2 = g2, stringsAsFactors = FALSE)
        row[[key_col]] <- key
        for (measure in measures) {
          va <- if (nrow(a) == 0L) 0 else a[[measure]][1L]
          vb <- if (nrow(b) == 0L) 0 else b[[measure]][1L]
          ratio <- if (vb == 0) {
            if (zero_policy == "infinite" && va > 0) Inf else NA_real_
          } else va / vb
          row[[paste0(measure, "_group_1")]] <- va
          row[[paste0(measure, "_group_2")]] <- vb
          row[[paste0(measure, "_difference")]] <- va - vb
          row[[paste0(measure, "_ratio")]] <- ratio
        }
        out[[h]] <- row
      }
    }
    do.call(rbind, out)
  }
  state_contrasts <- make_contrasts(state_summary, "state",
                                    c("event_share", "sequence_prevalence"))
  transition_contrasts <- make_contrasts(transition_summary, "transition",
                                         c("occurrence_share", "sequence_prevalence"))
  length_contrasts <- if (nrow(pair_matrix) == 0L) data.frame() else do.call(rbind, lapply(seq_len(nrow(pair_matrix)), function(r) {
    g1 <- groups[pair_matrix[r, 1L]]; g2 <- groups[pair_matrix[r, 2L]]
    a <- length_summary[length_summary$group == g1, , drop = FALSE]
    b <- length_summary[length_summary$group == g2, , drop = FALSE]
    data.frame(group_1 = g1, group_2 = g2,
               mean_length_group_1 = a$mean_length, mean_length_group_2 = b$mean_length,
               mean_length_difference = a$mean_length - b$mean_length,
               median_length_group_1 = a$median_length, median_length_group_2 = b$median_length,
               median_length_difference = a$median_length - b$median_length,
               stringsAsFactors = FALSE)
  }))
  result <- list(
    groups = data.frame(group = groups, n_sequences = as.integer(group_n), stringsAsFactors = FALSE),
    state_summary = if ("state" %in% metrics) state_summary else NULL,
    state_contrasts = if ("state" %in% metrics) state_contrasts else NULL,
    transition_summary = if ("transition" %in% metrics) transition_summary else NULL,
    transition_contrasts = if ("transition" %in% metrics) transition_contrasts else NULL,
    length_summary = if ("length" %in% metrics) length_summary else NULL,
    length_contrasts = if ("length" %in% metrics) length_contrasts else NULL,
    settings = list(reference = reference, metrics = metrics, include_self = include_self,
                    transition_separator = transition_separator, zero_policy = zero_policy)
  )
  class(result) <- c("gp3_sequence_group_comparison", "list")
  result
}

#' Plot a descriptive sequence-group comparison
#'
#' @param comparison A result from [compare_sequence_groups()].
#' @param component `"state"`, `"transition"`, or `"length"`.
#' @param measure Measure to display. Defaults depend on `component`.
#' @param top_n Maximum states or transitions to display.
#' @param main,xlab,ylab Plot labels.
#' @param ... Additional base-graphics arguments.
#'
#' @return The plotted summary data, invisibly.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' comparison <- compare_sequence_groups(sequences, group_col = "group")
#' plot_sequence_group_comparison(comparison, component = "state")
#'
#' @export
plot_sequence_group_comparison <- function(comparison,
                                           component = c("state", "transition", "length"),
                                           measure = NULL, top_n = 12L,
                                           main = NULL, xlab = NULL, ylab = NULL, ...) {
  component <- match.arg(component)
  .sequence_adv_scalar_number(top_n, "top_n", lower = 1, integer = TRUE)
  if (!inherits(comparison, "gp3_sequence_group_comparison")) {
    stop("`comparison` must be created by `compare_sequence_groups()`.", call. = FALSE)
  }
  if (component == "length") {
    data <- comparison$length_summary
    if (is.null(data)) stop("Length summaries were not requested.", call. = FALSE)
    if (is.null(measure)) measure <- "mean_length"
    if (!(measure %in% names(data))) stop("Unknown length measure.", call. = FALSE)
    if (is.null(main)) main <- "Sequence length by group"
    if (is.null(xlab)) xlab <- "Group"
    if (is.null(ylab)) ylab <- measure
    graphics::barplot(data[[measure]], names.arg = data$group, main = main,
                      xlab = xlab, ylab = ylab, ...)
    return(invisible(data))
  }
  data <- if (component == "state") comparison$state_summary else comparison$transition_summary
  if (is.null(data)) stop("The requested component was not calculated.", call. = FALSE)
  if (nrow(data) == 0L) stop("No comparison rows are available to plot.", call. = FALSE)
  key <- if (component == "state") "state" else "transition"
  if (is.null(measure)) measure <- "sequence_prevalence"
  if (!(measure %in% names(data))) stop("Unknown comparison measure.", call. = FALSE)
  totals <- stats::aggregate(data[[measure]], list(key = data[[key]]), max)
  names(totals)[2L] <- "score"
  selected <- utils::head(totals$key[order(-totals$score, totals$key, method = "radix")], top_n)
  data <- data[data[[key]] %in% selected, , drop = FALSE]
  if (nrow(data) == 0L) stop("No comparison rows are available to plot.", call. = FALSE)
  groups <- comparison$groups$group
  matrix_data <- matrix(0, nrow = length(selected), ncol = length(groups),
                        dimnames = list(selected, groups))
  for (i in seq_len(nrow(data))) matrix_data[data[[key]][i], data$group[i]] <- data[[measure]][i]
  if (is.null(main)) main <- paste("Sequence", component, "comparison")
  if (is.null(xlab)) xlab <- measure
  if (is.null(ylab)) ylab <- component
  graphics::barplot(t(matrix_data), beside = TRUE, horiz = TRUE, legend.text = groups,
                    main = main, xlab = xlab, ylab = ylab, ...)
  invisible(data)
}

test_that("consensus creation is deterministic and reports support", {
  data <- make_advanced_sequence_data()
  consensus <- create_consensus_sequence(data, group_cols = "group", tie_method = "first")
  expect_s3_class(consensus, "gp3_consensus_sequence")
  expect_true(all(c("consensus_state", "support_n", "agreement", "tie_n") %in% names(consensus)))
  expect_equal(nrow(consensus), 10L)
  expect_true(all(consensus$support_n == 4L))
  expect_true(all(consensus$agreement >= 0.5 & consensus$agreement <= 1))
  again <- create_consensus_sequence(data, group_cols = "group", tie_method = "first")
  expect_identical(consensus, again)
})

test_that("consensus tie policies and missing-state policies are explicit", {
  data <- data.frame(sequence_id = c("s1", "s2"), sequence_order = 1,
                     state = c("B", "A"), stringsAsFactors = FALSE)
  first <- create_consensus_sequence(data, tie_method = "first",
                                     state_levels = c("A", "B"))
  last <- create_consensus_sequence(data, tie_method = "last",
                                    state_levels = c("A", "B"))
  missing <- create_consensus_sequence(data, tie_method = "missing",
                                       state_levels = c("A", "B"))
  all <- create_consensus_sequence(data, tie_method = "all",
                                   state_levels = c("A", "B"))
  expect_equal(first$consensus_state, "A")
  expect_equal(last$consensus_state, "B")
  expect_true(is.na(missing$consensus_state))
  expect_equal(all$consensus_state, "A | B")
  data$state[2L] <- NA_character_
  expect_error(create_consensus_sequence(data, missing_state_policy = "error"))
  expect_equal(create_consensus_sequence(data, missing_state_policy = "exclude")$support_n, 1L)
})

test_that("consensus summaries, formatting and plotting work", {
  data <- make_advanced_sequence_data()
  consensus <- create_consensus_sequence(data, group_cols = "group")
  overall <- summarise_consensus_agreement(consensus, by = "overall")
  grouped <- summarise_consensus_agreement(consensus, by = "group")
  formatted <- format_consensus_sequence(consensus, include_agreement = TRUE)
  expect_equal(nrow(overall), 1L)
  expect_equal(nrow(grouped), 2L)
  expect_equal(nrow(formatted), 2L)
  file <- tempfile(fileext = ".pdf")
  grDevices::pdf(file)
  plotted <- plot_consensus_sequence(consensus, group = "g1")
  grDevices::dev.off()
  expect_true(file.exists(file))
  expect_true(is.data.frame(plotted))
})

test_that("descriptive group comparisons return expected components", {
  data <- make_advanced_sequence_data()
  comparison <- compare_sequence_groups(data, "group")
  expect_s3_class(comparison, "gp3_sequence_group_comparison")
  expect_equal(comparison$groups$n_sequences, c(4L, 4L))
  expect_true(all(c("state", "event_share", "sequence_prevalence") %in%
                    names(comparison$state_summary)))
  expect_true(all(c("transition", "occurrence_share") %in%
                    names(comparison$transition_summary)))
  expect_equal(nrow(comparison$length_summary), 2L)
  expect_false(any(grepl("p_value|statistic", names(comparison$state_contrasts))))
})

test_that("group comparison plotting returns plotted data", {
  comparison <- compare_sequence_groups(make_advanced_sequence_data(), "group")
  file <- tempfile(fileext = ".pdf")
  grDevices::pdf(file)
  state_data <- plot_sequence_group_comparison(comparison, "state", top_n = 4)
  length_data <- plot_sequence_group_comparison(comparison, "length")
  grDevices::dev.off()
  expect_true(nrow(state_data) > 0L)
  expect_equal(nrow(length_data), 2L)
})

test_that("grouped consensus requires explicit plot selection", {
  consensus <- create_consensus_sequence(
    make_advanced_sequence_data(),
    group_cols = "group"
  )
  expect_error(
    plot_consensus_sequence(consensus),
    "Select one consensus group"
  )

  ungrouped <- create_consensus_sequence(make_advanced_sequence_data())
  expect_error(
    summarise_consensus_agreement(ungrouped, by = "group"),
    "requires a consensus"
  )
})

test_that("group comparison rejects incomplete grouping metadata", {
  data <- make_advanced_sequence_data()
  data$group[data$sequence_id == "s01"] <- NA_character_
  expect_error(
    compare_sequence_groups(data, "group"),
    "non-missing and non-blank"
  )
})

test_that("advanced metadata mappings cannot duplicate core sequence columns", {
  data <- make_advanced_sequence_data()
  expect_error(
    create_consensus_sequence(data, group_cols = "sequence_id"),
    "must not repeat core sequence columns"
  )
})

test_that("zero-weight rows do not inflate consensus support", {
  data <- data.frame(
    sequence_id = c("s1", "s2"),
    sequence_order = c(1, 1),
    state = c("A", "B"),
    weight = c(1, 0),
    stringsAsFactors = FALSE
  )
  consensus <- create_consensus_sequence(data, weight_col = "weight")
  expect_equal(consensus$consensus_state, "A")
  expect_equal(consensus$support_n, 1L)
  expect_equal(consensus$support_weight, 1)
})

test_that("group comparison transition labels require an unambiguous separator", {
  data <- make_advanced_sequence_data()
  data$state[data$state == "A"] <- "A -> embedded"
  expect_error(
    compare_sequence_groups(data, "group"),
    "must not occur inside"
  )
})

test_that("reference-group contrasts use the reference as denominator", {
  comparison <- compare_sequence_groups(
    make_advanced_sequence_data(),
    "group",
    reference = "g1"
  )
  expect_true(all(comparison$state_contrasts$group_2 == "g1"))
  expect_true(all(comparison$transition_contrasts$group_2 == "g1"))
  expect_true(all(comparison$length_contrasts$group_2 == "g1"))
})

test_that("state plotting represents an unresolved consensus tie explicitly", {
  data <- data.frame(
    sequence_id = c("s1", "s2"),
    sequence_order = c(1, 1),
    state = c("A", "B"),
    stringsAsFactors = FALSE
  )
  consensus <- create_consensus_sequence(data, tie_method = "missing")
  file <- tempfile(fileext = ".pdf")
  grDevices::pdf(file)
  plotted <- plot_consensus_sequence(consensus, type = "states")
  grDevices::dev.off()
  expect_true(file.exists(file))
  expect_true(is.na(plotted$consensus_state))
})

test_that("empty transition comparisons fail clearly when plotted", {
  data <- data.frame(
    sequence_id = c("s1", "s2"),
    sequence_order = c(1, 1),
    state = c("A", "B"),
    group = c("g1", "g2"),
    stringsAsFactors = FALSE
  )
  comparison <- compare_sequence_groups(data, "group")
  expect_error(
    plot_sequence_group_comparison(comparison, component = "transition"),
    "No comparison rows"
  )
})

.sequence_package_version <- function() {
  description <- tryCatch(
    utils::packageDescription("gp3sequences"),
    error = function(e) NULL
  )

  if (
    is.list(description) &&
      !is.null(description$Version)
  ) {
    return(
      as.character(description$Version)
    )
  }

  if (file.exists("DESCRIPTION")) {
    description_file <- tryCatch(
      read.dcf("DESCRIPTION"),
      error = function(e) NULL
    )

    if (
      !is.null(description_file) &&
        "Version" %in% colnames(description_file)
    ) {
      return(
        as.character(
          description_file[1L, "Version"]
        )
      )
    }
  }

  NA_character_
}

.sequence_analysis_identifiers <- function(x) {
  ids <- attr(
    x,
    "sequence_ids",
    exact = TRUE
  )

  if (
    is.null(ids) &&
      is.list(x)
  ) {
    ids <- x$sequence_ids
  }

  if (
    is.null(ids) &&
      is.list(x) &&
      !is.null(x$assignments)
  ) {
    ids <- names(x$assignments)
  }

  if (
    is.null(ids) &&
      inherits(x, "dist")
  ) {
    ids <- attr(
      x,
      "Labels"
    )
  }

  if (
    is.null(ids) &&
      is.matrix(x)
  ) {
    ids <- rownames(x)
  }

  if (is.null(ids)) {
    return(character())
  }

  unique(
    as.character(ids)
  )
}

.sequence_analysis_state_levels <- function(x) {
  states <- attr(
    x,
    "state_levels",
    exact = TRUE
  )

  if (
    is.null(states) &&
      is.list(x)
  ) {
    states <- x$state_levels
  }

  if (
    is.null(states) &&
      is.list(x)
  ) {
    states <- x$symbol_names
  }

  if (is.null(states)) {
    return(character())
  }

  as.character(states)
}

.sequence_analysis_method <- function(x) {
  method <- attr(
    x,
    "method",
    exact = TRUE
  )

  if (
    is.null(method) &&
      is.list(x)
  ) {
    method <- x$method
  }

  if (
    is.null(method) &&
      inherits(x, "gp3_sequence_hmm")
  ) {
    method <- "categorical_hmm"
  }

  if (
    is.null(method) &&
      inherits(x, "gp3_transition_network")
  ) {
    method <- "transition_network"
  }

  if (is.null(method)) {
    return(NA_character_)
  }

  as.character(method)[1L]
}

.sequence_analysis_settings <- function(x) {
  settings <- attr(
    x,
    "settings",
    exact = TRUE
  )

  if (
    is.null(settings) &&
      is.list(x)
  ) {
    settings <- x$settings
  }

  if (
    is.null(settings) &&
      is.list(x)
  ) {
    candidates <- intersect(
      c(
        "k",
        "linkage",
        "seed",
        "order",
        "smoothing",
        "backoff",
        "tolerance",
        "pseudocount",
        "iterations",
        "converged"
      ),
      names(x)
    )

    if (length(candidates) > 0L) {
      settings <- x[candidates]
    }
  }

  if (is.null(settings)) {
    list()
  } else {
    settings
  }
}

.sequence_analysis_seed <- function(x) {
  provenance <- attr(
    x,
    "gp3sequences_provenance",
    exact = TRUE
  )

  if (
    is.list(provenance) &&
      !is.null(provenance$seed)
  ) {
    return(provenance$seed)
  }

  if (
    is.list(x) &&
      !is.null(x$seed)
  ) {
    return(x$seed)
  }

  NA_integer_
}

.sequence_analysis_provenance <- function(x) {
  provenance <- attr(
    x,
    "gp3sequences_provenance",
    exact = TRUE
  )

  if (is.null(provenance)) {
    provenance <- list()
  }

  if (is.null(provenance$package)) {
    provenance$package <- "gp3sequences"
  }

  if (is.null(provenance$package_version)) {
    provenance$package_version <- .sequence_package_version()
  }

  if (is.null(provenance$contract_version)) {
    provenance$contract_version <- .sequence_contract_version()
  }

  if (is.null(provenance$family)) {
    provenance$family <- .sequence_object_family(x)
  }

  if (is.null(provenance$method)) {
    provenance$method <- .sequence_analysis_method(x)
  }

  if (is.null(provenance$sequence_ids)) {
    provenance$sequence_ids <- .sequence_analysis_identifiers(x)
  }

  if (is.null(provenance$state_levels)) {
    provenance$state_levels <- .sequence_analysis_state_levels(x)
  }

  if (is.null(provenance$seed)) {
    provenance$seed <- .sequence_analysis_seed(x)
  }

  if (is.null(provenance$settings)) {
    provenance$settings <- .sequence_analysis_settings(x)
  }

  provenance
}

#' Audit a gp3sequences analysis object
#'
#' Inspects supported gp3sequences objects for structural contract validity,
#' recoverable provenance, identifiers, state levels, method settings, and
#' randomness metadata. The audit does not reinterpret sequence structure as a
#' psychological, diagnostic, cognitive, emotional, or causal construct.
#'
#' @param x A gp3sequences analysis result or supported structural object.
#' @param strict Logical; if `TRUE`, fail when the audit status is `"fail"`.
#' @param tolerance Numerical tolerance used for matrix/probability checks.
#'
#' @return An object of class `gp3_sequence_analysis_audit` containing
#' `summary`, `issues`, `provenance`, `contract`, and `status`.
#'
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3"), each = 3L),
#'   sequence_order = rep(1:3, times = 3L),
#'   state = c(
#'     "A", "B", "C",
#'     "A", "B", "B",
#'     "C", "B", "A"
#'   )
#' )
#' distance <- compute_sequence_distance(
#'   sequences,
#'   method = "levenshtein"
#' )
#' audit_sequence_analysis(distance)
#'
#' @export
audit_sequence_analysis <- function(
  x,
  strict = FALSE,
  tolerance = 1e-8
) {
  if (
    !is.logical(strict) ||
      length(strict) != 1L ||
      is.na(strict)
  ) {
    stop(
      "`strict` must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  if (
    !is.numeric(tolerance) ||
      length(tolerance) != 1L ||
      is.na(tolerance) ||
      !is.finite(tolerance) ||
      tolerance <= 0
  ) {
    stop(
      "`tolerance` must be one positive finite number.",
      call. = FALSE
    )
  }

  contract <- .sequence_object_contract(x)
  issues <- .sequence_validate_sequence_result(
    x,
    tolerance = tolerance
  )
  provenance <- .sequence_analysis_provenance(x)

  if (
    length(provenance$sequence_ids) == 0L &&
      contract$family %in% c(
        "distance",
        "clustering",
        "cluster_bootstrap",
        "cluster_ensemble",
        "hmm",
        "hmm_mixture"
      )
  ) {
    issues <- .sequence_contract_bind_issues(
      issues,
      .sequence_contract_issue(
        "sequence_ids_not_recoverable",
        "review",
        "sequence_ids",
        paste0(
          "Sequence identifiers could not be recovered from this ",
          "analysis object."
        )
      )
    )
  }

  status <- .sequence_contract_status(issues)

  summary <- data.frame(
    family = contract$family,
    primary_class = contract$primary_class,
    package_version = as.character(
      provenance$package_version
    ),
    contract_version = as.character(
      provenance$contract_version
    ),
    method = if (
      is.null(provenance$method)
    ) {
      NA_character_
    } else {
      as.character(provenance$method)[1L]
    },
    n_sequence_ids = length(
      provenance$sequence_ids
    ),
    n_state_levels = length(
      provenance$state_levels
    ),
    seed_recorded = !is.null(provenance$seed) &&
      length(provenance$seed) == 1L &&
      !is.na(provenance$seed),
    n_issues = nrow(issues),
    status = status,
    stringsAsFactors = FALSE
  )

  result <- list(
    summary = summary,
    issues = issues,
    provenance = provenance,
    contract = contract,
    status = status
  )

  class(result) <- c(
    "gp3_sequence_analysis_audit",
    "list"
  )

  if (
    strict &&
      identical(status, "fail")
  ) {
    stop(
      "Sequence-analysis audit failed: ",
      paste(
        unique(
          issues$message[
            issues$severity == "error"
          ]
        ),
        collapse = " "
      ),
      call. = FALSE
    )
  }

  result
}

.sequence_comparison_row <- function(
  field,
  x,
  y,
  equal
) {
  data.frame(
    field = field,
    x = paste(x, collapse = " | "),
    y = paste(y, collapse = " | "),
    equal = isTRUE(equal),
    stringsAsFactors = FALSE
  )
}

#' Compare two gp3sequences analysis results
#'
#' Compares structural contracts and recoverable provenance from two analysis
#' objects. This helper is intended for regression checks, sensitivity analyses,
#' and reproducibility audits; it does not select a statistically or
#' substantively "best" analysis.
#'
#' @param x,y Two analysis objects.
#' @param tolerance Numerical comparison tolerance.
#' @param compare_values Logical; when `TRUE`, also run `all.equal()` on the
#' complete objects after structural comparison.
#'
#' @return An object of class `gp3_sequence_analysis_comparison` containing
#' per-field comparisons, both audits, optional whole-object comparison, and
#' an overall equality flag.
#'
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3"), each = 3L),
#'   sequence_order = rep(1:3, times = 3L),
#'   state = c(
#'     "A", "B", "C",
#'     "A", "B", "B",
#'     "C", "B", "A"
#'   )
#' )
#' d1 <- compute_sequence_distance(
#'   sequences,
#'   method = "levenshtein"
#' )
#' d2 <- compute_sequence_distance(
#'   sequences,
#'   method = "levenshtein"
#' )
#' compare_sequence_analysis_results(d1, d2)
#'
#' @export
compare_sequence_analysis_results <- function(
  x,
  y,
  tolerance = 1e-8,
  compare_values = FALSE
) {
  if (
    !is.numeric(tolerance) ||
      length(tolerance) != 1L ||
      is.na(tolerance) ||
      !is.finite(tolerance) ||
      tolerance <= 0
  ) {
    stop(
      "`tolerance` must be one positive finite number.",
      call. = FALSE
    )
  }

  if (
    !is.logical(compare_values) ||
      length(compare_values) != 1L ||
      is.na(compare_values)
  ) {
    stop(
      "`compare_values` must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  x_audit <- audit_sequence_analysis(
    x,
    tolerance = tolerance
  )
  y_audit <- audit_sequence_analysis(
    y,
    tolerance = tolerance
  )

  x_provenance <- x_audit$provenance
  y_provenance <- y_audit$provenance

  comparisons <- do.call(
    rbind,
    list(
      .sequence_comparison_row(
        "family",
        x_audit$contract$family,
        y_audit$contract$family,
        identical(
          x_audit$contract$family,
          y_audit$contract$family
        )
      ),
      .sequence_comparison_row(
        "primary_class",
        x_audit$contract$primary_class,
        y_audit$contract$primary_class,
        identical(
          x_audit$contract$primary_class,
          y_audit$contract$primary_class
        )
      ),
      .sequence_comparison_row(
        "method",
        x_provenance$method,
        y_provenance$method,
        isTRUE(
          all.equal(
            x_provenance$method,
            y_provenance$method,
            tolerance = tolerance
          )
        )
      ),
      .sequence_comparison_row(
        "sequence_ids",
        x_provenance$sequence_ids,
        y_provenance$sequence_ids,
        identical(
          as.character(
            x_provenance$sequence_ids
          ),
          as.character(
            y_provenance$sequence_ids
          )
        )
      ),
      .sequence_comparison_row(
        "state_levels",
        x_provenance$state_levels,
        y_provenance$state_levels,
        identical(
          as.character(
            x_provenance$state_levels
          ),
          as.character(
            y_provenance$state_levels
          )
        )
      ),
      .sequence_comparison_row(
        "seed",
        x_provenance$seed,
        y_provenance$seed,
        isTRUE(
          all.equal(
            x_provenance$seed,
            y_provenance$seed,
            tolerance = tolerance
          )
        )
      ),
      .sequence_comparison_row(
        "settings",
        utils::capture.output(
          utils::str(
            x_provenance$settings,
            give.attr = FALSE
          )
        ),
        utils::capture.output(
          utils::str(
            y_provenance$settings,
            give.attr = FALSE
          )
        ),
        isTRUE(
          all.equal(
            x_provenance$settings,
            y_provenance$settings,
            tolerance = tolerance
          )
        )
      )
    )
  )

  row.names(comparisons) <- NULL

  value_comparison <- NULL

  if (compare_values) {
    value_comparison <- all.equal(
      x,
      y,
      tolerance = tolerance,
      check.attributes = TRUE
    )
  }

  result <- list(
    comparisons = comparisons,
    x_audit = x_audit,
    y_audit = y_audit,
    value_comparison = value_comparison,
    all_equal = all(comparisons$equal) &&
      (
        !compare_values ||
          isTRUE(value_comparison)
      )
  )

  class(result) <- c(
    "gp3_sequence_analysis_comparison",
    "list"
  )

  result
}

test_that("analysis audit recognises a native distance object", {
  x <- .sequence_test_case("minimal")

  distance <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )

  audit <- audit_sequence_analysis(
    distance
  )

  expect_s3_class(
    audit,
    "gp3_sequence_analysis_audit"
  )
  expect_identical(
    audit$contract$family,
    "distance"
  )
  expect_identical(
    audit$status,
    "pass"
  )
  expect_equal(
    audit$summary$n_sequence_ids,
    4L
  )
  expect_true(
    audit$summary$n_state_levels >= 4L
  )
})

test_that("analysis audit catches malformed distances", {
  x <- .sequence_test_case("minimal")

  distance <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )

  matrix_distance <- as.matrix(
    distance
  )

  matrix_distance[1L, 2L] <-
    matrix_distance[1L, 2L] + 1

  issues <- .sequence_validate_distance_matrix(
    matrix_distance
  )

  expect_true(
    any(
      issues$code == "asymmetric_distance"
    )
  )
  expect_true(
    any(
      issues$severity == "error"
    )
  )
})

test_that("analysis-result comparison separates structure from values", {
  x <- .sequence_test_case("minimal")

  d1 <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )
  d2 <- compute_sequence_distance(
    x,
    method = "levenshtein"
  )
  d3 <- compute_sequence_distance(
    x,
    method = "lcs"
  )

  same <- compare_sequence_analysis_results(
    d1,
    d2,
    compare_values = TRUE
  )
  different <- compare_sequence_analysis_results(
    d1,
    d3
  )

  expect_s3_class(
    same,
    "gp3_sequence_analysis_comparison"
  )
  expect_true(
    same$all_equal
  )
  expect_false(
    different$all_equal
  )
  expect_false(
    different$comparisons$equal[
      different$comparisons$field == "method"
    ]
  )
})

#' Compute pairwise sequence distances
#'
#' Provides transparent base-R implementations of Levenshtein edit distance,
#' longest-common-subsequence distance, optimal matching, and a first-order
#' transition-profile distance.
#'
#' @param data Long-format sequence data or a prepared gp3sequences result.
#' @param sequence_id_col,order_col,state_col Sequence columns.
#' @param method Distance method.
#' @param indel_cost Non-negative insertion/deletion cost used by
#'   `method = "optimal_matching"`.
#' @param substitution_cost Non-negative default substitution cost used by
#'   `method = "optimal_matching"`.
#' @param substitution_matrix Optional named square substitution-cost matrix for
#'   `method = "optimal_matching"`.
#' @param transition_smoothing Non-negative smoothing for transition profiles.
#' @param normalise One of `"none"`, `"max_length"`, or `"path_length"`.
#'
#' @return A `dist` object with method and preprocessing metadata attached.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' compute_sequence_distance(sequences, method = "lcs")
#'
#' @export
compute_sequence_distance <- function(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  method = c("levenshtein", "lcs", "optimal_matching", "transition"),
  indel_cost = 1,
  substitution_cost = 1,
  substitution_matrix = NULL,
  transition_smoothing = 0,
  normalise = c("none", "max_length", "path_length")
) {
  method <- match.arg(method)
  normalise <- match.arg(normalise)
  .sequence_adv_scalar_number(indel_cost, "indel_cost", lower = 0)
  .sequence_adv_scalar_number(substitution_cost, "substitution_cost", lower = 0)
  .sequence_adv_scalar_number(transition_smoothing, "transition_smoothing", lower = 0)
  x <- .sequence_adv_data(data, sequence_id_col, order_col, state_col,
                          missing_state_policy = "error")
  ids <- x$sequence_ids
  seqs <- x$sequences
  n <- length(seqs)
  matrix_out <- matrix(0, nrow = n, ncol = n, dimnames = list(ids, ids))
  if (!is.null(substitution_matrix)) {
    if (!is.matrix(substitution_matrix) || !is.numeric(substitution_matrix) ||
        nrow(substitution_matrix) != ncol(substitution_matrix) ||
        is.null(rownames(substitution_matrix)) || is.null(colnames(substitution_matrix)) ||
        anyNA(substitution_matrix) || any(!is.finite(substitution_matrix)) ||
        any(substitution_matrix < 0) || anyNA(rownames(substitution_matrix)) ||
        anyNA(colnames(substitution_matrix)) || anyDuplicated(rownames(substitution_matrix)) ||
        anyDuplicated(colnames(substitution_matrix)) ||
        !setequal(rownames(substitution_matrix), colnames(substitution_matrix))) {
      stop("`substitution_matrix` must be a finite, non-negative, named square matrix.",
           call. = FALSE)
    }
    missing_states <- setdiff(x$state_levels, rownames(substitution_matrix))
    if (length(missing_states) > 0L) {
      stop("The substitution matrix does not cover all observed states: ",
           paste(missing_states, collapse = ", "), ".", call. = FALSE)
    }
    substitution_matrix <- substitution_matrix[rownames(substitution_matrix),
                                               rownames(substitution_matrix), drop = FALSE]
    tolerance <- sqrt(.Machine$double.eps)
    if (max(abs(substitution_matrix - t(substitution_matrix))) > tolerance) {
      stop("`substitution_matrix` must be symmetric for a sequence distance.",
           call. = FALSE)
    }
    if (max(abs(diag(substitution_matrix))) > tolerance) {
      stop("The diagonal of `substitution_matrix` must be zero.", call. = FALSE)
    }
  }
  pairs <- .sequence_adv_pair_grid(ids)
  if (nrow(pairs) > 0L) {
    for (r in seq_len(nrow(pairs))) {
      i <- pairs[r, 1L]
      j <- pairs[r, 2L]
      a <- seqs[[i]]
      b <- seqs[[j]]
      raw <- switch(
        method,
        levenshtein = .sequence_adv_edit_distance(a, b, 1, 1, NULL),
        lcs = length(a) + length(b) - 2L * .sequence_adv_lcs_length(a, b),
        optimal_matching = .sequence_adv_edit_distance(
          a, b, indel_cost, substitution_cost, substitution_matrix
        ),
        transition = {
          pa <- .sequence_adv_transition_profile(a, x$state_levels, transition_smoothing)
          pb <- .sequence_adv_transition_profile(b, x$state_levels, transition_smoothing)
          sqrt(sum((pa - pb)^2))
        }
      )
      denominator <- switch(
        normalise,
        none = 1,
        max_length = max(length(a), length(b), 1L),
        path_length = max(length(a) + length(b), 1L)
      )
      value <- raw / denominator
      matrix_out[i, j] <- value
      matrix_out[j, i] <- value
    }
  }
  result <- stats::as.dist(matrix_out)
  attr(result, "method") <- method
  attr(result, "sequence_ids") <- ids
  attr(result, "state_levels") <- x$state_levels
  attr(result, "settings") <- list(
    indel_cost = indel_cost,
    substitution_cost = substitution_cost,
    substitution_matrix = substitution_matrix,
    transition_smoothing = transition_smoothing,
    normalise = normalise
  )
  class(result) <- unique(c("gp3_sequence_distance", class(result)))
  result
}

#' Summarise a sequence-distance object
#'
#' @param distance A distance object returned by [compute_sequence_distance()].
#'
#' @return A list containing an overall summary and per-sequence mean, median,
#' minimum, and maximum distances.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' distance <- compute_sequence_distance(sequences)
#' summarise_sequence_distance(distance)
#'
#' @export
summarise_sequence_distance <- function(distance) {
  matrix_distance <- .sequence_adv_distance_matrix(distance)
  n <- nrow(matrix_distance)
  values <- if (n > 1L) matrix_distance[upper.tri(matrix_distance)] else numeric()
  overall <- data.frame(
    n_sequences = as.integer(n),
    n_pairs = as.integer(length(values)),
    mean_distance = if (length(values)) mean(values) else NA_real_,
    median_distance = if (length(values)) stats::median(values) else NA_real_,
    min_distance = if (length(values)) min(values) else NA_real_,
    max_distance = if (length(values)) max(values) else NA_real_,
    stringsAsFactors = FALSE
  )
  per_sequence <- lapply(seq_len(n), function(i) {
    others <- matrix_distance[i, -i, drop = TRUE]
    data.frame(
      sequence_id = rownames(matrix_distance)[i],
      mean_distance = if (length(others)) mean(others) else NA_real_,
      median_distance = if (length(others)) stats::median(others) else NA_real_,
      min_distance = if (length(others)) min(others) else NA_real_,
      max_distance = if (length(others)) max(others) else NA_real_,
      stringsAsFactors = FALSE
    )
  })
  list(overall = overall, per_sequence = do.call(rbind, per_sequence),
       method = attr(distance, "method"), settings = attr(distance, "settings"))
}

#' Cluster sequences from a distance object
#'
#' @param distance A `dist` object or square matrix.
#' @param k Number of clusters.
#' @param method `"hierarchical"`, `"pam"`, or `"clara"`.
#' @param linkage Hierarchical linkage method.
#' @param seed Reproducibility seed for optional stochastic methods.
#' @param ... Additional arguments passed to the selected clustering function.
#'
#' @return A list of class `gp3_sequence_clustering` containing assignments,
#' model object, medoid identifiers where available, and settings.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' distance <- compute_sequence_distance(sequences)
#' cluster_sequences(distance, k = 2L)
#'
#' @export
cluster_sequences <- function(distance, k,
                              method = c("hierarchical", "pam", "clara"),
                              linkage = "average", seed = 1L, ...) {
  method <- match.arg(method)
  .sequence_adv_scalar_number(k, "k", lower = 2, integer = TRUE)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  if (!is.character(linkage) || length(linkage) != 1L || is.na(linkage) ||
      !(linkage %in% c("ward.D", "ward.D2", "single", "complete", "average",
                      "mcquitty", "median", "centroid"))) {
    stop("`linkage` is not a supported `stats::hclust()` method.", call. = FALSE)
  }
  matrix_distance <- .sequence_adv_distance_matrix(distance)
  n <- nrow(matrix_distance)
  if (k >= n) stop("`k` must be smaller than the number of sequences.", call. = FALSE)
  ids <- rownames(matrix_distance)
  dots <- list(...)
  if (length(dots) > 0L && (is.null(names(dots)) || any(!nzchar(names(dots))))) {
    stop("All additional clustering arguments in `...` must be named.",
         call. = FALSE)
  }
  protected <- intersect(names(dots), c("x", "k", "diss", "metric"))
  if (length(protected) > 0L) {
    stop("Do not supply protected clustering arguments through `...`: ",
         paste(protected, collapse = ", "), ".", call. = FALSE)
  }
  if (method == "hierarchical") {
    allowed <- intersect(names(dots), "members")
    unexpected <- setdiff(names(dots), allowed)
    if (length(unexpected) > 0L) {
      stop("Unsupported hierarchical clustering arguments: ",
           paste(unexpected, collapse = ", "), ".", call. = FALSE)
    }
    model <- do.call(
      stats::hclust,
      c(list(d = stats::as.dist(matrix_distance), method = linkage), dots)
    )
    assignments <- stats::cutree(model, k = k)
    assignments <- assignments[ids]
    medoids <- vapply(sort(unique(assignments)), function(cluster_id) {
      members <- which(assignments == cluster_id)
      if (length(members) == 1L) return(ids[members])
      ids[members[which.min(rowSums(matrix_distance[members, members, drop = FALSE]))]]
    }, character(1))
  } else if (method == "pam") {
    .sequence_adv_require("cluster", "run partitioning around medoids")
    model <- .sequence_adv_with_seed(
      seed,
      do.call(
        cluster::pam,
        c(list(x = stats::as.dist(matrix_distance), k = k, diss = TRUE), dots)
      )
    )
    assignments <- stats::setNames(as.integer(model$clustering), ids)
    medoids <- ids[model$id.med]
  } else {
    .sequence_adv_require("cluster", "run CLARA clustering")
    embedding_dimension <- min(n - 1L, max(2L, as.integer(k)))
    embedding_result <- stats::cmdscale(
      stats::as.dist(matrix_distance),
      k = embedding_dimension,
      eig = FALSE,
      add = TRUE
    )
    embedding <- if (is.list(embedding_result)) embedding_result$points else embedding_result
    if (!is.matrix(embedding) || nrow(embedding) != n) {
      stop("Classical multidimensional scaling did not return a usable CLARA embedding.",
           call. = FALSE)
    }
    if (is.null(dots$rngR)) {
      dots$rngR <- TRUE
    } else if (!isTRUE(dots$rngR)) {
      stop("CLARA reproducibility requires `rngR = TRUE`.", call. = FALSE)
    }
    model <- .sequence_adv_with_seed(
      seed,
      do.call(
        cluster::clara,
        c(list(x = embedding, k = k, metric = "euclidean"), dots)
      )
    )
    assignments <- stats::setNames(as.integer(model$clustering), ids)
    medoids <- ids[model$i.med]
  }
  result <- list(
    assignments = assignments,
    model = model,
    medoids = as.character(medoids),
    k = as.integer(k),
    method = method,
    linkage = if (method == "hierarchical") linkage else NULL,
    distance = stats::as.dist(matrix_distance),
    seed = as.integer(seed)
  )
  class(result) <- c("gp3_sequence_clustering", "list")
  result
}

#' Validate sequence clusters descriptively
#'
#' @param clustering A result from [cluster_sequences()] or a named assignment
#' vector.
#' @param distance Optional distance object when `clustering` is an assignment
#' vector.
#'
#' @return A list containing overall validation metrics, cluster sizes, and
#' per-sequence silhouette values.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' distance <- compute_sequence_distance(sequences)
#' fit <- cluster_sequences(distance, k = 2L)
#' validate_sequence_clusters(fit)
#'
#' @export
validate_sequence_clusters <- function(clustering, distance = NULL) {
  if (inherits(clustering, "gp3_sequence_clustering")) {
    assignments <- clustering$assignments
    matrix_distance <- .sequence_adv_distance_matrix(clustering$distance)
  } else {
    assignments <- clustering
    matrix_distance <- .sequence_adv_distance_matrix(distance)
  }
  if (is.null(names(assignments)) || anyNA(names(assignments)) ||
      any(!nzchar(names(assignments))) || anyDuplicated(names(assignments))) {
    stop("Cluster assignments must have unique sequence-ID names.", call. = FALSE)
  }
  ids <- rownames(matrix_distance)
  if (!setequal(names(assignments), ids)) {
    stop("Assignment names and distance-matrix identifiers must match.", call. = FALSE)
  }
  assignments <- assignments[ids]
  if (anyNA(assignments) || length(unique(assignments)) < 2L) {
    stop("At least two non-missing clusters are required.", call. = FALSE)
  }
  silhouette <- .sequence_adv_silhouette(assignments, matrix_distance)
  clusters <- sort(unique(assignments))
  cluster_sizes <- table(factor(assignments, levels = clusters))
  max_intra <- 0
  min_inter <- Inf
  within_values <- numeric()
  between_values <- numeric()
  for (i in seq_len(nrow(matrix_distance) - 1L)) {
    for (j in seq.int(i + 1L, nrow(matrix_distance))) {
      if (assignments[i] == assignments[j]) {
        within_values <- c(within_values, matrix_distance[i, j])
        max_intra <- max(max_intra, matrix_distance[i, j])
      } else {
        between_values <- c(between_values, matrix_distance[i, j])
        min_inter <- min(min_inter, matrix_distance[i, j])
      }
    }
  }
  dunn <- if (max_intra > 0 && is.finite(min_inter)) min_inter / max_intra else NA_real_
  ratio <- if (length(within_values) && length(between_values) && mean(between_values) > 0) {
    mean(within_values) / mean(between_values)
  } else NA_real_
  overall <- data.frame(
    n_sequences = length(assignments),
    n_clusters = length(clusters),
    average_silhouette = mean(silhouette),
    minimum_silhouette = min(silhouette),
    dunn_index = dunn,
    within_between_ratio = ratio,
    singleton_clusters = sum(cluster_sizes == 1L),
    stringsAsFactors = FALSE
  )
  per_sequence <- data.frame(
    sequence_id = ids,
    cluster = as.character(assignments),
    silhouette = as.numeric(silhouette),
    stringsAsFactors = FALSE
  )
  sizes <- data.frame(cluster = as.character(clusters), size = as.integer(cluster_sizes),
                      stringsAsFactors = FALSE)
  list(overall = overall, cluster_sizes = sizes, per_sequence = per_sequence)
}

#' Extract representative sequences from clusters
#'
#' @param clustering A clustering result.
#' @param distance Optional distance when absent from `clustering`.
#' @param n_per_cluster Number of representatives per cluster.
#'
#' @return A data frame with cluster, rank, representative ID, and mean
#' within-cluster distance.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' distance <- compute_sequence_distance(sequences)
#' fit <- cluster_sequences(distance, k = 2L)
#' extract_representative_sequences(fit)
#'
#' @export
extract_representative_sequences <- function(clustering, distance = NULL,
                                             n_per_cluster = 1L) {
  .sequence_adv_scalar_number(n_per_cluster, "n_per_cluster", lower = 1, integer = TRUE)
  if (inherits(clustering, "gp3_sequence_clustering")) {
    assignments <- clustering$assignments
    matrix_distance <- .sequence_adv_distance_matrix(clustering$distance)
  } else {
    assignments <- clustering
    matrix_distance <- .sequence_adv_distance_matrix(distance)
  }
  ids <- rownames(matrix_distance)
  if (is.null(names(assignments)) || !setequal(names(assignments), ids) || anyNA(assignments)) {
    stop("Cluster assignments must be non-missing and named for every sequence.",
         call. = FALSE)
  }
  assignments <- assignments[ids]
  pieces <- lapply(sort(unique(assignments)), function(k) {
    members <- which(assignments == k)
    scores <- if (length(members) == 1L) {
      0
    } else {
      rowSums(matrix_distance[members, members, drop = FALSE]) / (length(members) - 1L)
    }
    order_members <- members[order(scores, ids[members], method = "radix")]
    selected <- utils::head(order_members, n_per_cluster)
    data.frame(cluster = as.character(k), rank = seq_along(selected),
               sequence_id = ids[selected], mean_within_distance = scores[match(selected, members)],
               stringsAsFactors = FALSE)
  })
  result <- do.call(rbind, pieces)
  row.names(result) <- NULL
  result
}

#' Bootstrap sequence-cluster stability
#'
#' Uses repeated subsampling without replacement and records pairwise
#' co-clustering agreement relative to the full-data solution.
#'
#' @param distance Sequence distance object.
#' @param k Number of clusters.
#' @param method Clustering method.
#' @param n_boot Number of subsamples.
#' @param sample_fraction Fraction of sequences sampled in each iteration.
#' @param seed Reproducibility seed.
#' @param linkage Hierarchical linkage.
#' @param ... Additional clustering arguments.
#'
#' @return An object of class `gp3_sequence_cluster_bootstrap`.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' distance <- compute_sequence_distance(sequences)
#' bootstrap_sequence_clusters(distance, k = 2L, n_boot = 5L, seed = 1L)
#'
#' @export
bootstrap_sequence_clusters <- function(distance, k,
                                        method = c("hierarchical", "pam", "clara"),
                                        n_boot = 100L, sample_fraction = 0.8,
                                        seed = 1L, linkage = "average", ...) {
  method <- match.arg(method)
  .sequence_adv_scalar_number(k, "k", lower = 2, integer = TRUE)
  .sequence_adv_scalar_number(n_boot, "n_boot", lower = 1, integer = TRUE)
  .sequence_adv_scalar_number(sample_fraction, "sample_fraction", lower = 0.2, upper = 1)
  .sequence_adv_scalar_number(seed, "seed", lower = 0, integer = TRUE)
  matrix_distance <- .sequence_adv_distance_matrix(distance)
  ids <- rownames(matrix_distance)
  n <- length(ids)
  sample_n <- max(k + 1L, floor(n * sample_fraction))
  if (sample_n > n) stop("The requested subsample is too small for `k`.", call. = FALSE)
  original <- cluster_sequences(stats::as.dist(matrix_distance), k, method, linkage, seed, ...)
  original_same <- outer(original$assignments, original$assignments, `==`)
  evaluated <- matrix(0L, n, n, dimnames = list(ids, ids))
  matched <- matrix(0L, n, n, dimnames = list(ids, ids))
  iteration_rows <- vector("list", n_boot)
  .sequence_adv_with_seed(seed, {
    for (b in seq_len(n_boot)) {
      selected <- sort(sample(seq_len(n), sample_n, replace = FALSE))
      sub_matrix <- matrix_distance[selected, selected, drop = FALSE]
      fit <- cluster_sequences(stats::as.dist(sub_matrix), k, method, linkage,
                               seed = .sequence_adv_seed(seed, b), ...)
      current_same <- outer(fit$assignments, fit$assignments, `==`)
      evaluated[selected, selected] <- evaluated[selected, selected] + 1L
      matched[selected, selected] <- matched[selected, selected] +
        (current_same == original_same[selected, selected])
      iteration_rows[[b]] <- data.frame(
        iteration = b,
        n_sampled = length(selected),
        average_silhouette = validate_sequence_clusters(fit)$overall$average_silhouette,
        stringsAsFactors = FALSE
      )
    }
  })
  stability <- matrix(NA_real_, n, n, dimnames = list(ids, ids))
  valid <- evaluated > 0L
  stability[valid] <- matched[valid] / evaluated[valid]
  diag(stability) <- 1
  pair_values <- stability[upper.tri(stability)]
  finite_pair_values <- pair_values[is.finite(pair_values)]
  result <- list(
    original = original,
    pairwise_stability = stability,
    evaluated_counts = evaluated,
    iterations = do.call(rbind, iteration_rows),
    overall = data.frame(
      n_boot = as.integer(n_boot), sample_fraction = sample_fraction,
      mean_pairwise_stability = if (length(finite_pair_values)) mean(finite_pair_values) else NA_real_,
      min_pairwise_stability = if (length(finite_pair_values)) min(finite_pair_values) else NA_real_,
      stringsAsFactors = FALSE
    ),
    settings = list(
      k = as.integer(k),
      method = method,
      linkage = linkage,
      n_boot = as.integer(n_boot),
      sample_fraction = sample_fraction,
      seed = as.integer(seed)
    )
  )
  class(result) <- c("gp3_sequence_cluster_bootstrap", "list")
  result
}

#' Summarise sequence-cluster stability
#'
#' @param bootstrap A result from [bootstrap_sequence_clusters()].
#' @param threshold Pairwise stability threshold.
#'
#' @return Overall, cluster-level, and low-stability pair summaries.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' distance <- compute_sequence_distance(sequences)
#' boot <- bootstrap_sequence_clusters(distance, k = 2L, n_boot = 5L)
#' summarise_sequence_cluster_stability(boot)
#'
#' @export
summarise_sequence_cluster_stability <- function(bootstrap, threshold = 0.8) {
  if (!inherits(bootstrap, "gp3_sequence_cluster_bootstrap")) {
    stop("`bootstrap` must be created by `bootstrap_sequence_clusters()`.", call. = FALSE)
  }
  .sequence_adv_scalar_number(threshold, "threshold", lower = 0, upper = 1)
  assignments <- bootstrap$original$assignments
  matrix_stability <- bootstrap$pairwise_stability
  clusters <- sort(unique(assignments))
  cluster_summary <- do.call(rbind, lapply(clusters, function(k) {
    ids <- names(assignments)[assignments == k]
    values <- if (length(ids) > 1L) {
      current <- matrix_stability[ids, ids, drop = FALSE]
      current[upper.tri(current)]
    } else 1
    finite_values <- values[is.finite(values)]
    evaluated_pairs <- if (length(ids) > 1L) length(finite_values) else 0L
    data.frame(
      cluster = as.character(k),
      n_sequences = length(ids),
      mean_within_stability = if (length(finite_values)) mean(finite_values) else NA_real_,
      min_within_stability = if (length(finite_values)) min(finite_values) else NA_real_,
      n_evaluated_pairs = as.integer(evaluated_pairs),
      stringsAsFactors = FALSE
    )
  }))
  pairs <- which(upper.tri(matrix_stability) & matrix_stability < threshold, arr.ind = TRUE)
  low_pairs <- if (nrow(pairs) == 0L) {
    data.frame(sequence_id_1 = character(), sequence_id_2 = character(),
               stability = numeric(), stringsAsFactors = FALSE)
  } else {
    data.frame(sequence_id_1 = rownames(matrix_stability)[pairs[, 1L]],
               sequence_id_2 = colnames(matrix_stability)[pairs[, 2L]],
               stability = matrix_stability[pairs], stringsAsFactors = FALSE)
  }
  list(overall = bootstrap$overall, clusters = cluster_summary,
       low_stability_pairs = low_pairs, threshold = threshold)
}

#' Create a sequence-cluster ensemble
#'
#' Combines multiple named cluster assignments through a co-association matrix
#' and applies hierarchical clustering to `1 - co-association`.
#'
#' @param ... Two or more `gp3_sequence_clustering` objects or named assignment
#' vectors.
#' @param k Number of consensus clusters.
#' @param linkage Hierarchical linkage applied to the co-association distance.
#'
#' @return An object of class `gp3_sequence_cluster_ensemble` containing the
#' consensus assignments, co-association matrix, model, and source solutions.
#' @examples
#' sequences <- data.frame(
#'   sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
#'   sequence_order = rep(1:4, times = 4L),
#'   state = c("A", "B", "C", "D", "A", "B", "C", "C",
#'             "D", "C", "B", "A", "D", "C", "A", "A"),
#'   group = rep(c("g1", "g2"), each = 8L),
#'   stringsAsFactors = FALSE
#' )
#' d1 <- compute_sequence_distance(sequences, method = "lcs")
#' d2 <- compute_sequence_distance(sequences, method = "transition")
#' create_sequence_cluster_ensemble(cluster_sequences(d1, 2L),
#'                                  cluster_sequences(d2, 2L), k = 2L)
#'
#' @export
create_sequence_cluster_ensemble <- function(..., k, linkage = "average") {
  solutions <- list(...)
  if (length(solutions) < 2L) stop("Supply at least two clustering solutions.", call. = FALSE)
  .sequence_adv_scalar_number(k, "k", lower = 2, integer = TRUE)
  if (!is.character(linkage) || length(linkage) != 1L || is.na(linkage) ||
      !(linkage %in% c("ward.D", "ward.D2", "single", "complete", "average",
                       "mcquitty", "median", "centroid"))) {
    stop("`linkage` is not a supported `stats::hclust()` method.", call. = FALSE)
  }
  assignments <- lapply(solutions, function(solution) {
    if (inherits(solution, "gp3_sequence_clustering")) solution$assignments else solution
  })
  if (any(vapply(assignments, is.null, logical(1))) ||
      any(vapply(assignments, function(x) {
        is.null(names(x)) || anyNA(x) || anyNA(names(x)) || any(!nzchar(names(x))) ||
          anyDuplicated(names(x))
      }, logical(1)))) {
    stop("Every solution must provide non-missing assignments with unique sequence-ID names.",
         call. = FALSE)
  }
  ids <- names(assignments[[1L]])
  if (!all(vapply(assignments, function(x) setequal(names(x), ids), logical(1)))) {
    stop("All clustering solutions must cover the same sequence IDs.", call. = FALSE)
  }
  n <- length(ids)
  if (k >= n) stop("`k` must be smaller than the number of sequences.", call. = FALSE)
  coassociation <- matrix(0, n, n, dimnames = list(ids, ids))
  for (assignment in assignments) {
    assignment <- assignment[ids]
    coassociation <- coassociation + outer(assignment, assignment, `==`)
  }
  coassociation <- coassociation / length(assignments)
  diag(coassociation) <- 1
  ensemble_distance <- stats::as.dist(1 - coassociation)
  model <- stats::hclust(ensemble_distance, method = linkage)
  consensus <- stats::cutree(model, k = k)
  consensus <- consensus[ids]
  result <- list(
    assignments = consensus,
    coassociation = coassociation,
    distance = ensemble_distance,
    model = model,
    source_assignments = assignments,
    k = as.integer(k),
    linkage = linkage
  )
  class(result) <- c("gp3_sequence_cluster_ensemble", "list")
  result
}

test_that("all core sequence distances are symmetric and deterministic", {
  data <- make_advanced_sequence_data()
  methods <- c("levenshtein", "lcs", "optimal_matching", "transition")
  for (method in methods) {
    distance <- compute_sequence_distance(data, method = method)
    matrix_distance <- as.matrix(distance)
    expect_s3_class(distance, "gp3_sequence_distance")
    expect_equal(matrix_distance, t(matrix_distance), tolerance = 1e-12)
    expect_equal(unname(diag(matrix_distance)), rep(0, nrow(matrix_distance)))
    expect_true(all(matrix_distance >= 0))
    expect_identical(distance, compute_sequence_distance(data, method = method))
  }
})

test_that("distance normalisation and substitution matrices are explicit", {
  data <- make_advanced_sequence_data()
  states <- sort(unique(data$state))
  costs <- matrix(2, length(states), length(states), dimnames = list(states, states))
  diag(costs) <- 0
  raw <- compute_sequence_distance(data, method = "optimal_matching",
                                   substitution_matrix = costs, normalise = "none")
  norm <- compute_sequence_distance(data, method = "optimal_matching",
                                    substitution_matrix = costs, normalise = "max_length")
  expect_true(all(as.numeric(norm) <= as.numeric(raw)))
  expect_error(compute_sequence_distance(data, method = "optimal_matching",
                                         substitution_matrix = matrix(1, 2, 3)))
})

test_that("distance summaries include overall and per-sequence results", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  summary <- summarise_sequence_distance(distance)
  expect_equal(summary$overall$n_sequences, 8L)
  expect_equal(summary$overall$n_pairs, 28L)
  expect_equal(nrow(summary$per_sequence), 8L)
})

test_that("hierarchical clustering and validation are auditable", {
  distance <- compute_sequence_distance(make_advanced_sequence_data(), method = "lcs")
  fit <- cluster_sequences(distance, k = 2L, method = "hierarchical")
  expect_s3_class(fit, "gp3_sequence_clustering")
  expect_equal(length(fit$assignments), 8L)
  expect_equal(length(unique(fit$assignments)), 2L)
  validation <- validate_sequence_clusters(fit)
  expect_equal(validation$overall$n_sequences, 8L)
  expect_true(validation$overall$average_silhouette >= -1)
  expect_true(validation$overall$average_silhouette <= 1)
  representatives <- extract_representative_sequences(fit, n_per_cluster = 2L)
  expect_equal(nrow(representatives), 4L)
})

test_that("PAM is guarded by the optional cluster package", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  if (requireNamespace("cluster", quietly = TRUE)) {
    fit <- cluster_sequences(distance, 2L, method = "pam")
    expect_s3_class(fit, "gp3_sequence_clustering")
  } else {
    expect_error(cluster_sequences(distance, 2L, method = "pam"), "Optional package")
  }
})

test_that("bootstrap stability is reproducible", {
  distance <- compute_sequence_distance(make_advanced_sequence_data(), method = "lcs")
  first <- bootstrap_sequence_clusters(distance, k = 2L, n_boot = 8L,
                                       sample_fraction = 0.75, seed = 41L)
  second <- bootstrap_sequence_clusters(distance, k = 2L, n_boot = 8L,
                                        sample_fraction = 0.75, seed = 41L)
  expect_s3_class(first, "gp3_sequence_cluster_bootstrap")
  expect_equal(first$pairwise_stability, second$pairwise_stability)
  summary <- summarise_sequence_cluster_stability(first)
  expect_equal(nrow(summary$clusters), 2L)
  expect_true(summary$overall$mean_pairwise_stability >= 0)
  expect_true(summary$overall$mean_pairwise_stability <= 1)
})

test_that("cluster ensembles use transparent co-association", {
  distance_1 <- compute_sequence_distance(make_advanced_sequence_data(), method = "lcs")
  distance_2 <- compute_sequence_distance(make_advanced_sequence_data(), method = "transition")
  fit_1 <- cluster_sequences(distance_1, 2L)
  fit_2 <- cluster_sequences(distance_2, 2L)
  ensemble <- create_sequence_cluster_ensemble(fit_1, fit_2, k = 2L)
  expect_s3_class(ensemble, "gp3_sequence_cluster_ensemble")
  expect_equal(unname(diag(ensemble$coassociation)), rep(1, 8L))
  expect_true(all(ensemble$coassociation >= 0 & ensemble$coassociation <= 1))
  expect_equal(length(unique(ensemble$assignments)), 2L)
})

test_that("distance and substitution matrices enforce metric structure", {
  data <- make_advanced_sequence_data()
  states <- sort(unique(data$state))
  asymmetric <- matrix(1, length(states), length(states),
                       dimnames = list(states, states))
  diag(asymmetric) <- 0
  asymmetric[1L, 2L] <- 2
  expect_error(
    compute_sequence_distance(
      data,
      method = "optimal_matching",
      substitution_matrix = asymmetric
    ),
    "symmetric"
  )

  invalid_diagonal <- matrix(1, length(states), length(states),
                             dimnames = list(states, states))
  expect_error(
    compute_sequence_distance(
      data,
      method = "optimal_matching",
      substitution_matrix = invalid_diagonal
    ),
    "diagonal"
  )

  distance <- matrix(c(0, 1, 2, 0), 2L, 2L,
                     dimnames = list(c("a", "b"), c("a", "b")))
  expect_error(validate_sequence_clusters(c(a = 1L, b = 2L), distance),
               "symmetric")
})

test_that("CLARA is guarded and preserves sequence identifiers", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  if (requireNamespace("cluster", quietly = TRUE)) {
    fit <- cluster_sequences(distance, 2L, method = "clara", seed = 19L,
                             samples = 3L)
    expect_s3_class(fit, "gp3_sequence_clustering")
    expect_setequal(names(fit$assignments), attr(distance, "Labels"))
  } else {
    expect_error(cluster_sequences(distance, 2L, method = "clara"),
                 "Optional package")
  }
})

test_that("stochastic helpers restore the caller random-number state", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  set.seed(777L)
  before <- .Random.seed
  bootstrap_sequence_clusters(distance, 2L, n_boot = 3L,
                              sample_fraction = 0.75, seed = 9L)
  expect_identical(.Random.seed, before)
})

test_that("character cluster assignments are supported", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  assignments <- setNames(rep(c("left", "right"), each = 4L),
                          attr(distance, "Labels"))
  validation <- validate_sequence_clusters(assignments, distance)
  expect_equal(validation$overall$n_clusters, 2L)
  expect_setequal(validation$cluster_sizes$cluster, c("left", "right"))
})

test_that("unused factor levels do not become observed distance states", {
  data <- make_advanced_sequence_data()
  data$state <- factor(data$state, levels = c("A", "B", "C", "D", "UNUSED"))
  states <- c("A", "B", "C", "D")
  costs <- matrix(1, 4L, 4L, dimnames = list(states, states))
  diag(costs) <- 0
  distance <- compute_sequence_distance(
    data,
    method = "optimal_matching",
    substitution_matrix = costs
  )
  expect_false("UNUSED" %in% attr(distance, "state_levels"))
})

test_that("ensemble linkage and stability summaries are validated", {
  data <- make_advanced_sequence_data()
  d1 <- compute_sequence_distance(data, method = "lcs")
  d2 <- compute_sequence_distance(data, method = "transition")
  c1 <- cluster_sequences(d1, 2L)
  c2 <- cluster_sequences(d2, 2L)
  expect_error(
    create_sequence_cluster_ensemble(c1, c2, k = 2L, linkage = "unknown"),
    "not a supported"
  )
  boot <- bootstrap_sequence_clusters(d1, 2L, n_boot = 2L,
                                      sample_fraction = 0.5, seed = 4L)
  summary <- summarise_sequence_cluster_stability(boot)
  expect_true("n_evaluated_pairs" %in% names(summary$clusters))
  expect_true(all(summary$clusters$n_evaluated_pairs >= 0L))
})

test_that("optional CLARA clustering restores the caller RNG and is repeatable", {
  if (requireNamespace("cluster", quietly = TRUE)) {
    distance <- compute_sequence_distance(make_advanced_sequence_data(), method = "lcs")
    set.seed(321L)
    before <- .Random.seed
    first <- cluster_sequences(distance, 2L, method = "clara", seed = 12L,
                               samples = 4L, sampsize = 6L)
    expect_identical(.Random.seed, before)
    second <- cluster_sequences(distance, 2L, method = "clara", seed = 12L,
                                samples = 4L, sampsize = 6L)
    expect_identical(first$assignments, second$assignments)
    expect_error(
      cluster_sequences(distance, 2L, method = "clara", rngR = FALSE),
      "rngR = TRUE"
    )
  } else {
    expect_error(
      cluster_sequences(
        compute_sequence_distance(make_advanced_sequence_data()),
        2L,
        method = "clara"
      ),
      "Optional package"
    )
  }
})

test_that("integer controls reject values outside the R integer range", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  expect_error(
    cluster_sequences(distance, 2L, seed = .Machine$integer.max + 1),
    "invalid numeric"
  )
})

testthat::test_that("seed validation and bootstrap offsets are safe", {
  data <- make_advanced_sequence_data()
  distance <- compute_sequence_distance(data, method = "lcs")

  testthat::expect_error(
    cluster_sequences(distance, 2L, seed = -1L),
    "seed"
  )
  testthat::expect_silent(
    bootstrap_sequence_clusters(
      distance,
      k = 2L,
      n_boot = 1L,
      sample_fraction = 0.75,
      seed = .Machine$integer.max
    )
  )
})

test_that("additional clustering arguments must be named", {
  distance <- compute_sequence_distance(make_advanced_sequence_data())
  expect_error(
    cluster_sequences(distance, 2L, method = "hierarchical",
                      linkage = "average", seed = 1L, 1),
    "must be named"
  )
})

#' gp3sequences: Transparent Analysis of Ordered Categorical Sequences
#'
#' `gp3sequences` provides transparent, reproducible, and auditable tools
#' for working with ordered categorical sequences represented in ordinary
#' long-format data frames.
#'
#' The package is designed around explicit column mappings, deterministic
#' ordering, documented preprocessing policies, machine-readable diagnostics,
#' reproducible analysis contracts, and conservative structural summaries.
#' Current workflows cover motifs and subsequences, consensus and group
#' comparisons, distances and clustering, transition networks and higher-order
#' models, categorical hidden Markov models, longitudinal and time-varying
#' analyses, design-aware inference, provenance auditing, visual diagnostics,
#' and guarded interoperability with specialist sequence-analysis packages.
#' It is not restricted to eye-tracking data or to any particular device or
#' export format.
#'
#' @section Interpretation boundary:
#' Sequence outputs describe behavioural or structural patterns. They do not
#' independently establish emotion, cognition, comprehension, personality,
#' intention, diagnosis, deception, or other psychological attributes.
#' Substantive interpretation requires an appropriate study design and
#' external evidence.
#'
#' @keywords internal
"_PACKAGE"

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)
library(gp3sequences)

## ----case-data----------------------------------------------------------------
paths <- list(
  s01 = c("home", "search", "product", "cart", "checkout", "confirmation"),
  s02 = c("home", "search", "product", "reviews", "cart", "checkout"),
  s03 = c("home", "category", "product", "cart", "checkout", "confirmation"),
  s04 = c("home", "search", "category", "product", "cart", "checkout"),
  s05 = c("home", "category", "product", "reviews", "cart", "checkout"),
  s06 = c("home", "search", "product", "cart", "home", "search"),
  s07 = c("home", "category", "search", "product", "checkout", "confirmation"),
  s08 = c("home", "category", "product", "compare", "cart", "checkout"),
  s09 = c("home", "search", "compare", "product", "checkout", "home"),
  s10 = c("home", "category", "compare", "product", "cart", "checkout"),
  s11 = c("home", "search", "product", "compare", "cart", "checkout"),
  s12 = c("home", "category", "product", "checkout", "confirmation", "home")
)

case_data <- do.call(
  rbind,
  lapply(seq_along(paths), function(i) {
    data.frame(
      sequence_id = names(paths)[i],
      sequence_order = seq_along(paths[[i]]),
      state = paths[[i]],
      duration = 75 + 6 * seq_along(paths[[i]]) + 2 * i,
      participant_id = sprintf("p%02d", i),
      interface = if (i <= 6L) "interface_a" else "interface_b",
      stringsAsFactors = FALSE
    )
  })
)

head(case_data, 12L)

## ----case-prepare-------------------------------------------------------------
case_audit <- audit_sequence_data(
  case_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "interface")
)

case_prepared <- prepare_sequence_data(
  case_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "interface"),
  missing_state_policy = "error",
  duplicate_position_policy = "error",
  repeated_state_policy = "preserve",
  zero_duration_policy = "preserve",
  unknown_state_policy = "preserve",
  unused_state_levels = "preserve"
)

case_audit
case_prepared$status
case_prepared$decisions

## ----case-summaries-----------------------------------------------------------
state_summary <- summarise_sequence_states(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "interface")
)

transition_summary <- summarise_sequence_transitions(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "interface"),
  include_self = TRUE
)

path_summary <- format_sequence_paths(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "interface")
)

state_summary$overall
head(transition_summary$overall)
path_summary$paths

## ----case-motifs--------------------------------------------------------------
case_motifs <- extract_sequence_ngrams(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = "interface",
  min_length = 2L,
  max_length = 3L,
  overlap = "allow"
)

case_motif_summary <- summarise_sequence_motifs(case_motifs)
case_motif_filter <- filter_sequence_motifs(
  case_motif_summary,
  min_occurrences = 2L,
  min_sequences = 2L,
  min_prevalence = 0.15,
  motif_lengths = c(2L, 3L),
  top_n = 12L,
  rank_by = "sequence_prevalence",
  ties = "include"
)

format_sequence_motifs(
  case_motif_filter,
  prevalence = "percent",
  digits = 1L
)$table

## ----case-groups--------------------------------------------------------------
state_order <- sort(unique(case_prepared$data$state), method = "radix")

case_consensus <- create_consensus_sequence(
  case_prepared$data,
  group_cols = "interface",
  tie_method = "first",
  state_levels = state_order
)

case_comparison <- compare_sequence_groups(
  case_prepared$data,
  group_col = "interface"
)

format_consensus_sequence(case_consensus, include_agreement = TRUE)
summarise_consensus_agreement(case_consensus, by = "group")
head(case_comparison$state_contrasts)
head(case_comparison$transition_contrasts)
case_comparison$length_contrasts

## ----case-clustering----------------------------------------------------------
case_distance <- compute_sequence_distance(
  case_prepared$data,
  method = "lcs",
  normalise = "max_length"
)

case_cluster <- cluster_sequences(
  case_distance,
  k = 2L,
  method = "hierarchical",
  linkage = "average"
)

case_cluster_validation <- validate_sequence_clusters(case_cluster)
case_representatives <- extract_representative_sequences(case_cluster)

summarise_sequence_distance(case_distance)$overall
case_cluster$assignments
case_cluster_validation$overall
case_representatives

## ----case-network-------------------------------------------------------------
case_network <- create_transition_network(
  case_prepared$data,
  normalise = "from",
  include_self = TRUE
)

case_centrality <- summarise_transition_centrality(case_network)
case_communities <- detect_transition_communities(case_network)

case_order2 <- fit_higher_order_transition_model(
  case_prepared$data,
  order = 2L,
  smoothing = 0.5,
  backoff = TRUE
)

case_network
case_centrality
case_communities
predict_next_state(case_order2, c("home", "search"))
predict_next_state(case_order2, c("unseen"))

## ----case-hmm-----------------------------------------------------------------
one_state <- fit_sequence_hmm(
  case_prepared$data,
  n_states = 1L,
  max_iter = 30L,
  seed = 42L
)

two_state <- fit_sequence_hmm(
  case_prepared$data,
  n_states = 2L,
  max_iter = 50L,
  seed = 42L
)

summarise_sequence_hmm(two_state)$fit
head(decode_sequence_states(two_state, method = "viterbi"))
compare_sequence_hmms(one_state = one_state, two_state = two_state)

## ----case-report--------------------------------------------------------------
case_evidence <- list(
  preparation_status = case_prepared$status,
  preparation_decisions = case_prepared$decisions,
  state_summary = state_summary$overall,
  motif_summary = case_motif_filter$motifs,
  consensus = format_consensus_sequence(
    case_consensus,
    include_agreement = TRUE
  ),
  group_state_contrasts = case_comparison$state_contrasts,
  distance_summary = summarise_sequence_distance(case_distance)$overall,
  cluster_validation = case_cluster_validation$overall,
  representatives = case_representatives,
  network = case_network,
  centrality = case_centrality,
  hmm_comparison = compare_sequence_hmms(
    one_state = one_state,
    two_state = two_state
  )
)

names(case_evidence)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)

## ----data---------------------------------------------------------------------
paths <- list(
  s1 = c("home", "search", "product", "checkout"),
  s2 = c("home", "search", "product", "home"),
  s3 = c("home", "category", "product", "checkout"),
  s4 = c("home", "category", "home", "search")
)
sequence_data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(sequence_id = names(paths)[i],
             sequence_order = seq_along(paths[[i]]),
             state = paths[[i]], stringsAsFactors = FALSE)
}))

## ----network------------------------------------------------------------------
network <- create_transition_network(
  sequence_data,
  normalise = "from",
  include_self = TRUE
)
network
summarise_transition_centrality(network)
detect_transition_communities(network)

## ----higher-order-------------------------------------------------------------
model <- fit_higher_order_transition_model(
  sequence_data,
  order = 2L,
  smoothing = 0.5,
  backoff = TRUE
)
predict_next_state(model, c("home", "search"))
predict_next_state(model, c("unseen"))

## ----bootstrap----------------------------------------------------------------
boot <- bootstrap_transition_network(
  sequence_data,
  n_boot = 20L,
  level = 0.95,
  seed = 8L
)
head(boot)

---
title: "Contiguous Motif Workflow"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Contiguous Motif Workflow}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)
```

## Scope

This article demonstrates the restricted contiguous-motif workflow in
`gp3sequences`. The workflow accepts ordinary long-format data frames and does
not require Gazepoint software, hardware, exports, or `gp3tools`.

The functions describe recurring state windows and their locations. They do not
infer attention, cognition, emotion, intention, psychological state, or causal
mechanisms.

## Synthetic ordered-state data

The example contains three sequences and one preserved grouping variable.

```{r data}
library(gp3sequences)

sequence_data <- data.frame(
  sequence = c(
    rep("s1", 5L),
    rep("s2", 4L),
    rep("s3", 3L)
  ),
  order = c(1:5, 1:4, 1:3),
  state = c(
    "A", "B", "A", "B", "C",
    "A", "B", "C", "B",
    "B", "A", "B"
  ),
  group = c(
    rep("g1", 5L),
    rep("g1", 4L),
    rep("g2", 3L)
  ),
  stringsAsFactors = FALSE
)

sequence_data
```

## Extract contiguous motifs

`extract_sequence_ngrams()` enumerates contiguous state windows only. Minimum
and maximum motif lengths and the overlapping-occurrence policy are explicit.

```{r extract}
extracted <- extract_sequence_ngrams(
  sequence_data,
  sequence_id_col = "sequence",
  order_col = "order",
  state_col = "state",
  metadata_cols = "group",
  min_length = 2L,
  max_length = 3L,
  overlap = "allow"
)

head(extracted$occurrences)
```

## Summarise, filter, and format motifs

Sequence prevalence uses every validated sequence as its denominator. Filtering
is deterministic and retains explicit thresholds and tie handling.

```{r summarise-filter-format}
motif_summary <- summarise_sequence_motifs(extracted)

motif_filter <- filter_sequence_motifs(
  motif_summary,
  min_occurrences = 2L,
  min_sequences = 2L,
  min_prevalence = 0.50,
  motif_lengths = c(2L, 3L),
  top_n = 10L,
  rank_by = "sequence_prevalence",
  ties = "include"
)

motif_table <- format_sequence_motifs(
  motif_filter,
  prevalence = "percent",
  digits = 1L
)

motif_table$table
```

## Summarise motif positions

Positions may represent the start, centre, or end of each motif occurrence.
Absolute positions use one-based state indices. Relative positions range from 0
to 1 across each sequence.

```{r position-summary}
position_summary <- summarise_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  by = "group"
)

position_summary$summary
```

`format_sequence_motif_positions()` changes display precision and units without
modifying the underlying analytical object.

```{r position-format}
position_table <- format_sequence_motif_positions(
  position_summary,
  digits = 1L,
  position_units = "percent",
  include_rank = TRUE
)

position_table$table
```

## Plot motif prevalence

`plot_sequence_motifs()` uses base R graphics. The returned data frame contains
the exact motifs and values used in the plot.

```{r motif-plot, fig.cap="Sequence prevalence for the most frequent contiguous motifs."}
plotted_motifs <- plot_sequence_motifs(
  motif_summary,
  metric = "sequence_prevalence",
  top_n = 8L,
  motif_lengths = c(2L, 3L),
  ties = "include",
  horizontal = TRUE
)

plotted_motifs[
  c("plot_rank", "motif", "motif_length", "plot_value")
]
```

## Plot motif locations

The strip display shows individual occurrence positions with deterministic
stacking. No random jitter is used.

```{r position-strip, fig.cap="Relative centre positions of selected motif occurrences."}
strip_data <- plot_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  top_n = 5L,
  display = "strip"
)

head(
  strip_data[
    c(
      "sequence_id",
      "motif",
      "position_value",
      "plot_rank"
    )
  ]
)
```

The distribution display provides a compact base-R summary for the same
structural positions.

```{r position-distribution, fig.cap="Distributions of relative motif centre positions."}
plot_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  top_n = 5L,
  display = "distribution"
)
```

## Interpretation boundary

The reported counts, prevalence values, and positions describe the supplied
ordered states under the declared preparation, motif-length, overlap, filtering,
and position rules. Any substantive interpretation belongs to the research
design and cannot be inferred automatically from motif structure alone.

---
title: "Sequence Distances, Clustering, and Stability"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Sequence Distances, Clustering, and Stability}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Synthetic paths

```{r data}
paths <- list(
  s1 = c("A", "B", "C", "D"),
  s2 = c("A", "B", "C", "C"),
  s3 = c("A", "C", "C", "D"),
  s4 = c("D", "C", "B", "A"),
  s5 = c("D", "C", "A", "A"),
  s6 = c("D", "B", "B", "A")
)
sequence_data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(sequence_id = names(paths)[i],
             sequence_order = seq_along(paths[[i]]),
             state = paths[[i]], stringsAsFactors = FALSE)
}))
```

## Transparent distance families

```{r distances}
levenshtein <- compute_sequence_distance(sequence_data, method = "levenshtein")
lcs <- compute_sequence_distance(sequence_data, method = "lcs")
om <- compute_sequence_distance(
  sequence_data,
  method = "optimal_matching",
  indel_cost = 1,
  substitution_cost = 2,
  normalise = "max_length"
)
transition <- compute_sequence_distance(sequence_data, method = "transition")
summarise_sequence_distance(lcs)$overall
```

The distance method and all costs are retained as object attributes. The
transition method compares first-order transition-probability profiles; it is
not a general stochastic-process model.

## Clustering and validation

```{r clustering}
fit <- cluster_sequences(lcs, k = 2L, method = "hierarchical", linkage = "average")
fit$assignments
validate_sequence_clusters(fit)$overall
extract_representative_sequences(fit)
```

## Subsampling stability

```{r stability}
stability <- bootstrap_sequence_clusters(
  lcs,
  k = 2L,
  n_boot = 20L,
  sample_fraction = 0.8,
  seed = 100L
)
summarise_sequence_cluster_stability(stability)$overall
```

Cluster stability describes reproducibility under the selected resampling and
clustering settings. It does not establish that the clusters are natural,
causal, or substantively meaningful.

## Co-association ensemble

```{r ensemble}
transition_fit <- cluster_sequences(
  transition,
  k = 2L,
  method = "hierarchical",
  linkage = "average"
)
ensemble <- create_sequence_cluster_ensemble(
  fit,
  transition_fit,
  k = 2L
)
ensemble$assignments
ensemble$coassociation
```

The ensemble records how often pairs are assigned together across supplied
solutions. It does not automatically validate the number or meaning of
clusters.

## Optional PAM and CLARA interfaces

```{r optional-cluster}
if (requireNamespace("cluster", quietly = TRUE)) {
  pam_fit <- cluster_sequences(lcs, k = 2L, method = "pam", seed = 11L)
  clara_fit <- cluster_sequences(
    lcs,
    k = 2L,
    method = "clara",
    seed = 11L,
    samples = 5L,
    sampsize = 5L
  )
  list(pam = pam_fit$assignments, clara = clara_fit$assignments)
}
```

PAM uses the supplied dissimilarities directly. CLARA uses a documented
classical multidimensional-scaling embedding because `cluster::clara()`
expects observations rather than a dissimilarity object.

---
title: "Latent Sequence Models and Optional Adapters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Latent Sequence Models and Optional Adapters}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Interpretation boundary

Hidden states and mixture components are statistical constructs. They must not
be labelled as emotions, cognitive states, diagnoses, intentions, or causal
mechanisms without independent theory, design, and validation.

## Synthetic categorical sequences

```{r data}
paths <- list(
  s1 = c("A", "A", "B", "B", "C"),
  s2 = c("A", "B", "B", "C", "C"),
  s3 = c("A", "A", "B", "C", "C"),
  s4 = c("C", "C", "B", "B", "A"),
  s5 = c("C", "B", "B", "A", "A"),
  s6 = c("C", "C", "B", "A", "A")
)
sequence_data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(sequence_id = names(paths)[i],
             sequence_order = seq_along(paths[[i]]),
             state = paths[[i]], stringsAsFactors = FALSE)
}))
```

## Categorical HMM

```{r hmm}
hmm <- fit_sequence_hmm(
  sequence_data,
  n_states = 2L,
  max_iter = 60L,
  seed = 10L
)
summarise_sequence_hmm(hmm)$fit
head(decode_sequence_states(hmm, method = "viterbi"))

one_state <- fit_sequence_hmm(
  sequence_data,
  n_states = 1L,
  max_iter = 30L,
  seed = 10L
)
compare_sequence_hmms(one_state = one_state, two_state = hmm)
```

## Mixture HMM

```{r mixture}
mixture <- fit_sequence_hmm_mixture(
  sequence_data,
  n_components = 2L,
  n_states = 2L,
  max_iter = 40L,
  inner_initial_iter = 5L,
  seed = 12L
)
summarise_sequence_hmm(mixture)$mixture
mixture$responsibilities
```


## Estimation limitations

The native estimators are compact, dependency-light, time-homogeneous
categorical HMM workflows. EM estimation can converge to local optima, latent
state labels are exchangeable, and AIC or BIC differences do not validate a
substantive interpretation. Analysts should inspect convergence histories, fit
multiple seeded specifications when the result matters, and use a specialist
package such as `seqHMM` for multichannel, covariate-dependent, or more complex
models.

## Optional ecosystem adapters

The adapters are guarded by `requireNamespace()` and do not make specialist
packages mandatory dependencies.

```{r adapters}
grp_input <- as_grpstring_data(sequence_data)
grp_input$key
grp_input$strings

if (requireNamespace("TraMineR", quietly = TRUE)) {
  traminer_sequences <- as_traminer_sequences(sequence_data)
  class(traminer_sequences)
}

if (requireNamespace("TraMineR", quietly = TRUE) &&
    requireNamespace("seqHMM", quietly = TRUE)) {
  seqhmm_sequences <- as_seqhmm_sequences(sequence_data)
  class(seqhmm_sequences)
}

if (requireNamespace("arules", quietly = TRUE) &&
    requireNamespace("arulesSequences", quietly = TRUE)) {
  cspade_input <- as_arules_sequences(sequence_data)
  arules::transactionInfo(cspade_input)
}

network <- create_transition_network(sequence_data)
if (requireNamespace("igraph", quietly = TRUE)) {
  graph <- as_igraph_transition_network(network)
  class(graph)
}

renamed <- sequence_data
names(renamed)[names(renamed) == "sequence_order"] <- "position"
names(renamed)[names(renamed) == "state"] <- "aoi_label"
prepared <- prepare_gp3tools_sequences(renamed)
prepared$status
```

---
title: "Bounded Non-Contiguous Subsequence Mining"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Bounded Non-Contiguous Subsequence Mining}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Why bounded subsequences?

A non-contiguous subsequence preserves order while allowing intervening states.
The implementation requires explicit motif length, maximum gap, maximum span,
and combination limits. These constraints keep enumeration auditable and avoid
silently searching an unbounded combinatorial space.

## Synthetic sequences

```{r data}
sequences <- data.frame(
  sequence_id = rep(paste0("s", 1:6), each = 5L),
  sequence_order = rep(1:5, times = 6L),
  state = c(
    "A", "B", "C", "D", "E",
    "A", "C", "B", "D", "E",
    "A", "B", "D", "C", "E",
    "E", "D", "C", "B", "A",
    "E", "C", "D", "B", "A",
    "E", "D", "B", "C", "A"
  ),
  group = rep(rep(c("forward", "reverse"), each = 3L), each = 5L),
  stringsAsFactors = FALSE
)
```

## Enumerate occurrences

```{r extract}
occurrences <- extract_sequence_subsequences(
  sequences,
  metadata_cols = "group",
  min_length = 2L,
  max_length = 3L,
  max_gap = 2L,
  max_span = 4L,
  repeated_state_policy = "preserve"
)
head(occurrences)
attributes(occurrences)[c("n_sequences", "settings")]
```

## Sequence-level prevalence

```{r summary}
subsequence_summary <- summarise_sequence_subsequences(occurrences)
head(subsequence_summary, 10L)

frequent <- filter_sequence_subsequences(
  subsequence_summary,
  min_sequences = 2L,
  min_prevalence = 0.25,
  top_n = 12L
)
frequent
```

## Group comparison

```{r compare}
comparison <- compare_sequence_subsequences(
  occurrences,
  group_col = "group",
  p_adjust = "holm"
)
head(comparison, 10L)
```

The comparison is based on sequence-level presence, not occurrence multiplicity.
Adjusted p-values do not turn an observational grouping into a causal design.

```{r plot, fig.width=7, fig.height=5}
plot_sequence_subsequences(frequent, metric = "sequence_prevalence")
```

## Relation to specialist packages

The bounded enumerator is intentionally narrow. Large-scale frequent sequence
mining, event-sequence constraint systems, and discriminating-subsequence
algorithms remain appropriate uses of specialist packages through explicit
adapters.

---
title: "Consensus Sequences and Descriptive Group Comparisons"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Consensus Sequences and Descriptive Group Comparisons}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Scope

This workflow describes aligned states and differences in observed sequence
structure. A consensus is not a behavioural norm, and a between-group
difference is not evidence of a psychological or causal mechanism.

## Synthetic sequences

```{r data}
paths <- list(
  s1 = c("home", "search", "product", "checkout"),
  s2 = c("home", "search", "product", "home"),
  s3 = c("home", "category", "product", "checkout"),
  s4 = c("home", "category", "home", "home")
)
sequence_data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(
    sequence_id = names(paths)[i],
    sequence_order = seq_along(paths[[i]]),
    state = paths[[i]],
    group = rep(c("interface_a", "interface_b"), each = 2L)[i],
    stringsAsFactors = FALSE
  )
}))
sequence_data
```

## Aligned-position consensus

```{r consensus}
consensus <- create_consensus_sequence(
  sequence_data,
  group_cols = "group",
  tie_method = "first",
  state_levels = c("home", "search", "category", "product", "checkout")
)
consensus
summarise_consensus_agreement(consensus, by = "group")
format_consensus_sequence(consensus, include_agreement = TRUE)
```

```{r consensus-plot, fig.width=7, fig.height=4}
plot_consensus_sequence(consensus, type = "agreement", group = "interface_a")
```

## Descriptive group comparison

```{r comparison}
comparison <- compare_sequence_groups(
  sequence_data,
  group_col = "group"
)
comparison$groups
head(comparison$state_contrasts)
head(comparison$transition_contrasts)
comparison$length_contrasts
```

```{r comparison-plot, fig.width=7, fig.height=5}
plot_sequence_group_comparison(comparison, component = "state", top_n = 5L)
```

The output reports counts, shares, prevalence, differences, and ratios. It does
not compute a significance test or automatically rank one group as preferable.

---
title: "Extended Sequence Visualisations"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Extended Sequence Visualisations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Synthetic data

```{r data}
data <- data.frame(
  sequence_id = rep(paste0("s", 1:8), each = 6L),
  sequence_order = rep(1:6, times = 8L),
  state = c(
    rep(c("A", "B", "B", "C", "D", "D"), 4L),
    rep(c("D", "C", "C", "B", "A", "A"), 4L)
  ),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(data, method = "levenshtein")
clustering <- cluster_sequences(distance, k = 2L, method = "hierarchical")
network <- create_transition_network(data)
```

## Sequence index

```{r index, fig.width=7, fig.height=5}
plot_sequence_index(data)
```

## State distribution and entropy

```{r distribution, fig.width=7, fig.height=4}
plot_sequence_state_distribution(data)
```

```{r entropy, fig.width=7, fig.height=4}
plot_sequence_entropy(data)
```

Entropy is a structural diversity summary at each aligned position. It is not a
measure of participant uncertainty or cognition.

## Distance and clustering diagnostics

```{r distance, fig.width=6, fig.height=5}
plot_sequence_distance_heatmap(distance)
```

```{r silhouette, fig.width=7, fig.height=4}
plot_sequence_cluster_silhouette(clustering, distance)
```

## Transition network

```{r network, fig.width=6, fig.height=6}
plot_transition_network(network)
```

These base-R plots are intentionally focused on package-native audited objects.
They complement, rather than replace, specialist visualisation ecosystems.

---
title: "Longitudinal and Panel Sequence Workflows"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Longitudinal and Panel Sequence Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Scope

Panel sequences are repeated ordered-state records from the same independent
unit. The workflow preserves the panel identifier, occasion, sequence identity,
and preprocessing decisions. Distance between occasions is a structural change
measure; it is not evidence of learning, adaptation, or causality by itself.

## Synthetic data

```{r data}
base <- data.frame(
  participant_id = rep(paste0("p", 1:4), each = 8L),
  occasion = rep(rep(c(1, 2), each = 4L), times = 4L),
  sequence_id = rep(paste0("s", 1:8), each = 4L),
  sequence_order = rep(1:4, times = 8L),
  state = c(
    "A", "B", "C", "D", "A", "B", "D", "D",
    "A", "C", "C", "D", "A", "C", "D", "D",
    "D", "C", "B", "A", "D", "C", "A", "A",
    "D", "B", "B", "A", "D", "B", "A", "A"
  ),
  stringsAsFactors = FALSE
)
head(base)
```

## Prepare and audit the panel

```{r prepare}
panel <- prepare_sequence_panel(
  base,
  panel_id_col = "participant_id",
  occasion_col = "occasion"
)
panel$index
```

A unique panel/occasion combination is required by default. This prevents two
sequences from being silently treated as the same repeated observation.

## Summarise occasions and states

```{r summaries}
panel_summary <- summarise_sequence_panel(panel)
panel_summary$occasions
head(panel_summary$states)
```

## Quantify within-panel change

```{r change}
changes <- compare_sequence_panel_changes(
  panel,
  method = "levenshtein",
  normalise = "max_length"
)
changes
```

Alternative distance methods use the same explicit arguments as
`compute_sequence_distance()`. The result compares consecutive occasions within
each panel only.

```{r plot, fig.width=7, fig.height=4}
plot_sequence_panel_changes(changes, metric = "distance", type = "individual")
plot_sequence_panel_changes(changes, metric = "distance", type = "summary")
```

## Reporting

Report the panel unit, occasion ordering, distance method, normalisation,
sequence counts at each occasion, and any missing occasions. Treat change as a
structural description unless a separate design supports stronger inference.

---
title: "Sequence Data Validation and Preparation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Sequence Data Validation and Preparation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Why preparation is explicit

Ordered categorical data can contain missing states, duplicated positions,
unsorted rows, consecutive repeats, zero durations, unknown states, and
inconsistent metadata. Silent repair can change the analytical object.
`gp3sequences` therefore separates non-modifying audit and validation from
policy-driven preparation.

## A deliberately problematic synthetic input

The example includes an unsorted sequence, a duplicated position, a missing
state, a consecutive repeat, a zero duration, an unexpected state, and an
unused factor level. Participant and group metadata remain constant within
each sequence.

```{r problematic-data}
problem_data <- data.frame(
  sequence_id = c("s2", "s1", "s1", "s1", "s1", "s2", "s2", "s2", "s2"),
  sequence_order = c(2, 2, 1, 2, 3, 1, 2, 3, 4),
  state = factor(
    c("search", "search", "home", "search", NA,
      "home", "home", "product", "other"),
    levels = c("home", "search", "product", "checkout", "other", "unused")
  ),
  duration = c(120, 100, 90, 110, 80, 100, 0, 150, 130),
  participant_id = c("p2", "p1", "p1", "p1", "p1", "p2", "p2", "p2", "p2"),
  group = c("interface_b", "interface_a", "interface_a", "interface_a",
            "interface_a", "interface_b", "interface_b", "interface_b",
            "interface_b"),
  stringsAsFactors = FALSE
)

expected_states <- c("home", "search", "product", "checkout")
problem_data
```

## Audit without modification

`audit_sequence_data()` reports one row per issue using stable issue codes and
severity values. It does not repair the data.

```{r audit}
audit <- audit_sequence_data(
  problem_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states
)

audit
as.data.frame(table(audit$severity), stringsAsFactors = FALSE)
as.data.frame(table(audit$issue_code), stringsAsFactors = FALSE)
```

## Compact validation contract

A review-level issue does not automatically invalidate an input. Error-level
issues must be resolved through source correction or an explicit supported
policy.

```{r validation}
validation <- validate_sequence_data(
  problem_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states
)

validation[c("valid", "status", "n_errors", "n_reviews", "n_info")]
validation$mapping
```

## Apply explicit preparation policies

This example deliberately chooses to:

- drop rows with missing states;
- retain the first row at duplicated positions;
- collapse consecutive repeated states;
- drop zero-duration rows;
- drop states absent from the declared state set;
- drop unused factor levels.

These are analytical choices, not universal defaults.

```{r prepare}
prepared <- prepare_sequence_data(
  problem_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states,
  missing_state_policy = "drop",
  duplicate_position_policy = "first",
  repeated_state_policy = "collapse",
  zero_duration_policy = "drop",
  unknown_state_policy = "drop",
  unused_state_levels = "drop"
)

prepared$status
prepared$decisions
prepared$data
prepared$audit
```

## Revalidate the canonical result

The prepared table uses stable canonical columns while preserving unmapped
metadata and original-row provenance.

```{r revalidate}
revalidation <- validate_sequence_data(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  expected_states = expected_states
)

revalidation[c("valid", "status", "n_errors", "n_reviews", "n_info")]
prepared$mapping
prepared$state_levels
```

## Errors that require source correction

Some conditions are intentionally not repaired automatically. Examples include
missing sequence identifiers, missing or non-finite order values, negative or
non-finite durations, absent mapped columns, duplicated column names, invalid
column types, and metadata that varies within a sequence. These conditions
require correction or an explicit redefinition of the sequence unit.

## Reporting recommendations

A reproducible report should record the input mapping, expected states, every
preparation policy, the audit table, the decision log, original and prepared
row counts, and the final state levels. These records describe data handling;
they do not validate a substantive interpretation of the resulting sequence
patterns.

---
title: "Transition Networks and Higher-Order Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Transition Networks and Higher-Order Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Synthetic navigation sequences

```{r data}
paths <- list(
  s1 = c("home", "search", "product", "checkout"),
  s2 = c("home", "search", "product", "home"),
  s3 = c("home", "category", "product", "checkout"),
  s4 = c("home", "category", "home", "search")
)
sequence_data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(sequence_id = names(paths)[i],
             sequence_order = seq_along(paths[[i]]),
             state = paths[[i]], stringsAsFactors = FALSE)
}))
```

## First-order transition network

```{r network}
network <- create_transition_network(
  sequence_data,
  normalise = "from",
  include_self = TRUE
)
network
summarise_transition_centrality(network)
detect_transition_communities(network)
```

Centrality values are graph-structural descriptors. They do not independently
measure attention, importance, intent, or influence.

## Higher-order transition model

```{r higher-order}
model <- fit_higher_order_transition_model(
  sequence_data,
  order = 2L,
  smoothing = 0.5,
  backoff = TRUE
)
predict_next_state(model, c("home", "search"))
predict_next_state(model, c("unseen"))
```

## Whole-sequence bootstrap

```{r bootstrap}
boot <- bootstrap_transition_network(
  sequence_data,
  n_boot = 20L,
  level = 0.95,
  seed = 8L
)
head(boot)
```

Bootstrap samples are drawn at the sequence level, preserving within-sequence
transition dependence.

---
title: "Reproducible Sequence Analysis: A Synthetic Case Study"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reproducible Sequence Analysis: A Synthetic Case Study}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)
library(gp3sequences)
```

## Study objective

This synthetic example examines navigation-path structure under two assigned
interface conditions. It demonstrates an auditable workflow rather than a
claim that sequence structure reveals hidden attention, preference, cognition,
emotion, intention, or causality.

## Synthetic study data

The data contain 12 independent participant-level sequences of equal maximum
length. Interface condition is assigned as sequence-level metadata and all
state labels are directly observed navigation locations.

```{r case-data}
paths <- list(
  s01 = c("home", "search", "product", "cart", "checkout", "confirmation"),
  s02 = c("home", "search", "product", "reviews", "cart", "checkout"),
  s03 = c("home", "category", "product", "cart", "checkout", "confirmation"),
  s04 = c("home", "search", "category", "product", "cart", "checkout"),
  s05 = c("home", "category", "product", "reviews", "cart", "checkout"),
  s06 = c("home", "search", "product", "cart", "home", "search"),
  s07 = c("home", "category", "search", "product", "checkout", "confirmation"),
  s08 = c("home", "category", "product", "compare", "cart", "checkout"),
  s09 = c("home", "search", "compare", "product", "checkout", "home"),
  s10 = c("home", "category", "compare", "product", "cart", "checkout"),
  s11 = c("home", "search", "product", "compare", "cart", "checkout"),
  s12 = c("home", "category", "product", "checkout", "confirmation", "home")
)

case_data <- do.call(
  rbind,
  lapply(seq_along(paths), function(i) {
    data.frame(
      sequence_id = names(paths)[i],
      sequence_order = seq_along(paths[[i]]),
      state = paths[[i]],
      duration = 75 + 6 * seq_along(paths[[i]]) + 2 * i,
      participant_id = sprintf("p%02d", i),
      interface = if (i <= 6L) "interface_a" else "interface_b",
      stringsAsFactors = FALSE
    )
  })
)

head(case_data, 12L)
```

## Prespecified preparation

The synthetic input is expected to be complete and uniquely ordered. Policies
therefore refuse missing states and duplicate positions while preserving
repeated states and positive durations.

```{r case-prepare}
case_audit <- audit_sequence_data(
  case_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "interface")
)

case_prepared <- prepare_sequence_data(
  case_data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "interface"),
  missing_state_policy = "error",
  duplicate_position_policy = "error",
  repeated_state_policy = "preserve",
  zero_duration_policy = "preserve",
  unknown_state_policy = "preserve",
  unused_state_levels = "preserve"
)

case_audit
case_prepared$status
case_prepared$decisions
```

## Structural summaries

```{r case-summaries}
state_summary <- summarise_sequence_states(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "interface")
)

transition_summary <- summarise_sequence_transitions(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "interface"),
  include_self = TRUE
)

path_summary <- format_sequence_paths(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "interface")
)

state_summary$overall
head(transition_summary$overall)
path_summary$paths
```

## Recurring contiguous motifs

```{r case-motifs}
case_motifs <- extract_sequence_ngrams(
  case_prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = "interface",
  min_length = 2L,
  max_length = 3L,
  overlap = "allow"
)

case_motif_summary <- summarise_sequence_motifs(case_motifs)
case_motif_filter <- filter_sequence_motifs(
  case_motif_summary,
  min_occurrences = 2L,
  min_sequences = 2L,
  min_prevalence = 0.15,
  motif_lengths = c(2L, 3L),
  top_n = 12L,
  rank_by = "sequence_prevalence",
  ties = "include"
)

format_sequence_motifs(
  case_motif_filter,
  prevalence = "percent",
  digits = 1L
)$table
```

## Consensus and condition contrasts

```{r case-groups}
state_order <- sort(unique(case_prepared$data$state), method = "radix")

case_consensus <- create_consensus_sequence(
  case_prepared$data,
  group_cols = "interface",
  tie_method = "first",
  state_levels = state_order
)

case_comparison <- compare_sequence_groups(
  case_prepared$data,
  group_col = "interface"
)

format_consensus_sequence(case_consensus, include_agreement = TRUE)
summarise_consensus_agreement(case_consensus, by = "group")
head(case_comparison$state_contrasts)
head(case_comparison$transition_contrasts)
case_comparison$length_contrasts
```

## Distance, clustering, and representatives

The clustering layer is declared in advance as normalised LCS distance,
two-cluster average-linkage hierarchical clustering, and standard structural
validation summaries.

```{r case-clustering}
case_distance <- compute_sequence_distance(
  case_prepared$data,
  method = "lcs",
  normalise = "max_length"
)

case_cluster <- cluster_sequences(
  case_distance,
  k = 2L,
  method = "hierarchical",
  linkage = "average"
)

case_cluster_validation <- validate_sequence_clusters(case_cluster)
case_representatives <- extract_representative_sequences(case_cluster)

summarise_sequence_distance(case_distance)$overall
case_cluster$assignments
case_cluster_validation$overall
case_representatives
```

## Transition network and recent-context model

```{r case-network}
case_network <- create_transition_network(
  case_prepared$data,
  normalise = "from",
  include_self = TRUE
)

case_centrality <- summarise_transition_centrality(case_network)
case_communities <- detect_transition_communities(case_network)

case_order2 <- fit_higher_order_transition_model(
  case_prepared$data,
  order = 2L,
  smoothing = 0.5,
  backoff = TRUE
)

case_network
case_centrality
case_communities
predict_next_state(case_order2, c("home", "search"))
predict_next_state(case_order2, c("unseen"))
```

## Compact categorical HMM sensitivity description

The native HMM is included as a compact statistical summary, not as a source of
substantive state labels. A one-state and two-state model are compared
descriptively using the same observations and symbol coding.

```{r case-hmm}
one_state <- fit_sequence_hmm(
  case_prepared$data,
  n_states = 1L,
  max_iter = 30L,
  seed = 42L
)

two_state <- fit_sequence_hmm(
  case_prepared$data,
  n_states = 2L,
  max_iter = 50L,
  seed = 42L
)

summarise_sequence_hmm(two_state)$fit
head(decode_sequence_states(two_state, method = "viterbi"))
compare_sequence_hmms(one_state = one_state, two_state = two_state)
```

## Assemble report-ready evidence

```{r case-report}
case_evidence <- list(
  preparation_status = case_prepared$status,
  preparation_decisions = case_prepared$decisions,
  state_summary = state_summary$overall,
  motif_summary = case_motif_filter$motifs,
  consensus = format_consensus_sequence(
    case_consensus,
    include_agreement = TRUE
  ),
  group_state_contrasts = case_comparison$state_contrasts,
  distance_summary = summarise_sequence_distance(case_distance)$overall,
  cluster_validation = case_cluster_validation$overall,
  representatives = case_representatives,
  network = case_network,
  centrality = case_centrality,
  hmm_comparison = compare_sequence_hmms(
    one_state = one_state,
    two_state = two_state
  )
)

names(case_evidence)
```

## Interpretation boundary

The workflow documents recurring paths, aligned-position support, descriptive
condition contrasts, dissimilarity, clustering reproducibility, transition
structure, recent-context probabilities, and latent statistical summaries.
None of these outputs independently identifies attention, preference,
comprehension, emotion, cognition, intention, diagnosis, deception, or causal
mechanisms. Such interpretation requires an appropriate design, external
measurement, and independent validation.

---
title: "Time-Varying Condition Comparisons"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Time-Varying Condition Comparisons}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Model target

`fit_time_varying_sequence_model()` estimates the probability of a declared
state or transition over aligned sequence time. It uses group-specific smooths
and can include a participant random-effect smooth. The model concerns a
predeclared structural outcome, not an unobserved psychological state.

## Synthetic repeated sequences

```{r data}
set.seed(1)
participants <- paste0("p", 1:24)
x <- do.call(
  rbind,
  lapply(seq_along(participants), function(i) {
    time <- 1:12
    group <- if (i <= 12L) "control" else "treatment"
    linear <-
      -0.4 +
      0.06 * time +
      0.35 * (group == "treatment") * sin(time / 3)

    data.frame(
      participant_id = participants[i],
      sequence_id = participants[i],
      sequence_order = time,
      group = group,
      state = ifelse(
        stats::runif(length(time)) < stats::plogis(linear),
        "A",
        "B"
      ),
      stringsAsFactors = FALSE
    )
  })
)
```

## Fit and inspect

```{r fit}
if (requireNamespace("mgcv", quietly = TRUE)) {
  model <- fit_time_varying_sequence_model(
    x,
    group_col = "group",
    participant_id_col = "participant_id",
    target_state = "A",
    k = 5L
  )
  model_summary <- summarise_time_varying_sequence_model(model)
  model_summary$metadata
  model_summary$parametric_terms
  model_summary$smooth_terms
}
```

## Predictions

```{r predict}
if (requireNamespace("mgcv", quietly = TRUE)) {
  predictions <- predict_time_varying_sequence_model(
    model,
    time = seq(1, 12, length.out = 60L),
    level = 0.95
  )
  head(predictions)
}
```

```{r plot, fig.width=7, fig.height=4}
if (requireNamespace("mgcv", quietly = TRUE)) {
  plot_time_varying_sequence_model(model)
}
```

## Transition outcomes

Use `outcome = "transition"` together with `from_state` and `to_state` to model
a predeclared transition. The time coordinate refers to the origin position.

## Interpretation

Pointwise intervals describe uncertainty conditional on the fitted model. A
time-varying association is not automatically a causal condition effect. Causal
language requires valid assignment, implementation, estimand definition, and an
analysis aligned with the experimental design.

---
title: "Design-Aware Sequence Group Inference"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Design-Aware Sequence Group Inference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(gp3sequences)
```

## Why declare the design?

Sequence rows are rarely independent. The comparison API requires the group
column and independent unit, and can additionally record pairs or assignment
clusters. It aggregates the selected metric before permutation or bootstrap
resampling.

## Synthetic randomized groups

```{r data}
paths <- replicate(20, sample(c("A", "B", "C"), 6L, replace = TRUE), simplify = FALSE)
data <- do.call(rbind, lapply(seq_along(paths), function(i) {
  data.frame(
    participant_id = paste0("p", i),
    sequence_id = paste0("s", i),
    sequence_order = seq_along(paths[[i]]),
    state = paths[[i]],
    group = if (i <= 10L) "control" else "treatment",
    stringsAsFactors = FALSE
  )
}))
```

## Declare and test

```{r test}
design <- declare_sequence_comparison_design(
  group_col = "group",
  unit_col = "participant_id",
  design = "randomized"
)

result <- test_sequence_group_difference(
  data,
  design,
  metric = "state_prevalence",
  target_state = "A",
  n_permutations = 999L,
  seed = 10L
)
result$estimate
```

Supported metrics are deliberately limited to transparent quantities:
sequence length, transition count, state prevalence, and declared subsequence
presence.

## Bootstrap interval

```{r bootstrap}
result <- bootstrap_sequence_group_difference(
  result,
  n_boot = 999L,
  level = 0.95,
  seed = 11L
)
summarise_sequence_group_inference(result)
```

```{r plots, fig.width=7, fig.height=4}
plot_sequence_group_inference(result, type = "permutation")
plot_sequence_group_inference(result, type = "group_means")
```

## Causal language

For `design = "randomized"`, causal interpretation still depends on valid
assignment, implementation, attrition handling, and an estimand consistent with
the design. For `design = "observational"`, the output explicitly describes an
associational exchangeability-based contrast and does not license causal claims.

---
title: "Getting Started with gp3sequences"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with gp3sequences}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)
library(gp3sequences)
```

## Purpose

`gp3sequences` accepts ordinary long-format ordered categorical data. This
article introduces the complete structural workflow: audit, preparation,
encoding, descriptive summaries, contiguous motifs, distances, clustering,
consensus, group comparison, and transition structure.

All outputs are structural or statistical. They do not independently establish
attention, cognition, emotion, comprehension, intention, diagnosis, causality,
or other psychological attributes.

## Synthetic long-format data

Each sequence has an identifier, an explicit order, a categorical state, a
positive duration, participant metadata, and an assigned interface group.

```{r data}
paths <- list(
  s1 = c("home", "search", "product", "cart", "checkout"),
  s2 = c("home", "search", "product", "cart", "home"),
  s3 = c("home", "category", "product", "cart", "checkout"),
  s4 = c("home", "category", "product", "search", "checkout"),
  s5 = c("home", "category", "search", "product", "checkout"),
  s6 = c("home", "search", "category", "product", "home"),
  s7 = c("home", "category", "product", "cart", "home"),
  s8 = c("home", "search", "product", "checkout", "home")
)

raw_sequences <- do.call(
  rbind,
  lapply(seq_along(paths), function(i) {
    data.frame(
      sequence_id = names(paths)[i],
      sequence_order = seq_along(paths[[i]]),
      state = paths[[i]],
      duration = 80 + 10 * seq_along(paths[[i]]) + i,
      participant_id = sprintf("p%02d", i),
      group = if (i <= 4L) "interface_a" else "interface_b",
      stringsAsFactors = FALSE
    )
  })
)

raw_sequences
```

## Audit, validate, and prepare

`audit_sequence_data()` returns a machine-readable issue table without
modifying the input. `validate_sequence_data()` adds a compact status contract.
`prepare_sequence_data()` applies explicit policies and returns canonical data,
an audit trail, and a decision log.

```{r contract}
audit <- audit_sequence_data(
  raw_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

validation <- validate_sequence_data(
  raw_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

prepared <- prepare_sequence_data(
  raw_sequences,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group"),
  missing_state_policy = "error",
  duplicate_position_policy = "error",
  repeated_state_policy = "preserve",
  zero_duration_policy = "preserve",
  unknown_state_policy = "preserve",
  unused_state_levels = "preserve"
)

validation$status
prepared$status
prepared$mapping
prepared$decisions
head(prepared$data)
```

## Encode states and inspect basic structure

State codes are transparent identifiers derived from a deterministic state
ordering. The state, transition, and path helpers describe the observed
structure without assigning substantive meaning to the labels.

```{r summaries}
encoded <- encode_sequence_data(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

state_summary <- summarise_sequence_states(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  duration_col = "duration",
  metadata_cols = c("participant_id", "group")
)

transition_summary <- summarise_sequence_transitions(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "group"),
  include_self = TRUE
)

paths_table <- format_sequence_paths(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = c("participant_id", "group")
)

encoded$dictionary
state_summary$overall
head(transition_summary$overall)
paths_table$paths
```

## Discover contiguous motifs

Motifs are exact contiguous state windows. Their lengths, overlap rule,
prevalence denominator, filtering thresholds, and tie policy remain explicit.

```{r motifs}
motif_occurrences <- extract_sequence_ngrams(
  prepared$data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = "group",
  min_length = 2L,
  max_length = 3L,
  overlap = "allow"
)

motif_summary <- summarise_sequence_motifs(motif_occurrences)

motif_filter <- filter_sequence_motifs(
  motif_summary,
  min_occurrences = 2L,
  min_sequences = 2L,
  min_prevalence = 0.20,
  motif_lengths = c(2L, 3L),
  top_n = 10L,
  rank_by = "sequence_prevalence",
  ties = "include"
)

motif_table <- format_sequence_motifs(
  motif_filter,
  prevalence = "percent",
  digits = 1L
)

motif_table$table
```

## Compare sequences through distances and clustering

Distance choice is part of the analysis specification. This example uses LCS
distance followed by average-linkage hierarchical clustering. Validation
summaries describe the supplied solution; they do not prove that the clusters
are natural or substantively meaningful.

```{r distance-clustering}
lcs_distance <- compute_sequence_distance(
  prepared$data,
  method = "lcs",
  normalise = "max_length"
)

cluster_fit <- cluster_sequences(
  lcs_distance,
  k = 2L,
  method = "hierarchical",
  linkage = "average"
)

cluster_validation <- validate_sequence_clusters(cluster_fit)
representatives <- extract_representative_sequences(cluster_fit)

summarise_sequence_distance(lcs_distance)$overall
cluster_fit$assignments
cluster_validation$overall
representatives
```

## Consensus and descriptive group comparison

Aligned-position consensus and group contrasts remain descriptive. The
consensus is not a behavioural norm, and contrasts do not establish a causal
mechanism.

```{r consensus-groups}
consensus <- create_consensus_sequence(
  prepared$data,
  group_cols = "group",
  tie_method = "first",
  state_levels = encoded$dictionary$state
)

group_comparison <- compare_sequence_groups(
  prepared$data,
  group_col = "group"
)

summarise_consensus_agreement(consensus, by = "group")
format_consensus_sequence(consensus, include_agreement = TRUE)
head(group_comparison$state_contrasts)
head(group_comparison$transition_contrasts)
group_comparison$length_contrasts
```

## Transition structure

A first-order transition network summarises observed state-to-state movement.
Centrality and community outputs are graph descriptors, not measures of
attention, influence, preference, or intention.

```{r networks}
network <- create_transition_network(
  prepared$data,
  normalise = "from",
  include_self = TRUE
)

network
summarise_transition_centrality(network)
detect_transition_communities(network)
```

## Continue with focused articles

The package website contains focused articles for motif positions, consensus
and groups, distances and clustering, transition networks, latent models, and
optional ecosystem adapters. The method-selection article provides a compact
guide to choosing among them.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-latent-models.R
\name{compare_sequence_hmms}
\alias{compare_sequence_hmms}
\title{Compare fitted sequence HMMs descriptively}
\usage{
compare_sequence_hmms(...)
}
\arguments{
\item{...}{Fitted HMM or mixture-HMM objects.}
}
\value{
A data frame containing log likelihood, AIC, BIC, parameter count,
convergence status, and delta criteria. Criteria are descriptive and do not
automatically select a substantive model.
}
\description{
Compare fitted sequence HMMs descriptively
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
model_1 <- fit_sequence_hmm(sequences, 1L, max_iter = 5L)
model_2 <- fit_sequence_hmm(sequences, 2L, max_iter = 5L)
compare_sequence_hmms(one = model_1, two = model_2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-inference.R
\name{plot_sequence_group_inference}
\alias{plot_sequence_group_inference}
\title{Plot sequence group inference}
\usage{
plot_sequence_group_inference(
  inference,
  type = c("permutation", "group_means"),
  ...
)
}
\arguments{
\item{inference}{A sequence group-inference object.}

\item{type}{\code{"permutation"} or \code{"group_means"}.}

\item{...}{Additional base-graphics arguments.}
}
\value{
The inference object, invisibly.
}
\description{
Plot sequence group inference
}
\examples{
# See `test_sequence_group_difference()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motif-visualisation.R
\name{summarise_sequence_motif_positions}
\alias{summarise_sequence_motif_positions}
\title{Summarise Sequence Motif Positions}
\usage{
summarise_sequence_motif_positions(
  x,
  position = c("start", "centre", "end"),
  scale = c("absolute", "relative"),
  by = NULL
)
}
\arguments{
\item{x}{An object returned by \code{extract_sequence_ngrams()}.}

\item{position}{Position represented by each occurrence: motif \code{"start"},
\code{"centre"}, or \code{"end"}.}

\item{scale}{Position scale: one-based \code{"absolute"} sequence positions or
\code{"relative"} positions from 0 to 1.}

\item{by}{Optional character vector naming preserved metadata columns used
to produce separate summaries, such as \code{"group"} or \code{"condition"}.}
}
\value{
A named list containing:
\itemize{
\item \code{summary}: one row per motif and optional metadata group;
\item \code{occurrences}: occurrence-level absolute, relative, and selected-scale
positions;
\item \code{sequences}, validation metadata, and extraction settings;
\item \code{settings}: the resolved position basis, scale, and grouping columns.
}

The summary reports motif identifiers and labels, motif length, occurrence
and sequence counts, and minimum, maximum, mean, and median positions.
}
\description{
Summarises where contiguous motif occurrences appear within sequences.
}
\details{
Absolute positions use the one-based state index within each validated
sequence. Relative positions are calculated as
\code{(absolute_position - 1) / (n_states - 1)} and are constrained to the
interval from 0 to 1. A sequence containing one state is assigned relative
position 0.

Grouping is descriptive. The function does not test differences or attach
behavioural, psychological, cognitive, or causal interpretations to motif
location.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "C", "A", "B", "C", "B"),
  group = c(rep("g1", 5L), rep("g2", 4L))
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state",
  metadata_cols = "group",
  min_length = 2,
  max_length = 3
)

positions <- summarise_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  by = "group"
)

positions$summary

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{compute_sequence_distance}
\alias{compute_sequence_distance}
\title{Compute pairwise sequence distances}
\usage{
compute_sequence_distance(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  method = c("levenshtein", "lcs", "optimal_matching", "transition"),
  indel_cost = 1,
  substitution_cost = 1,
  substitution_matrix = NULL,
  transition_smoothing = 0,
  normalise = c("none", "max_length", "path_length")
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared gp3sequences result.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{method}{Distance method.}

\item{indel_cost}{Non-negative insertion/deletion cost used by
\code{method = "optimal_matching"}.}

\item{substitution_cost}{Non-negative default substitution cost used by
\code{method = "optimal_matching"}.}

\item{substitution_matrix}{Optional named square substitution-cost matrix for
\code{method = "optimal_matching"}.}

\item{transition_smoothing}{Non-negative smoothing for transition profiles.}

\item{normalise}{One of \code{"none"}, \code{"max_length"}, or \code{"path_length"}.}
}
\value{
A \code{dist} object with method and preprocessing metadata attached.
}
\description{
Provides transparent base-R implementations of Levenshtein edit distance,
longest-common-subsequence distance, optimal matching, and a first-order
transition-profile distance.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
compute_sequence_distance(sequences, method = "lcs")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motifs.R
\name{summarise_sequence_motifs}
\alias{summarise_sequence_motifs}
\title{Summarise Contiguous Sequence Motifs}
\usage{
summarise_sequence_motifs(x)
}
\arguments{
\item{x}{An object returned by \code{extract_sequence_ngrams()}.}
}
\value{
A named list containing:
\itemize{
\item \code{by_sequence}: occurrence counts for each sequence-motif pair;
\item \code{overall}: total occurrence counts, sequence counts, sequence prevalence,
occurrence share, and mean occurrence rates;
\item \code{sequences} and \code{state_dictionary} from extraction;
\item \code{audit}, \code{status}, \code{mapping}, and extraction settings;
\item scalar counts for sequences, occurrences, and distinct motifs.
}
}
\description{
Aggregates extracted contiguous motif occurrences by sequence and overall.
}
\details{
Sequence prevalence uses every validated sequence in the extraction object as
its denominator, including sequences too short to contain a requested motif.
Results are sorted deterministically by sequence prevalence, occurrence
count, motif length, and motif key.

The function reports structural recurrence only. It does not perform
significance testing or infer psychological, cognitive, emotional, or
diagnostic attributes.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state",
  min_length = 2,
  max_length = 3
)

summaries <- summarise_sequence_motifs(extracted)
summaries$by_sequence
summaries$overall

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-consensus-groups.R
\name{compare_sequence_groups}
\alias{compare_sequence_groups}
\title{Compare sequence groups descriptively}
\usage{
compare_sequence_groups(
  data,
  group_col,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  reference = NULL,
  metrics = c("state", "transition", "length"),
  include_self = TRUE,
  transition_separator = " -> ",
  zero_policy = c("missing", "infinite")
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{group_col}{Grouping column that is constant within sequence.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{reference}{Optional reference group. When supplied, each other
group is reported as \code{group_1} and the reference as \code{group_2}; differences
and ratios are therefore other-minus-reference and other/reference. When
omitted, all pairwise group contrasts are returned.}

\item{metrics}{Any of \code{"state"}, \code{"transition"}, and \code{"length"}.}

\item{include_self}{Include self-transitions.}

\item{transition_separator}{Separator used in reported transition labels. It
must not occur inside an observed state label.}

\item{zero_policy}{Ratio policy when the reference value is zero.}
}
\value{
A list of class \code{gp3_sequence_group_comparison} containing group
summaries and pairwise descriptive contrasts.
}
\description{
Produces descriptive state, transition and sequence-length comparisons.
No hypothesis tests or causal interpretations are produced.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
compare_sequence_groups(sequences, group_col = "group")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-networks.R
\name{bootstrap_transition_network}
\alias{bootstrap_transition_network}
\title{Bootstrap transition-network edge weights}
\usage{
bootstrap_transition_network(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  n_boot = 100L,
  level = 0.95,
  seed = 1L,
  include_self = TRUE
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{n_boot}{Number of bootstrap samples of whole sequences.}

\item{level}{Confidence level for percentile intervals.}

\item{seed}{Reproducibility seed.}

\item{include_self}{Include self-transitions.}
}
\value{
A data frame containing observed edge weights, bootstrap means,
standard deviations, and percentile intervals.
}
\description{
Bootstrap transition-network edge weights
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
bootstrap_transition_network(sequences, n_boot = 5L, seed = 1L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-data.R
\name{prepare_sequence_data}
\alias{prepare_sequence_data}
\title{Prepare Long-Format Sequence Data}
\usage{
prepare_sequence_data(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL,
  missing_state_policy = c("error", "drop"),
  duplicate_position_policy = c("error", "first", "last"),
  repeated_state_policy = c("preserve", "collapse"),
  zero_duration_policy = c("preserve", "drop", "error"),
  unknown_state_policy = c("preserve", "drop", "error"),
  unused_state_levels = c("preserve", "drop")
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{duration_col}{Optional name of a numeric duration column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}

\item{missing_state_policy}{Policy for missing states: \code{"error"} or
\code{"drop"}.}

\item{duplicate_position_policy}{Policy for duplicated sequence positions:
\code{"error"}, \code{"first"}, or \code{"last"}.}

\item{repeated_state_policy}{Policy for consecutive repeated states:
\code{"preserve"} or \code{"collapse"}.}

\item{zero_duration_policy}{Policy for zero durations: \code{"preserve"},
\code{"drop"}, or \code{"error"}.}

\item{unknown_state_policy}{Policy for states absent from
\code{expected_states}: \code{"preserve"}, \code{"drop"}, or \code{"error"}.}

\item{unused_state_levels}{Policy for unused factor levels: \code{"preserve"}
or \code{"drop"}.}
}
\value{
A named list containing:
\itemize{
\item \code{data}: canonical prepared data, or \code{NULL} when unresolved errors remain;
\item \code{audit}: input- and output-stage diagnostics;
\item \code{decisions}: a machine-readable preprocessing decision log;
\item \code{mapping}: source-to-contract column mappings;
\item \code{status}: \code{pass}, \code{review}, or \code{fail};
\item row counts and final state levels.
}

The canonical columns are \code{sequence_id}, \code{sequence_order}, \code{state},
\code{original_row}, and optional \code{duration}. Unmapped columns are preserved.
}
\description{
Applies explicit preprocessing policies and returns a deterministic,
canonical long-format representation.
}
\details{
Rows are sorted deterministically by sequence identifier, sequence order,
and original row number. When consecutive repeats are collapsed, the first
row supplies non-duration values and available durations are summed.

Unresolved errors produce \code{status = "fail"} and \code{data = NULL}; diagnostics
and decision records remain available.
}
\examples{
sequences <- data.frame(
  id = c("s2", "s1", "s1", "s2"),
  position = c(2, 2, 1, 1),
  state = c("product", "search", "home", "home")
)

prepared <- prepare_sequence_data(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

prepared$status
prepared$data

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-covariate-hmm.R
\name{fit_covariate_sequence_hmm}
\alias{fit_covariate_sequence_hmm}
\title{Fit a covariate-dependent categorical hidden Markov model}
\usage{
fit_covariate_sequence_hmm(
  data,
  n_states,
  initial_covariate_cols = NULL,
  transition_covariate_cols = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  symbol_levels = NULL,
  state_names = NULL,
  emission_probs = NULL,
  max_iter = 100L,
  inner_maxit = 100L,
  tolerance = 1e-06,
  pseudocount = 1e-06,
  ridge = 1e-06,
  seed = 1L,
  keep_posteriors = FALSE
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{n_states}{Number of latent states.}

\item{initial_covariate_cols}{Numeric sequence-constant covariates for
initial-state probabilities.}

\item{transition_covariate_cols}{Numeric row-level covariates for transition
probabilities.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{symbol_levels}{Optional observed-symbol order.}

\item{state_names}{Optional latent-state names.}

\item{emission_probs}{Optional starting emission matrix.}

\item{max_iter}{Maximum EM iterations.}

\item{inner_maxit}{Maximum BFGS iterations in each multinomial M-step.}

\item{tolerance}{Relative log-likelihood tolerance.}

\item{pseudocount}{Emission smoothing count.}

\item{ridge}{Non-negative coefficient penalty.}

\item{seed}{Reproducibility seed.}

\item{keep_posteriors}{Retain final posteriors.}
}
\value{
An object of class \code{gp3_covariate_sequence_hmm}.
}
\description{
Fits a categorical HMM whose initial-state and transition probabilities may
depend on explicitly declared numeric covariates. Multinomial-logit
coefficients are estimated inside the EM algorithm with a small ridge
penalty. Emission probabilities remain time-homogeneous.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(paste0("s", 1:8), each = 5L),
  sequence_order = rep(1:5, times = 8L),
  state = rep(c("A", "B", "C", "B", "A"), times = 8L),
  condition = rep(rep(c(0, 1), each = 4L), each = 5L),
  time_scaled = rep(seq(-1, 1, length.out = 5L), times = 8L)
)
fit_covariate_sequence_hmm(
  sequences, 2L,
  initial_covariate_cols = "condition",
  transition_covariate_cols = c("condition", "time_scaled"),
  max_iter = 3L, inner_maxit = 10L, seed = 1L
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-panel.R
\name{prepare_sequence_panel}
\alias{prepare_sequence_panel}
\title{Prepare longitudinal or panel sequence data}
\usage{
prepare_sequence_panel(
  data,
  panel_id_col,
  occasion_col,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = NULL,
  require_unique_occasions = TRUE
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared gp3sequences result.}

\item{panel_id_col}{Column identifying the repeatedly observed unit.}

\item{occasion_col}{Column identifying the wave, visit, or occasion.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{metadata_cols}{Optional columns that must remain constant within each
sequence.}

\item{require_unique_occasions}{Require at most one sequence per panel unit
and occasion.}
}
\value{
An object of class \code{gp3_sequence_panel} containing canonical data,
a panel index, state levels, and mapping information.
}
\description{
Validates a collection of sequences observed repeatedly for the same panel
units and creates an auditable panel index. Each sequence must map to exactly
one panel unit and one occasion.
}
\examples{
panel <- data.frame(
  participant_id = rep(c("p1", "p2"), each = 8L),
  occasion = rep(rep(c(1, 2), each = 4L), times = 2L),
  sequence_id = rep(c("p1_w1", "p1_w2", "p2_w1", "p2_w2"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "C", "C", "D",
            "D", "C", "B", "A", "D", "B", "B", "A")
)
prepare_sequence_panel(panel, "participant_id", "occasion")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-encoding-summaries.R
\name{summarise_sequence_states}
\alias{summarise_sequence_states}
\title{Summarise Sequence States}
\usage{
summarise_sequence_states(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{duration_col}{Optional name of a numeric duration column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}
}
\value{
A named list containing:
\itemize{
\item \code{by_sequence}: state counts and proportions within each sequence;
\item \code{overall}: state counts and proportions across all sequences;
\item \code{audit}, \code{status}, and \code{mapping} from input validation.
}

When \code{duration_col} is supplied, both tables also include duration sums,
duration proportions, and mean durations.
}
\description{
Produces per-sequence and overall state-frequency summaries from validated
ordered sequence data.
}
\details{
Observation proportions use state rows as the denominator. Sequence
proportions report the proportion of sequences in which each state occurs.
Missing durations are excluded from duration calculations; an all-missing
duration group returns \code{NA}.
}
\examples{
sequences <- data.frame(
  id = c("s1", "s1", "s1", "s2", "s2"),
  position = c(1, 2, 3, 1, 2),
  state = c("A", "B", "A", "B", "C")
)

summaries <- summarise_sequence_states(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

summaries$by_sequence
summaries$overall

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-inference.R
\name{summarise_sequence_group_inference}
\alias{summarise_sequence_group_inference}
\title{Summarise sequence group inference}
\usage{
summarise_sequence_group_inference(inference)
}
\arguments{
\item{inference}{A sequence group-inference object.}
}
\value{
A list containing the estimate, optional interval, design, and
interpretation statement.
}
\description{
Summarise sequence group inference
}
\examples{
# See `test_sequence_group_difference()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{extract_representative_sequences}
\alias{extract_representative_sequences}
\title{Extract representative sequences from clusters}
\usage{
extract_representative_sequences(
  clustering,
  distance = NULL,
  n_per_cluster = 1L
)
}
\arguments{
\item{clustering}{A clustering result.}

\item{distance}{Optional distance when absent from \code{clustering}.}

\item{n_per_cluster}{Number of representatives per cluster.}
}
\value{
A data frame with cluster, rank, representative ID, and mean
within-cluster distance.
}
\description{
Extract representative sequences from clusters
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(sequences)
fit <- cluster_sequences(distance, k = 2L)
extract_representative_sequences(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{cluster_sequences}
\alias{cluster_sequences}
\title{Cluster sequences from a distance object}
\usage{
cluster_sequences(
  distance,
  k,
  method = c("hierarchical", "pam", "clara"),
  linkage = "average",
  seed = 1L,
  ...
)
}
\arguments{
\item{distance}{A \code{dist} object or square matrix.}

\item{k}{Number of clusters.}

\item{method}{\code{"hierarchical"}, \code{"pam"}, or \code{"clara"}.}

\item{linkage}{Hierarchical linkage method.}

\item{seed}{Reproducibility seed for optional stochastic methods.}

\item{...}{Additional arguments passed to the selected clustering function.}
}
\value{
A list of class \code{gp3_sequence_clustering} containing assignments,
model object, medoid identifiers where available, and settings.
}
\description{
Cluster sequences from a distance object
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(sequences)
cluster_sequences(distance, k = 2L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-inference.R
\name{bootstrap_sequence_group_difference}
\alias{bootstrap_sequence_group_difference}
\title{Bootstrap a sequence group difference}
\usage{
bootstrap_sequence_group_difference(
  inference,
  n_boot = 999L,
  level = 0.95,
  seed = 1L
)
}
\arguments{
\item{inference}{A result from \code{\link[=test_sequence_group_difference]{test_sequence_group_difference()}}.}

\item{n_boot}{Number of bootstrap samples.}

\item{level}{Confidence level.}

\item{seed}{Reproducibility seed.}
}
\value{
The inference object with a \code{bootstrap} component.
}
\description{
Resamples declared independent units within groups and returns a percentile
interval for the mean difference. This interval does not by itself license
causal interpretation.
}
\examples{
# See `test_sequence_group_difference()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-multichannel-hmm.R
\name{fit_multichannel_sequence_hmm}
\alias{fit_multichannel_sequence_hmm}
\title{Fit a multichannel categorical hidden Markov model}
\usage{
fit_multichannel_sequence_hmm(
  data,
  n_states,
  channel_cols,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  symbol_levels = NULL,
  state_names = NULL,
  initial_probs = NULL,
  transition_probs = NULL,
  emission_probs = NULL,
  max_iter = 200L,
  tolerance = 1e-06,
  pseudocount = 1e-06,
  seed = 1L,
  keep_posteriors = FALSE
)
}
\arguments{
\item{data}{Long-format multichannel sequence data.}

\item{n_states}{Number of latent states.}

\item{channel_cols}{Names of categorical observation channels.}

\item{sequence_id_col, order_col}{Core sequence columns.}

\item{symbol_levels}{Optional named list of symbol orders by channel.}

\item{state_names}{Optional latent-state names.}

\item{initial_probs, transition_probs, emission_probs}{Optional starting values.}

\item{max_iter}{Maximum EM iterations.}

\item{tolerance}{Relative log-likelihood tolerance.}

\item{pseudocount}{Non-negative smoothing count.}

\item{seed}{Reproducibility seed.}

\item{keep_posteriors}{Retain final forward-backward results.}
}
\value{
An object of class \code{gp3_multichannel_sequence_hmm}.
}
\description{
Fits a finite-state, time-homogeneous HMM to two or more categorical
channels under conditional independence of channels given the latent state.
Latent states are statistical model states only.
}
\examples{
multichannel <- data.frame(
  sequence_id = rep(paste0("s", 1:4), each = 5L),
  sequence_order = rep(1:5, times = 4L),
  action = c("A", "B", "C", "C", "D", "A", "B", "B", "C", "D",
             "D", "C", "B", "A", "A", "D", "C", "C", "B", "A"),
  context = rep(c("x", "x", "y", "y", "z"), times = 4L)
)
fit_multichannel_sequence_hmm(multichannel, 2L,
                              channel_cols = c("action", "context"),
                              max_iter = 5L, seed = 1L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-capabilities.R
\name{sequence_capabilities}
\alias{sequence_capabilities}
\title{Report gp3sequences capabilities and optional integrations}
\usage{
sequence_capabilities(include_optional = TRUE, check_versions = TRUE)
}
\arguments{
\item{include_optional}{Logical; include optional/reference capabilities.}

\item{check_versions}{Logical; report installed optional-package versions.}
}
\value{
A data frame with capability family, capability, implementation role,
backend, availability, and version information. The function never installs,
attaches, or loads optional packages.
}
\description{
Returns a deterministic, machine-readable inventory of native analytical
capabilities, optional adapters, reference implementations, and specialist
handoffs relevant to the current gp3sequences development series.
}
\examples{
capabilities <- sequence_capabilities()
capabilities[c("family", "capability", "role", "available")]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{create_sequence_cluster_ensemble}
\alias{create_sequence_cluster_ensemble}
\title{Create a sequence-cluster ensemble}
\usage{
create_sequence_cluster_ensemble(..., k, linkage = "average")
}
\arguments{
\item{...}{Two or more \code{gp3_sequence_clustering} objects or named assignment
vectors.}

\item{k}{Number of consensus clusters.}

\item{linkage}{Hierarchical linkage applied to the co-association distance.}
}
\value{
An object of class \code{gp3_sequence_cluster_ensemble} containing the
consensus assignments, co-association matrix, model, and source solutions.
}
\description{
Combines multiple named cluster assignments through a co-association matrix
and applies hierarchical clustering to \code{1 - co-association}.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
d1 <- compute_sequence_distance(sequences, method = "lcs")
d2 <- compute_sequence_distance(sequences, method = "transition")
create_sequence_cluster_ensemble(cluster_sequences(d1, 2L),
                                 cluster_sequences(d2, 2L), k = 2L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{summarise_sequence_cluster_stability}
\alias{summarise_sequence_cluster_stability}
\title{Summarise sequence-cluster stability}
\usage{
summarise_sequence_cluster_stability(bootstrap, threshold = 0.8)
}
\arguments{
\item{bootstrap}{A result from \code{\link[=bootstrap_sequence_clusters]{bootstrap_sequence_clusters()}}.}

\item{threshold}{Pairwise stability threshold.}
}
\value{
Overall, cluster-level, and low-stability pair summaries.
}
\description{
Summarise sequence-cluster stability
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(sequences)
boot <- bootstrap_sequence_clusters(distance, k = 2L, n_boot = 5L)
summarise_sequence_cluster_stability(boot)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-time-models.R
\name{plot_time_varying_sequence_model}
\alias{plot_time_varying_sequence_model}
\title{Plot predicted time-varying sequence probabilities}
\usage{
plot_time_varying_sequence_model(
  model,
  time = NULL,
  level = 0.95,
  show_interval = TRUE,
  ...
)
}
\arguments{
\item{model}{A fitted time-varying sequence model.}

\item{time}{Optional prediction grid.}

\item{level}{Pointwise confidence level.}

\item{show_interval}{Draw pointwise confidence ribbons.}

\item{...}{Additional arguments passed to \code{\link[graphics:plot]{graphics::plot()}}.}
}
\value{
Prediction data, invisibly.
}
\description{
Plot predicted time-varying sequence probabilities
}
\examples{
# See `fit_time_varying_sequence_model()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-adapters.R
\name{as_grpstring_data}
\alias{as_grpstring_data}
\title{Create GrpString-compatible event and string inputs}
\usage{
as_grpstring_data(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  alphabet = NULL
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{alphabet}{Optional single-character symbols. When omitted, printable
ASCII characters are assigned deterministically.}
}
\value{
A list of class \code{gp3_grpstring_input} containing a wide event data
frame, event-name vector, character vector, conversion key, and string vector.
}
\description{
Create GrpString-compatible event and string inputs
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
as_grpstring_data(sequences)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-subsequences.R
\name{summarise_sequence_subsequences}
\alias{summarise_sequence_subsequences}
\title{Summarise non-contiguous subsequences}
\usage{
summarise_sequence_subsequences(occurrences)
}
\arguments{
\item{occurrences}{A result from \code{\link[=extract_sequence_subsequences]{extract_sequence_subsequences()}}.}
}
\value{
A data frame with occurrence counts, sequence prevalence, and span
diagnostics.
}
\description{
Summarise non-contiguous subsequences
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "B", "D"))
summarise_sequence_subsequences(extract_sequence_subsequences(data))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-multichannel-hmm.R
\name{plot_multichannel_sequence_hmm}
\alias{plot_multichannel_sequence_hmm}
\title{Plot multichannel HMM emission profiles}
\usage{
plot_multichannel_sequence_hmm(model, channel = model$channel_names[1L], ...)
}
\arguments{
\item{model}{A fitted multichannel HMM.}

\item{channel}{Channel name to plot.}

\item{...}{Additional arguments passed to \code{\link[graphics:barplot]{graphics::barplot()}}.}
}
\value{
The selected emission matrix, invisibly.
}
\description{
Plot multichannel HMM emission profiles
}
\examples{
# See `fit_multichannel_sequence_hmm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-adapters.R
\name{as_igraph_transition_network}
\alias{as_igraph_transition_network}
\title{Convert a transition network to an igraph object}
\usage{
as_igraph_transition_network(network, directed = TRUE)
}
\arguments{
\item{network}{A first-order transition network.}

\item{directed}{Whether the resulting graph is directed.}
}
\value{
An \code{igraph} graph with edge attributes copied from the network.
}
\description{
Convert a transition network to an igraph object
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
network <- create_transition_network(sequences)
if (requireNamespace("igraph", quietly = TRUE)) {
  as_igraph_transition_network(network)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-latent-models.R
\name{fit_sequence_hmm_mixture}
\alias{fit_sequence_hmm_mixture}
\title{Fit a mixture of categorical hidden Markov models}
\usage{
fit_sequence_hmm_mixture(
  data,
  n_components,
  n_states,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  symbol_levels = NULL,
  max_iter = 200L,
  inner_initial_iter = 20L,
  tolerance = 1e-06,
  pseudocount = 1e-06,
  seed = 1L
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{n_components}{Number of mixture components.}

\item{n_states}{Number of hidden states per component; scalar or vector.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{symbol_levels}{Optional symbol ordering.}

\item{max_iter}{Maximum mixture-EM iterations.}

\item{inner_initial_iter}{Initial single-HMM iterations per component.}

\item{tolerance}{Relative log-likelihood tolerance.}

\item{pseudocount}{Smoothing count.}

\item{seed}{Reproducibility seed.}
}
\value{
An object of class \code{gp3_sequence_hmm_mixture}.
}
\description{
Fits sequence-level mixture components, each containing a categorical HMM.
Component membership is a statistical clustering device and should not be
interpreted as a substantive latent type without external validation.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
fit_sequence_hmm_mixture(sequences, n_components = 2L, n_states = 2L,
                         max_iter = 5L, inner_initial_iter = 2L, seed = 1L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-covariate-hmm.R
\name{decode_covariate_sequence_states}
\alias{decode_covariate_sequence_states}
\title{Decode states from a covariate-dependent HMM}
\usage{
decode_covariate_sequence_states(
  model,
  data = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  method = c("viterbi", "posterior")
)
}
\arguments{
\item{model}{A fitted covariate HMM.}

\item{data}{Optional new data. Training data are used when omitted.}

\item{sequence_id_col, order_col, state_col}{Core columns for new data.}

\item{method}{\code{"viterbi"} or \code{"posterior"}.}
}
\value{
A long decoded-state data frame.
}
\description{
Decode states from a covariate-dependent HMM
}
\examples{
# See `fit_covariate_sequence_hmm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-subsequences.R
\name{plot_sequence_subsequences}
\alias{plot_sequence_subsequences}
\title{Plot non-contiguous subsequence summaries}
\usage{
plot_sequence_subsequences(
  x,
  metric = "sequence_prevalence",
  top_n = 10L,
  decreasing = TRUE,
  ...
)
}
\arguments{
\item{x}{A summary or group-comparison table.}

\item{metric}{Numeric column to plot.}

\item{top_n}{Maximum number of subsequences.}

\item{decreasing}{Sort metric in decreasing order.}

\item{...}{Additional arguments passed to \code{\link[graphics:barplot]{graphics::barplot()}}.}
}
\value{
The plotted rows, invisibly.
}
\description{
Plot non-contiguous subsequence summaries
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "B", "D"))
summary <- summarise_sequence_subsequences(extract_sequence_subsequences(data))
plot_sequence_subsequences(summary, top_n = 5L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-networks.R
\name{predict_next_state}
\alias{predict_next_state}
\title{Predict the next state from a transition model}
\usage{
predict_next_state(model, history, top_n = NULL)
}
\arguments{
\item{model}{A higher-order transition model.}

\item{history}{Character vector of observed recent states.}

\item{top_n}{Optional number of states to retain. Returned probabilities
remain on the full-model scale and are not renormalised after truncation.}
}
\value{
A probability table ordered from highest to lowest probability,
with the context order actually used.
}
\description{
Predict the next state from a transition model
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
model <- fit_higher_order_transition_model(sequences, order = 2L)
predict_next_state(model, c("A", "B"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{summarise_sequence_distance}
\alias{summarise_sequence_distance}
\title{Summarise a sequence-distance object}
\usage{
summarise_sequence_distance(distance)
}
\arguments{
\item{distance}{A distance object returned by \code{\link[=compute_sequence_distance]{compute_sequence_distance()}}.}
}
\value{
A list containing an overall summary and per-sequence mean, median,
minimum, and maximum distances.
}
\description{
Summarise a sequence-distance object
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(sequences)
summarise_sequence_distance(distance)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-adapters.R
\name{prepare_gp3tools_sequences}
\alias{prepare_gp3tools_sequences}
\title{Prepare common gp3tools-style sequence outputs}
\usage{
prepare_gp3tools_sequences(
  data,
  sequence_id_col = NULL,
  order_col = NULL,
  state_col = NULL,
  duration_col = NULL,
  metadata_cols = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame or list with a data-frame \code{data} component.}

\item{sequence_id_col, order_col, state_col}{Optional explicit mappings.}

\item{duration_col}{Optional duration mapping.}

\item{metadata_cols}{Optional constant-within-sequence metadata columns.}

\item{...}{Additional arguments passed to \code{\link[=prepare_sequence_data]{prepare_sequence_data()}}.}
}
\value{
A standard gp3sequences preparation result.
}
\description{
This optional compatibility helper recognises ordinary data frames or lists
containing a data-frame component and maps common sequence-column names to
the neutral gp3sequences contract. It does not require a gp3tools class.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
names(sequences)[names(sequences) == "sequence_order"] <- "position"
names(sequences)[names(sequences) == "state"] <- "aoi_label"
prepare_gp3tools_sequences(sequences)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-latent-models.R
\name{fit_sequence_hmm}
\alias{fit_sequence_hmm}
\title{Fit a categorical hidden Markov model}
\usage{
fit_sequence_hmm(
  data,
  n_states,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  symbol_levels = NULL,
  state_names = NULL,
  initial_probs = NULL,
  transition_probs = NULL,
  emission_probs = NULL,
  max_iter = 200L,
  tolerance = 1e-06,
  pseudocount = 1e-06,
  seed = 1L,
  keep_posteriors = FALSE
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{n_states}{Number of latent states.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{symbol_levels}{Optional observed-symbol ordering.}

\item{state_names}{Optional latent-state names.}

\item{initial_probs, transition_probs, emission_probs}{Optional starting values.}

\item{max_iter}{Maximum EM iterations.}

\item{tolerance}{Relative log-likelihood tolerance.}

\item{pseudocount}{Non-negative smoothing count.}

\item{seed}{Reproducibility seed.}

\item{keep_posteriors}{Retain final forward-backward results.}
}
\value{
An object of class \code{gp3_sequence_hmm} containing fitted parameters,
log likelihood, convergence diagnostics, symbol coding, and optional
posteriors.
}
\description{
Fits a finite-state, time-homogeneous categorical HMM by Baum-Welch EM.
Latent states are statistical model states only; they are not psychological,
diagnostic, or causal constructs.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
fit_sequence_hmm(sequences, n_states = 2L, max_iter = 5L, seed = 1L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-consensus-groups.R
\name{summarise_consensus_agreement}
\alias{summarise_consensus_agreement}
\title{Summarise consensus agreement}
\usage{
summarise_consensus_agreement(
  consensus,
  by = c("overall", "group", "position"),
  threshold = 0.5
)
}
\arguments{
\item{consensus}{A result from \code{\link[=create_consensus_sequence]{create_consensus_sequence()}}.}

\item{by}{Summary level: \code{"overall"}, \code{"group"}, or \code{"position"}.}

\item{threshold}{Agreement threshold used to count low-agreement positions.}
}
\value{
A data frame with position counts, mean, median, minimum and maximum
agreement, weighted agreement, tie counts, and low-agreement counts.
}
\description{
Summarise consensus agreement
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
consensus <- create_consensus_sequence(sequences)
summarise_consensus_agreement(consensus)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-latent-models.R
\name{summarise_sequence_hmm}
\alias{summarise_sequence_hmm}
\title{Summarise a fitted sequence HMM}
\usage{
summarise_sequence_hmm(model)
}
\arguments{
\item{model}{A single or mixture HMM.}
}
\value{
A list of convergence, fit, initial, transition, emission, and
mixture summaries.
}
\description{
Summarise a fitted sequence HMM
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
model <- fit_sequence_hmm(sequences, 2L, max_iter = 5L)
summarise_sequence_hmm(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-latent-models.R
\name{decode_sequence_states}
\alias{decode_sequence_states}
\title{Decode hidden states from a fitted HMM}
\usage{
decode_sequence_states(
  model,
  data = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  method = c("viterbi", "posterior"),
  component = NULL
)
}
\arguments{
\item{model}{A fitted single HMM or HMM mixture.}

\item{data}{Optional new long-format data. Training sequences are used when
omitted.}

\item{sequence_id_col, order_col, state_col}{Sequence columns for new data.}

\item{method}{\code{"viterbi"} or \code{"posterior"}.}

\item{component}{Mixture component to decode. When omitted for a mixture,
each sequence uses its highest-responsibility component.}
}
\value{
A long data frame containing decoded latent states and posterior
probabilities where available.
}
\description{
Decode hidden states from a fitted HMM
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
model <- fit_sequence_hmm(sequences, 2L, max_iter = 5L)
decode_sequence_states(model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-data.R
\name{audit_sequence_data}
\alias{audit_sequence_data}
\title{Audit Long-Format Sequence Data}
\usage{
audit_sequence_data(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{duration_col}{Optional name of a numeric duration column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}
}
\value{
A data frame with one row per detected issue and the stable columns
\code{sequence_id}, \code{row}, \code{column}, \code{issue_code}, \code{severity}, \code{value},
\code{message}, and \code{action}. Severity values are \code{error}, \code{review}, and
\code{info}.
}
\description{
Examines a long-format data frame against the neutral \code{gp3sequences}
sequence-data contract without modifying the input.
}
\details{
The audit checks column mappings, empty inputs, missing identifiers,
missing or non-numeric order values, duplicated positions, integer order
gaps, unordered rows, missing states, consecutive repeated states,
single-row sequences, invalid durations, inconsistent metadata, unexpected
states, and unused factor levels.

The function reports structural properties only. It does not infer
psychological, cognitive, emotional, or diagnostic states.
}
\examples{
sequences <- data.frame(
  id = rep(c("s1", "s2"), each = 3L),
  position = rep(1:3, times = 2L),
  state = c("home", "search", "product", "home", "category", "product")
)

audit_sequence_data(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-encoding-summaries.R
\name{summarise_sequence_transitions}
\alias{summarise_sequence_transitions}
\title{Summarise Adjacent Sequence Transitions}
\usage{
summarise_sequence_transitions(
  data,
  sequence_id_col,
  order_col,
  state_col,
  metadata_cols = NULL,
  expected_states = NULL,
  include_self = TRUE
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}

\item{include_self}{Logical value indicating whether transitions from a
state to the same state should be included.}
}
\value{
A named list containing:
\itemize{
\item \code{by_sequence}: transition counts, proportions within each sequence, and
conditional proportions within each origin state;
\item \code{overall}: transition counts, sequence coverage, global proportions, and
conditional origin-state proportions;
\item \code{audit}, \code{status}, \code{mapping}, and the resolved \code{include_self} setting.
}
}
\description{
Counts transitions between adjacent ordered states for each sequence and
across the complete data set.
}
\details{
A transition is defined only between adjacent rows after deterministic
ordering by sequence identifier, sequence order, and original row.
Sequences with one state contribute no transitions.
}
\examples{
sequences <- data.frame(
  id = c("s1", "s1", "s1", "s2", "s2"),
  position = c(1, 2, 3, 1, 2),
  state = c("A", "B", "C", "A", "C")
)

transitions <- summarise_sequence_transitions(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

transitions$by_sequence
transitions$overall

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motifs.R
\name{extract_sequence_ngrams}
\alias{extract_sequence_ngrams}
\title{Extract Contiguous Sequence N-Grams}
\usage{
extract_sequence_ngrams(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL,
  min_length = 2L,
  max_length = 3L,
  overlap = c("allow", "disallow"),
  separator = " > ",
  state_levels = NULL
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{duration_col}{Optional name of a numeric duration column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}

\item{min_length}{Positive whole number giving the shortest motif length.}

\item{max_length}{Positive whole number giving the longest motif length.}

\item{overlap}{Character value specifying whether overlapping occurrences of
the same motif within the same sequence are \code{"allow"}ed or
\code{"disallow"}ed.}

\item{separator}{Character value used only to display state labels in the
human-readable \code{motif} column.}

\item{state_levels}{Optional atomic vector defining the complete state
ordering. When omitted, factor levels are respected; otherwise observed
labels are sorted deterministically.}
}
\value{
A named list containing:
\itemize{
\item \code{occurrences}: one row per retained contiguous motif occurrence;
\item \code{motifs}: the distinct motif dictionary;
\item \code{sequences}: sequence-level counts of candidate and retained occurrences;
\item \code{state_dictionary}: the deterministic state dictionary;
\item \code{audit}, \code{status}, and \code{mapping} from input validation;
\item \code{settings}: the resolved motif extraction settings.
}
}
\description{
Enumerates contiguous state motifs from validated long-format sequence data.
}
\details{
Motifs are contiguous windows only. No subsequence gaps, edit distances,
statistical tests, or substantive interpretations are introduced.
Consecutive repeated states are used exactly as supplied; any repeat
collapsing should be performed explicitly with \code{prepare_sequence_data()}
before extraction.

With \code{overlap = "disallow"}, overlap is resolved independently within each
sequence-motif pair by a deterministic left-to-right greedy rule. Different
motifs and different motif lengths do not compete for positions.

The collision-resistant \code{motif_id} and \code{motif_key} columns are derived from
deterministic state codes. The display separator may therefore also occur
inside a state label without changing motif identity.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
)

ngrams <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state",
  min_length = 2,
  max_length = 3,
  overlap = "allow"
)

ngrams$occurrences
ngrams$motifs

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-visualisations-extended.R
\name{plot_sequence_state_distribution}
\alias{plot_sequence_state_distribution}
\title{Plot state distributions over aligned positions}
\usage{
plot_sequence_state_distribution(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  proportion = TRUE,
  state_levels = NULL,
  palette = "Dark 3",
  ...
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared result.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{proportion}{Plot position-wise proportions rather than counts.}

\item{state_levels}{Optional state ordering.}

\item{palette}{Base HCL palette.}

\item{...}{Additional arguments passed to \code{\link[graphics:matplot]{graphics::matplot()}}.}
}
\value{
A position-by-state matrix, invisibly.
}
\description{
Plot state distributions over aligned positions
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "C", "D"))
plot_sequence_state_distribution(data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-time-models.R
\name{predict_time_varying_sequence_model}
\alias{predict_time_varying_sequence_model}
\title{Predict a time-varying sequence model}
\usage{
predict_time_varying_sequence_model(
  model,
  time = NULL,
  groups = NULL,
  level = 0.95
)
}
\arguments{
\item{model}{A fitted time-varying sequence model.}

\item{time}{Optional numeric prediction grid.}

\item{groups}{Optional subset of fitted groups.}

\item{level}{Confidence level for pointwise intervals.}
}
\value{
A data frame with fitted probabilities and pointwise intervals.
}
\description{
Predict a time-varying sequence model
}
\examples{
# See `fit_time_varying_sequence_model()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-analysis-audit.R
\name{compare_sequence_analysis_results}
\alias{compare_sequence_analysis_results}
\title{Compare two gp3sequences analysis results}
\usage{
compare_sequence_analysis_results(
  x,
  y,
  tolerance = 1e-08,
  compare_values = FALSE
)
}
\arguments{
\item{x, y}{Two analysis objects.}

\item{tolerance}{Numerical comparison tolerance.}

\item{compare_values}{Logical; when \code{TRUE}, also run \code{all.equal()} on the
complete objects after structural comparison.}
}
\value{
An object of class \code{gp3_sequence_analysis_comparison} containing
per-field comparisons, both audits, optional whole-object comparison, and
an overall equality flag.
}
\description{
Compares structural contracts and recoverable provenance from two analysis
objects. This helper is intended for regression checks, sensitivity analyses,
and reproducibility audits; it does not select a statistically or
substantively "best" analysis.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3"), each = 3L),
  sequence_order = rep(1:3, times = 3L),
  state = c(
    "A", "B", "C",
    "A", "B", "B",
    "C", "B", "A"
  )
)
d1 <- compute_sequence_distance(
  sequences,
  method = "levenshtein"
)
d2 <- compute_sequence_distance(
  sequences,
  method = "levenshtein"
)
compare_sequence_analysis_results(d1, d2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gp3sequences-package.R
\docType{package}
\name{gp3sequences-package}
\alias{gp3sequences}
\alias{gp3sequences-package}
\title{gp3sequences: Transparent Analysis of Ordered Categorical Sequences}
\description{
\code{gp3sequences} provides transparent, reproducible, and auditable tools
for working with ordered categorical sequences represented in ordinary
long-format data frames.
}
\details{
The package is designed around explicit column mappings, deterministic
ordering, documented preprocessing policies, machine-readable diagnostics,
reproducible analysis contracts, and conservative structural summaries.
Current workflows cover motifs and subsequences, consensus and group
comparisons, distances and clustering, transition networks and higher-order
models, categorical hidden Markov models, longitudinal and time-varying
analyses, design-aware inference, provenance auditing, visual diagnostics,
and guarded interoperability with specialist sequence-analysis packages.
It is not restricted to eye-tracking data or to any particular device or
export format.
}
\section{Interpretation boundary}{

Sequence outputs describe behavioural or structural patterns. They do not
independently establish emotion, cognition, comprehension, personality,
intention, diagnosis, deception, or other psychological attributes.
Substantive interpretation requires an appropriate study design and
external evidence.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://stefanosbalaskas.github.io/gp3sequences/}
  \item \url{https://github.com/stefanosbalaskas/gp3sequences}
  \item Report bugs at \url{https://github.com/stefanosbalaskas/gp3sequences/issues}
}

}
\author{
\strong{Maintainer}: Stefanos Balaskas \email{s.balaskas@ac.upatras.gr} (\href{https://orcid.org/0000-0003-2444-9796}{ORCID})

Authors:
\itemize{
  \item Stefanos Balaskas \email{s.balaskas@ac.upatras.gr} (\href{https://orcid.org/0000-0003-2444-9796}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-networks.R
\name{create_transition_network}
\alias{create_transition_network}
\title{Create a transition network from ordered sequences}
\usage{
create_transition_network(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  group_cols = NULL,
  order = 1L,
  include_self = TRUE,
  normalise = c("count", "from", "global"),
  smoothing = 0,
  context_separator = " > "
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared result.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{group_cols}{Optional grouping columns constant within sequence.}

\item{order}{Markov order. \code{1} creates ordinary state-to-state edges;
larger values create context-to-next-state edges.}

\item{include_self}{Include first-order self-transitions.}

\item{normalise}{Edge weight scale: counts, conditional probabilities from
each context, or global shares.}

\item{smoothing}{Non-negative additive smoothing applied to observed edges.}

\item{context_separator}{Separator used for higher-order contexts.}
}
\value{
A data frame of class \code{gp3_transition_network} containing context,
next state, counts, weights, sequence prevalence, and group columns.
}
\description{
Create a transition network from ordered sequences
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
create_transition_network(sequences, normalise = "from")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-subsequences.R
\name{filter_sequence_subsequences}
\alias{filter_sequence_subsequences}
\title{Filter non-contiguous subsequence summaries}
\usage{
filter_sequence_subsequences(
  summary,
  min_sequences = 1L,
  min_prevalence = 0,
  max_mean_gap = Inf,
  top_n = NULL,
  ties = c("include", "exclude")
)
}
\arguments{
\item{summary}{A data frame from \code{\link[=summarise_sequence_subsequences]{summarise_sequence_subsequences()}}.}

\item{min_sequences}{Minimum sequence count.}

\item{min_prevalence}{Minimum sequence prevalence.}

\item{max_mean_gap}{Optional maximum mean gap.}

\item{top_n}{Optional number of rows retained after deterministic sorting.}

\item{ties}{Include or exclude ties at the \code{top_n} boundary.}
}
\value{
A filtered summary data frame.
}
\description{
Filter non-contiguous subsequence summaries
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "B", "D"))
x <- summarise_sequence_subsequences(extract_sequence_subsequences(data))
filter_sequence_subsequences(x, min_sequences = 2L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-visualisations-extended.R
\name{plot_sequence_distance_heatmap}
\alias{plot_sequence_distance_heatmap}
\title{Plot a sequence-distance heatmap}
\usage{
plot_sequence_distance_heatmap(
  distance,
  order_by = NULL,
  palette = "Viridis",
  show_labels = TRUE,
  ...
)
}
\arguments{
\item{distance}{A sequence distance object or square matrix.}

\item{order_by}{Optional clustering or named assignment vector used to order
the matrix.}

\item{palette}{Base HCL palette.}

\item{show_labels}{Draw sequence identifiers.}

\item{...}{Additional arguments passed to \code{\link[graphics:image]{graphics::image()}}.}
}
\value{
The ordered distance matrix, invisibly.
}
\description{
Plot a sequence-distance heatmap
}
\examples{
data <- data.frame(sequence_id = rep(paste0("s", 1:4), each = 4L),
                   sequence_order = rep(1:4, 4L),
                   state = c("A", "B", "C", "D", "A", "B", "C", "C",
                             "D", "C", "B", "A", "D", "C", "A", "A"))
plot_sequence_distance_heatmap(compute_sequence_distance(data))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-consensus-groups.R
\name{plot_consensus_sequence}
\alias{plot_consensus_sequence}
\title{Plot a consensus sequence}
\usage{
plot_consensus_sequence(
  consensus,
  type = c("agreement", "states"),
  group = NULL,
  main = NULL,
  xlab = "Sequence position",
  ylab = NULL,
  ...
)
}
\arguments{
\item{consensus}{A consensus result.}

\item{type}{\code{"agreement"} or \code{"states"}.}

\item{group}{Optional group value, encoded group key, or named list of
values when grouped consensus was created. It is required when more than one
consensus group is present.}

\item{main, xlab, ylab}{Plot labels.}

\item{...}{Additional arguments passed to base graphics.}
}
\value{
The plotted data, invisibly.
}
\description{
Plot a consensus sequence
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
consensus <- create_consensus_sequence(sequences)
plot_consensus_sequence(consensus)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-consensus-groups.R
\name{plot_sequence_group_comparison}
\alias{plot_sequence_group_comparison}
\title{Plot a descriptive sequence-group comparison}
\usage{
plot_sequence_group_comparison(
  comparison,
  component = c("state", "transition", "length"),
  measure = NULL,
  top_n = 12L,
  main = NULL,
  xlab = NULL,
  ylab = NULL,
  ...
)
}
\arguments{
\item{comparison}{A result from \code{\link[=compare_sequence_groups]{compare_sequence_groups()}}.}

\item{component}{\code{"state"}, \code{"transition"}, or \code{"length"}.}

\item{measure}{Measure to display. Defaults depend on \code{component}.}

\item{top_n}{Maximum states or transitions to display.}

\item{main, xlab, ylab}{Plot labels.}

\item{...}{Additional base-graphics arguments.}
}
\value{
The plotted summary data, invisibly.
}
\description{
Plot a descriptive sequence-group comparison
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
comparison <- compare_sequence_groups(sequences, group_col = "group")
plot_sequence_group_comparison(comparison, component = "state")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-networks.R
\name{fit_higher_order_transition_model}
\alias{fit_higher_order_transition_model}
\title{Fit a higher-order transition model}
\usage{
fit_higher_order_transition_model(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  order = 2L,
  smoothing = 0.5,
  backoff = TRUE,
  context_separator = " > "
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{order}{Context order.}

\item{smoothing}{Additive smoothing over observed next states.}

\item{backoff}{Retain lower-order context tables for prediction backoff.}

\item{context_separator}{Context separator.}
}
\value{
An object of class \code{gp3_higher_order_transition_model}.
}
\description{
Fit a higher-order transition model
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
fit_higher_order_transition_model(sequences, order = 2L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-networks.R
\name{detect_transition_communities}
\alias{detect_transition_communities}
\title{Detect descriptive transition communities}
\usage{
detect_transition_communities(
  network,
  method = c("label_propagation", "components"),
  max_iter = 100L,
  seed = 1L
)
}
\arguments{
\item{network}{A first-order transition network.}

\item{method}{\code{"label_propagation"} or \code{"components"}.}

\item{max_iter}{Maximum label-propagation iterations.}

\item{seed}{Reproducibility seed used only to rotate update order.}
}
\value{
A data frame with states and deterministic community labels.
}
\description{
Detect descriptive transition communities
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
network <- create_transition_network(sequences)
detect_transition_communities(network)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motif-visualisation.R
\name{plot_sequence_motif_positions}
\alias{plot_sequence_motif_positions}
\title{Plot Sequence Motif Positions}
\usage{
plot_sequence_motif_positions(
  x,
  motifs = NULL,
  position = c("start", "centre", "end"),
  scale = c("absolute", "relative"),
  top_n = 10L,
  display = c("strip", "distribution")
)
}
\arguments{
\item{x}{An object returned by \code{extract_sequence_ngrams()} or
\code{summarise_sequence_motif_positions()}.}

\item{motifs}{Optional character vector of motif identifiers, motif keys, or
displayed motif labels. When omitted, motifs are selected deterministically
by occurrence count, sequence count, motif length, and motif key.}

\item{position}{Occurrence position represented by motif \code{"start"},
\code{"centre"}, or \code{"end"}.}

\item{scale}{Position scale: one-based \code{"absolute"} sequence positions or
\code{"relative"} positions from 0 to 1.}

\item{top_n}{Positive whole number giving the number of motifs selected when
\code{motifs} is \code{NULL}.}

\item{display}{Base-R display type: occurrence \code{"strip"} or
\code{"distribution"} boxplots.}
}
\value{
Invisibly returns the exact occurrence-level data used by the plot,
including deterministic motif ranks and plotting coordinates where relevant.
}
\description{
Shows occurrence locations for selected contiguous motifs.
}
\details{
The strip display uses deterministic vertical stacking rather than random
jitter. The distribution display uses one horizontal boxplot per motif.
Empty inputs produce an informative blank plot and return an empty table.
The function provides structural location summaries only.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "C", "A", "B", "C", "B")
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

plot_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative",
  top_n = 5,
  display = "strip"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motifs.R
\name{filter_sequence_motifs}
\alias{filter_sequence_motifs}
\title{Filter Sequence Motif Summaries}
\usage{
filter_sequence_motifs(
  x,
  min_occurrences = 1L,
  min_sequences = 1L,
  min_prevalence = 0,
  motif_lengths = NULL,
  top_n = NULL,
  rank_by = c("sequence_prevalence", "n_occurrences", "n_sequences"),
  ties = c("include", "first")
)
}
\arguments{
\item{x}{A motif extraction, motif summary, or filtered-motif object.}

\item{min_occurrences}{Non-negative whole number giving the minimum total
occurrence count.}

\item{min_sequences}{Non-negative whole number giving the minimum number of
sequences containing the motif.}

\item{min_prevalence}{Minimum sequence prevalence between 0 and 1.}

\item{motif_lengths}{Optional vector of positive whole-number motif lengths
to retain.}

\item{top_n}{Optional positive whole number giving the requested number of
highest-ranked motifs.}

\item{rank_by}{Metric used for deterministic top-ranking: one of
\code{"sequence_prevalence"}, \code{"n_occurrences"}, or \code{"n_sequences"}.}

\item{ties}{Character value controlling the top-\code{n} boundary. \code{"include"}
retains every motif tied on \code{rank_by} with the final selected motif;
\code{"first"} retains exactly \code{top_n} motifs after deterministic secondary
sorting.}
}
\value{
A named list containing the filtered \code{motifs} table, the matching
sequence-level rows in \code{by_sequence}, sequence and state dictionaries,
validation metadata, filter settings, and counts before and after filtering.
}
\description{
Applies transparent count, prevalence, length, and top-ranking filters to
sequence motif summaries.
}
\details{
Filtering is descriptive and deterministic. When \code{ties = "first"}, ties are
resolved by sequence prevalence, occurrence count, sequence count, shorter
motif length, and finally the collision-resistant motif key.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

filtered <- filter_sequence_motifs(
  extracted,
  min_sequences = 2,
  top_n = 5,
  ties = "include"
)

filtered$motifs

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-encoding-summaries.R
\name{format_sequence_paths}
\alias{format_sequence_paths}
\title{Format Ordered Sequence Paths}
\usage{
format_sequence_paths(
  data,
  sequence_id_col,
  order_col,
  state_col,
  metadata_cols = NULL,
  expected_states = NULL,
  separator = " > ",
  collapse_repeats = FALSE
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}

\item{separator}{Character value inserted between adjacent state labels.}

\item{collapse_repeats}{Logical value indicating whether consecutive
repeated states should be collapsed for path display.}
}
\value{
A named list containing:
\itemize{
\item \code{paths}: one row per sequence with observation counts, formatted-state
counts, unique-state counts, start and end states, and the path string;
\item \code{audit}, \code{status}, and \code{mapping} from input validation;
\item \code{settings}: the path separator and repeat-collapsing choice.
}
}
\description{
Creates a compact one-row-per-sequence representation of ordered state
paths.
}
\details{
Repeat collapsing affects only the formatted representation. It does not
modify the supplied data or alter non-consecutive repeated states.
Input rows are ordered deterministically for formatting, but review-level
diagnostics such as \code{unordered_rows} remain reflected in the returned
\code{status} and \code{audit}.
}
\examples{
sequences <- data.frame(
  id = c("s1", "s1", "s1", "s2", "s2"),
  position = c(1, 2, 3, 1, 2),
  state = c("A", "B", "C", "A", "C")
)

paths <- format_sequence_paths(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

paths$paths

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-panel.R
\name{summarise_sequence_panel}
\alias{summarise_sequence_panel}
\title{Summarise a sequence panel}
\usage{
summarise_sequence_panel(panel)
}
\arguments{
\item{panel}{A result from \code{\link[=prepare_sequence_panel]{prepare_sequence_panel()}}.}
}
\value{
A list with occasion-level sequence summaries and state prevalence.
}
\description{
Summarise a sequence panel
}
\examples{
panel_data <- data.frame(
  participant_id = rep(c("p1", "p2"), each = 6L),
  occasion = rep(rep(c(1, 2), each = 3L), times = 2L),
  sequence_id = rep(c("a", "b", "c", "d"), each = 3L),
  sequence_order = rep(1:3, times = 4L),
  state = c("A", "B", "C", "A", "C", "C", "C", "B", "A", "C", "B", "B")
)
summarise_sequence_panel(prepare_sequence_panel(panel_data, "participant_id", "occasion"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-covariate-hmm.R
\name{predict_covariate_transition_probabilities}
\alias{predict_covariate_transition_probabilities}
\title{Predict covariate-dependent transition probabilities}
\usage{
predict_covariate_transition_probabilities(model, newdata)
}
\arguments{
\item{model}{A fitted covariate HMM.}

\item{newdata}{Data frame containing transition covariates.}
}
\value{
A long data frame with one row per input row, origin state, and
destination state.
}
\description{
Predict covariate-dependent transition probabilities
}
\examples{
# See `fit_covariate_sequence_hmm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-networks.R
\name{summarise_transition_centrality}
\alias{summarise_transition_centrality}
\title{Summarise transition-network centrality}
\usage{
summarise_transition_centrality(
  network,
  directed = TRUE,
  pagerank_damping = 0.85,
  pagerank_tolerance = 1e-10,
  pagerank_max_iter = 1000L
)
}
\arguments{
\item{network}{A first-order transition network.}

\item{directed}{Treat the network as directed.}

\item{pagerank_damping}{Damping factor for PageRank.}

\item{pagerank_tolerance}{Convergence tolerance.}

\item{pagerank_max_iter}{Maximum PageRank iterations.}
}
\value{
A data frame containing degree, strength, weighted closeness,
unweighted betweenness, and PageRank centrality.
}
\description{
Summarise transition-network centrality
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
network <- create_transition_network(sequences)
summarise_transition_centrality(network)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-visualisations-extended.R
\name{plot_transition_network}
\alias{plot_transition_network}
\title{Plot a first-order transition network}
\usage{
plot_transition_network(
  network,
  weight_col = "weight",
  minimum_weight = 0,
  vertex_cex = 1,
  edge_scale = 5,
  ...
)
}
\arguments{
\item{network}{A first-order network from \code{\link[=create_transition_network]{create_transition_network()}}.}

\item{weight_col}{Edge-weight column.}

\item{minimum_weight}{Minimum plotted edge weight.}

\item{vertex_cex}{Vertex label size.}

\item{edge_scale}{Edge-width multiplier.}

\item{...}{Additional arguments passed to \code{\link[graphics:plot.window]{graphics::plot.window()}}.}
}
\value{
The plotted network rows, invisibly.
}
\description{
Plot a first-order transition network
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "C", "D"))
plot_transition_network(create_transition_network(data, normalise = "from"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-time-models.R
\name{fit_time_varying_sequence_model}
\alias{fit_time_varying_sequence_model}
\title{Fit a time-varying sequence condition model}
\usage{
fit_time_varying_sequence_model(
  data,
  group_col,
  participant_id_col,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  time_col = NULL,
  outcome = c("state", "transition"),
  target_state = NULL,
  from_state = NULL,
  to_state = NULL,
  k = 5L,
  method = "REML",
  include_random_effect = TRUE
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared result.}

\item{group_col}{Group or condition column.}

\item{participant_id_col}{Participant or repeated-unit column.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{time_col}{Optional numeric time column; defaults to \code{order_col}.}

\item{outcome}{\code{"state"} or \code{"transition"}.}

\item{target_state}{Target state for occupancy models.}

\item{from_state, to_state}{Target transition for transition models.}

\item{k}{Basis dimension for time smooths.}

\item{method}{Smoothing-parameter estimation method passed to
\code{\link[mgcv:gam]{mgcv::gam()}}.}

\item{include_random_effect}{Include a participant random-effect smooth.}
}
\value{
An object of class \code{gp3_sequence_time_model}.
}
\description{
Fits a binary generalized additive model for either occupancy of a target
state or occurrence of a target transition over aligned sequence time. Group
specific smooths model time-varying condition differences, and a participant
random-effect smooth accounts for repeated observations. The optional \code{mgcv}
package is required.
}
\examples{
set.seed(2026)

n_participants <- 24L
n_positions <- 12L

participant_index <- rep(
  seq_len(n_participants),
  each = n_positions
)

sequence_order <- rep(
  seq_len(n_positions),
  times = n_participants
)

group_by_participant <- rep(
  c("control", "treatment"),
  each = n_participants \%/\% 2L
)

group <- rep(
  group_by_participant,
  each = n_positions
)

probability <- stats::plogis(
  -0.8 +
    0.08 * sequence_order +
    0.4 * (group == "treatment")
)

observed_b <- stats::rbinom(
  length(probability),
  size = 1L,
  prob = probability
)

sequences <- data.frame(
  participant_id = paste0("p", participant_index),
  sequence_id = paste0("s", participant_index),
  sequence_order = sequence_order,
  state = ifelse(observed_b == 1L, "B", "A"),
  group = group
)

if (requireNamespace("mgcv", quietly = TRUE)) {
  fit <- fit_time_varying_sequence_model(
    sequences,
    group_col = "group",
    participant_id_col = "participant_id",
    target_state = "B",
    k = 4L
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{validate_sequence_clusters}
\alias{validate_sequence_clusters}
\title{Validate sequence clusters descriptively}
\usage{
validate_sequence_clusters(clustering, distance = NULL)
}
\arguments{
\item{clustering}{A result from \code{\link[=cluster_sequences]{cluster_sequences()}} or a named assignment
vector.}

\item{distance}{Optional distance object when \code{clustering} is an assignment
vector.}
}
\value{
A list containing overall validation metrics, cluster sizes, and
per-sequence silhouette values.
}
\description{
Validate sequence clusters descriptively
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(sequences)
fit <- cluster_sequences(distance, k = 2L)
validate_sequence_clusters(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-visualisations-extended.R
\name{plot_sequence_entropy}
\alias{plot_sequence_entropy}
\title{Plot position-wise state entropy}
\usage{
plot_sequence_entropy(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  base = 2,
  normalise = TRUE,
  ...
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared result.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{base}{Logarithm base.}

\item{normalise}{Divide entropy by the maximum entropy for the observed
alphabet.}

\item{...}{Additional arguments passed to \code{\link[graphics:plot]{graphics::plot()}}.}
}
\value{
A data frame of position-wise entropy values, invisibly.
}
\description{
Plot position-wise state entropy
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "C", "D"))
plot_sequence_entropy(data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-consensus-groups.R
\name{create_consensus_sequence}
\alias{create_consensus_sequence}
\title{Create an aligned-position consensus sequence}
\usage{
create_consensus_sequence(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  group_cols = NULL,
  weight_col = NULL,
  missing_state_policy = c("exclude", "state", "error"),
  missing_state_label = "<MISSING>",
  tie_method = c("first", "last", "missing", "all"),
  state_levels = NULL,
  min_support = 1L
)
}
\arguments{
\item{data}{A long-format data frame, a prepared gp3sequences result, or a
data frame containing canonical columns.}

\item{sequence_id_col, order_col, state_col}{Column names defining sequences.}

\item{group_cols}{Optional columns defining independent consensus groups.}

\item{weight_col}{Optional non-negative row-weight column.}

\item{missing_state_policy}{One of \code{"exclude"}, \code{"state"}, or \code{"error"}.}

\item{missing_state_label}{Label used when missing states are retained.}

\item{tie_method}{Deterministic tie policy: \code{"first"}, \code{"last"},
\code{"missing"}, or \code{"all"}.}

\item{state_levels}{Optional preferred state ordering used for ties.}

\item{min_support}{Minimum number of contributing sequences at a position.}
}
\value{
A data frame of class \code{gp3_consensus_sequence} containing group
columns, sequence position, consensus state, support counts and weights,
agreement proportion, tie count, tied states, and total group sequences.
}
\description{
Computes a descriptive consensus state at each observed sequence position.
The function does not treat the consensus as a behavioural norm and does not
infer psychological or causal meaning.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
create_consensus_sequence(sequences, group_cols = "group")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-inference.R
\name{test_sequence_group_difference}
\alias{test_sequence_group_difference}
\title{Test a sequence group difference}
\usage{
test_sequence_group_difference(
  data,
  design,
  metric = c("sequence_length", "transition_count", "state_prevalence",
    "subsequence_presence"),
  target_state = NULL,
  target_subsequence = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  separator = " > ",
  n_permutations = 999L,
  alternative = c("two.sided", "greater", "less"),
  seed = 1L
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{design}{A comparison design.}

\item{metric}{\code{"sequence_length"}, \code{"transition_count"},
\code{"state_prevalence"}, or \code{"subsequence_presence"}.}

\item{target_state}{Required for state prevalence.}

\item{target_subsequence}{Required for subsequence presence, expressed using
\code{separator}.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{separator}{Subsequence label separator.}

\item{n_permutations}{Number of permutations.}

\item{alternative}{Alternative hypothesis.}

\item{seed}{Reproducibility seed.}
}
\value{
An object of class \code{gp3_sequence_group_inference}.
}
\description{
Aggregates sequence metrics to the declared independent unit and performs a
permutation or randomization test. For observational data the p-value tests
exchangeability-based association only; it is not a causal estimate.
}
\examples{
data <- data.frame(
  participant_id = rep(paste0("p", 1:8), each = 4L),
  sequence_id = rep(paste0("s", 1:8), each = 4L),
  sequence_order = rep(1:4, times = 8L),
  state = c(rep(c("A", "B", "C", "D"), 4L),
            rep(c("A", "A", "C", "D"), 4L)),
  group = rep(rep(c("control", "treatment"), each = 4L), each = 4L)
)
design <- declare_sequence_comparison_design("group", "participant_id",
                                             design = "randomized")
test_sequence_group_difference(data, design, metric = "state_prevalence",
                               target_state = "A", n_permutations = 99L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-distances-clustering.R
\name{bootstrap_sequence_clusters}
\alias{bootstrap_sequence_clusters}
\title{Bootstrap sequence-cluster stability}
\usage{
bootstrap_sequence_clusters(
  distance,
  k,
  method = c("hierarchical", "pam", "clara"),
  n_boot = 100L,
  sample_fraction = 0.8,
  seed = 1L,
  linkage = "average",
  ...
)
}
\arguments{
\item{distance}{Sequence distance object.}

\item{k}{Number of clusters.}

\item{method}{Clustering method.}

\item{n_boot}{Number of subsamples.}

\item{sample_fraction}{Fraction of sequences sampled in each iteration.}

\item{seed}{Reproducibility seed.}

\item{linkage}{Hierarchical linkage.}

\item{...}{Additional clustering arguments.}
}
\value{
An object of class \code{gp3_sequence_cluster_bootstrap}.
}
\description{
Uses repeated subsampling without replacement and records pairwise
co-clustering agreement relative to the full-data solution.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
distance <- compute_sequence_distance(sequences)
bootstrap_sequence_clusters(distance, k = 2L, n_boot = 5L, seed = 1L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-consensus-groups.R
\name{format_consensus_sequence}
\alias{format_consensus_sequence}
\title{Format consensus sequences as paths}
\usage{
format_consensus_sequence(
  consensus,
  separator = " -> ",
  include_order = FALSE,
  include_agreement = FALSE,
  digits = 3L
)
}
\arguments{
\item{consensus}{A consensus result.}

\item{separator}{State separator.}

\item{include_order}{Include position labels.}

\item{include_agreement}{Include rounded agreement values.}

\item{digits}{Number of decimal places for agreement values.}
}
\value{
One row per consensus group with a formatted path and position count.
}
\description{
Format consensus sequences as paths
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
consensus <- create_consensus_sequence(sequences)
format_consensus_sequence(consensus)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-data.R
\name{validate_sequence_data}
\alias{validate_sequence_data}
\title{Validate Long-Format Sequence Data}
\usage{
validate_sequence_data(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{duration_col}{Optional name of a numeric duration column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}
}
\value{
A named list containing \code{valid}, \code{status}, issue counts, the
complete \code{audit} table, the column \code{mapping}, row and sequence counts,
and observed \code{state_levels}. A result is valid when no error-severity
issue is present. Review-severity issues do not automatically invalidate
the input.
}
\description{
Produces a compact validation result based on
\code{audit_sequence_data()} without modifying the input.
}
\examples{
sequences <- data.frame(
  id = rep(c("s1", "s2"), each = 2L),
  position = rep(1:2, times = 2L),
  state = c("home", "search", "home", "product")
)

validation <- validate_sequence_data(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

validation$status
validation$valid

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-encoding-summaries.R
\name{encode_sequence_data}
\alias{encode_sequence_data}
\title{Encode Ordered Sequence States}
\usage{
encode_sequence_data(
  data,
  sequence_id_col,
  order_col,
  state_col,
  duration_col = NULL,
  metadata_cols = NULL,
  expected_states = NULL,
  state_levels = NULL,
  prefix = "S",
  width = NULL
)
}
\arguments{
\item{data}{A data frame containing ordered state observations.}

\item{sequence_id_col}{Name of the sequence identifier column.}

\item{order_col}{Name of the numeric sequence-order column.}

\item{state_col}{Name of the categorical state column.}

\item{duration_col}{Optional name of a numeric duration column.}

\item{metadata_cols}{Optional character vector naming columns that should
remain constant within each sequence.}

\item{expected_states}{Optional vector of known or permitted state values.}

\item{state_levels}{Optional atomic vector defining the complete state
ordering. When omitted, factor levels are respected; otherwise observed
state labels are sorted alphabetically.}

\item{prefix}{Character prefix used for labelled codes.}

\item{width}{Optional positive integer width for the numeric part of each
labelled code. The default is determined from the dictionary size.}
}
\value{
A named list containing:
\itemize{
\item \code{data}: deterministically sorted canonical data with \code{state_index} and
\code{state_code};
\item \code{dictionary}: state labels, integer indices, labelled codes, and an
observed-state indicator;
\item \code{audit}, \code{status}, and \code{mapping} from input validation;
\item \code{settings}: the resolved code prefix and width.
}
}
\description{
Creates a deterministic dictionary and adds integer and labelled state
codes to long-format sequence data.
}
\details{
The function does not reinterpret states. State codes are transparent
identifiers derived from an explicit or deterministic state ordering.
}
\examples{
sequences <- data.frame(
  id = c("s1", "s1", "s2", "s2"),
  position = c(1, 2, 1, 2),
  state = c("home", "search", "home", "product")
)

encoded <- encode_sequence_data(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

encoded$dictionary
encoded$data

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-panel.R
\name{compare_sequence_panel_changes}
\alias{compare_sequence_panel_changes}
\title{Compare within-panel sequence changes}
\usage{
compare_sequence_panel_changes(
  panel,
  method = c("levenshtein", "lcs", "optimal_matching", "transition"),
  normalise = c("none", "max_length", "path_length"),
  indel_cost = 1,
  substitution_cost = 1,
  substitution_matrix = NULL,
  transition_smoothing = 0
)
}
\arguments{
\item{panel}{A sequence panel.}

\item{method}{Distance method supported by \code{\link[=compute_sequence_distance]{compute_sequence_distance()}}.}

\item{normalise}{Distance normalisation.}

\item{indel_cost, substitution_cost}{Costs for optimal matching.}

\item{substitution_matrix}{Optional named substitution matrix.}

\item{transition_smoothing}{Smoothing for transition-profile distance.}
}
\value{
A data frame of class \code{gp3_sequence_panel_changes}.
}
\description{
Computes distance and simple structural changes between consecutive panel
occasions. The output is descriptive and does not establish a causal or
psychological change.
}
\examples{
panel_data <- data.frame(
  participant_id = rep(c("p1", "p2"), each = 8L),
  occasion = rep(rep(c(1, 2), each = 4L), times = 2L),
  sequence_id = rep(c("a", "b", "c", "d"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "C", "C", "D",
            "D", "C", "B", "A", "D", "B", "B", "A")
)
compare_sequence_panel_changes(
  prepare_sequence_panel(panel_data, "participant_id", "occasion"),
  method = "lcs"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-analysis-audit.R
\name{audit_sequence_analysis}
\alias{audit_sequence_analysis}
\title{Audit a gp3sequences analysis object}
\usage{
audit_sequence_analysis(x, strict = FALSE, tolerance = 1e-08)
}
\arguments{
\item{x}{A gp3sequences analysis result or supported structural object.}

\item{strict}{Logical; if \code{TRUE}, fail when the audit status is \code{"fail"}.}

\item{tolerance}{Numerical tolerance used for matrix/probability checks.}
}
\value{
An object of class \code{gp3_sequence_analysis_audit} containing
\code{summary}, \code{issues}, \code{provenance}, \code{contract}, and \code{status}.
}
\description{
Inspects supported gp3sequences objects for structural contract validity,
recoverable provenance, identifiers, state levels, method settings, and
randomness metadata. The audit does not reinterpret sequence structure as a
psychological, diagnostic, cognitive, emotional, or causal construct.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3"), each = 3L),
  sequence_order = rep(1:3, times = 3L),
  state = c(
    "A", "B", "C",
    "A", "B", "B",
    "C", "B", "A"
  )
)
distance <- compute_sequence_distance(
  sequences,
  method = "levenshtein"
)
audit_sequence_analysis(distance)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-panel.R
\name{plot_sequence_panel_changes}
\alias{plot_sequence_panel_changes}
\title{Plot longitudinal sequence changes}
\usage{
plot_sequence_panel_changes(
  changes,
  metric = c("distance", "length_change", "transition_change"),
  type = c("individual", "summary"),
  ...
)
}
\arguments{
\item{changes}{A result from \code{\link[=compare_sequence_panel_changes]{compare_sequence_panel_changes()}}.}

\item{metric}{One of \code{"distance"}, \code{"length_change"}, or
\code{"transition_change"}.}

\item{type}{Plot individual panel trajectories or occasion-transition means.}

\item{...}{Additional graphical arguments passed to base plotting functions.}
}
\value{
The plotted data, invisibly.
}
\description{
Plot longitudinal sequence changes
}
\examples{
panel_data <- data.frame(
  participant_id = rep(c("p1", "p2"), each = 6L),
  occasion = rep(rep(c(1, 2), each = 3L), times = 2L),
  sequence_id = rep(c("a", "b", "c", "d"), each = 3L),
  sequence_order = rep(1:3, times = 4L),
  state = c("A", "B", "C", "A", "C", "C", "C", "B", "A", "C", "B", "B")
)
changes <- compare_sequence_panel_changes(
  prepare_sequence_panel(panel_data, "participant_id", "occasion")
)
plot_sequence_panel_changes(changes)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-subsequences.R
\name{extract_sequence_subsequences}
\alias{extract_sequence_subsequences}
\title{Extract bounded non-contiguous sequence subsequences}
\usage{
extract_sequence_subsequences(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  metadata_cols = NULL,
  min_length = 2L,
  max_length = 5L,
  max_gap = Inf,
  max_span = Inf,
  repeated_state_policy = c("preserve", "collapse"),
  separator = " > ",
  max_combinations_per_sequence = 100000L
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared result.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{metadata_cols}{Optional sequence-constant metadata retained as
attributes.}

\item{min_length, max_length}{Minimum and maximum subsequence lengths.}

\item{max_gap}{Maximum number of skipped positions between adjacent selected
states. Use \code{Inf} for no restriction.}

\item{max_span}{Maximum difference between the first and last selected
sequence positions. Use \code{Inf} for no restriction.}

\item{repeated_state_policy}{Preserve or collapse consecutive repeated
states before mining.}

\item{separator}{Separator used in stable motif labels.}

\item{max_combinations_per_sequence}{Safety limit for the number of index
combinations considered for any one sequence.}
}
\value{
A data frame of class \code{gp3_sequence_subsequences}, with one row per
qualifying occurrence.
}
\description{
Enumerates ordered non-contiguous subsequences under explicit length, gap,
span, and safety limits. The implementation is transparent and intended for
modest sequence collections; specialist mining packages remain preferable
for very large search spaces.
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2"), each = 5L),
  sequence_order = rep(1:5, times = 2L),
  state = c("A", "B", "C", "D", "E", "A", "C", "B", "D", "E")
)
extract_sequence_subsequences(sequences, min_length = 2L, max_length = 3L,
                              max_gap = 2L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-inference.R
\name{declare_sequence_comparison_design}
\alias{declare_sequence_comparison_design}
\title{Declare a sequence group-comparison design}
\usage{
declare_sequence_comparison_design(
  group_col,
  unit_col,
  design = c("observational", "randomized", "paired_randomized"),
  pair_col = NULL,
  cluster_col = NULL
)
}
\arguments{
\item{group_col}{Group or treatment column.}

\item{unit_col}{Independent assignment or analysis-unit column.}

\item{design}{\code{"observational"}, \code{"randomized"}, or
\code{"paired_randomized"}.}

\item{pair_col}{Pair or block column required for paired randomization.}

\item{cluster_col}{Optional higher-level assignment cluster.}
}
\value{
An object of class \code{gp3_sequence_comparison_design}.
}
\description{
Records the unit of assignment and the design under which a sequence-group
contrast will be evaluated. Causal interpretation is permitted only for an
explicitly randomized design and still depends on the validity of the study
implementation.
}
\examples{
declare_sequence_comparison_design("group", "participant_id",
                                   design = "randomized")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-covariate-hmm.R
\name{summarise_covariate_sequence_hmm}
\alias{summarise_covariate_sequence_hmm}
\title{Summarise a covariate-dependent HMM}
\usage{
summarise_covariate_sequence_hmm(model)
}
\arguments{
\item{model}{A fitted covariate HMM.}
}
\value{
A list of fit, coefficient, and emission summaries.
}
\description{
Summarise a covariate-dependent HMM
}
\examples{
# See `fit_covariate_sequence_hmm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motif-visualisation.R
\name{plot_sequence_motifs}
\alias{plot_sequence_motifs}
\title{Plot Sequence Motif Summaries}
\usage{
plot_sequence_motifs(
  x,
  metric = c("sequence_prevalence", "n_occurrences", "n_sequences", "occurrence_share"),
  top_n = 20L,
  motif_lengths = NULL,
  ties = c("include", "first"),
  horizontal = TRUE
)
}
\arguments{
\item{x}{A motif extraction, summary, or filtered-motif object.}

\item{metric}{Structural metric to plot: \code{"sequence_prevalence"},
\code{"n_occurrences"}, \code{"n_sequences"}, or \code{"occurrence_share"}.}

\item{top_n}{Positive whole number giving the requested number of motifs.}

\item{motif_lengths}{Optional vector of positive whole-number motif lengths
to retain.}

\item{ties}{Top-\code{n} boundary policy. \code{"include"} retains all motifs tied on
the selected metric; \code{"first"} retains exactly \code{top_n} after deterministic
secondary sorting.}

\item{horizontal}{Logical value indicating whether bars should be horizontal.}
}
\value{
Invisibly returns the exact motif table used for plotting, including
\code{plot_rank}, \code{plot_value}, \code{plot_label}, and \code{bar_midpoint}.
}
\description{
Draws a dependency-free base-R bar chart of structural motif measurements.
}
\details{
The plot is descriptive. It does not perform inferential testing or assign
substantive meaning to motif frequency or prevalence. Empty inputs produce
an informative blank plot and return an empty table.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "C", "A", "B", "C", "B")
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

plot_sequence_motifs(
  extracted,
  metric = "sequence_prevalence",
  top_n = 10
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-subsequences.R
\name{compare_sequence_subsequences}
\alias{compare_sequence_subsequences}
\title{Compare subsequence prevalence between groups}
\usage{
compare_sequence_subsequences(
  occurrences,
  group_col,
  test = c("auto", "chisq", "fisher"),
  p_adjust = "BH",
  min_sequence_count = 1L
)
}
\arguments{
\item{occurrences}{A non-contiguous subsequence occurrence table.}

\item{group_col}{Sequence-constant group column retained through
\code{metadata_cols} during extraction.}

\item{test}{\code{"auto"}, \code{"chisq"}, or \code{"fisher"}.}

\item{p_adjust}{Multiple-testing adjustment method.}

\item{min_sequence_count}{Minimum total sequence count per subsequence.}
}
\value{
A data frame of prevalence contrasts and adjusted p-values.
}
\description{
Performs transparent contingency-table tests on sequence-level presence.
Multiple testing is adjusted explicitly. Results are associational unless a
randomized design independently justifies causal interpretation.
}
\examples{
data <- data.frame(
  sequence_id = rep(paste0("s", 1:4), each = 4L),
  sequence_order = rep(1:4, 4L),
  state = c("A", "B", "C", "D", "A", "B", "D", "D",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L)
)
occurrences <- extract_sequence_subsequences(data, metadata_cols = "group")
compare_sequence_subsequences(occurrences, "group")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-visualisations-extended.R
\name{plot_sequence_cluster_silhouette}
\alias{plot_sequence_cluster_silhouette}
\title{Plot sequence-cluster silhouette values}
\usage{
plot_sequence_cluster_silhouette(clustering, distance = NULL, ...)
}
\arguments{
\item{clustering}{A clustering result or named assignment vector.}

\item{distance}{Optional distance when \code{clustering} is an assignment vector.}

\item{...}{Additional arguments passed to \code{\link[graphics:barplot]{graphics::barplot()}}.}
}
\value{
The ordered per-sequence silhouette table, invisibly.
}
\description{
Plot sequence-cluster silhouette values
}
\examples{
data <- data.frame(sequence_id = rep(paste0("s", 1:4), each = 4L),
                   sequence_order = rep(1:4, 4L),
                   state = c("A", "B", "C", "D", "A", "B", "C", "C",
                             "D", "C", "B", "A", "D", "C", "A", "A"))
distance <- compute_sequence_distance(data)
plot_sequence_cluster_silhouette(cluster_sequences(distance, 2L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-multichannel-hmm.R
\name{decode_multichannel_sequence_states}
\alias{decode_multichannel_sequence_states}
\title{Decode latent states from a multichannel HMM}
\usage{
decode_multichannel_sequence_states(
  model,
  data = NULL,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  channel_cols = NULL,
  method = c("viterbi", "posterior")
)
}
\arguments{
\item{model}{A fitted multichannel HMM.}

\item{data}{Optional new long-format multichannel data.}

\item{sequence_id_col, order_col}{Core sequence columns for new data.}

\item{channel_cols}{Channel columns for new data; defaults to training names.}

\item{method}{\code{"viterbi"} or \code{"posterior"}.}
}
\value{
A long data frame of decoded states and posterior probabilities.
}
\description{
Decode latent states from a multichannel HMM
}
\examples{
# See `fit_multichannel_sequence_hmm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-time-models.R
\name{summarise_time_varying_sequence_model}
\alias{summarise_time_varying_sequence_model}
\title{Summarise a time-varying sequence model}
\usage{
summarise_time_varying_sequence_model(model)
}
\arguments{
\item{model}{A fitted time-varying sequence model.}
}
\value{
A list containing model metadata, parametric terms, smooth terms,
deviance explained, and convergence information.
}
\description{
Summarise a time-varying sequence model
}
\examples{
# See `fit_time_varying_sequence_model()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motif-visualisation.R
\name{format_sequence_motif_positions}
\alias{format_sequence_motif_positions}
\title{Format Sequence Motif Position Summaries}
\usage{
format_sequence_motif_positions(
  x,
  digits = 3L,
  position_units = c("proportion", "percent"),
  include_rank = TRUE
)
}
\arguments{
\item{x}{An object returned by \code{summarise_sequence_motif_positions()}.}

\item{digits}{Whole number from 0 to 15 controlling numeric rounding.}

\item{position_units}{Display units for relative positions: \code{"proportion"}
or \code{"percent"}. Absolute positions remain one-based sequence indices.}

\item{include_rank}{Logical value indicating whether to include an
earlier-to-later rank based on mean position. Ranking is performed within
each supplied \code{by} group.}
}
\value{
A named list containing the formatted \code{table}, validation metadata,
and formatting settings.
}
\description{
Produces a deterministic report-ready table from positional motif summaries.
}
\details{
Formatting copies and transforms the summary table only. The input object is
not modified. Rows are ordered deterministically by grouping values, mean
and median position, occurrence and sequence counts, motif length, and motif
key. Equal mean positions receive the same minimum rank.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "C", "A", "B", "C", "B")
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

positions <- summarise_sequence_motif_positions(
  extracted,
  position = "centre",
  scale = "relative"
)

formatted <- format_sequence_motif_positions(
  positions,
  position_units = "percent",
  digits = 1
)

formatted$table

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-multichannel-hmm.R
\name{summarise_multichannel_sequence_hmm}
\alias{summarise_multichannel_sequence_hmm}
\title{Summarise a multichannel sequence HMM}
\usage{
summarise_multichannel_sequence_hmm(model)
}
\arguments{
\item{model}{A fitted multichannel HMM.}
}
\value{
A list of convergence, fit, transition, initial, and channel-specific
emission summaries.
}
\description{
Summarise a multichannel sequence HMM
}
\examples{
# See `fit_multichannel_sequence_hmm()`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-motifs.R
\name{format_sequence_motifs}
\alias{format_sequence_motifs}
\title{Format Sequence Motif Summaries}
\usage{
format_sequence_motifs(
  x,
  digits = 3L,
  prevalence = c("proportion", "percent"),
  include_rank = TRUE,
  rank_by = c("sequence_prevalence", "n_occurrences", "n_sequences"),
  ties = c("min", "first"),
  include_ids = TRUE
)
}
\arguments{
\item{x}{A motif extraction, motif summary, or filtered-motif object.}

\item{digits}{Whole number from 0 to 15 controlling numeric rounding.}

\item{prevalence}{Character value specifying whether prevalence and
occurrence share are shown as \code{"proportion"}s or \code{"percent"}ages.}

\item{include_rank}{Logical value indicating whether to add a rank column.}

\item{rank_by}{Metric used to order and rank motifs: one of
\code{"sequence_prevalence"}, \code{"n_occurrences"}, or \code{"n_sequences"}.}

\item{ties}{Character value specifying \code{"min"} shared ranks or
deterministic \code{"first"} ranks.}

\item{include_ids}{Logical value indicating whether \code{motif_id} and
\code{motif_key} should be included in the formatted table.}
}
\value{
A named list containing \code{table}, validation metadata, and formatting
settings. The table contains only structural motif measurements.
}
\description{
Produces a stable, report-ready table from motif extraction, summary, or
filtered-motif output.
}
\details{
Formatting changes display precision and units only. It does not change the
underlying motif counts or introduce substantive interpretation.
}
\examples{
sequences <- data.frame(
  id = c(rep("s1", 5L), rep("s2", 4L)),
  position = c(1:5, 1:4),
  state = c("A", "B", "A", "B", "A", "A", "B", "A", "C")
)

extracted <- extract_sequence_ngrams(
  sequences,
  sequence_id_col = "id",
  order_col = "position",
  state_col = "state"
)

formatted <- format_sequence_motifs(
  extracted,
  prevalence = "percent",
  digits = 1
)

formatted$table

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-visualisations-extended.R
\name{plot_sequence_index}
\alias{plot_sequence_index}
\title{Plot a sequence index heatmap}
\usage{
plot_sequence_index(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  sort_by = c("input", "length", "path"),
  state_levels = NULL,
  palette = "Dark 3",
  show_sequence_labels = TRUE,
  ...
)
}
\arguments{
\item{data}{Long-format sequence data or a prepared result.}

\item{sequence_id_col, order_col, state_col}{Core sequence columns.}

\item{sort_by}{\code{"input"}, \code{"length"}, or \code{"path"}.}

\item{state_levels}{Optional state ordering.}

\item{palette}{Base HCL palette.}

\item{show_sequence_labels}{Draw sequence identifiers.}

\item{...}{Additional arguments passed to \code{\link[graphics:image]{graphics::image()}}.}
}
\value{
The plotted state-code matrix, invisibly.
}
\description{
Plot a sequence index heatmap
}
\examples{
data <- data.frame(sequence_id = rep(c("a", "b"), each = 4L),
                   sequence_order = rep(1:4, 2L),
                   state = c("A", "B", "C", "D", "A", "C", "C", "D"))
plot_sequence_index(data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence-adapters.R
\name{as_seqhmm_sequences}
\alias{as_seqhmm_sequences}
\title{Convert sequence data to seqHMM observations}
\usage{
as_seqhmm_sequences(
  data,
  sequence_id_col = "sequence_id",
  order_col = "sequence_order",
  state_col = "state",
  ...
)
}
\arguments{
\item{data}{Long-format sequence data.}

\item{sequence_id_col, order_col, state_col}{Sequence columns.}

\item{...}{Additional arguments passed to \code{\link[=as_traminer_sequences]{as_traminer_sequences()}}.}
}
\value{
A TraMineR \code{stslist} suitable for \code{seqHMM::build_hmm()}.
}
\description{
Convert sequence data to seqHMM observations
}
\examples{
sequences <- data.frame(
  sequence_id = rep(c("s1", "s2", "s3", "s4"), each = 4L),
  sequence_order = rep(1:4, times = 4L),
  state = c("A", "B", "C", "D", "A", "B", "C", "C",
            "D", "C", "B", "A", "D", "C", "A", "A"),
  group = rep(c("g1", "g2"), each = 8L),
  stringsAsFactors = FALSE
)
if (requireNamespace("TraMineR", quietly = TRUE) &&
    requireNamespace("seqHMM", quietly = TRUE)) {
  as_seqhmm_sequences(sequences)
}

}
